/*
    File:               argonAdvapi32Dll.h
    Description:        Dynamical loading of some functions of the
                        system's ADVAPI32.DLL.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         26.11.2005
    last updated:       09.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonAdvapi32Dll_h
#define __argonAdvapi32Dll_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




/*
    -----------------------------------------------------------------
    Usage:

    Simply call the functions defined in here their usual way like
    #OpenProcessToken()# or #LookupAccountName()#, etc.,
    in your source code as you need it.

    All these functions defined in here are emulating the system behavior
    that way, that they load the functions during runtime from the system
    and only if this did succeed, they will be called.

    The first time you call any of these functions, the library will try
    to be loaded and the function pointers will try to be retrieved.

    The unloading of the library will be done automatically
    during the deleting of the global variables.
    If you included this file in a DLL project, you will not want this
    behavior to happen because it should be prevented to call FreeLibrary()
    when a DLL is detached from a process (where global variables will be
    deleted).
    So in the case of DLL project, it is recommended to export an own
    DLL-deinit-function which you should redirect to a call to
    #Advapi32Dll::UnloadAPI()#. If the user then calls this deinit-function,
    the ADVAPI32.DLL will be unloaded and later on, when your DLL is detachted
    from a process, it is harmless to call #Advapi32Dll::UnloadAPI()# again
    because it will not do anything any more.
*/


#define ADVAPI32DLL_NAME            "ADVAPI32.DLL"


#if defined (__cplusplus)
extern "C" {
#endif

typedef BOOL (WINAPI *TfnOpenProcessToken)(
    IN HANDLE ProcessHandle,
    IN DWORD DesiredAccess,
    OUT PHANDLE TokenHandle
);

BOOL WINAPI argonOpenProcessToken(
    IN HANDLE ProcessHandle,
    IN DWORD DesiredAccess,
    OUT PHANDLE TokenHandle
);

typedef BOOL (WINAPI *TfnLookupAccountSidA)(
    IN LPCSTR lpSystemName,
    IN PSID Sid,
    OUT LPSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

BOOL WINAPI argonLookupAccountSidA(
    IN LPCSTR lpSystemName,
    IN PSID Sid,
    OUT LPSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

typedef BOOL (WINAPI *TfnLookupAccountSidW)(
    IN LPCWSTR lpSystemName,
    IN PSID Sid,
    OUT LPWSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

BOOL WINAPI argonLookupAccountSidW(
    IN LPCWSTR lpSystemName,
    IN PSID Sid,
    OUT LPWSTR Name,
    IN OUT LPDWORD cbName,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

typedef BOOL (WINAPI *TfnLookupAccountNameA)(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

BOOL WINAPI argonLookupAccountNameA(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

typedef BOOL (WINAPI *TfnLookupAccountNameW)(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

BOOL WINAPI argonLookupAccountNameW(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpAccountName,
    OUT PSID Sid,
    IN OUT LPDWORD cbSid,
    OUT LPWSTR ReferencedDomainName,
    IN OUT LPDWORD cbReferencedDomainName,
    OUT PSID_NAME_USE peUse
);

typedef BOOL (WINAPI *TfnLookupPrivilegeValueA)(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpName,
    OUT PLUID lpLuid
);

BOOL WINAPI argonLookupPrivilegeValueA(
    IN LPCSTR lpSystemName,
    IN LPCSTR lpName,
    OUT PLUID lpLuid
);

typedef BOOL (WINAPI *TfnLookupPrivilegeValueW)(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpName,
    OUT PLUID lpLuid
);

BOOL WINAPI argonLookupPrivilegeValueW(
    IN LPCWSTR lpSystemName,
    IN LPCWSTR lpName,
    OUT PLUID lpLuid
);

typedef BOOL (WINAPI *TfnGetTokenInformation)(
    IN HANDLE TokenHandle,
    IN TOKEN_INFORMATION_CLASS TokenInformationClass,
    OUT LPVOID TokenInformation,
    IN DWORD TokenInformationLength,
    OUT PDWORD ReturnLength
);

BOOL WINAPI argonGetTokenInformation(
    IN HANDLE TokenHandle,
    IN TOKEN_INFORMATION_CLASS TokenInformationClass,
    OUT LPVOID TokenInformation,
    IN DWORD TokenInformationLength,
    OUT PDWORD ReturnLength
);

typedef BOOL (WINAPI *TfnAdjustTokenPrivileges)(
    IN HANDLE TokenHandle,
    IN BOOL DisableAllPrivileges,
    IN PTOKEN_PRIVILEGES NewState,
    IN DWORD BufferLength,
    OUT PTOKEN_PRIVILEGES PreviousState,
    OUT PDWORD ReturnLength
);

BOOL WINAPI argonAdjustTokenPrivileges(
    IN HANDLE TokenHandle,
    IN BOOL DisableAllPrivileges,
    IN PTOKEN_PRIVILEGES NewState,
    IN DWORD BufferLength,
    OUT PTOKEN_PRIVILEGES PreviousState,
    OUT PDWORD ReturnLength
);

#define OpenProcessToken      ArgonLibrary::argonOpenProcessToken
#define LookupAccountSidA     ArgonLibrary::argonLookupAccountSidA
#define LookupAccountSidW     ArgonLibrary::argonLookupAccountSidW
#define LookupAccountNameA    ArgonLibrary::argonLookupAccountNameA
#define LookupAccountNameW    ArgonLibrary::argonLookupAccountNameW
#define LookupPrivilegeValueA ArgonLibrary::argonLookupPrivilegeValueA
#define LookupPrivilegeValueW ArgonLibrary::argonLookupPrivilegeValueW
#define GetTokenInformation   ArgonLibrary::argonGetTokenInformation
#define AdjustTokenPrivileges ArgonLibrary::argonAdjustTokenPrivileges

#ifdef UNICODE
    #define argonLookupAccountSid     argonLookupAccountSidW
    #define argonLookupAccountName    argonLookupAccountNameW
    #define argonLookupPrivilegeValue argonLookupPrivilegeValueW
#else
    #define argonLookupAccountSid     argonLookupAccountSidA
    #define argonLookupAccountName    argonLookupAccountNameA
    #define argonLookupPrivilegeValue argonLookupPrivilegeValueA
#endif

#if defined (__cplusplus)
}
#endif


class Advapi32Dll
{
    protected :
        class Unloader
        {
            public :
                Unloader() {  }
                ~Unloader() { if( Advapi32Dll::s_bAutoUnload ) Advapi32Dll::UnloadAPI(); }
        };

        static Unloader s_Unloader;

    public :
        static HMODULE s_hModule;
        static TfnOpenProcessToken      s_pfnOpenProcessToken;
        static TfnLookupAccountSidA     s_pfnLookupAccountSidA;
        static TfnLookupAccountSidW     s_pfnLookupAccountSidW;
        static TfnLookupAccountNameA    s_pfnLookupAccountNameA;
        static TfnLookupAccountNameW    s_pfnLookupAccountNameW;
        static TfnLookupPrivilegeValueA s_pfnLookupPrivilegeValueA;
        static TfnLookupPrivilegeValueW s_pfnLookupPrivilegeValueW;
        static TfnGetTokenInformation   s_pfnGetTokenInformation;
        static TfnAdjustTokenPrivileges s_pfnAdjustTokenPrivileges;
        static bool s_bAutoUnload;  // true = automatically unloads the Dll when global objects are destructed.

        Advapi32Dll();

        static bool hasAPI();

        static void LoadAPI();
        static void UnloadAPI();

    protected :
        static void Nullify();
};




}

#endif
