/*
    File:               argonControlGrouping.cpp
    Description:        These utils allow grouping of controls or any other
                        rectangular based items and causing them automatically
                        to be respositioned and resizes according to a new
                        outer container size.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         22.10.2005
    last updated:       26.10.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonControlGrouping_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonControlGrouping.h"

namespace ArgonLibrary
{




bool ControlGrouping::ItemOrGroup::Init()
{
    if( m_bInitialized ) return true;

    if( m_firstSubGroup )
    {
        m_bInitialized = true;
        // Note: We set the initialized flag to prevent endless loops
        // if the user accidentially formed a hierarchical loop
        // (which is invalid, but need to be caught).

        ItemOrGroup *p = m_firstSubGroup;

        if( !p->Init() )
            { m_bInitialized = false; return false; }
        _Rect_ rectBound = p->m_rectItemInitial;

        while( p = p->m_next )
        {
            if( !p->Init() )
                { m_bInitialized = false; return false; }
            rectBound.UnionRectEmulRet( p->m_rectItemInitial ); // note: we expect the initial rects never to have negative width or height.
        }

        if( rectBound.IsRectEmpty() )
        {
            m_bInitialized = false;
            String err;
            err.Sprintf( _T("The group \"%s\" is empty."), m_strItemOrGroupName );
            m_cg->onComposeError( err );
            return false;
        }

        m_rectItemCurrent = m_rectItemInitial = rectBound;

        // Now we check if this is a strip and if so, we have to calc some values.

        if( m_firstChild )
        {
            p = m_firstChild;

            if( !p->Init() )
                { m_bInitialized = false; return false; }
            _Size_ sizExtents = p->m_rectItemInitial.Size().MaxSize(_Size_(0, 0));

            while( p = p->m_nextSibling )
            {
                if( !p->Init() )
                    { m_bInitialized = false; return false; }
                sizExtents += p->m_rectItemInitial.Size().MaxSize(_Size_(0, 0));
            }

            if( sizExtents.cx < 1 ) sizExtents.cx = 1;
            if( sizExtents.cy < 1 ) sizExtents.cy = 1;

            m_nStripExtentInitial = m_GroupType == grpt_HorzGroupStrip ?
                sizExtents.cx : sizExtents.cy;  // note: we expect the initial rects never to have negative width or height.
        }
    }
    else
    if( m_hwndAttachedControl )
    {
        m_rectItemInitial.GetControlRect( m_hwndAttachedControl );
        m_rectItemCurrent = m_rectItemInitial;

        if( m_rectItemInitial.IsRectEmpty() )
        {
            String err;
            err.Sprintf( _T("The control item \"%s\" is empty."), m_strItemOrGroupName );
            m_cg->onComposeError( err );
            return false;
        }

        m_bInitialized = true;
    }
    else
    if( m_bUserContent )
    {
        if( !m_cg->onInitialItem( m_strItemOrGroupName, m_pUserData, m_rectItemInitial ) )
            return false;

        if( m_rectItemInitial.IsRectEmpty() )
        {
            String err;
            err.Sprintf( _T("The user content item \"%s\" is empty."), m_strItemOrGroupName );
            m_cg->onComposeError( err );
            return false;
        }

        m_bInitialized = true;
    }
    else
    if( m_GroupType == grpt_Separator )
    {
        ItemOrGroup *strip = NULL, *tl = NULL, *br = NULL;
        getGroupNeighbours( strip, tl, br );

        if( !strip )
        {
            String err;
            err.Sprintf( _T("The separator item \"%s\" is not placed in a strip."), m_strItemOrGroupName );
            m_cg->onComposeError( err );
            return false;
        }

        if( !tl && !br )
        {
            String err;
            err.Sprintf( _T("The separator item \"%s\" is alone in the strip."), m_strItemOrGroupName );
            m_cg->onComposeError( err );
            return false;
        }

        // IMPORTANT:
        // In the following init calls, we would cause an endless loop
        //  (or a stack-overflow) when one of the returned neighbour
        //  items (not the parent strip) contains also a separator.
        //  So 2 separators must not be direct neighbours.
        if( tl ) if( !tl->Init() ) return false;
        if( br ) if( !br->Init() ) return false;
        // Note: We cannot init the strip, because it would need all
        // children to be inited which is not possible since one
        // of its far descendents is #this# which is currently initing.

        if( strip->m_GroupType == grpt_HorzGroupStrip )
        {
            if( tl && br ) {
                m_rectItemInitial.left   = tl->m_rectItemInitial.right;
                m_rectItemInitial.right  = br->m_rectItemInitial.left;
                m_rectItemInitial.top    = tmin( tl->m_rectItemInitial.top,    br->m_rectItemInitial.top    );
                m_rectItemInitial.bottom = tmin( tl->m_rectItemInitial.bottom, br->m_rectItemInitial.bottom );
            } else if( tl ) {
                m_rectItemInitial.left   =
                m_rectItemInitial.right  = tl->m_rectItemInitial.right;
                m_rectItemInitial.top    = tl->m_rectItemInitial.top;
                m_rectItemInitial.bottom = tl->m_rectItemInitial.bottom;
            } else if( br ) {
                m_rectItemInitial.left   =
                m_rectItemInitial.right  = br->m_rectItemInitial.left;
                m_rectItemInitial.top    = br->m_rectItemInitial.top;
                m_rectItemInitial.bottom = br->m_rectItemInitial.bottom;
            }
        }
        else
        {
            if( tl && br ) {
                m_rectItemInitial.top    = tl->m_rectItemInitial.bottom;
                m_rectItemInitial.bottom = br->m_rectItemInitial.top;
                m_rectItemInitial.left   = tmin( tl->m_rectItemInitial.left,   br->m_rectItemInitial.left   );
                m_rectItemInitial.right  = tmin( tl->m_rectItemInitial.right,  br->m_rectItemInitial.right  );
            } else if( tl ) {
                m_rectItemInitial.top    =
                m_rectItemInitial.bottom = tl->m_rectItemInitial.bottom;
                m_rectItemInitial.left   = tl->m_rectItemInitial.left;
                m_rectItemInitial.right  = tl->m_rectItemInitial.right;
            } else if( br ) {
                m_rectItemInitial.top    =
                m_rectItemInitial.bottom = br->m_rectItemInitial.top;
                m_rectItemInitial.left   = br->m_rectItemInitial.left;
                m_rectItemInitial.right  = br->m_rectItemInitial.right;
            }
        }

        m_rectItemCurrent = m_rectItemInitial;

        m_bInitialized = true;
    }
    else
    {
        String err;
        err.Sprintf( _T("The item or group \"%s\" cannot be initialized."), m_strItemOrGroupName );
        m_cg->onComposeError( err );

        return false;
    }

    // So for all successfully initialized groups,
    // we check if to set additional values.
    if( m_bInitialized )
    {
        m_fItemAdoptedInitWidth  = (double) m_rectItemInitial.Width();
        m_fItemAdoptedInitHeight = (double) m_rectItemInitial.Height();
    }

    return true;
}

///<summary>
///  <b>WalkParentMinSize()</b> calculates the minimum size for all
///  groups in the parent tree beginning with <code>this</code> parent group.
///</summary>
void ControlGrouping::ItemOrGroup::WalkParentMinSize( bool bDoWidths, bool bDoHeights )
{
    if( !bDoWidths && !bDoHeights ) return;

    _Size_ minSize(0, 0);

    // First we check for a strip.

    if( m_GroupType == grpt_HorzGroupStrip )
        minSize.cx = tmax( m_rectItemInitial.Width() - m_nStripExtentInitial, (int) minSize.cx );

    if( m_GroupType == grpt_VertGroupStrip )
        minSize.cy = tmax( m_rectItemInitial.Height() - m_nStripExtentInitial, (int) minSize.cy );

    // Now we check for subgroups not part of a strip.

    for( ItemOrGroup *p = m_firstSubGroup ; p ; p = p->m_next )
    {
        if( p->m_parentStrip == this ) continue;
        minSize.MaxSize( p->m_minSize );
    }

    // Now we set the new minimum extents for this group.
    // We overwrite them (not combining taking the maximum)
    // because it can be that the new minimum size of a subgroup
    // has been made smaller than it was before.
    if( bDoWidths  ) m_minSize.cx = m_fixedWidth  == fixedt_Fixed ? m_rectItemInitial.Cx() : minSize.cx;
    if( bDoHeights ) m_minSize.cy = m_fixedHeight == fixedt_Fixed ? m_rectItemInitial.Cy() : minSize.cy;

    // Now we walk upwards.

    if( m_parent )
        m_parent->WalkParentMinSize( bDoWidths, bDoHeights );
}

namespace privateControlGrouping
{
    struct _ChildInfo
    {
        bool    m_bFixed;
        int     m_nExtent;
    };
}

///<summary>
///  <b>getGroupNeighbours()</b> Gets the desired neighbours in the nearest
///  parent horz- or vert-strip.
///</summary>
bool ControlGrouping::ItemOrGroup::getGroupNeighbours( ItemOrGroup *& strip, ItemOrGroup *& left_or_top, ItemOrGroup *& right_or_bottom )
{
    if( !m_parentStrip )
        return m_parent ? m_parent->getGroupNeighbours( strip, left_or_top, right_or_bottom ) : false;

    strip = m_parentStrip;

    for( ItemOrGroup *p = m_parentStrip->m_firstChild ; p ; p = p->m_nextSibling )
        if( p->m_nextSibling == this )
            { left_or_top = p; break; }

    if( m_nextSibling ) right_or_bottom = m_nextSibling;

    return true;
}

///<summary>
///  <b>ArrangeItems()</b> arranges an attached control and/or
///  sub-groups. It expects <code>this</code> to have been placed
///  already (so its m_rectItemCurrent should have the correct value).
///</summary>
bool ControlGrouping::ItemOrGroup::ArrangeItems()
{
    WindowDeferrer *wd = m_cg->m_pWindowDeferrer;
    bool bWasDeferring = wd ? wd->deferring() : false;

    bool ret = iterateArrangeItems();

    // At the end, we defer window positions, but only if
    // we had to call the deferrer's Begin() somewhen in the iteration here.

    bool bStillDeferring = wd ? wd->deferring() : false;

    if( !bWasDeferring && bStillDeferring )
        if( !wd->End() )
            ret = false;

    return ret;
}

bool ControlGrouping::ItemOrGroup::iterateArrangeItems()
{
    // Position control and user item.

    if( m_hwndAttachedControl )
    {
        // set new pos and size
        _Rect_ rectOld;
        rectOld.GetControlRect( m_hwndAttachedControl );
        m_cg->SetWindowRect( m_hwndAttachedControl, m_rectItemCurrent );

        if( m_rectItemCurrent != rectOld )
        {
            // now we invalidate the control if desired.
            bool bInvalidate =
                m_bAutoInvalidateAttachedControl || (m_cg->m_dwProcFlags & ControlGrouping::procf_RedrawAllControls);

            if( !bInvalidate &&
                (m_cg->m_dwProcFlags & ControlGrouping::procf_RedrawStaticControls) )
            {
                TCHAR cls[64];
                // #::GetClassName()# will *always* append the null-terminator
                // even if the string buffer is too small.
                if( ::GetClassName( m_hwndAttachedControl, cls, _countof(cls) ) == 0 ) cls[0] = (TCHAR) 0;
                bool bStaticCtrl = _striis_( cls, _T("Static") );
                if( bStaticCtrl ) bInvalidate = true;
            }

            if( bInvalidate )
                InvalidateRect( m_hwndAttachedControl, NULL, TRUE );
        }

        m_cg->onArrangedControl( m_strItemOrGroupName, m_hwndAttachedControl, m_pUserData, m_rectItemInitial, m_rectItemCurrent );
    }

    if( m_bUserContent )
        m_cg->onArrangeUserItem( m_strItemOrGroupName, m_pUserData, m_rectItemInitial, m_rectItemCurrent );

    if( m_hwndInform && m_messageidInform != 0 )
    {
        GroupRearranged gr;
        memset( &gr, 0, sizeof(gr) );
        gr.m_cg                     = m_cg;
        _strcpy_( gr.m_strItemOrGroupName, m_strItemOrGroupName, _countof(gr.m_strItemOrGroupName) );
        gr.m_GroupType              = m_GroupType;
        gr.m_hwndAttachedControl    = m_hwndAttachedControl;
        gr.m_bUserContent           = m_bUserContent;
        gr.m_pUserData              = m_pUserData;
        gr.m_rectItemInitial        = m_rectItemInitial;
        gr.m_rectItemCurrent        = m_rectItemCurrent;

        SendMessage( m_hwndInform, m_messageidInform, 0, (LPARAM) &gr );
    }

    // First we treat each subgroup as not being part of a strip.

    ItemOrGroup *p;
    for( p = m_firstSubGroup ; p ; p = p->m_next )
    {
        _Rect_ rectScaled = m_rectItemCurrent;

        rectScaled.left   += p->m_rectItemInitial.left   - m_rectItemInitial.left;
        rectScaled.top    += p->m_rectItemInitial.top    - m_rectItemInitial.top;
        rectScaled.right  += p->m_rectItemInitial.right  - m_rectItemInitial.right;
        rectScaled.bottom += p->m_rectItemInitial.bottom - m_rectItemInitial.bottom;

        _Rect_ rect = rectScaled;
        rect.LimitSizeMin( p->m_minSize );
        if( p->m_fixedWidth  != fixedt_Dynamic ) rect.SetWidth ( p->m_rectItemInitial.Width () );
        if( p->m_fixedHeight != fixedt_Dynamic ) rect.SetHeight( p->m_rectItemInitial.Height() );

        DWORD mrtf = p->m_align + (DWORD) alnstartbit_Horz + (DWORD) alnstartbit_Vert;
        rect.MoveRectToFrame( rectScaled, mrtf );

        p->m_rectItemCurrent = rect;
        p->m_rectItemCurrent &= m_rectItemCurrent;
    }

    // Now we handle subgroups of a strip.
    // There can be only one strip.

    if( m_GroupType == grpt_HorzGroupStrip && m_firstChild && m_cStripSubGroups > 0 )
    {
        Array<privateControlGrouping::_ChildInfo> aChildsInfo( m_cStripSubGroups );
        privateControlGrouping::_ChildInfo *pChildInfo = aChildsInfo.Ptr();

        int resizeOffset = m_rectItemCurrent.Width() - m_rectItemInitial.Width();

        double fStripExtentInitial = m_nStripExtentInitial >= 1 ? (double) m_nStripExtentInitial : 1.0;
        int    nStripExtentCurrent = (int) fStripExtentInitial + resizeOffset;

        // First we get the fixed extent child groups.
        privateControlGrouping::_ChildInfo *pChildInfo_LastScalable = NULL;
        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; p = p->m_nextSibling, ++pChildInfo )
        {
            if( pChildInfo->m_bFixed = p->m_fixedWidth != fixedt_Dynamic )
            {
                pChildInfo->m_nExtent = (int) type<double>::round( p->m_fItemAdoptedInitWidth );
                fStripExtentInitial -= pChildInfo->m_nExtent;
                nStripExtentCurrent -= pChildInfo->m_nExtent;
                if( fStripExtentInitial <= 0.0 ) fStripExtentInitial = 1.0;
            }
            else
            {
                pChildInfo_LastScalable = pChildInfo;
            }
        }

        double fRestExtentInitial = fStripExtentInitial;
        int    nRestExtentCurrent = nStripExtentCurrent;

        // Then we get the child groups where their minimum extent will take effect.
        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; )
        {
            if( pChildInfo->m_bFixed )
                { p = p->m_nextSibling; ++pChildInfo; continue; }

            int ext = pChildInfo == pChildInfo_LastScalable || fRestExtentInitial == 0.0 ?
                nRestExtentCurrent :
                (int) type<double>::round( (p->m_fItemAdoptedInitWidth * nRestExtentCurrent) / fRestExtentInitial );

            if( ext < p->m_minSize.cx )
            {
                pChildInfo->m_bFixed = true;
                pChildInfo->m_nExtent = p->m_minSize.cx;
                fStripExtentInitial -= p->m_fItemAdoptedInitWidth;
                nStripExtentCurrent -= pChildInfo->m_nExtent;
                if( fStripExtentInitial <= 0.0 ) fStripExtentInitial = 1.0;

                // Since we have limited a child to its minimum extent,
                // the other siblings will get smaller, so we have
                // to restart from the beginning.
                fRestExtentInitial = fStripExtentInitial;
                nRestExtentCurrent = nStripExtentCurrent;
                p = m_firstChild;
                pChildInfo = aChildsInfo.Ptr();
                continue;
            }

            pChildInfo->m_nExtent = ext;

            fRestExtentInitial -= p->m_fItemAdoptedInitWidth;
            nRestExtentCurrent -= ext;
            if( fRestExtentInitial <= 0.0 ) fRestExtentInitial = 1.0;
            p = p->m_nextSibling;
            ++pChildInfo;
        }

        _Range_ rng;
        rng.m_start = m_rectItemCurrent.left + (m_firstChild->m_rectItemInitial.left - m_rectItemInitial.left);

        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; p = p->m_nextSibling, ++pChildInfo )
        {
            rng.m_length = pChildInfo->m_nExtent;

            p->m_rectItemCurrent.SetHorzRange( rng );
            p->m_rectItemCurrent &= m_rectItemCurrent;

            rng.m_start = rng.GetEnd();
            if( p->m_nextSibling ) rng.m_start += p->m_nextSibling->m_rectItemInitial.left - p->m_rectItemInitial.right;
        }
    }

    if( m_GroupType == grpt_VertGroupStrip && m_firstChild && m_cStripSubGroups > 0 )
    {
        Array<privateControlGrouping::_ChildInfo> aChildsInfo( m_cStripSubGroups );
        privateControlGrouping::_ChildInfo *pChildInfo = aChildsInfo.Ptr();

        int resizeOffset = m_rectItemCurrent.Height() - m_rectItemInitial.Height();

        double fStripExtentInitial = m_nStripExtentInitial >= 1 ? (double) m_nStripExtentInitial : 1.0;
        int    nStripExtentCurrent = (int) fStripExtentInitial + resizeOffset;

        // First we get the fixed extent child groups.
        privateControlGrouping::_ChildInfo *pChildInfo_LastScalable = NULL;
        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; p = p->m_nextSibling, ++pChildInfo )
        {
            if( pChildInfo->m_bFixed = p->m_fixedHeight != fixedt_Dynamic )
            {
                pChildInfo->m_nExtent = (int) type<double>::round( p->m_fItemAdoptedInitHeight );
                fStripExtentInitial -= pChildInfo->m_nExtent;
                nStripExtentCurrent -= pChildInfo->m_nExtent;
                if( fStripExtentInitial <= 0.0 ) fStripExtentInitial = 1.0;
            }
            else
            {
                pChildInfo_LastScalable = pChildInfo;
            }
        }

        double fRestExtentInitial = fStripExtentInitial;
        int    nRestExtentCurrent = nStripExtentCurrent;

        // Then we get the child groups where their minimum extent will take effect.
        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; )
        {
            if( pChildInfo->m_bFixed )
                { p = p->m_nextSibling; ++pChildInfo; continue; }

            int ext = pChildInfo == pChildInfo_LastScalable || fRestExtentInitial == 0.0 ?
                nRestExtentCurrent :
                (int) type<double>::round( (p->m_fItemAdoptedInitHeight * nRestExtentCurrent) / fRestExtentInitial );

            if( ext < p->m_minSize.cy )
            {
                pChildInfo->m_bFixed = true;
                pChildInfo->m_nExtent = p->m_minSize.cy;
                fStripExtentInitial -= p->m_fItemAdoptedInitHeight;
                nStripExtentCurrent -= pChildInfo->m_nExtent;
                if( fStripExtentInitial <= 0.0 ) fStripExtentInitial = 1.0;

                // Since we have limited a child to its minimum extent,
                // the other siblings will get smaller, so we have
                // to restart from the beginning.
                fRestExtentInitial = fStripExtentInitial;
                nRestExtentCurrent = nStripExtentCurrent;
                p = m_firstChild;
                pChildInfo = aChildsInfo.Ptr();
                continue;
            }

            pChildInfo->m_nExtent = ext;

            fRestExtentInitial -= p->m_fItemAdoptedInitHeight;
            nRestExtentCurrent -= ext;
            if( fRestExtentInitial <= 0.0 ) fRestExtentInitial = 1.0;
            p = p->m_nextSibling;
            ++pChildInfo;
        }

        _Range_ rng;
        rng.m_start = m_rectItemCurrent.top + (m_firstChild->m_rectItemInitial.top - m_rectItemInitial.top);

        for( p = m_firstChild, pChildInfo = aChildsInfo.Ptr() ; p ; p = p->m_nextSibling, ++pChildInfo )
        {
            rng.m_length = pChildInfo->m_nExtent;

            p->m_rectItemCurrent.SetVertRange( rng );
            p->m_rectItemCurrent &= m_rectItemCurrent;

            rng.m_start = rng.GetEnd();
            if( p->m_nextSibling ) rng.m_start += p->m_nextSibling->m_rectItemInitial.top - p->m_rectItemInitial.bottom;
        }
    }

    // Now go down the tree

    for( p = m_firstSubGroup ; p ; p = p->m_next )
        if( !p->iterateArrangeItems() ) return false;

    return true;
}




bool ControlGrouping::Store_Header::CheckHeaderIntegrity()
{
    Store_Header UNALIGNED *I = (Store_Header UNALIGNED*) this;

    if( I->m_4cc != StoreHeaderFourcc ) return false;

    if( I->m_wVersion != 1 || I->m_wRevision != 0 ) return false;

    if( I->m_GroupType != grpt_HorzGroupStrip && I->m_GroupType != grpt_VertGroupStrip ) return false;
    if( I->m_cStripSubGroups < 1 ) return false;

    _size_t_ calcedsize = sizeof(Store_Header) + (I->m_cStripSubGroups * sizeof(Store_Group));
    if( calcedsize != (_size_t_) I->m_dwContentSize ) return false;

    return true;
}




bool ControlGrouping::Store::SetHeader( PCTSTR pstrStripName, eGroupType grpt, int nStripExtentInitial )
{
    if( !pstrStripName || !(grpt == grpt_HorzGroupStrip || grpt == grpt_VertGroupStrip) )
        return false;

    bool bHadHeader = GetHeader() != NULL;

    if( !bHadHeader )
        m_aData.SetSize( sizeof(Store_Header) );

    Store_Header UNALIGNED *hdr = GetHeader();
    if( !hdr ) return false;

    if( !bHadHeader )
    {
        memset( hdr, 0, sizeof(Store_Header) );
        hdr->m_4cc              = StoreHeaderFourcc;
        hdr->m_wVersion         = 1;
        hdr->m_wRevision        = 0;
        hdr->m_dwContentSize    = sizeof(Store_Header);
    }

    StringW wstr = pstrStripName;
    memset            ( hdr->m_strStripName, 0,    sizeof  (hdr->m_strStripName) );
    _strcpy_unaligned_( hdr->m_strStripName, wstr, _countof(hdr->m_strStripName) );
    hdr->m_GroupType            = grpt;
    hdr->m_nStripExtentInitial  = nStripExtentInitial;

    return true;
}

bool ControlGrouping::Store::AddGroup( eFixedType fixedExtent, double adoptedInitExtent )
{
    Store_Header UNALIGNED *hdr = GetHeader();
    if( !hdr ) return false;

    DWORD bytesize = (DWORD) sizeof(Store_Header) + ((hdr->m_cStripSubGroups + 1) * (DWORD) sizeof(Store_Group));
    bool bHadGroup = m_aData.GetSize() >= bytesize;
    if( bHadGroup ) return false;

    m_aData.SetSize( bytesize );

    hdr = GetHeader();
    if( !hdr ) { m_aData.Empty(); return false; }

    Store_Group UNALIGNED *grp = (Store_Group UNALIGNED*) ( ((PBYTE) hdr) + (sizeof(Store_Header) + hdr->m_cStripSubGroups * sizeof(Store_Group)) );
    grp->m_dwFixedExtent        = fixedExtent;
    grp->m_fAdoptedInitExtent   = adoptedInitExtent;

    hdr->m_cStripSubGroups++;
    hdr->m_dwContentSize = (DWORD) m_aData.GetSize();

    return true;
}



ControlGrouping::Store_Header UNALIGNED *ControlGrouping::Store::GetHeader()
{
    if( m_aData.GetSize() < sizeof(Store_Header) )
        return NULL;

    return (Store_Header UNALIGNED*) m_aData.Ptr();
}

ControlGrouping::Store_Group UNALIGNED *ControlGrouping::Store::GetGroupAt( _idx_ idx )
{
    if( idx < 0 || m_aData.GetSize() < sizeof(Store_Header) + (idx + 1) * sizeof(Store_Group) )
        return NULL;

    Store_Header UNALIGNED *hdr = GetHeader();
    if( !hdr ) return NULL;

    if( idx >= (_idx_) hdr->m_cStripSubGroups ) return NULL;

    return ((Store_Group UNALIGNED*) (hdr + 1)) + idx;
}



///<returns> True if contents are valid and correct, false otherwise. </returns>
bool ControlGrouping::Store::CheckIntegrity()
{
    Store_Header UNALIGNED *hdr = GetHeader();
    if( !hdr ) return false;

    if( hdr->m_4cc != StoreHeaderFourcc )
        return false;

    if( !(hdr->m_GroupType == grpt_HorzGroupStrip || hdr->m_GroupType == grpt_VertGroupStrip) )
        return false;

    if( (_size_t_) hdr->m_dwContentSize != m_aData.GetSize() ) return false;

    _size_t_ calcedsize = sizeof(Store_Header) + (hdr->m_cStripSubGroups * sizeof(Store_Group));
    if( calcedsize != (_size_t_) hdr->m_dwContentSize ) return false;

    // Now we check the data integrity also by the sum of the extents.
    double sum = 0.0;
    for( _idx_ i = 0 ; i < (_idx_) hdr->m_cStripSubGroups ; ++i )
    {
        Store_Group UNALIGNED *grp = ((Store_Group UNALIGNED*) (hdr + 1)) + i;
        sum += grp->m_fAdoptedInitExtent;
    }

    double epsilon = 1e-10;
    if( (double) hdr->m_nStripExtentInitial < sum - epsilon ) return false;
    if( (double) hdr->m_nStripExtentInitial > sum + epsilon ) return false;

    return true;
}



bool ControlGrouping::Store::GetArrayData( PVOID& ptr, UINT& bytesize )
{
    Store_Header UNALIGNED *hdr = GetHeader();
    if( !hdr ) return false;

    PVOID _ptr      = (PVOID) m_aData.Ptr();
    UINT  _bytesize = (UINT)  m_aData.GetSize();

    if( !_ptr || _bytesize < sizeof(Store_Header) ) return false;

    ptr      = _ptr;
    bytesize = _bytesize;

    return true;
}

bool ControlGrouping::Store::SetArrayData( PVOID ptr, UINT bytesize )
{
    if( !ptr || bytesize < sizeof(Store_Header) ) return false;

    Store_Header UNALIGNED *hdr = (Store_Header UNALIGNED*) ptr;
    if( !hdr ) return false;

    if( hdr->m_4cc != StoreHeaderFourcc )
        return false;

    DWORD contentsize = GetContentSize( ptr );
    if( contentsize < sizeof(Store_Header) ) return false;
    if( (UINT) contentsize > bytesize ) return false;

    DWORD cSubGroups = (contentsize - (DWORD) sizeof(Store_Header)) / (DWORD) sizeof(Store_Group);
    if( hdr->m_cStripSubGroups != cSubGroups ) return false;

    UINT calcedsize = (UINT) sizeof(Store_Header) + (hdr->m_cStripSubGroups * (UINT) sizeof(Store_Group));
    if( calcedsize != contentsize ) return false;

    // Now we check the data integrity also by the sum of the extents.
    double sum = 0.0;
    for( _idx_ i = 0 ; i < (_idx_) cSubGroups ; ++i )
    {
        Store_Group UNALIGNED *grp = ((Store_Group UNALIGNED*) (hdr + 1)) + i;
        sum += grp->m_fAdoptedInitExtent;
    }

    double epsilon = 1e-10;
    if( (double) hdr->m_nStripExtentInitial < sum - epsilon ) return false;
    if( (double) hdr->m_nStripExtentInitial > sum + epsilon ) return false;

    m_aData.SetData( (PBYTE) ptr, contentsize );

    return m_aData.GetSize() == (_size_t_) contentsize;
}




ControlGrouping::ControlGrouping()
                :m_htItemsAndGroups( 17 )
{
    Ctor();
}

ControlGrouping::ControlGrouping( DWORD dwProcFlags )
                :m_htItemsAndGroups( 17 )
{
    Ctor();
    m_dwProcFlags = dwProcFlags;
}

void ControlGrouping::Ctor()
{
    m_bComposing = false;
    m_dwProcFlags = procf__Default;
    m_pWindowDeferrer = NULL;
    m_bAllowDragMoveFixedGroups = true;
    m_hcrsrHorzDrag = LoadCursor( NULL, IDC_SIZEWE );
    m_hcrsrVertDrag = LoadCursor( NULL, IDC_SIZENS );
    m_bDragging = false;
    m_hgrpBorderBeingDragged = NULL;
    m_ptMouseOnBeginDrag.SetEmpty();
    m_rectBorderBeginDrag.SetRectEmpty();
    m_rectDraggingInArea.SetRectEmpty();
}



void ControlGrouping::ComposeStart()
{
    m_bComposing = true;
    m_htItemsAndGroups.Empty();
    m_aTopOrLeftDragableBorders.Empty();
    m_strComposeErrors.Empty();
}

///<summary>
///  <b>ComposeEnd()</b> will do the final calculations like
///  initializing the added items and groups, retrieving and
///  calculating its initial rectangles and resolving pointers.
///  <br/>
///  Because of this, it is not necessary in which order to
///  add controls, items and groups during composing mode.
///  This call will automatically recognize the hierarchie.
///  But it will fail if groups or items (especially parent
///  groups) are missing.
///</summary>
bool ControlGrouping::ComposeEnd()
{
    const String *key;
    ItemOrGroup *grp;

    Iterator iter;
    while( m_htItemsAndGroups.GetNextPtr(iter, key, grp) )
        if( !grp->Init() )
            { m_bComposing = false; return false; }

    m_bComposing = false;
    return true;
}

///<summary>
///  <b>AddRootGroup()</b> adds a group that acts as a root group.
///  Normally you would give it an empty name, but you can
///  choose any name you like.
///</summary>
///<param name="rectInitial"> the initial rectangle </param>
///<param name="pstrGroupName"> name of the new group.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddRootGroup( _Rect_ rectInitial, PCTSTR pstrGroupName )
{
    if( !pstrGroupName ) pstrGroupName = _T("");

    ItemOrGroup *ptr = _AddGroup( NULL, pstrGroupName, grpt_Single, false );
    if( !ptr ) return false;

    ptr->m_rectItemCurrent = ptr->m_rectItemInitial = rectInitial;
    ptr->m_bInitialized = true;

    return true;
}

///<summary>
///  <b>AddRootGroup()</b> adds a group that acts as a root group.
///  Normally you would give it an empty name, but you can
///  choose any name you like.
///</summary>
///<param name="hwndContainer"> the dialog or any other window that
///  contains the controls to be rearranged. Its client area
///  will be used as initial rectangle. </param>
///<param name="pstrGroupName"> name of the new group.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddRootGroup( HWND hwndContainer, PCTSTR pstrGroupName )
{
    if( !hwndContainer ) return false;
    return AddRootGroup( _Rect_().GetClientRect(hwndContainer), pstrGroupName );
}

///<summary>
///  <b>AddGroup()</b> adds a group.
///  If an item or group with that name did already exist, it will be overwritten.
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the group to add does
///  not exist yet, the new group will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the group to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrGroupName"> name of the new group.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddGroup( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName )
{
    ItemOrGroup *ptr = _AddGroup( pstrParentGroupName, pstrGroupName, grpt_Single, false );
    return ptr != NULL;
}

///<summary>
///  <b>AddGrid()</b> adds a new bounding group with the
///  specified group name. And below it, it will create the specified
///  number of m by n sub-groups arranged. <br/><br/>
///
///  Their names will be exactly the same as the specified bounding
///  group name, but appended by a - character followed by an
///  integer decimal number (without preceding zeros) that
///  indicates its number. This number is one-based and therefore
///  is its index + 1.
///  The order of these numbers will go from left to right and
///  continues in the following row, independent of the value of
///  <b>bRowsBeforeColumns</b>. <br/><br/>
///
///  Internally, this call will create vertical and horizontal group strips.
///  A strip is a strip of sub-groups.
///  <br/>
///  Exactly: If <b>bRowsBeforeColumns</b> is true, then the bounding group
///  first will be partitioned into a vertical strip (so each sub-group
///  is a row) and then each sub-group of the strip will be partitioned
///  into a horizontal strip. <br/>
///  In that case, the names of the sub-groups of the vertical strip
///  will be the same as for the bounding group, except that it will
///  be appended by "-row" followed by a one-based number.
///  <br/>
///  But if <b>bRowsBeforeColumns</b> is false, then the bounding group
///  first will be partitioned into a horizontal strip (so each sub-group
///  is a column) and then each sub-group of the strip will be partitioned
///  into a vertical strip. <br/>
///  In that case, the names of the sub-groups of the vertical strip
///  will be the same as for the bounding group, except that it will
///  be appended by "-column" followed by a one-based number. <br/><br/>
///
///  Sample: <br/>
///  If you create a grid of 3 rows by 2 columns with <b>bRowsBeforeColumns</b>
///  set to false, then this method will create the following:
///  (in each group, its name is written)
///  <pre>
///
///      Bounding Group:
///
///      +-----------------------------------------------------+
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      |                     MyGridGroup                     |
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      |                                                     |
///      +-----------------------------------------------------+
///
///
///      First rows are generated as vertical strip:
///
///      +-----------------------------------------------------+
///      |                                                     |
///      |                  MyGridGroup-row1                   |
///      |                                                     |
///      +-----------------------------------------------------+
///      |                                                     |
///      |                  MyGridGroup-row2                   |
///      |                                                     |
///      +-----------------------------------------------------+
///      |                                                     |
///      |                  MyGridGroup-row3                   |
///      |                                                     |
///      +-----------------------------------------------------+
///
///
///      Then, in each of the rows, a horizontal strip is generated
///      to create the columns for each row:
///
///      IMPORTANT: The numbers of the final 6 groups will always
///      be arranged, the it goes from left to right, and then
///      continues in the following row, *independet* of the value of
///      bRowsBeforeColumns !!!
///
///      +--------------------------+--------------------------+
///      |                          |                          |
///      |      MyGridGroup-1       |      MyGridGroup-2       |
///      |                          |                          |
///      +--------------------------+--------------------------+
///      |                          |                          |
///      |      MyGridGroup-3       |      MyGridGroup-4       |
///      |                          |                          |
///      +--------------------------+--------------------------+
///      |                          |                          |
///      |      MyGridGroup-5       |      MyGridGroup-6       |
///      |                          |                          |
///      +--------------------------+--------------------------+
///
///  </pre>
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the group to add does
///  not exist yet, the new group will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the group to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrGroupName"> name of the new bounding group.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<param name="sizGrid"> m by n final sub-groups in the grid to create.
///  See descriptions in the summary. </param>
///<param name="bRowsBeforeColumns"> if to partition the bounding group
///  into rows and then each row into columns (true) or vice versa (false). </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddGrid( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName, _Size_ sizGrid, bool bRowsBeforeColumns )
{
    if( sizGrid.IsEmpty() ) return false;

    eGroupType grptOuterStrip = bRowsBeforeColumns ? grpt_VertGroupStrip : grpt_HorzGroupStrip;
    eGroupType grptInnerStrip = bRowsBeforeColumns ? grpt_HorzGroupStrip : grpt_VertGroupStrip;

    ItemOrGroup *ptr = _AddGroup( pstrParentGroupName, pstrGroupName, grptOuterStrip );
    if( !ptr ) return false;

    PCTSTR pstrOuterFormat = bRowsBeforeColumns ? _T("%s-row%d") : _T("%s-column%d");
    PCTSTR pstrInnerFormat = _T("%s-%d");

    int cOuterGroups = bRowsBeforeColumns ? (int) sizGrid.cy : (int) sizGrid.cx;
    int cInnerGroups = bRowsBeforeColumns ? (int) sizGrid.cx : (int) sizGrid.cy;

    int nOuterGroupNrInc = bRowsBeforeColumns ? cInnerGroups : 1;
    int nInnerGroupNrInc = bRowsBeforeColumns ? 1 : cOuterGroups;

    ptr->m_firstChild = NULL;
    ptr->m_cStripSubGroups = cOuterGroups;

    // Now we create the sub-groups.
    ItemOrGroup *pouterPrev = NULL;
    String inner, outer;
    int nrCurrOuterGroup = 1;
    for( int iouter = 0 ; iouter < cOuterGroups ; ++iouter )
    {
        // First we create an outer sub-group, that's a sub-group
        // of the outer strips.
        outer.Sprintf( pstrOuterFormat, pstrGroupName, iouter + 1 );
        ItemOrGroup *pouter = _AddGroup( pstrGroupName, outer, grptInnerStrip );
        if( !pouter ) return false;
        if( pouter->m_parent != ptr ) return false; // just for savety

        pouter->m_parentStrip = ptr;
        pouter->m_nextSibling = NULL;
        pouter->m_firstChild = NULL;
        if( pouterPrev ) pouterPrev->m_nextSibling = pouter;
        if( !pouter->m_parentStrip->m_firstChild ) pouter->m_parentStrip->m_firstChild = pouter;
        pouter->m_cStripSubGroups = cInnerGroups;

        // Now we create the strip's new sub-groups in where the parent strip is the above sub-group.
        ItemOrGroup *pinnerPrev = NULL;
        int nrCurrInnerGroup = nrCurrOuterGroup;
        for( int iinner = 0 ; iinner < cInnerGroups ; ++iinner )
        {
            inner.Sprintf( pstrInnerFormat, pstrGroupName, nrCurrInnerGroup );
            ItemOrGroup *pinner = _AddGroup( outer, inner, grpt_Single );
            if( !pinner ) return false;
            if( pinner->m_parent != pouter ) return false; // just for savety

            pinner->m_parentStrip = pouter;
            pinner->m_nextSibling = NULL;
            if( pinnerPrev ) pinnerPrev->m_nextSibling = pinner;
            if( !pinner->m_parentStrip->m_firstChild ) pinner->m_parentStrip->m_firstChild = pinner;

            nrCurrInnerGroup += nInnerGroupNrInc;
            pinnerPrev = pinner;
        }

        nrCurrOuterGroup += nOuterGroupNrInc;
        pouterPrev = pouter;
    }

    return true;
}

///<summary>
///  <b>AddControl()</b> adds an existing HWND as a group item.
///  If an item or group with that name did already exist, it will be overwritten.
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the item to add does
///  not exist yet, the new item will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the item to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrItemName"> name of the new item.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<param name="hwndCtrl"> existing HWND to attach </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddControl( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, HWND hwndCtrl )
{
    if( !hwndCtrl ) return false;

    ItemOrGroup *ptr = _AddGroup( pstrParentGroupName, pstrItemName, grpt_Single, false );
    if( !ptr ) return false;

    ptr->m_hwndAttachedControl = hwndCtrl;

    return true;
}

///<summary>
///  <b>AddControlID()</b> adds a control by its ID as a group item.
///  If an item or group with that name did already exist, it will be overwritten.
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the item to add does
///  not exist yet, the new item will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the item to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrItemName"> name of the new item.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<param name="hwndParent"> existing parent HWND where the control is in </param>
///<param name="idCtrl"> ID of the control (which does not necessarily need to exist yet) </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddControlID( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, HWND hwndParent, UINT idCtrl )
{
    if( !hwndParent || idCtrl == 0 ) return false;
    return AddControl( pstrParentGroupName, pstrItemName, GetDlgItem(hwndParent, idCtrl) );
}

///<summary>
///  <b>AddContentItem()</b> adds an item that contains user specific
///  content. The content can be set also afterwards using the
///  <code>SetUserData()</code> call.
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the item to add does
///  not exist yet, the new item will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the item to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrItemName"> name of the new item.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<returns> True on success, false on error. </returns>
///<remarks>
///  To handle user content items, you will need to override <code>onInitialItem()</code>
///  and <code>onArrangeItem()</code>.
///</remarks>
bool ControlGrouping::AddContentItem( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, void *pUserData )
{
    ItemOrGroup *ptr = _AddGroup( pstrParentGroupName, pstrItemName, grpt_Single, false );
    if( !ptr ) return false;

    ptr->m_bUserContent = true;
    ptr->m_pUserData = pUserData;

    return true;
}

///<summary>
///  <b>AddSeparatorItem()</b> adds an item like a control but without
///  a control attached. The item can be used as separator in a grid.
///  The separator is an item without any content, but the grid can stretch
///  and squeeze it. With that, you can realize stretchable distances
///  between grid's groups with contents. <br/>
///  CAUTION: Do *never* position 2 separators as neighbours. Between 2
///  separator groups, there must be at least one item group with content
///  (a control or user content) or an endless loop or even stack overflow
///  could occur.
///</summary>
///<param name="pstrParentGroupName"> name of the parent group. <br/>
///  If this is NULL (or an empty string) and the item to add does
///  not exist yet, the new item will have an empty string as
///  parent name. Usually, you will want to use an empty name
///  as root group name. <br/>
///  If this is NULL and the item to add did already exist,
///  its old parent group name will not be changed. </param>
///<param name="pstrItemName"> name of the new item.
///  This name needs to be unique of all item and group names within
///  <code>this</code> object. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::AddSeparatorItem( PCTSTR pstrParentGroupName, PCTSTR pstrItemName )
{
    ItemOrGroup *ptr = _AddGroup( pstrParentGroupName, pstrItemName, grpt_Separator, false );
    return ptr != NULL;
}

///<param name="pstrParentGroupName"> If NULL, and if the group to create does
///  exist already, then its parent will cannot be changed, otherwise an error (NULL) is returned. </param>
///<param name="bOverwriteGroupType"> If true, the group type and user data will be set,
///  otherwise on false, they will not be set. <br/>
///  Except: If this param is false, but no group type has been set yet,
///  the new group type will be set nevertheless. </param>
ControlGrouping::ItemOrGroup *ControlGrouping::_AddGroup( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName, eGroupType grpType, bool bOverwriteGroupType )
{
    if( !m_bComposing || !pstrGroupName ) return NULL;

    ItemOrGroup *pparent = NULL;
    if( pstrParentGroupName )
    {
        pparent = _AddGroup( NULL, pstrParentGroupName, grpt_Single, false );
        if( !pparent ) return NULL;

        _strcpy_( pparent->m_strItemOrGroupName, pstrParentGroupName, _countof(pparent->m_strItemOrGroupName) );
    }

    ItemOrGroup *ptr = &m_htItemsAndGroups[pstrGroupName];
    if( !ptr ) return NULL;

    ptr->m_cg = this;

    bool group_did_exist = ptr->m_GroupType != grpt_none;

    // Now we return a failure if we want to set a different parent.
    // The new group #ptr# must be independent, so no parent (except the desired one),
    // no siblings.
    // But sub-groups it is allowed to have already.
    if( pparent && group_did_exist )
        if( ptr->m_parent != pparent || ptr->m_next )
        {
            String err;
            err.Sprintf( _T("The item or group \"%s\" to be added does already exist\nand it cannot be placed below the new parent group \"%s\"."),
                         ptr->m_strItemOrGroupName, pparent->m_strItemOrGroupName );
            onComposeError( err );
            return NULL;
        }

    _strcpy_( ptr->m_strItemOrGroupName, pstrGroupName, _countof(ptr->m_strItemOrGroupName) );

    // Now we add the new group to the beginning of the parent's sub-group-list.
    if( pparent )
    {
        ptr->m_parent = pparent;
        ptr->m_next = pparent->m_firstSubGroup;
        pparent->m_firstSubGroup = ptr;
    }

    if( bOverwriteGroupType || ptr->m_GroupType == grpt_none )
        ptr->m_GroupType = grpType;

    return ptr;
}



bool ControlGrouping::SetFixedSize( PCTSTR pstrItemOrGroupName, eFixedType fixedWidth, eFixedType fixedHeight )
{
    if( m_bComposing || !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    ptr->m_fixedWidth  = fixedWidth;
    ptr->m_fixedHeight = fixedHeight;

    return true;
}

bool ControlGrouping::SetAlign( PCTSTR pstrItemOrGroupName, DWORD align )
{
    if( m_bComposing || !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    ptr->m_align = align;

    return true;
}

///<summary>
///  <b>SetMinWidth()</b> sets the minimum width for the specified item or group.
///  This call can either combine the specified minimum width with the existing one,
///  or overwrite the old one.
///  In case of combination, the new minimum width will be the greater value of
///  the 2 values the specified minimum width and the existing one.
///</summary>
///<param name="pstrItemOrGroupName"> name of the item or group </param>
///<param name="minWidth"> minimum width value </param>
///<param name="bAndInitialWidth"> if this is true, then the minimum width
///  will be either the specified <b>minWidth</b> or the initial width
///  (automatically calculated), whichever is the greater value. </param>
///<param name="bCombineExisting"> If true, the specified and the existing
///  minimum width values will be combined. If false, the old one will be
///  overwritten. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::SetMinWidth( PCTSTR pstrItemOrGroupName, int minWidth, bool bAndInitialWidth, bool bCombineExisting )
{
    if( m_bComposing || !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    long new_min = bAndInitialWidth ?
        tmax( minWidth, ptr->m_rectItemInitial.Width() ) : minWidth;

    ptr->m_minSize.cx = bCombineExisting ?
        tmax( new_min, ptr->m_minSize.cx ) : new_min;

    // Walk the parent tree upwards.
    if( ptr->m_parent )
        ptr->m_parent->WalkParentMinSize( true, false );

    return true;
}

///<summary>
///  <b>SetMinHeight()</b> sets the minimum height for the specified item or group.
///  This call can either combine the specified minimum height with the existing one,
///  or overwrite the old one.
///  In case of combination, the new minimum height will be the greater value of
///  the 2 values the specified minimum height and the existing one.
///</summary>
///<param name="pstrItemOrGroupName"> name of the item or group </param>
///<param name="minHeight"> minim height value </param>
///<param name="bAndInitialHeight"> if this is true, then the minimum height
///  will be either the specified <b>minHeight</b> or the initial height
///  (automatically calculated), whichever is the greater value. </param>
///<param name="bCombineExisting"> If true, the specified and the existing
///  minimum height values will be combined. If false, the old one will be
///  overwritten. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::SetMinHeight( PCTSTR pstrItemOrGroupName, int minHeight, bool bAndInitialHeight, bool bCombineExisting )
{
    if( m_bComposing || !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    long new_min = bAndInitialHeight ?
        tmax( minHeight, ptr->m_rectItemInitial.Height() ) : minHeight;

    ptr->m_minSize.cy = bCombineExisting ?
        tmax( new_min, ptr->m_minSize.cy ) : new_min;

    // Walk the parent tree upwards.
    if( ptr->m_parent )
        ptr->m_parent->WalkParentMinSize( false, true );

    return true;
}

///<summary>
///  <b>GetMinSize()</b> gets the minimum width and height for the specified item or group.
///</summary>
///<param name="pstrItemOrGroupName"> name of the item or group </param>
///<param name="minSize"> on return, will contain the minimum size </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::GetMinSize( PCTSTR pstrItemOrGroupName, _Size_& minSize )
{
    if( m_bComposing || !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    minSize = ptr->m_minSize;

    return true;
}

///<summary>
///  <b>SetUserData()</b> sets or removes user data from an item or group. <br/>
///  Note: This method can be called during composing as well as afterwards.
///</summary>
///<param name="pstrItemOrGroupName"> name of the item or group </param>
///<param name="pUserData"> the new user data or NULL if to remove old user data pointer </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::SetUserData( PCTSTR pstrItemOrGroupName, void *pUserData )
{
    if( !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    ptr->m_pUserData = pUserData;

    return true;
}

///<summary>
///  <b>SetInformMessage()</b> will cause to send the specified message
///  to the specified inform window when the group or item has been
///  repositioned. <br/>
///  The message's LPARAM will be a pointer to a <code>GroupRearranged</code> struct.
///</summary>
///<param name="pstrItemOrGroupName"> name of the item or group </param>
///<param name="hwndInform"> window to send the message to </param>
///<param name="messageidInform"> message id </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::SetInformMessage( PCTSTR pstrItemOrGroupName, HWND hwndInform, UINT messageidInform )
{
    if( !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    ptr->m_hwndInform      = hwndInform;
    ptr->m_messageidInform = messageidInform;

    return true;
}

///<summary>
///  <b>SetAutoInvalidateControl()</b> enables or disables calling of
///  <b>InvalidateRect()</b> for a control attached to the specified group
///  when the group is rearranged.
///  This is useful for e.g. some static text controls which do not repaint
///  themselves correctly all the time when they are resized.
///</summary>
bool ControlGrouping::SetAutoInvalidateControl( PCTSTR pstrItemOrGroupName, bool bSet )
{
    if( !pstrItemOrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName );
    if( !ptr ) return false;

    ptr->m_bAutoInvalidateAttachedControl = bSet;

    return true;
}

bool ControlGrouping::SetTopOrLeftBorderAutoDragable( PCTSTR pstrGroupName, bool bSet, const RECT *prectTopOrLeftBorderInflate )
{
    if( !pstrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrGroupName );
    if( !ptr ) return false;

    ptr->m_bTopOrLeftBorderAutoDragable = bSet;

    if( prectTopOrLeftBorderInflate )
        ptr->m_rectTopOrLeftBorderInflate = *prectTopOrLeftBorderInflate;

    if( bSet ) {
        if( m_aTopOrLeftDragableBorders.FindByVal( (HGROUP) ptr ) < 0 )
            m_aTopOrLeftDragableBorders.AddByVal( (HGROUP) ptr );
    } else {
        m_aTopOrLeftDragableBorders.Remove( m_aTopOrLeftDragableBorders.FindByVal( (HGROUP) ptr ) );
    }

    return true;
}

///<summary>
///  <b>GetTopOrLeftBorderAutoDragable()</b> returns the <code>m_bTopOrLeftBorderAutoDragable</code>
///  flag for a group. <br/>
///  Note: This flag will be returned even if the group is not a strip's child or
///  even if it cannot be dragged at all.
///  To determine if the border really may be dragged (for the current size of the
///  root group), do the following:
///  First call <b>GetTopOrLeftBorderAutoDragable()</b> to determine if the user
///  enabled dragging for that border, and then call <b>GetTopOrLeftDragBorder()</b>
///  to determine if the border currently really can be dragged.
///</summary>
bool ControlGrouping::GetTopOrLeftBorderAutoDragable( PCTSTR pstrGroupName ) const
{
    if( !pstrGroupName ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrGroupName );
    if( !ptr ) return false;

    return ptr->m_bTopOrLeftBorderAutoDragable;
}

bool ControlGrouping::GetTopOrLeftBorderInflateRect( PCTSTR pstrGroupName, RECT *prectTopOrLeftBorderInflate ) const
{
    if( !pstrGroupName || !prectTopOrLeftBorderInflate ) return false;

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrGroupName );
    if( !ptr ) return false;

    *prectTopOrLeftBorderInflate = ptr->m_rectTopOrLeftBorderInflate;

    return true;
}



bool ControlGrouping::hasStripBordersAutoDragable( HGROUP hStripGroup ) const
{
    if( !hStripGroup ) return false;

    ItemOrGroup *pStrip = (ItemOrGroup*) hStripGroup;
    if( !pStrip ) return false;

    if( pStrip->m_GroupType != grpt_HorzGroupStrip && pStrip->m_GroupType != grpt_VertGroupStrip )
        return false;

    if( pStrip->m_cStripSubGroups < 1 || !pStrip->m_firstChild )
        return false;

    for( ItemOrGroup *child = pStrip->m_firstChild ; child ; child = child->m_nextSibling )
        if( child->m_bTopOrLeftBorderAutoDragable )
            return true;

    return false;
}



///<summary>
///  Call <b>Rearrange()</b> to perform repositioning and resizing
///  of all groups below the specified root group according
///  to the specified new position and size of the root group.
///</summary>
///<param name="pNewRect"> new position and size for the root group
///  (negative positions and sizes are supported). </param>
///<param name="pstrRootGroupName"> name of the root group (usually an empty string)
///  If NULL -> will be treated like an empty string in here. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::Rearrange( const RECT *pNewRect, PCTSTR pstrRootGroupName )
{
    if( m_bComposing || !pNewRect ) return false;
    if( !pstrRootGroupName ) pstrRootGroupName = _T("");

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrRootGroupName );
    if( !ptr ) return false;

    ptr->m_rectItemCurrent = pNewRect;

    return ptr->ArrangeItems();
}

///<summary>
///  Call <b>InformRearrange()</b> to re-perform the repositioning and resizing
///  of all groups below the specified root group.
///  This will not change the positions/sizes of the groups since a new
///  rect is not specified in here. But this method can be used just to
///  force sending the inform-messages again.
///</summary>
///<param name="pstrRootGroupName"> name of the root group (usually an empty string)
///  If NULL -> will be treated like an empty string in here. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::InformRearrange( PCTSTR pstrRootGroupName )
{
    if( m_bComposing ) return false;
    if( !pstrRootGroupName ) pstrRootGroupName = _T("");

    ItemOrGroup *ptr = m_htItemsAndGroups.GetPtrAt( pstrRootGroupName );
    if( !ptr ) return false;

    return ptr->ArrangeItems();
}



///<summary>
///  <b>GetTopOrLeftDragBorder()</b> gets the border to the left or on top of
///  a strip's child group. It will get the left border if the child group
///  is in a horizontal strip. It will get the top border if the child
///  group is in a vertical strip. <br/>
///  Notice that the border rectangle can be empty or even have negative sizes.
///</summary>
///<param name="hStripChildGroup"> Handle of the child group of a strip.
///  If this group is not a strip's child, false will be returned. </param>
///<param name="prectBorder"> The rectangle (must not be NULL) where to
///  receive the border rectangle. </param>
///<param name="prectGrabBorder"> The rectangle where the user
///  may move the mouse in so that he can start dragging the border.
///  Normally, this is identical to the <b>prectBorder</b> rectangle coords,
///  except for the case when a top-or-left-border-Inflate-Rectangle has
///  been set in <b>SetTopOrLeftBorderAutoDragable()</b>. In that case then,
///  the grab-border-rect will be made thicker compared to the border-rect
///  according to the inflate-rect's coords. </param>
///<param name="prectDragArea"> If not NULL, this rectangle will be the
///  bounding rect wherein the border may be dragged around. If this rect
///  is an empty rect, then this means that the border is not dragable
///  (maybe because all child groups are of fixed size or they all reached
///  their minimum sizes for the current size of the root group). </param>
///<returns> True on success and if the border may be dragged,
///  false on error or if the border may not/cannot be dragged.
///  A border cannot be dragged even if it has stretchable neighbouring
///  groups but they both have reached their minimum sizes for the
///  current size of the root group. </returns>
bool ControlGrouping::GetTopOrLeftDragBorder( HGROUP hStripChildGroup, PRECT prectBorder, PRECT prectGrabBorder, PRECT prectDragArea )
{
    if( !prectBorder ) return false;

    ItemOrGroup *pChild;
    ItemOrGroup *pLeftScalable, *pRightScalable, *pDirectLeftSibling;
    bool bHorzStrip, bVertStrip;
    long currextL, currextR, minL, minR;

    if( !_CalcTopOrLeftDragBorder( hStripChildGroup,
                                   pChild,
                                   pLeftScalable, pRightScalable, pDirectLeftSibling,
                                   bHorzStrip, bVertStrip,
                                   currextL, currextR, minL, minR ) )
        return false;

    // So now we really know that the border may be dragged.
    // Let's get the border rect.

    _Rect_ rectBorderCurrent = pChild->m_parentStrip->m_rectItemCurrent;

    if( bHorzStrip ) {
        rectBorderCurrent.left   = pDirectLeftSibling->m_rectItemCurrent.right;
        rectBorderCurrent.right  = pChild->m_rectItemCurrent.left;
    } else {
        rectBorderCurrent.top    = pDirectLeftSibling->m_rectItemCurrent.bottom;
        rectBorderCurrent.bottom = pChild->m_rectItemCurrent.top;
    }

    if( prectBorder )
        *prectBorder = rectBorderCurrent;

    if( prectGrabBorder )
    {
        *prectGrabBorder = rectBorderCurrent;

        if( bHorzStrip ) {
            prectGrabBorder->left   -= pChild->m_rectTopOrLeftBorderInflate.left;
            prectGrabBorder->right  += pChild->m_rectTopOrLeftBorderInflate.right;
        } else {
            prectGrabBorder->top    -= pChild->m_rectTopOrLeftBorderInflate.top;
            prectGrabBorder->bottom += pChild->m_rectTopOrLeftBorderInflate.bottom;
        }
    }

    if( prectDragArea )
    {
        _Rect_ rectDragAreaCurrent = rectBorderCurrent;

        if( bHorzStrip ) {
            rectDragAreaCurrent.left   -= pLeftScalable->m_rectItemCurrent.Width();
            rectDragAreaCurrent.right  += pRightScalable->m_rectItemCurrent.Width();
        } else {
            rectDragAreaCurrent.top    -= pLeftScalable->m_rectItemCurrent.Height();
            rectDragAreaCurrent.bottom += pRightScalable->m_rectItemCurrent.Height();
        }

        *prectDragArea = rectDragAreaCurrent;
    }

    return true;
}

///<summary>
///  <b>DragTopOrLeftBorderTo()</b> will set the left or top border
///  of the specified child group (of a strip) at the specified
///  position causing the child group and its top or left neighbour
///  to be stretched and squeezed. <br/>
///  Since dragging is possible only in a specific valid range, this method
///  can return the corrected value which has been bound to its valid range. <br/>
///  If one neighbour is of fixed size, then the next outer most neighbour(s)
///  will be searched until stretchable one(s) have been found. The result
///  of this is that dragging will also cause the fixed-sized groups in between
///  to be moved around. But this effect can be disabled if not desired.
///</summary>
///<param name="hStripChildGroup"> Handle of the child group of a strip.
///  If this group is not a strip's child, false will be returned. </param>
///<param name="pos"> Desired coordinate. If the child group is in a horizontal
///  strip, then left coordinate of the child group will be set. If it is
///  in a vertical strip, its top coordinate will be set. </param>
///<param name="adjustedPos"> receives the adjusted coordinate bound to the
///  valid range for the dragging operation. <br/>
///  Can be NULL if this value is not desired. </param>
///<returns> True on success, false on error. </returns>
///<remarks>
///  This call will be used mainly when dragging the border with the mouse.
///</remarks>
bool ControlGrouping::DragTopOrLeftBorderTo( HGROUP hStripChildGroup, long pos, long *adjustedPos )
{
    ItemOrGroup *pChild;
    ItemOrGroup *pLeftScalable, *pRightScalable, *pDirectLeftSibling;
    bool bHorzStrip, bVertStrip;
    long currextL, currextR, minL, minR;

    if( !_CalcTopOrLeftDragBorder( hStripChildGroup,
                                   pChild,
                                   pLeftScalable, pRightScalable, pDirectLeftSibling,
                                   bHorzStrip, bVertStrip,
                                   currextL, currextR, minL, minR ) )
        return false;

    // So now we really know that the border may be dragged.

    long oldpos = bHorzStrip ? pChild->m_rectItemCurrent.left  : pChild->m_rectItemCurrent.top;
    long ofs = pos - oldpos;

    if( currextL + ofs < minL ) ofs = minL - currextL;
    if( currextR - ofs < minR ) ofs = currextR - minR;

    if( adjustedPos ) *adjustedPos = oldpos + ofs;

    //long newextL = currextL + ofs;
    long newextR = currextR - ofs;

    double oldAdoptedextL = bHorzStrip ? pLeftScalable->m_fItemAdoptedInitWidth  : pLeftScalable->m_fItemAdoptedInitHeight;
    double oldAdoptedextR = bHorzStrip ? pRightScalable->m_fItemAdoptedInitWidth : pRightScalable->m_fItemAdoptedInitHeight;

    double oldAdoptedsum  = oldAdoptedextL + oldAdoptedextR;
    double oldsum         = currextL + currextR;

    double newAdoptedextL, newAdoptedextR;

    eFixedType fixedLeft  = bHorzStrip ? pLeftScalable->m_fixedWidth  : pLeftScalable->m_fixedHeight;
    eFixedType fixedRight = bHorzStrip ? pRightScalable->m_fixedWidth : pRightScalable->m_fixedHeight;

    if( fixedLeft  == fixedt_FixedDragable ||
        fixedRight == fixedt_FixedDragable )
    {
        newAdoptedextL = oldAdoptedextL + ofs;
        newAdoptedextR = oldAdoptedextR - ofs;
    }
    else
    {
        newAdoptedextR = (oldAdoptedsum * newextR) / (oldsum != 0.0 ? oldsum : 1.0);
        newAdoptedextL = oldAdoptedsum - newAdoptedextR;
    }

    if( bHorzStrip )
    {
        pLeftScalable->m_fItemAdoptedInitWidth  = newAdoptedextL;
        pRightScalable->m_fItemAdoptedInitWidth = newAdoptedextR;
    }
    else
    {
        pLeftScalable->m_fItemAdoptedInitHeight  = newAdoptedextL;
        pRightScalable->m_fItemAdoptedInitHeight = newAdoptedextR;
    }
    // Note: We do not need to set the current rects, because
    // this will be done in the strip's ArrangeItems() below.

    pChild->m_parentStrip->ArrangeItems();

    return true;
}

///<summary>
///  <b>SetRelativeTopOrLeftBorderPos()</b> acts similar to
///  <b>DragTopOrLeftBorderTo()</b> except that it takes the position
///  as a relative position value in the range 0.0 to 1.0 rather
///  than a pixel position value.
///</summary>
///<param name="hStripChildGroup"> Handle of the child group of a strip.
///  If this group is not a strip's child, false will be returned. </param>
///<param name="relativePos"> Desired relative coordinate (range 0.0 to 1.0,
///  will be clipped automatically). If the child group is in a horizontal
///  strip, then left coordinate of the child group will be set. If it is
///  in a vertical strip, its top coordinate will be set. </param>
///<returns> True on success, false on error. </returns>
///<remarks>
///  This call will be used mainly to set the border position according
///  to a relative value from 0.0 to 1.0 e.g. entered by the user. <br/>
///  Do *not* use this to restore positions stored in a storage media
///  like the registry. For this, you should use the other methods
///  that allow storing and restoring a whole strip. See remarks there
///  why this is necessary.
///</remarks>
bool ControlGrouping::SetRelativeTopOrLeftBorderPos( HGROUP hStripChildGroup, double relativePos )
{
    relativePos = tbound( 0.0, relativePos, 1.0 );

    ItemOrGroup *pChild;
    ItemOrGroup *pLeftScalable, *pRightScalable, *pDirectLeftSibling;
    bool bHorzStrip, bVertStrip;
    long currextL, currextR, minL, minR;

    if( !_CalcTopOrLeftDragBorder( hStripChildGroup,
                                   pChild,
                                   pLeftScalable, pRightScalable, pDirectLeftSibling,
                                   bHorzStrip, bVertStrip,
                                   currextL, currextR, minL, minR ) )
        return false;

    // So now we really know that the border may be dragged.

    double oldAdoptedextL = bHorzStrip ? pLeftScalable->m_fItemAdoptedInitWidth : pLeftScalable->m_fItemAdoptedInitHeight;
    double oldAdoptedextR = bHorzStrip ? pRightScalable->m_fItemAdoptedInitWidth : pRightScalable->m_fItemAdoptedInitHeight;

    double oldAdoptedsum  = oldAdoptedextL + oldAdoptedextR;

    double newAdoptedextR = oldAdoptedsum * relativePos;
    double newAdoptedextL = oldAdoptedsum - newAdoptedextR;

    if( bHorzStrip )
    {
        pLeftScalable->m_fItemAdoptedInitWidth  = newAdoptedextL;
        pRightScalable->m_fItemAdoptedInitWidth = newAdoptedextR;
    }
    else
    {
        pLeftScalable->m_fItemAdoptedInitHeight  = newAdoptedextL;
        pRightScalable->m_fItemAdoptedInitHeight = newAdoptedextR;
    }
    // Note: We do not need to set the current rects, because
    // this will be done in the strip's ArrangeItems() below.

    pChild->m_parentStrip->ArrangeItems();

    return true;
}

///<summary>
///  <b>GetRelativeTopOrLeftBorderPos()</b> gets the relative position
///  value that can be set using <b>SetRelativeTopOrLeftBorderPos()</b>.
///</summary>
///<param name="hStripChildGroup"> Handle of the child group of a strip.
///  If this group is not a strip's child, false will be returned. </param>
///<param name="relativePos"> The returned relative coordinate (range 0.0 to 1.0,
///  will be clipped automatically). If the child group is in a horizontal
///  strip, then left coordinate of the child group will be retrieved. If it is
///  in a vertical strip, its top coordinate will be retrieved. </param>
///<returns> True on success, false on error. </returns>
///<remarks>
///  This call will be used mainly to get the border position according
///  to a relative value from 0.0 to 1.0. <br/>
///  Do *not* use this to store positions in a storage media
///  like the registry. For this, you should use the other methods
///  that allow storing and restoring a whole strip. See remarks there
///  why this is necessary.
///</remarks>
bool ControlGrouping::GetRelativeTopOrLeftBorderPos( HGROUP hStripChildGroup, double& relativePos )
{
    ItemOrGroup *pChild;
    ItemOrGroup *pLeftScalable, *pRightScalable, *pDirectLeftSibling;
    bool bHorzStrip, bVertStrip;
    long currextL, currextR, minL, minR;

    if( !_CalcTopOrLeftDragBorder( hStripChildGroup,
                                   pChild,
                                   pLeftScalable, pRightScalable, pDirectLeftSibling,
                                   bHorzStrip, bVertStrip,
                                   currextL, currextR, minL, minR ) )
        return false;

    // So now we really know that the border may be dragged.

    double oldAdoptedextL = bHorzStrip ? pLeftScalable->m_fItemAdoptedInitWidth : pLeftScalable->m_fItemAdoptedInitHeight;
    double oldAdoptedextR = bHorzStrip ? pRightScalable->m_fItemAdoptedInitWidth : pRightScalable->m_fItemAdoptedInitHeight;

    double oldAdoptedsum  = oldAdoptedextL + oldAdoptedextR;

    relativePos = oldAdoptedextR / (oldAdoptedsum != 0.0 ? oldAdoptedsum : 1.0);
    relativePos = tbound( 0.0, relativePos, 1.0 );

    return true;
}

///<summary>
///  <b>GetNextDragableBorder()</b> gets the dragable border group
///  from an index into the array of dragable border groups.
///  This array includes all groups's borders having been enabled as
///  dragable using <b>SetTopOrLeftBorderAutoDragable()</b>.
///</summary>
///<returns> False if the index is out of range. </returns>
bool ControlGrouping::GetNextDragableBorder( _idx_ idx, HGROUP& hGrp )
{
    if( idx < 0 || idx >= (_idx_) m_aTopOrLeftDragableBorders.GetSize() )
        { hGrp = NULL; return false; }

    hGrp = m_aTopOrLeftDragableBorders[idx];

    return true;
}

// We enum all dragable borders and check, if the mouse is in.
ControlGrouping::HGROUP ControlGrouping::IsPtInDragableBorder( POINT pt, PRECT prectBorder, PRECT prectDragArea )
{
    HGROUP h;
    for( _idx_ i = 0 ; GetNextDragableBorder(i, h) ; ++i )
    {
        _Rect_ rectBorder, rectGrabBorder;
        if( GetTopOrLeftDragBorder( h, rectBorder, rectGrabBorder, prectDragArea ) )
            if( rectGrabBorder.PtInRect(pt) )
            {
                if( prectBorder ) *prectBorder = rectBorder;
                return h;
            }
    }

    return NULL;
}



bool ControlGrouping::_CalcTopOrLeftDragBorder( HGROUP hStripChildGroup,
                                                ItemOrGroup *& pChild,
                                                ItemOrGroup *& pLeftScalable, ItemOrGroup *& pRightScalable, ItemOrGroup *& pDirectLeftSibling,
                                                bool& bHorzStrip, bool& bVertStrip,
                                                long& currextL, long& currextR, long& minL, long& minR )
{
    if( !hStripChildGroup ) return false;

    pChild = (ItemOrGroup*) hStripChildGroup;
    if( !pChild ) return false;

    // We check if the child group is part of a strip.

    if( !pChild->m_parentStrip ) return false;
    if( pChild->m_parentStrip->m_cStripSubGroups < 2 ) return false;
    bHorzStrip = pChild->m_parentStrip->m_GroupType == grpt_HorzGroupStrip;
    bVertStrip = pChild->m_parentStrip->m_GroupType == grpt_VertGroupStrip;
    if( !bHorzStrip && !bVertStrip ) return false;

    // Now let's find the top/left and bottom/right nearest scalable neighbour.
    //  Notice that scalable in the meaning of dragging borders include groups
    //  of type fixedt_Dynamic as well as fixedt_FixedDragable.

    pLeftScalable = pRightScalable = pDirectLeftSibling = NULL;

    ItemOrGroup *p = pChild->m_parentStrip->m_firstChild;
    while( p && p != pChild )
    {
        if( bHorzStrip ) { if( p->m_fixedWidth  != fixedt_Fixed ) pLeftScalable = p; }
        else             { if( p->m_fixedHeight != fixedt_Fixed ) pLeftScalable = p; }
        if( p->m_nextSibling == pChild ) pDirectLeftSibling = p;
        p = p->m_nextSibling;
    }

    if( p != pChild )   // this just means some hierarchy problem: #pChild# was not found in the strip although it should be there.
        return false;

    while( p )
    {
        if( bHorzStrip ) { if( p->m_fixedWidth  != fixedt_Fixed ) { pRightScalable = p; break; } }
        else             { if( p->m_fixedHeight != fixedt_Fixed ) { pRightScalable = p; break; } }
        p = p->m_nextSibling;
    }

    if( !pLeftScalable || !pRightScalable || !pDirectLeftSibling ) // no scalable neighbours found
        return false;

    if( !m_bAllowDragMoveFixedGroups )
    {
        if( pLeftScalable->m_nextSibling != pChild ) return false;
        if( pRightScalable != pChild ) return false;
    }

    currextL = bHorzStrip ? pLeftScalable->m_rectItemCurrent.Width()  : pLeftScalable->m_rectItemCurrent.Height();
    currextR = bHorzStrip ? pRightScalable->m_rectItemCurrent.Width() : pRightScalable->m_rectItemCurrent.Height();

    minL = bHorzStrip ? pLeftScalable->m_minSize.cx  : pLeftScalable->m_minSize.cy;
    minR = bHorzStrip ? pRightScalable->m_minSize.cx  : pRightScalable->m_minSize.cy;

    // Let's check if the scalable groups have reached their minimums but only *together*.
    if( (currextL + currextR) <= (minL + minR) )
        return false;

    // So now we really know that the border may be dragged.

    return true;
}



///<summary>
///  <b>SetWindowRect()</b> will set the specified window to the rect but
///  using window pos deferring if enabled. Warning: You will have to call
///  the current <b>WindowDeferrer</b>'s <b>End()</b> method yourself afterwards.
///</summary>
bool ControlGrouping::SetWindowRect( HWND hwnd, _Rect_ rect, UINT swpFlags )
{
    if( m_pWindowDeferrer )
    {
        if( m_pWindowDeferrer->deferring() ? true : m_pWindowDeferrer->Begin( 128 ) )
            return m_pWindowDeferrer->Deferwindowpos( hwnd, HWND_TOP, rect.left, rect.top, rect.Width(), rect.Height(),
                                                      swpFlags );
    }

    return rect.SetWindowRect( hwnd, HWND_TOP, swpFlags );
}



///<summary>
///  <b>HandleWndMessage()</b> handles window messages of the parent window
///  whos controls have been added to <code>this</code> object.
///  <br/>
///  After or before calling this method, you should call your own window
///  procedure. So therefore, the return value returned here can be ignored
///  and need not be returned by your window proc.
///</summary>
///<returns> True if the message has been executed by this method
///  and it *must not* be used in further message handling.
///  In that case, return the <i>result</i> value in your message proc.
///  False if the message may be used for further processing. </returns>
bool ControlGrouping::HandleWndMessage( LRESULT& result, HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam, PCTSTR pstrRootGroupName )
{
    result = 0;
    if( !hwnd ) return false;

    switch( message )
    {
        // The following messages are used when resizing this
        // window so that the controls in it are rearranged.

        case WM_WINDOWPOSCHANGED :
        {
            WINDOWPOS *wp = (WINDOWPOS*) lParam;

            //_Rect_ rect( wp );
            //rect.MoveRectToOrigin();
            //Rearrange( rect, pstrRootGroupName );

            if( (wp->flags & (SWP_NOMOVE | SWP_NOSIZE)) != (SWP_NOMOVE | SWP_NOSIZE) )
                Rearrange( _Rect_().GetClientRect(hwnd), pstrRootGroupName );

            break;
        }

        #ifdef WM_GETMINMAXINFO     // not defined for WindowsCE
        case WM_GETMINMAXINFO :
        {
            MINMAXINFO *p = (MINMAXINFO*) lParam;
            _Size_ minSize;
            if( GetMinSize( pstrRootGroupName, minSize ) )
            {
                minSize.MaxSize( _Size_(p->ptMinTrackSize) );
                p->ptMinTrackSize = _Pt_(minSize);
            }
            break;
        }
        #endif

        // The following messages are used to support
        // dragging of dragable borders in strips.

        case WM_SETCURSOR :
        {
            _Pt_ mouse;
            mouse.GetCursorPos().ScreenToClient( hwnd );
            //mouse.GetMessagePos().ScreenToClient( hwnd );
            // Note: We use GetCursorPos() instead of GetMessagePos() because
            // otherwise, the previous mouse position before the mousemove
            // will be returned (don't ask me why).

            HGROUP hGrp = IsPtInDragableBorder( mouse );
            if( hGrp )
            {
                SetCursor( getGroupType( getParentStripGroup(hGrp) ) == grpt_HorzGroupStrip ?
                    m_hcrsrHorzDrag : m_hcrsrVertDrag );
                result = TRUE;
                return true;
            }

            break;
        }

        case WM_LBUTTONDOWN :
        case WM_LBUTTONDBLCLK :
        {
            _Pt_ mouse( (DWORD) lParam );

            if( m_bDragging ) break;

            HGROUP hGrp = IsPtInDragableBorder( mouse, m_rectBorderBeginDrag, m_rectDraggingInArea );
            if( hGrp )
            {
                m_hgrpBorderBeingDragged = hGrp;
                m_bDragging = true;
                m_ptMouseOnBeginDrag = mouse;
                SetCapture( hwnd );
                SetCursor( getGroupType( getParentStripGroup(hGrp) ) == grpt_HorzGroupStrip ?
                    m_hcrsrHorzDrag : m_hcrsrVertDrag );
                result = 0;
                return true;
            }

            break;
        }

        case WM_CAPTURECHANGED :
        {
            if( m_bDragging )
            {
                m_bDragging = false;
                m_hgrpBorderBeingDragged = NULL;
                m_ptMouseOnBeginDrag.SetEmpty();
                m_rectBorderBeginDrag.SetRectEmpty();
                m_rectDraggingInArea.SetRectEmpty();
                result = 0;
                return true;
            }
            break;
        }

        case WM_LBUTTONUP :
        {
            if( m_bDragging )
            {
                ReleaseCapture();
                result = 0;
                return true;
            }

            break;
        }

        case WM_MOUSEMOVE :
        {
            if( m_bDragging && m_hgrpBorderBeingDragged && GetCapture() == hwnd )
            {
                _Pt_ mouse;
                mouse.GetCursorPos().ScreenToClient( hwnd );

                HGROUP hGrp = m_hgrpBorderBeingDragged;
                bool bHorzStrip = getGroupType( getParentStripGroup(hGrp) ) == grpt_HorzGroupStrip;

                long mouseOfs;  // the offset of the mouse to the group whos top or left border is being dragged
                if( bHorzStrip ) mouseOfs = m_ptMouseOnBeginDrag.x - m_rectBorderBeginDrag.right;
                else             mouseOfs = m_ptMouseOnBeginDrag.y - m_rectBorderBeginDrag.bottom;

                long newMousePos = bHorzStrip ? mouse.x : mouse.y;

                DragTopOrLeftBorderTo( hGrp, newMousePos - mouseOfs );

                result = 0;
                return true;
            }

            break;
        }
    }

    return false;
}



bool ControlGrouping::GetStripStore( Store& store, PCTSTR pstrStripGroupName )
{
    if( !pstrStripGroupName ) return false;

    ItemOrGroup *pGrp = m_htItemsAndGroups.GetPtrAt( pstrStripGroupName );
    if( !pGrp ) return false;

    if( pGrp->m_GroupType != grpt_HorzGroupStrip && pGrp->m_GroupType != grpt_VertGroupStrip )
        return false;

    if( pGrp->m_cStripSubGroups <= 0 || !pGrp->m_firstChild )
        return false;

    bool bHorzStrip = pGrp->m_GroupType == grpt_HorzGroupStrip;

    if( !store.SetHeader( pstrStripGroupName, pGrp->m_GroupType, pGrp->m_nStripExtentInitial ) )
        return false;

    for( ItemOrGroup *child = pGrp->m_firstChild ; child ; child = child->m_nextSibling )
    {
        if( !store.AddGroup( bHorzStrip ? child->m_fixedWidth : child->m_fixedHeight,
                             bHorzStrip ? child->m_fItemAdoptedInitWidth : child->m_fItemAdoptedInitHeight ) )
            return false;
    }

    return store.CheckIntegrity();
}

bool ControlGrouping::SetStripStore( Store *pStore, bool bAutoRearrangeNow )
{
    if( !pStore ) return false;
    if( !pStore->CheckIntegrity() ) return false;

    Store_Header UNALIGNED *hdr = pStore->GetHeader();
    if( !hdr ) return false;

    StringW strStripGroupName; strStripGroupName.Strcpy_unaligned( hdr->m_strStripName );

    ItemOrGroup *pGrp = m_htItemsAndGroups.GetPtrAt( String(strStripGroupName) );
    if( !pGrp ) return false;

    if( pGrp->m_cStripSubGroups != (int) hdr->m_cStripSubGroups ) return false;
    if( (DWORD) pGrp->m_GroupType != hdr->m_GroupType ) return false;
    if( pGrp->m_nStripExtentInitial != hdr->m_nStripExtentInitial ) return false;

    bool bHorzStrip = pGrp->m_GroupType == grpt_HorzGroupStrip;

    _idx_ i = 0;
    ItemOrGroup *child;
    for( child = pGrp->m_firstChild ; child ; child = child->m_nextSibling, ++i )
    {
        if( i >= pGrp->m_cStripSubGroups ) return false;

        Store_Group UNALIGNED *stgrp = pStore->GetGroupAt( i );
        if( !stgrp ) return false;

        eFixedType fixed = bHorzStrip ? child->m_fixedWidth : child->m_fixedHeight;
        if( (DWORD) fixed != stgrp->m_dwFixedExtent ) return false;

        double adoptedInitExt = bHorzStrip ? child->m_fItemAdoptedInitWidth : child->m_fItemAdoptedInitHeight;

        if( fixed == fixedt_Fixed )
            if( stgrp->m_fAdoptedInitExtent != adoptedInitExt )
                return false;
    }

    if( i != pGrp->m_cStripSubGroups ) return false;

    i = 0;
    for( child = pGrp->m_firstChild ; child ; child = child->m_nextSibling, ++i )
    {
        Store_Group UNALIGNED *stgrp = pStore->GetGroupAt( i );
        if( !stgrp ) return false;

        if( bHorzStrip ) child->m_fItemAdoptedInitWidth  = stgrp->m_fAdoptedInitExtent;
        else             child->m_fItemAdoptedInitHeight = stgrp->m_fAdoptedInitExtent;
    }

    if( bAutoRearrangeNow )
    {
        pGrp->ArrangeItems();
    }

    return true;
}

bool ControlGrouping::StoreStrip( RegKey& rk, PCTSTR pstrRegkeyName, PCTSTR pstrStripGroupName )
{
    if( !rk.isOpen() || _isstrempty_(pstrRegkeyName) ) return false;

    Store store;
    if( !GetStripStore( store, pstrStripGroupName ) )
        return false;

    PVOID ptr;
    UINT  bytesize;
    if( !store.GetArrayData( ptr, bytesize ) )
        return false;

    return rk.SetBinary( pstrRegkeyName, ptr, bytesize );
}

bool ControlGrouping::RestoreStrip( RegKey& rk, PCTSTR pstrRegkeyName, PCTSTR pstrStripGroupName, bool bAutoRearrangeNow )
{
    if( !rk.isOpen() || _isstrempty_(pstrRegkeyName) ) return false;

    Store_Header hdr;
    if( !rk.GetBinaryPart( pstrRegkeyName, &hdr, sizeof(hdr) ) )
        return false;

    if( !hdr.CheckHeaderIntegrity() )
        return false;

    Store_Header UNALIGNED *phdr = (Store_Header UNALIGNED*) &hdr;

    Store store;
    store.m_aData.SetSize( phdr->m_dwContentSize );
    if( !rk.GetBinary( pstrRegkeyName, store.m_aData.Ptr(), store.m_aData.GetSize() ) )
        return false;

    if( !store.CheckIntegrity() )
        return false;

    return SetStripStore( &store, bAutoRearrangeNow );
}

///<summary>
///  <b>StoreAutoDragableStrips()</b> saves all strips that have at least one
///  auto-dragable border to the registry. A border is auto-dragable, when its
///  flag has been enabled using <b>SetTopOrLeftBorderAutoDragable()</b>.
///</summary>
///<param name="rk"> The RegKey object for storing data in. </param>
///<param name="pstrRegkeyNamePrefix"> A name prefix. For each strip, this
///  method will produce a registry key name that starts with that prefix
///  followed immediately by the strip's group name. </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::StoreAutoDragableStrips( RegKey& rk, PCTSTR pstrRegkeyNamePrefix )
{
    if( !rk.isOpen() || _isstrempty_(pstrRegkeyNamePrefix) ) return false;
    bool ok = true;

    const String *key;
    ItemOrGroup *grp;

    Iterator iter;
    while( m_htItemsAndGroups.GetNextPtr(iter, key, grp) )
    {
        if( !hasStripBordersAutoDragable( (HGROUP) grp ) )
            continue;

        String regkeyname = pstrRegkeyNamePrefix;
        regkeyname += grp->m_strItemOrGroupName;

        if( !StoreStrip( rk, regkeyname, grp->m_strItemOrGroupName ) )
            ok = false;
    }

    return ok;
}

///<summary>
///  <b>RestoreAutoDragableStrips()</b> loads all strips that have at least one
///  auto-dragable border to the registry. A border is auto-dragable, when its
///  flag has been enabled using <b>SetTopOrLeftBorderAutoDragable()</b>.
///</summary>
///<param name="rk"> The RegKey object for storing data in. </param>
///<param name="pstrRegkeyNamePrefix"> A name prefix. For each strip, this
///  method will produce a registry key name that starts with that prefix
///  followed immediately by the strip's group name. </param>
///<param name="bAutoRearrangeStripsNow"> If true, all the strips that have
///  been loaded will be rearranged (only those strips). </param>
///<returns> True on success, false on error. </returns>
bool ControlGrouping::RestoreAutoDragableStrips( RegKey& rk, PCTSTR pstrRegkeyNamePrefix, bool bAutoRearrangeStripsNow )
{
    if( !rk.isOpen() || _isstrempty_(pstrRegkeyNamePrefix) ) return false;
    bool ok = true;

    const String *key;
    ItemOrGroup *grp;

    Iterator iter;
    while( m_htItemsAndGroups.GetNextPtr(iter, key, grp) )
    {
        if( !hasStripBordersAutoDragable( (HGROUP) grp ) )
            continue;

        String regkeyname = pstrRegkeyNamePrefix;
        regkeyname += grp->m_strItemOrGroupName;

        if( !RestoreStrip( rk, regkeyname, grp->m_strItemOrGroupName, bAutoRearrangeStripsNow ) )
            ok = false;
    }

    return ok;
}




}
