/*
    File:               argonControlGrouping.h
    Description:        These utils allow grouping of controls or any other
                        rectangular based items and causing them automatically
                        to be respositioned and resizes according to a new
                        outer container size.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         22.10.2005
    last updated:       11.10.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonControlGrouping_h
#define __argonControlGrouping_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonRect.h"
    #include "argonArray.h"
    #include "argonBase.h"
    #include "argonString.h"
    #include "argonRegistry.h"
    #include "argonWindowDeferrer.h"
#endif

namespace ArgonLibrary
{




/*
    How to use:
    ===============================================

    The principal functioning of ControlGrouping is to position
    and resize controls in a dialog or other window when it is resized.

    The class is kept as simple as possible compared to functionality.
    Therefore it can cover many of the desired wishes when repositioning
    controls, but others cannot be realized with this class.

    The basic idea is to build groups where each group can contain
    one or more child groups so forming a hierarchical tree of groups.
    A group can have attached a control or a so called user-content
    item which is not a control but something else the user can
    define by deriving from the ControlGrouping class.

    The result is a hierarchical tree of group items. The top most one
    will be treated as the root group. Its size and positioned can
    be specified (by a call to Rearrange()) and all child groups and
    controls will be repositioned and resized accordingly.

    Normally a child group will be related to its parent group independent
    from any other of its possibly existing sibling groups.
    The relation can be that a child group has a dynamic size and therefore
    will be resized when its parent group is resized. In that case, a
    minimum size can be specified for the child group.
    Or the relation can be that a child group has a fixed size and when
    its parent group is resized, the child group is positioned docked
    to borders of the parent group or to its center.
    These relations can be selected for x and y directions independently.

    There is a special form of relation for child groups:
    The Grid.
    A group can be created as grid that way, that a parent group is created, then
    below it several groups as rows will be created (called a vertical group strip),
    and below each of the row groups, several groups as columns will be
    created (called a horizontal group strip).

    A group strip treats its dependent child groups as neighbours and
    keeps their distances between them, either in horizontal or vertical
    direction.
    Note: To a group strip you can add additional child groups which are not
    part of the strip. These child groups will treat its parent group
    like a normal parent group not knowing anything that the parent group
    also acts as a strip.

    Don't forget, the grid described above is created by partitioning a group
    into rows and each row into columns.
    (Let's call this grid-type-1).
    But the other way is also possible. A grid can be created by partitioning
    a group into columns and each column into rows.
    (Let's call this grid-type-2).

    Normally, the child groups of a parent strip group are resized equally.
    But for each child group you can specify their width and/or height to
    be fixed or have a minimum size.

    Important: Imagine you have a grid of grid-type-1 created (so first rows,
    then columns) and you want a whole row to be of fixed height, then it
    is *not* sufficient to set a fixed height for each of the row's column
    child groups. You will have to set the fixed height for the row group item
    itself. (See the remarks to AddGrid() on how names for row or column
    groups are created so that you can address them).
    Here you see, the whole tree architecture is one-directional. That means
    that a child group can relate its position and size according to its
    parent group, but a parent group does not know anything of relation settings
    regarding its child groups. The only exception is for a strip.
    This is done for simplicity.

    One thing to say about strip groups:
    Imagine a horizontal strip (so e.g. one row of a grid of grid-type-1
    where this row has several child groups working as column fields).
    If some of them have been set a fixed width, then just the remaining
    sibling groups will be stretched in their width.
    In all cases, the strip will keep the distances between neighbour
    groups.
    But what if you want this distance to be stretched, too. (This can be
    a wish if you do not want any of the child groups to be stretched,
    but only their distances to each other).
    To realize this, special separator groups have been implemented.
    So for each stretchable distance, you will have to create an additional
    child group in the strip and add a separator group below that child group.
    So if you had a horizontal strip of 4 groups but you want them all of
    fixed width and just their 3 borders (between each of 2 neighbours)
    to be stretched, you will have to create in fact a strip of 7 (4 + 3)
    groups.
    Below each of the 2nd, 4th and 6th child groups, you will have to
    add a separator group. You can set a minimum width (or height) for
    a separator, by setting the minimum for the separator item.
    (Setting a minimum width or height for a group will also automatically
    update the minimum sizes for all its parent and parent-parent groups
    so that you can set the minimum for a leaf group itself. As mentioned
    above, this does not work for setting fixed sizes for a group).

    General Usage:
    ---------------------------------------

    First do some basic general settings.
    E.g. it is recommended to set a WindowDeferrer so that the controls
    in the ControlGrouping are rearranged using DeferWindowPos()
    instead of SetWindowPos(). If you have a child-dialog in a dialog
    and both of them use a ControlGrouping, it is recommended to
    set one WindowDeferrer for all of them, because the WindowDeferrer
    will handle groups of windows of different parent windows
    automatically separately.

    Then call ComposeStart().
    Then form a tree by adding groups, grids and controls.
    Then call ComposeEnd() which will go through all groups and
    determines their initial sizes and positions according to the
    leaf groups which should contain controls, be user-content items
    or be separator-items.
    If even one of all leaf groups is none of these 3 types, the group
    will be seen as empty and ComposeEnd() will fail. This will result
    in incorrect rearranging of controls later. So if ComposeEnd()
    returns an error, you made a design failure in the group-tree.

    But there can be group-tree design failures you can make which
    ComposeEnd() will not detect (and therefore will not return an error).
    See below.

    After composing, you can set parameters for specific groups
    like setting fixed sizes, minimum sizes, alignment for fixed
    sized groups.

    In your dialog (or other window) that contains the controls,
    you can call HandleWndMessage() from your window procedure to
    automatically have the groups rearranged according to window
    resizing.

    Also dragging support for dragging borders between groups
    of a strip is implemented as well as storing and restoring
    them so that dragged positions can be saved on storage media
    like the registry etc.
    Note: When restoring strips or one strip from the registry or
    any other storage media, be sure to call it *after* all setting
    calls have been made to your groups. That means, restoring
    should be the last call when initializing your group tree.

    Note on dragging: There is a 2nd kind of fixed size for a sub-group:
    FixedDragable. This works just the same as a fixed groups when
    the parent strip is resized. When when the border of FixedDragable
    sub-group is being dragged, then this does work and it is stretched
    as if the sub-group would not be of fixed size.
    The advantage:
    This way, you can realize the behaviour of commonly known toolbars and toolboxes
    of IDEs: their thickness stays the same when the IDE window is resized, but their
    borders can be dragged.
    The disadvantage:
    The internal calculations of the resizing and dragging behaviour in this
    implementation supports correct handling only for up to 1 stretchable
    (fixedt_Dynamic) sub-groups in a strip, when the strip also contains at
    least 1 sub-group of type fixedt_FixedDragable.
    If you include 2 or stretchable sub-groups beside a FixedDragable sub-group,
    then dragging a border of the FixedDragable sub-group can result in strange
    behaviour.
    So keep in mind to include only 1 sub-group of type fixedt_Dynamic into a
    strip when there are also sub-groups of type fixedt_FixedDragable.
    Well, this would be sufficient to support the behaviour as known from
    common IDEs as desribed above: The MDI-window in the middle should be the
    only sub-group of type fixedt_Dynamic and the toolboxes around can be of
    type fixedt_FixedDragable.

    Group-tree design failures:
    ---------------------------------------

    As mentioned above, leaf groups must not be empty of ComposeEnd()
    will return an error since it does not know how to determine
    the initial size and position of that empty group.

    Also prevent 2 separator group items to be neighbours in a strip.
    ComposeEnd() will not be able to detect this design failure,
    but an endless loop (with a possible stack overflow) will be
    the result.
    So you must ensure yourself not to have 2 separator items being
    direct neighbours in any strip.

    Don't build loops in the hierarchy.
    This means if you have e.g. a group named "A" and this has a child
    group named "B", then the parent group of "A" must not be "B",
    because this will not describe a tree any more.
    So never break the heirarchical architecture of a tree, never build
    loops or links between independent branches.

    Things to think of:
    ---------------------------------------

    Usually, you would want to set the WS_CLIPCHIDREN (and maybe even
    WS_CLIPSIBLINGS in some cases of overlapping controls) to
    prevent the whole dialog from flickering when its controls are
    moved and sized.
    Unfortunately, some controls (e.g. mainly static controls) will not
    redraw themselves completely when resized.
    So you can specify flags (e.g. procf_RedrawAllControls) to force all
    controls being redrawn when the dialog is resized, or only static
    controls (procf_RedrawStaticControls).
    If you do not want that, you can force only some specific controls
    (grouped items) to invalidate on move/size automatically by using
    SetAutoInvalidateControl().
*/

class ControlGrouping
{
    public :
        #define max_ControlGrouping_GroupNameLen    80

        enum eGroupType {
            grpt_none                   = 0,
            grpt_Single,
            grpt_HorzGroupStrip,
            grpt_VertGroupStrip,
            grpt_Separator
        };

        enum eAlign {
            aln_Left                    = 0x00,
            aln_Right                   = 0x01,
            aln_Center                  = 0x02,
            aln_Top                     = 0x00,
            aln_Bottom                  = 0x04,
            aln_Vcenter                 = 0x08,
            alnmask_Horz                = 0x03,
            alnmask_Vert                = 0x0c,
            alnstartbit_Horz            = 0x01,
            alnstartbit_Vert            = 0x04
        };

            ///<summary> Enum <b>eProcFlags</b> specifies processing flags when controls are (re)grouped (moved/sized). </summary>
        enum eProcFlags {
            procf_RedrawAllControls     = 0x00000001,   // if set, all (grouped) controls will be redrawn every time they are moved/sized
            procf_RedrawStaticControls  = 0x00000002,   // if set, (grouped) static controls will be redrawn every time they are moved/sized
            procf__Default              = 0
        };

            ///<summary> Enum <b>eFixedType</b> specifies the kind of dynamic-ness or fixed-ness of a sub-group of a parent strip. <br/>
            ///  WARNING: When using <b>fixedt_FixedDragable</b>, the strip should contain only one (or none) sub-groups of
            ///  type <b>fixedt_Dynamic</b> or else dragging borders can result in strange behaviours. </summary>
        enum eFixedType {
            fixedt_Dynamic              = 0,            // resizes when the parent strip is resized, and also resizes according to dragging a border
            fixedt_FixedDragable,                       // resizes only when dragging a border, but not when its parent strip is resized
            fixedt_Fixed                                // permanently fixed in all cases
        };

        typedef UINT_PTR HGROUP;

        struct ItemOrGroup
        {
            public :
                ControlGrouping *m_cg;
                TCHAR       m_strItemOrGroupName [max_ControlGrouping_GroupNameLen];    // name of this item or group
                eGroupType  m_GroupType;

                // attached control
                HWND        m_hwndAttachedControl;
                bool        m_bAutoInvalidateAttachedControl;   // if true, InvalidateRect() will be called for the attached control when it is rearranged. (useful for e.g. some static text controls)

                // user data
                bool        m_bUserContent;         // true if this is an item that contains user specific content
                void       *m_pUserData;            // you can use this pointer on your own how you like

                // Message to send when the item has been rearranged.
                HWND        m_hwndInform;
                UINT        m_messageidInform;

                // positions
                _Rect_      m_rectItemInitial;      // the initial rect of the item or group (if #m_bInitialized# is true)
                _Rect_      m_rectItemCurrent;

                // relation parameters
                int         m_cStripSubGroups;
                _Size_      m_minSize;
                eFixedType  m_fixedWidth;
                eFixedType  m_fixedHeight;
                DWORD       m_align;
                double      m_fItemAdoptedInitWidth;    // These values indicate the part of the strip's width or height in initial space.
                double      m_fItemAdoptedInitHeight;   //  Normally this will be the initial width or height. But when the user wants to drag a border between
                                                        //  2 neighbour strip child groups, then one of these values for both child groups will be changed
                                                        //  but as if we were in initial dialog size (so in initial space).
                int         m_nStripExtentInitial;  // sum of the initial widths or heights of all child groups of the strip.

                // internal stuff
                bool        m_bInitialized;
                ItemOrGroup *m_parent;          // parent group
                ItemOrGroup *m_next;            // next Group that has been added to the parent.
                ItemOrGroup *m_firstSubGroup;   // first sub-group of the list of sub-groups currently having been added as child to #this# group
                ItemOrGroup *m_parentStrip;     // parent strip group. if NULL -> #this# group is *not* in a strip.
                ItemOrGroup *m_nextSibling;     // next Group if #this# is part of a horz or vert strip
                ItemOrGroup *m_firstChild;      // first child group of a horz or vert strip of groups when #this# is the parent group (and therefore is the strip's bounding group)

                // dragging
                bool        m_bTopOrLeftBorderAutoDragable; // This is used only by the automatic dragging. Manual handling of dragging will ignore this flag.
                _Rect_      m_rectTopOrLeftBorderInflate;   // makes the draggable border rect thicker (only the direction of dragging will be used, the other 2 rect coords will be ignored)

            public :
                bool Init();
                bool isStrip() const { return m_GroupType == grpt_HorzGroupStrip || m_GroupType == grpt_VertGroupStrip; }
                void WalkParentMinSize( bool bDoWidths, bool bDoHeights );
                bool getGroupNeighbours( ItemOrGroup *& strip, ItemOrGroup *& left_or_top, ItemOrGroup *& right_or_bottom );
                bool ArrangeItems();
            protected :
                bool iterateArrangeItems();
        };

        struct GroupRearranged  // used when information message is sent
        {
            friend struct ItemOrGroup;

            private :
                ControlGrouping *m_cg;

            public :
                TCHAR       m_strItemOrGroupName [max_ControlGrouping_GroupNameLen];    // name of this item or group
                eGroupType  m_GroupType;

                // attached control
                HWND        m_hwndAttachedControl;

                // user data
                bool        m_bUserContent;         // true if this is an item that contains user specific content
                void       *m_pUserData;            // you can use this pointer on your own how you like

                // positions
                _Rect_      m_rectItemInitial;      // the initial rect of the item or group (if #m_bInitialized# is true)
                _Rect_      m_rectItemCurrent;

                // Positioning other windows (automatically using defer if enabled)
                bool SetWindowRect( HWND hwnd, UINT swpFlags = SWP_NOACTIVATE | SWP_NOZORDER )
                    { return m_cg ? m_cg->SetWindowRect(hwnd, m_rectItemCurrent, swpFlags) : false; }
        };

        #pragma pack(1)
        struct Store_Header
        {
            public :
                #define StoreHeaderFourcc       makeFourcc('S', 't', 'r', 'p')
                DWORD   m_4cc;
                WORD    m_wVersion;             // version nr. (1 for now)
                WORD    m_wRevision;            // revision (subversion) nr. (0 for now)
                DWORD   m_dwContentSize;        // size in bytes of the whole content data (header plus all Store_Group items).
                WCHAR   m_strStripName [max_ControlGrouping_GroupNameLen];  // name of the strip group
                DWORD   m_GroupType;            // strip's group type
                DWORD   m_cStripSubGroups;
                int     m_nStripExtentInitial;  // sum of the initial widths or heights of all child groups of the strip.

                bool CheckHeaderIntegrity();
        };

        struct Store_Group
        {
            public :
                DWORD   m_dwFixedExtent;        // a eFixedType value
                double  m_fAdoptedInitExtent;
        };
        #pragma pack()

        // This class is used when storing groups to or restoring from memory.
        class Store
        {
            friend class ControlGrouping;

            private :
                Array<BYTE> m_aData;

            public :
                Store() {  }

            // Construct the data array
                bool SetHeader( PCTSTR pstrStripName, eGroupType grpt, int nStripExtentInitial );
                bool AddGroup( eFixedType fixedExtent, double adoptedInitExtent );

                Store_Header UNALIGNED *GetHeader();
                Store_Group  UNALIGNED *GetGroupAt( _idx_ idx );

                bool CheckIntegrity();

            // Retrieve or set the whole data from or to memory
            // that is to be stored in or retrieved from storage media (like registry etc).
                bool GetArrayData( PVOID& ptr, UINT& bytesize );
                bool SetArrayData( PVOID  ptr, UINT  bytesize );
                static DWORD GetContentSize( PVOID ptr )
                    { return ptr ? ((Store_Header UNALIGNED*) ptr)->m_dwContentSize : 0; }
        };

    protected :
        HashTable<String, ItemOrGroup>  m_htItemsAndGroups;
        bool                            m_bComposing;

        Array<HGROUP>                   m_aTopOrLeftDragableBorders;
        DWORD                           m_dwProcFlags;  // eProcFlags flags

    public :
        String                          m_strComposeErrors;


    public :
        ControlGrouping();
        ControlGrouping( DWORD dwProcFlags );
        virtual ~ControlGrouping() {  }
    protected :
        ControlGrouping( const ControlGrouping& src );
    private :
        void Ctor();


    // Composing
    public :
        bool isComposing() const { return m_bComposing; }
        void ComposeStart();
        bool ComposeEnd();
        bool AddRootGroup( _Rect_ rectInitial, PCTSTR pstrGroupName = NULL );
        bool AddRootGroup( HWND hwndContainer, PCTSTR pstrGroupName = NULL );
        bool AddGroup( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName );
        bool AddGrid( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName, _Size_ sizGrid, bool bRowsBeforeColumns = true );
        bool AddControl( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, HWND hwndCtrl );
        bool AddControlID( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, HWND hwndParent, UINT idCtrl );
        bool AddContentItem( PCTSTR pstrParentGroupName, PCTSTR pstrItemName, void *pUserData );
        bool AddSeparatorItem( PCTSTR pstrParentGroupName, PCTSTR pstrItemName );
    protected :
        ItemOrGroup *_AddGroup( PCTSTR pstrParentGroupName, PCTSTR pstrGroupName, eGroupType grpType, bool bOverwriteGroupType = true );
            ///<summary>
            ///  <b>onComposeError()</b> will be called whenever a hierarchical or other relevant
            ///  problem occured. Override this method if you want to catch error messages during compose mode.
            ///</summary>
            ///<param name="pstrError"> The error message. </param>
        virtual void onComposeError( PCTSTR pstrError ) { m_strComposeErrors += pstrError; m_strComposeErrors += _T("\n\n"); }

    // Settings (after composing only, except where stated differently)
    public :
        DWORD GetProcFlags() { return m_dwProcFlags; }
        void  SetProcFlags( DWORD dwProcFlags ) { m_dwProcFlags = dwProcFlags; }
        bool  SetFixedSize( PCTSTR pstrItemOrGroupName, eFixedType fixedWidth, eFixedType fixedHeight );
        bool  SetFixedSize( PCTSTR pstrItemOrGroupName, bool bFixedWidth, bool bFixedHeight )
            { return SetFixedSize( pstrItemOrGroupName, bFixedWidth ? fixedt_Fixed : fixedt_Dynamic, bFixedHeight ? fixedt_Fixed : fixedt_Dynamic ); }
        bool  SetAlign    ( PCTSTR pstrItemOrGroupName, DWORD align );
        bool  SetMinWidth ( PCTSTR pstrItemOrGroupName, int minWidth,  bool bAndInitialWidth  = false, bool bCombineExisting = true );
        bool  SetMinHeight( PCTSTR pstrItemOrGroupName, int minHeight, bool bAndInitialHeight = false, bool bCombineExisting = true );
        bool  GetMinSize  ( PCTSTR pstrItemOrGroupName, _Size_& minSize );
        bool  SetUserData ( PCTSTR pstrItemOrGroupName, void *pUserData );
        bool  SetInformMessage( PCTSTR pstrItemOrGroupName, HWND hwndInform, UINT messageidInform );
        bool  SetAutoInvalidateControl( PCTSTR pstrItemOrGroupName, bool bSet );
        bool  SetTopOrLeftBorderAutoDragable( PCTSTR pstrGroupName, bool bSet, const RECT *prectTopOrLeftBorderInflate = NULL );
        bool  GetTopOrLeftBorderAutoDragable( PCTSTR pstrGroupName ) const;
        bool  GetTopOrLeftBorderInflateRect( PCTSTR pstrGroupName, RECT *prectTopOrLeftBorderInflate ) const;

    // Infos, Parameters, Relations.
    public :
        HGROUP getGroupHandle( PCTSTR pstrItemOrGroupName ) { return (HGROUP) m_htItemsAndGroups.GetPtrAt( pstrItemOrGroupName ); }
        String getGroupName( HGROUP hGrp ) const { return hGrp ? ((ItemOrGroup*) hGrp)->m_strItemOrGroupName : String(); }
        bool   isItemOrGroup( PCTSTR pstrGroupName ) const
            { ItemOrGroup *p; return m_htItemsAndGroups.LookupPtr(pstrGroupName, p); }
        eGroupType getGroupType( HGROUP hGrp ) const { return hGrp ? ((ItemOrGroup*) hGrp)->m_GroupType : grpt_none; }
        HGROUP getParentGroup( HGROUP hGrp ) const { return hGrp ? (HGROUP) ((ItemOrGroup*) hGrp)->m_parent : NULL; }
        HGROUP getParentStripGroup( HGROUP hGrp ) const { return hGrp ? (HGROUP) ((ItemOrGroup*) hGrp)->m_parentStrip : NULL; }
        bool hasStripBordersAutoDragable( HGROUP hStripGroup ) const;

    // General Settings.
    protected :
        WindowDeferrer *m_pWindowDeferrer;
        bool            m_bAllowDragMoveFixedGroups;   // if true -> when dragging a border, also fixed-sized sibling groups between scalable groups may be moved
    public :
            ///<summary>
            ///  <b>setWindowDeferrer()</b> can be set to use window position deferring
            ///  (DeferWindowPos()) or normal positioning (SetWindowPos()).
            ///  The passed object must exist until setting another or till #this# is destructed.
            ///  The window deferrer's Begin() and End() methods, you can call on your own
            ///  or let it be done by the call automatically. If you call Begin() on your own,
            ///  you must also call End() on your own.
            ///</summary>
        void            setWindowDeferrer( WindowDeferrer *wd ) { m_pWindowDeferrer = wd; }
        WindowDeferrer *getWindowDeferrer() { return m_pWindowDeferrer; }
            ///<summary>
            ///  <b>AllowDragMoveFixedGroups()</b> enables if in a strip, groups
            ///  of fixed size lying between groups of dynamic size may be dragged
            ///  so that the fixed group is moved while the dynamic ones are restretched.
            ///</summary>
        void            AllowDragMoveFixedGroups( bool bEnable ) { m_bAllowDragMoveFixedGroups = bEnable; }
        bool            isDragMoveFixedGroupsAllowed() const { return m_bAllowDragMoveFixedGroups; }


    // Rearrange groups, items and controls.
    public :
        bool Rearrange( const RECT *pNewRect, PCTSTR pstrRootGroupName = NULL );
        bool InformRearrange( PCTSTR pstrRootGroupName = NULL );

    // For dragging of borders between neighbour groups in a strip.
    //  Note: These are for manual checking of if and which borders to drag.
    //  They ignore any #m_bTopOrLeftBorderAutoDragable# flag settings.
    public :
        bool GetTopOrLeftDragBorder( HGROUP hStripChildGroup, PRECT prectBorder, PRECT prectGrabBorder = NULL, PRECT prectDragArea = NULL );
        bool DragTopOrLeftBorderTo( HGROUP hStripChildGroup, long pos, long *adjustedPos = NULL );

        bool SetRelativeTopOrLeftBorderPos( HGROUP hStripChildGroup, double  relativePos );
        bool GetRelativeTopOrLeftBorderPos( HGROUP hStripChildGroup, double& relativePos );

        bool GetNextDragableBorder( _idx_ idx, HGROUP& hGrp );
        HGROUP IsPtInDragableBorder( POINT pt, PRECT prectBorder = NULL, PRECT prectDragArea = NULL );

    protected :
        bool _CalcTopOrLeftDragBorder( HGROUP hStripChildGroup,
                                       ItemOrGroup *& pChild,
                                       ItemOrGroup *& pLeftScalable, ItemOrGroup *& pRightScalable, ItemOrGroup *& pDirectLeftSibling,
                                       bool& bHorzStrip, bool& bVertStrip,
                                       long& currextL, long& currextR, long& minL, long& minR );

    // helper methods
    public :
        bool SetWindowRect( HWND hwnd, _Rect_ rect, UINT swpFlags = SWP_NOACTIVATE | SWP_NOZORDER );


    // (User) Content Items
    protected :
            ///<summary>
            ///  <b>onInitialItem()</b> will be called when a content item (that contains
            ///  user content or a control) is to be initialized.
            ///  Override this method and return the appropriate initial rectangle. <br/>
            ///  This method will be called during processing of the <code>ComposeEnd()</code> call.
            ///</summary>
            ///<param name="pstrItemName"> name of the item </param>
            ///<param name="pUserData"> the user data if it has been set </param>
            ///<param name="rectInitial"> in here, return the initial rectangle </param>
            ///<returns> Return true on success, false on error. </returns>
            ///<remarks>
            ///  In here, you can call <code>onComposeError()</code> if you need to present
            ///  an error (during compose mode) to the user.
            ///</remarks>
        virtual bool onInitialItem( PCTSTR pstrItemName, void *pUserData, _Rect_& rectInitial ) { return false; }
            ///<summary>
            ///  <b>onArrangedControl()</b> will be called when an attached control
            ///  has been repositioned.
            ///  This method will be called during processing of the <code>Rearrange()</code> call.
            ///</summary>
            ///<param name="pstrItemName"> name of the item </param>
            ///<param name="hwndAttachedControl"> the control window having been repositioned </param>
            ///<param name="pUserData"> the user data if it has been set </param>
            ///<param name="rectInitial"> the initial item rectangle </param>
            ///<param name="rectCurrent"> the current item rectangle </param>
        virtual void onArrangedControl( PCTSTR pstrItemName, HWND hwndAttachedControl, void *pUserData, _Rect_ rectInitial, _Rect_ rectCurrent ) {  }
            ///<summary>
            ///  <b>onArrangeUserItem()</b> will be called when a user content item
            ///  is to be repositioned.
            ///  Override this method and use the passed current rectangle to place
            ///  your control, item etc. <br/>
            ///  This method will be called during processing of the <code>Rearrange()</code> call.
            ///</summary>
            ///<param name="pstrItemName"> name of the item </param>
            ///<param name="pUserData"> the user data if it has been set </param>
            ///<param name="rectInitial"> the initial item rectangle </param>
            ///<param name="rectCurrent"> the current item rectangle </param>
        virtual void onArrangeUserItem( PCTSTR pstrItemName, void *pUserData, _Rect_ rectInitial, _Rect_ rectCurrent ) {  }


    // Message handling
    protected :
        HCURSOR m_hcrsrHorzDrag;
        HCURSOR m_hcrsrVertDrag;
        bool    m_bDragging;
        HGROUP  m_hgrpBorderBeingDragged;
        _Pt_    m_ptMouseOnBeginDrag;
        _Rect_  m_rectBorderBeginDrag;
        _Rect_  m_rectDraggingInArea;
    public :
        bool HandleWndMessage( LRESULT& result, HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam, PCTSTR pstrRootGroupName = NULL );


    // Storing and Restoring.
    public :
        // To get or set the storing data of one strip.
        bool GetStripStore( Store& store, PCTSTR pstrStripGroupName );
        bool SetStripStore( Store *pStore, bool bAutoRearrangeNow = true );

        // To store or restore one strip in the registry.
        bool StoreStrip( RegKey& rk, PCTSTR pstrRegkeyName, PCTSTR pstrStripGroupName );
        bool RestoreStrip( RegKey& rk, PCTSTR pstrRegkeyName, PCTSTR pstrStripGroupName, bool bAutoRearrangeNow = true );

        // To store or restore all strips with auto-dragable enabled borders (only).
        bool StoreAutoDragableStrips( RegKey& rk, PCTSTR pstrRegkeyNamePrefix );
        bool RestoreAutoDragableStrips( RegKey& rk, PCTSTR pstrRegkeyNamePrefix, bool bAutoRearrangeStripsNow = true );
};




}

#endif
