/*
    File:               argonInstance.cpp
    Description:        Module and Process instance simplifications.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       24.10.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonInstance_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonInstance.h"

#ifndef _argon_no_includes
    #include <shlwapi.h>            // for DLLGETVERSIONPROC
    #include <stddef.h>             // for offsetof macro
#endif

namespace ArgonLibrary
{




// And now __ImageBase for function #GetCurrentModuleHandle()# taken from delayhlp.cpp.
// This works only on VC++ 7.0 compilers and higher.

#if defined(_M_IA64)
#pragma section(".base", long, read)
extern "C"
__declspec(allocate(".base"))
const IMAGE_DOS_HEADER __ImageBase;
#else
extern "C"
const IMAGE_DOS_HEADER __ImageBase;
#endif

///<summary>
///  With this function, you can retrieve the module or instance
///  handle of the current module (dll or exe file). <br/>
///  <u>Caution:</u> It will work on VC++ 7.0 compilers and higher.
///</summary>
///<remarks>
///  When using the Argon Library, it is recommended
///   to use <b>Get_Module_Instance()</b> and <b>Get_Resource_Instance()</b>
///   if you need the current module's handle.
///  Although, <b>GetCurrentModuleHandle()</b> will work, the other methods are
///   designed to handle specific purposes of the handle and may return
///   another handle if it makes sense. (E.g. if a resource is in some
///   other module, <b>Get_Resource_Instance()</b> could return a different
///   handle if it will be implemented to do so in future versions of
///   the Argon Library).
///</remarks>
HMODULE GetCurrentModuleHandle()
{
    #ifndef _WIN32_WCE

    // Now the HMODULE as I found it somewhere in delayhlp.cpp or _vcclrit.h.
    // This works only on VC++ 7.0 compilers and higher.

    return (HMODULE) (&__ImageBase);

    #else

    // Note: I do not have a code yet to get the current module handle
    // on WindowsCE platforms.

    #ifndef _MFC_VER
    return NULL;
    #else
    AFX_MODULE_STATE *ptr = AfxGetModuleState();
    return ptr ? ptr->m_hCurrentInstanceHandle : NULL;
    #endif

    #endif
}




///<summary> Gets the version of a dll in a packed form, packed by
///  the PACKDLLVERSION macro. </summary>
///<remarks> Interresting dlls could be "Shell32.dll" and "Comctl32.dll". </remarks>
DWORD GetDllVersion( PCTSTR pstrDllName )
{
    DWORD dwVersion = 0;

    HINSTANCE hinstDll = LoadLibrary( pstrDllName );
    if( !hinstDll ) return dwVersion;

    #ifndef _WIN32_WCE
    DLLGETVERSIONPROC pDllGetVersion = (DLLGETVERSIONPROC) GetProcAddress( hinstDll, "DllGetVersion" );
    #else
    DLLGETVERSIONPROC pDllGetVersion = (DLLGETVERSIONPROC) GetProcAddressA( hinstDll, "DllGetVersion" );
    #endif

    if( pDllGetVersion )
    {
        DLLVERSIONINFO dvi;
        memset( &dvi, 0, sizeof(dvi) );
        dvi.cbSize = sizeof(dvi);

        HRESULT hr = (*pDllGetVersion)( &dvi );

        if( SUCCEEDED(hr) )
            dwVersion = PACKDLLVERSION( dvi.dwMajorVersion, dvi.dwMinorVersion );
    }

    FreeLibrary( hinstDll );

    return dwVersion;
}

///<summary> Checks if on the current system in the current application
///  the Windows XP theme support is available and in use (e.g. by
///  a manifest in the resources). </summary>
BOOL IsThemingInAppEnabled()
{
    // Comctl32.dll must have version 6 at minimum
    if( GetDllVersion( _T("Comctl32.dll") ) < PACKDLLVERSION(6, 0) )
        return FALSE;

    // We try to load the Uxtheme.dll
    HINSTANCE hinstThemeDll = LoadLibrary( _T("Uxtheme.dll") );
    if( !hinstThemeDll ) return FALSE;

    typedef BOOL (CALLBACK *t_IsAppThemed)();
    #ifndef _WIN32_WCE
    t_IsAppThemed pfnIsAppThemed = (t_IsAppThemed) GetProcAddress( hinstThemeDll, "IsAppThemed" );
    #else
    t_IsAppThemed pfnIsAppThemed = (t_IsAppThemed) GetProcAddressA( hinstThemeDll, "IsAppThemed" );
    #endif

    BOOL bThemed = pfnIsAppThemed ? pfnIsAppThemed() : FALSE;

    FreeLibrary( hinstThemeDll );

    return bThemed;
}

HRESULT Enable_Theme_Dialog_Texture( HWND hwnd, DWORD dwFlags )
{
    if( !IsThemingInAppEnabled() )
        return E_FAIL;

    // We try to load the Uxtheme.dll
    HINSTANCE hinstThemeDll = LoadLibrary( _T("Uxtheme.dll") );
    if( !hinstThemeDll ) return E_FAIL;

    typedef BOOL (CALLBACK *t_EnableThemeDialogTexture)( HWND hwnd, DWORD dwFlags );
    #ifndef _WIN32_WCE
    t_EnableThemeDialogTexture pfnEnableThemeDialogTexture = (t_EnableThemeDialogTexture) GetProcAddress( hinstThemeDll, "EnableThemeDialogTexture" );
    #else
    t_EnableThemeDialogTexture pfnEnableThemeDialogTexture = (t_EnableThemeDialogTexture) GetProcAddressA( hinstThemeDll, "EnableThemeDialogTexture" );
    #endif

    HRESULT hr = pfnEnableThemeDialogTexture ? pfnEnableThemeDialogTexture( hwnd, dwFlags ) : E_FAIL;

    FreeLibrary( hinstThemeDll );

    return hr;
}




static HMODULE s_hModule = NULL;

class ModuleIniter
{
    public :
        ModuleIniter() { s_hModule = GetCurrentModuleHandle(); }
};

static ModuleIniter s_ModuleIniter;

///<summary>
///  <b>Get_Module_Instance()</b> returns the instance handle of
///  the module (dll or exe) this .cpp file is in.
///</summary>
///<returns> instance handle </returns>
///<remarks>
///  Use this function if you need an module instance handle.
///  But if you want to load resources from this module,
///  it is recommended to use the Get_Resource_Instance() call.
///</remarks>
HINSTANCE Get_Module_Instance()
{
    if( s_hModule ) return s_hModule;
    return s_hModule = GetCurrentModuleHandle();
}

///<summary>
///  <b>Get_Resource_Instance()</b> returns the instance handle
///  of the module (dll or exe) this .cpp file is in when used to load
///  resources from this module.
///  At the moment, this function is the same as Get_Module_Instance(),
///  but this may (or may not) change in the future.
///</summary>
///<returns> instance handle </returns>
HINSTANCE Get_Resource_Instance()
{
    if( s_hModule ) return s_hModule;
    return s_hModule = GetCurrentModuleHandle();
}




///<summary>
///  <b>SingleInstanceDetection()</b> creates the single instance detection
///  object. Pass a name that uniquely shall define the instance throughout
///  the local (session) space or global (system) space.
///</summary>
///<param name="pstrInstanceName"> unique name to identify different instances
///  of the same application or module.
///  <br/>
///  IMPORTANT: This name will be internally used as name for a file mapping object.
///  Do not use this name for other kernel objects especially not for other
///  file mapping objects.
///  <br/><br/>
///  According to the usage of Terminal Services and kernel object names,
///  you can prefix the passed instance name by "Local\" or "Global\"
///  to recognize single instance detection within the local (session) space
///  (e.g. only within the currently logged in user) or within the global (system) space
///  (within all currently logged in users alltogether and the system
///  like for service applications).
///  When no prefix specified, the kernel object name space will be
///  same like "Local\" for applications started in local or session space, or
///  it will be like "Global\" for applications started in global space.
///  See the Terminal Services documentations for more on this. </param>
///<remarks>
///  Initially, the contact window handle will be set to NULL (so no window set).
///</remarks>
SingleInstanceDetection::SingleInstanceDetection( PCTSTR pstrInstanceName )
{
    m_hMemory = ::CreateFileMapping( INVALID_HANDLE_VALUE,
                                     NULL,
                                     PAGE_READWRITE,
                                     0,
                                     sizeof(MemData),
                                     pstrInstanceName );
    m_bMemoryCreated = m_hMemory && GetLastError() != ERROR_ALREADY_EXISTS;

    if( m_bMemoryCreated )
    {
        MemData *p = (MemData*) MapViewOfFile( m_hMemory, FILE_MAP_WRITE, 0, 0, sizeof(MemData) );
        if( p )
        {
            PBYTE pId = (PBYTE) &p->m_hdr.m_wHdrId;
            p->m_hdr.m_wHdrId           = 0;    // will be set at the end of initialization
            p->m_hdr.m_wStructVersion   = 1;
            p->m_hdr.m_cHeader          = sizeof(MemHdr);
            p->m_hdr.m_cInstanceData    = sizeof(SingleInstanceData);
            p->m_hdr.m_ofsInstanceData  = (DWORD) offsetof( MemData, m_dat );

            p->m_dat.Empty();
            p->m_dat.m_dwStructSize     = sizeof(SingleInstanceData);
            p->m_dat.m_dwPid1stInstance = GetCurrentProcessId();

            pId[0] = 'M';
            pId[1] = 'D';

            UnmapViewOfFile( p );
        }
    }
}

SingleInstanceDetection::~SingleInstanceDetection()
{
    if( m_hMemory )
        { CloseHandle( m_hMemory ); m_hMemory = NULL; }
        // Note: The savest way now would to be to clear the m_wHdrId member in the memory
        // before releasing it (which should only the instance to that has the m_bMemoryCreated
        // set to true). But we hope that next time the app is started, the memory is another
        // so that the m_wHdrId is not "MD" yet before it is set explicitely in the ctor.
}



bool SingleInstanceDetection::getInstanceData( SingleInstanceData& dat ) const
{
    dat.Empty();
    if( !m_hMemory ) return false;

    MemHdr *p = (MemHdr*) MapViewOfFile( m_hMemory, FILE_MAP_READ, 0, 0, sizeof(MemHdr) );
    if( !p ) return false;
    WORD hdrid          = p->m_wHdrId;     // we check the header as the first one, because it is initialized as the last item in the ctor. so we do not need to sync the memory access
    PBYTE pId           = (PBYTE) &hdrid;
    bool bValidHeader   = pId[0] == 'M' && pId[1] == 'D';
    int cData           = (int) p->m_cInstanceData;
    int ofsData         = (int) p->m_ofsInstanceData;
    if( p->m_wStructVersion < 1 || (int) p->m_cHeader <= 0 || cData <= 0 || ofsData < 0 ) bValidHeader = false;
    UnmapViewOfFile( p );

    if( !bValidHeader ) return false;

    SingleInstanceData *ptr = (SingleInstanceData*) MapViewOfFile( m_hMemory, FILE_MAP_READ, 0, ofsData, cData );
    if( !ptr ) return false;

    // Now we read the memory.
    // Since only the HWND can be changed, we do not synchronize the access
    // but we compare the HWNDs after copying if they are equal and if not,
    // we repeat the copying. We do this a limited number of times.
    for( int i = 0 ; i < 8 ; ++i )
    {
        CopyMemory( &dat, ptr, cData );

        if( dat.m_hwndContactWindow )
            if( !IsWindow(dat.m_hwndContactWindow) )
                Sleep( 10 );    // We let other threads in the system run so that the HWND is written into the memory finally.

        Sleep( 0 );             // We let other threads in the system run so that the HWND is written into the memory finally.

        if( dat.m_hwndContactWindow == ptr->m_hwndContactWindow )
            break;

        Sleep( 10 );            // We let other threads in the system run so that the HWND is written into the memory finally.
    }

    UnmapViewOfFile( ptr );
    return true;
}

///<summary>
///  <b>setContactWindow()</b> sets the contact window. The contact window
///  can be used to send messages to the first application/module instance.
///  This call will fail if called by an instance of the application/module that
///  is not the first one.
///</summary>
bool SingleInstanceDetection::setContactWindow( HWND hwnd )
{
    if( !m_hMemory ) return false;
    if( !m_bMemoryCreated ) return false;

    MemHdr *p = (MemHdr*) MapViewOfFile( m_hMemory, FILE_MAP_READ, 0, 0, sizeof(MemHdr) );
    if( !p ) return false;
    WORD hdrid          = p->m_wHdrId;     // we check the header as the first one, because it is initialized as the last item in the ctor. so we do not need to sync the memory access
    PBYTE pId           = (PBYTE) &hdrid;
    bool bValidHeader   = pId[0] == 'M' && pId[1] == 'D';
    int cData           = (int) p->m_cInstanceData;
    int ofsData         = (int) p->m_ofsInstanceData;
    if( p->m_wStructVersion < 1 || (int) p->m_cHeader <= 0 || cData <= 0 || ofsData < 0 ) bValidHeader = false;
    UnmapViewOfFile( p );

    if( !bValidHeader ) return false;

    SingleInstanceData *ptr = (SingleInstanceData*) MapViewOfFile( m_hMemory, FILE_MAP_WRITE, 0, ofsData, cData );
    if( !ptr ) return false;

    ptr->m_hwndContactWindow = hwnd;

    UnmapViewOfFile( ptr );
    return true;
}




}
