/*
    File:               argonLinkedList.h
    Description:        Linked List.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         06.04.2005
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonLinkedList_h
#define __argonLinkedList_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4706 )    // assignment within conditional expression

    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




template< class T > class LinkedList;

///<summary> <b>T</b> should be the class name of the
///  objects to handle in the linked list. </summary>
template< class T >
class Node
{
    friend class Node<T>;

    public :
        T *m_prev;
        T *m_next;
        LinkedList<T> *m_pList;

    public :
        Node() { m_prev = m_next = NULL; m_pList = NULL; }
        ~Node() { Unlink(); }

        bool IsLinked() const { return m_pList != NULL; } // if this node has been added to a linked list

        void LinkTo( LinkedList<T> *list ) { list->AddNode( this ); }
        void Unlink()                      { if( m_pList ) m_pList->RemoveNode( this ); }

        bool IsFirst() const { return m_pList ? (m_pList->m_p1 == m_pList->ObjPtr(this)) : false; }
        bool IsLast () const { return m_pList ? (m_pList->m_pL == m_pList->ObjPtr(this)) : false; }

        bool Exchange( Node<T> *node2 ); // can exchange even between different lists or when one is not linked at all

        int GetIndex() const;
};

///<summary> <b>T</b> should be the class name of the
///  objects to handle in the linked list. </summary>
template< class T >
class LinkedList
{
    friend class Node<T>;

    public :
        T *m_p1;            // first and
        T *m_pL;            // last item of the list
    protected :
        int m_nNodeOfs;     // offset to a node relative to the object <T> containing the node

    public :
        LinkedList( int nodeofs ) { m_p1 = m_pL = NULL; m_nNodeOfs = nodeofs; }
        // Note: Set #nodeofs# to the offset of the node to the address of its object.
        // For this, use the #offsetof(class,node)# macro, where #class# should be the
        // class of the objects (= T) and #node# the member name of the node.
        ~LinkedList() { RemoveAll(); } // if ever the list obj is deleted when still
                                       // some nodes are linked -> unlink them all
                                       // on list destruction

        void Add   ( T *p ) { if( p ) AddNode   ( NodePtr( p ) ); }
        void Remove( T *p ) { if( p ) RemoveNode( NodePtr( p ) ); }
        void RemoveAll()    { while( m_pL ) Remove( m_pL ); }

        T *prev( T *p ) const { return p ? NodePtr( p )->m_prev : NULL; }
        T *next( T *p ) const { return p ? NodePtr( p )->m_next : NULL; }

        bool IsEmpty() const { return( !m_p1 ); }
        int  Count  () const;

        T *GetAt( int idx );

    protected :
        Node<T> *NodePtr( const T *p ) const { return (Node<T>*)( (BYTE*)p + m_nNodeOfs ); }
        T *ObjPtr ( const Node<T> *p ) const { return (T*)( (BYTE*)p - m_nNodeOfs ); }

        void AddNode   ( Node<T> *node );
        void RemoveNode( Node<T> *node );
};




///<returns> true only if successfully exchanged. </returns>
template< class T >
bool Node<T>::Exchange( Node<T> *node2 )
{
    Node<T> *node1 = this;
    if( !node1 || !node2 ) return false;

    // If the nodes both are linked,
    // the lists must have same node-byte-offset.
    if( node1->m_pList && node2->m_pList )
        if( node1->m_pList->m_nNodeOfs != node2->m_pList->m_nNodeOfs )
            return false;

    if( node1->m_pList && node2->m_pList )
        if( node1->m_pList->ObjPtr( node1 ) == node2->m_prev )
        {
            // If node2 is directly following node1.
            // Note: Both nodes are already in the same list.

            // Store old pointers.
            T *node1_old_prev = node1->m_prev;
            T *node2_old_next = node2->m_next;

            T *obj1 = node1->m_pList->ObjPtr( node1 );
            T *obj2 = node2->m_pList->ObjPtr( node2 );

            // Set node1 related pointers.
            node1->m_prev = obj2;
            node1->m_next = node2_old_next;
            if( node1->m_next ) node1->m_pList->NodePtr( node1->m_next )->m_prev = obj1;
            else                node1->m_pList->m_pL = obj1;

            // Set node2 related pointers.
            node2->m_prev = node1_old_prev;
            node2->m_next = obj1;
            if( node2->m_prev ) node2->m_pList->NodePtr( node2->m_prev )->m_next = obj2;
            else                node2->m_pList->m_p1 = obj2;

            return true;
        }

    if( node1->m_pList && node2->m_pList )
        if( node1->m_pList->ObjPtr( node1 ) == node2->m_next )
        {
            // If node2 is directly preceding node1.
            // Note: Both nodes are already in the same list.

            // Store old pointers.
            T *node2_old_prev = node2->m_prev;
            T *node1_old_next = node1->m_next;

            T *obj1 = node1->m_pList->ObjPtr( node1 );
            T *obj2 = node2->m_pList->ObjPtr( node2 );

            // Set node1 related pointers.
            node1->m_prev = node2_old_prev;
            node1->m_next = obj2;
            if( node1->m_prev ) node1->m_pList->NodePtr( node1->m_prev )->m_next = obj1;
            else                node1->m_pList->m_p1 = obj1;

            // Set node2 related pointers.
            node2->m_prev = obj1;
            node2->m_next = node1_old_next;
            if( node2->m_next ) node2->m_pList->NodePtr( node2->m_next )->m_prev = obj2;
            else                node2->m_pList->m_pL = obj2;

            return true;
        }

    // If node1 and node2 are not direct neighbours.
    // Note: The nodes are either in the same list,
    //       or in different lists,
    //       or only one of them is in a list,
    //       or none of them is linked to a list.

    // Store all old pointers related to node1.
    T             *node1_old_prev  = node1->m_prev;
    T             *node1_old_next  = node1->m_next;
    LinkedList<T> *node1_old_pList = node1->m_pList;

    // Store all old pointers related to node2.
    T             *node2_old_prev  = node2->m_prev;
    T             *node2_old_next  = node2->m_next;
    LinkedList<T> *node2_old_pList = node2->m_pList;

    // Remove nodes from their current positions.
    node1->Unlink();
    node2->Unlink();

    // Insert node1 at its new position.
    node1->m_prev  = node2_old_prev;
    node1->m_next  = node2_old_next;
    node1->m_pList = node2_old_pList;
    if( node1->m_pList ) {
        T *obj1 = node1->m_pList->ObjPtr( node1 );
        if( node1->m_prev ) node1->m_pList->NodePtr( node1->m_prev )->m_next = obj1;
        else                node1->m_pList->m_p1 = obj1;
        if( node1->m_next ) node1->m_pList->NodePtr( node1->m_next )->m_prev = obj1;
        else                node1->m_pList->m_pL = obj1;
    }

    // Insert node2 at its new position.
    node2->m_prev  = node1_old_prev;
    node2->m_next  = node1_old_next;
    node2->m_pList = node1_old_pList;
    if( node2->m_pList ) {
        T *obj2 = node2->m_pList->ObjPtr( node2 );
        if( node2->m_prev ) node2->m_pList->NodePtr( node2->m_prev )->m_next = obj2;
        else                node2->m_pList->m_p1 = obj2;
        if( node2->m_next ) node2->m_pList->NodePtr( node2->m_next )->m_prev = obj2;
        else                node2->m_pList->m_pL = obj2;
    }

    return true;
}



template< class T >
int Node<T>::GetIndex() const
{
    int idx = 0;
    for( T *p = m_prev ; p ; p = m_prev )
        ++idx;
    return idx;
}




template< class T >
int LinkedList<T>::Count() const
{
    int cnt = 0;
    for( T *p = m_p1 ; p ; p = next( p ) )
        ++cnt;
    return cnt;
}



template< class T >
T *LinkedList<T>::GetAt( int idx )
{
    if( idx < 0 ) return NULL;
    for( T *p = m_p1 ; p ; p = next( p ) )
        if( --idx < 0 ) return p;
    return NULL;
}



template< class T >
void LinkedList<T>::AddNode( Node<T> *node )
{
    if( node->IsLinked() ) return;

    node->m_pList = this;

    node->m_next = NULL;
    if( node->m_prev = m_pL ) NodePtr( node->m_prev )->m_next = ObjPtr( node );

    m_pL = ObjPtr( node );
    if( !m_p1 ) m_p1 = m_pL;
}

template< class T >
void LinkedList<T>::RemoveNode( Node<T> *node )
{
    if( !node->IsLinked() ) return;

    if( node->m_prev ) NodePtr( node->m_prev )->m_next = node->m_next;
    else m_p1 = node->m_next;
    if( node->m_next ) NodePtr( node->m_next )->m_prev = node->m_prev;
    else m_pL = node->m_prev;

    node->m_pList = NULL;
}




}

#endif
