/*
    File:               argonLocking.cpp
    Description:        Useful extensions to locks and synchronization.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonLocking_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonLocking.h"

namespace ArgonLibrary
{




LockModuleWide::Creator LockModuleWide::s_Creator;
bool                    LockModuleWide::s_bDeletingGlobalObjects = false;

#ifdef _argonLocking_use_moduleMutex
HANDLE LockModuleWide::s_hLockMutex = NULL;
#else
CRITICAL_SECTION LockModuleWide::s_crit_sec;
bool             LockModuleWide::s_crit_sec_initialized = false;
#endif

LockModuleWide::Creator::Creator()
{
    // Note: The Creator is used just for one purpose: To create the
    //  mutex somewhen when the global objects are initialized.
    //  For this, we have a static member object #s_Creator#.
    //  But nevertheless, when we need the mutex earlier than the
    //  #s_Creator# object is created, then the constructor of
    //  #LockModuleWide# will create the mutex already then.

    LockModuleWide::Prepare();
}

LockModuleWide::Creator::~Creator()
{
    LockModuleWide::s_bDeletingGlobalObjects = true;
    LockModuleWide::OnExit();
}

void LockModuleWide::Prepare()
{
    if( s_bDeletingGlobalObjects )
        return;

    // Note: Below, #atexit()# is marked out and
    // the Creator-dtor is used instead. This is because
    // when we are in a Dll and in DLL_PROCESS_DETACH,
    // then the exit function registered with #atexit()#
    // will be called first before globa object dtors.
    // Since #Thread_Local_Storage::DeinitTls()# is called
    // by having been registered using #atexit()#, we
    // want to be sure that the #LockModuleWide::OnExit()#
    // is called after #Thread_Local_Storage::DeinitTls()#.

    #ifdef _argonLocking_use_moduleMutex

    if( !s_hLockMutex )
    {
        s_hLockMutex = CreateMutex( NULL, FALSE, NULL );
        //atexit( OnExit );
    }

    #else

    if( !s_crit_sec_initialized )
    {
        ::InitializeCriticalSection( &s_crit_sec );
        s_crit_sec_initialized = true;
        //atexit( OnExit );
    }

    #endif
}

void LockModuleWide::OnExit()
{
    #ifdef _argonLocking_use_moduleMutex

    if( s_hLockMutex )
    {
        HANDLE hmtx = s_hLockMutex;
        s_hLockMutex = NULL;
        CloseHandle( hmtx );
    }

    #else

    if( s_crit_sec_initialized )
    {
        ::DeleteCriticalSection( &s_crit_sec );
        s_crit_sec_initialized = false;
    }

    #endif
}

void LockModuleWide::Lock()
{
    #ifdef _argonLocking_use_moduleMutex

    if( !m_bLocked && s_hLockMutex )
    {
        if( WaitForSingleObject( s_hLockMutex, INFINITE ) == WAIT_OBJECT_0 )
            m_bLocked = true;
    }

    #else

    if( !m_bLocked && s_crit_sec_initialized )
    {
        ::EnterCriticalSection( &s_crit_sec );
        m_bLocked = true;
    }

    #endif
}

void LockModuleWide::Unlock()
{
    #ifdef _argonLocking_use_moduleMutex

    if( m_bLocked )
    {
        if( s_hLockMutex )
            ReleaseMutex( s_hLockMutex );

        m_bLocked = false;
    }

    #else

    if( m_bLocked )
    {
        if( s_crit_sec_initialized )
            ::LeaveCriticalSection( &s_crit_sec );

        m_bLocked = false;
    }

    #endif
}

void LockModuleWide::Enter()
{
    #ifdef _argonLocking_use_moduleMutex

    Prepare();
    if( s_hLockMutex )
        WaitForSingleObject( s_hLockMutex, INFINITE );

    #else

    Prepare();
    if( s_crit_sec_initialized )
        ::EnterCriticalSection( &s_crit_sec );

    #endif
}

void LockModuleWide::Leave()
{
    #ifdef _argonLocking_use_moduleMutex

    if( s_hLockMutex )
        ReleaseMutex( s_hLockMutex );

    #else

    if( s_crit_sec_initialized )
        ::LeaveCriticalSection( &s_crit_sec );

    #endif
}




}
