/*
    File:               argonSync.h
    Description:        Simple synchronization classes, Semaphores.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         26.08.2005
    last updated:       26.10.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argonSync_use_lockerMutex          If you #define this, then Locker
                                        will use a mutex rather than a critical section.
*/

#ifndef __argonSync_h
#define __argonSync_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonArray.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4100 )    // unreferenced parameter

    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




#ifndef _countof
#define _countof(array) (sizeof(array) / sizeof((array)[0]))
#endif

class singleLock;

class SyncObject
{
    friend class singleLock;

    protected :
        HANDLE m_hObject;

    public :
        SyncObject() { m_hObject = NULL; }
        virtual ~SyncObject() { if( m_hObject ) { ::CloseHandle( m_hObject ); m_hObject = NULL; } }

        virtual HANDLE get() const { return m_hObject; }
        operator HANDLE() const { return get(); }

            ///<summary> Used by the <b>SyncWrapper</b> class. </summary>
        virtual SyncObject *deref() { return this; }

            ///<remarks> If you set <b>dwTimeout</b> to zero -> you can check the signalled state
            /// of the syncobject. If True is returned, then the syncobject is signalled. </remarks>
        virtual bool Lock( DWORD dwTimeout = INFINITE )
            { return ::WaitForSingleObject(m_hObject, dwTimeout) == WAIT_OBJECT_0; }
        virtual bool Unlock() = 0;

        bool signalled() { return Lock( 0 ); }
};

class Semaphore : public SyncObject
{
    public :
        #ifndef _WIN32_WCE
        Semaphore( long lInitialCount,     long lMaxCount,     PCSTR  pstrName,        PSECURITY_ATTRIBUTES psaAttributes = NULL );
        Semaphore( long lInitialCount = 1, long lMaxCount = 1, PCWSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #else
        Semaphore( long lInitialCount = 1, long lMaxCount = 1, PCTSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #endif
        virtual ~Semaphore() {  }

        virtual bool Unlock() { return Unlock( 1, NULL ); }
        virtual bool Unlock( long lCount, LONG *pPrevCount = NULL )
            { return ::ReleaseSemaphore( m_hObject, lCount, pPrevCount ) != FALSE; }
};

class Mutex : public SyncObject
{
    public :
        #ifndef _WIN32_WCE
        Mutex( bool bInitiallyOwn,         PCSTR  pstrName,        PSECURITY_ATTRIBUTES psaAttributes = NULL );
        Mutex( bool bInitiallyOwn = false, PCWSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #else
        Mutex( bool bInitiallyOwn = false, PCTSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #endif
        virtual ~Mutex() {  }

        virtual bool Unlock() { return ::ReleaseMutex( m_hObject ) != FALSE; }
};

///<remarks> For an <b>Ev</b> which has Manual-Reset enabled, you can test if it
///  is set by calling <b>Lock(0)</b> with a time-elapse of zero milliseconds.
///  If TRUE is returned then -> the <b>Ev</b> is set to signaled. </remarks>
class Ev : public SyncObject
{
    public :
        #ifndef _WIN32_WCE
        Ev( bool bInitialState,         bool bManualReset,         PCSTR  pstrName,        PSECURITY_ATTRIBUTES psaAttributes = NULL );
        Ev( bool bInitialState = false, bool bManualReset = false, PCWSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #else
        Ev( bool bInitialState = false, bool bManualReset = false, PCTSTR pstrName = NULL, PSECURITY_ATTRIBUTES psaAttributes = NULL );
        #endif
        virtual ~Ev() {  }

        bool SetEvent()   { return m_hObject ? ::SetEvent  ( m_hObject ) != FALSE : false; }
        bool PulseEvent() { return m_hObject ? ::PulseEvent( m_hObject ) != FALSE : false; }
        bool ResetEvent() { return m_hObject ? ::ResetEvent( m_hObject ) != FALSE : false; }

        virtual bool Unlock() { return true; } // Note: Unlocking an #Ev# will do nothing.
};

class CritSec : public SyncObject
{
    protected :
        CRITICAL_SECTION m_crit_sec;
        // IMPORTANT: Because this object stores a CRITICAL_SECTION,
        // you *must never* use #CritSec# objects in #Array<># (or
        // create it using the #new(_size_t_, void *)# operator) because
        // in arrays, the objects can be moved from one to another
        // memory-address, but this is not allowed for initialized
        // CRITICAL_SECTIONs.
        // To explicitely cause a compiler error if a #CritSec# is used
        // in an #Array<>#, we do not implement the copy-ctor and operator=().

    public :
        CritSec() { ::InitializeCriticalSection( &m_crit_sec ); }
        virtual ~CritSec() { ::DeleteCriticalSection( &m_crit_sec ); }

        CRITICAL_SECTION *get() { return &m_crit_sec; }

        virtual bool Lock() { ::EnterCriticalSection( &m_crit_sec ); return true; }
        virtual bool Lock( DWORD dwTimeout ) { return Lock(); }
        virtual bool Unlock() { ::LeaveCriticalSection( &m_crit_sec ); return true; }

    protected :
        CritSec( const CritSec& src );
        CritSec& operator =( const CritSec& src );
};

#ifndef _argonSync_use_lockerMutex
typedef CritSec Locker;
#else
typedef Mutex Locker;
#endif

///<summary>
///  Use this to class to have an object that acts as a SyncObject
///  but in reality references (through a pointer) to another
///  SyncObject. <br/>
///  Beware, the referenced SyncObject stays valid as long as this
///  wrapper is referencing it.
///  <br/><br/>
///  This class allows cascading. You can pass a <b>SyncWrapper</b>
///  object into another <b>SyncWrapper</b> object.
///  This way, it is possible to use <b>SyncWrapper</b> as a wrapper
///  for another SyncObject and treat the <b>SyncWrapper</b> object
///  as a SyncObject itself. This will be useful if you need to have SyncObjects
///  on different locations but reference the same SyncObject in behind.
///</summary>
class SyncWrapper : public SyncObject
{
    protected :
        SyncObject *m_pObject;

    public :
        SyncWrapper() { m_pObject = NULL; }
        SyncWrapper( SyncObject& object ) { m_pObject = &object; }
        virtual ~SyncWrapper() {  }

        void Prepare( SyncObject& object ) { m_pObject = &object; }

        virtual HANDLE get() const
            { return m_pObject ? m_pObject->get() : SyncObject::get(); }

            ///<summary> Gets the SyncObject behind the scenes by dereferencing the wrapper chain. </summary>
        virtual SyncObject *deref()
            { return m_pObject ? m_pObject->deref() : deref(); }

        virtual bool Lock( DWORD dwTimeout = INFINITE ) { return m_pObject ? m_pObject->Lock( dwTimeout ) : false; }
        virtual bool Unlock() { return m_pObject ? m_pObject->Unlock() : false; }
};



///<summary>
///  Use this to cause one of the <b>SyncObject</b>
///  derived classes above to be locked and unlocked.
///  <br/><br/>
///  With creating this object on the stack, the sync object
///  will be unlocked automatically when the scope is left.
///</summary>
class singleLock
{
    protected :
        SyncObject *m_pObject;
        bool        m_bAcquired;

    public :
        singleLock( SyncObject& object, bool bInitialLock = true );
        virtual ~singleLock() { Unlock(); }

        bool Lock( DWORD dwTimeout = INFINITE );
        bool Unlock();
        bool IsLocked() const { return m_bAcquired; }

            ///<summary> <b>_HackAcquired()</b> can be used to make a hardcore
            ///  change of the internally stored acquired (locked) state.
            ///  Treat this with caution. It can be useful for Semaphore objects
            ///  e.g. to just perform an Unlock without first doing a Lock. </summary>
        void _HackAcquired( bool bAcquired ) { m_bAcquired = bAcquired; }
};



///<summary> <b>waitResult</b> can be used to hold and check the return value
///  of one of the following calls:              <br/>
///      <b>WaitForSingleObject()</b>            <br/>
///      <b>WaitForSingleObjectEx()</b>          <br/>
///      <b>WaitForMultipleObjects()</b>         <br/>
///      <b>WaitForMultipleObjectsEx()</b>       <br/>
///      <b>MsgWaitForMultipleObjects()</b>      <br/>
///      <b>MsgWaitForMultipleObjectsEx()</b>    <br/>
///      <b>multiWait::Wait()</b>                </summary>
class waitResult
{
    public :
        DWORD   m_dwResult;
        int     m_cHandles;     // number of handles that was used in the wait call.
        bool    m_bWaitedAll;   // if waiting for all was enabled in the wait call.

    public :
        waitResult( DWORD result, int num_handles, bool bWaitedAll )
            { m_dwResult = result; m_cHandles = num_handles; m_bWaitedAll = bWaitedAll; }

        operator DWORD() { return m_dwResult; }
        DWORD    get() const { return m_dwResult; }

            ///<summary> The wait failed. </summary>
        bool Failed() const { return m_dwResult == WAIT_FAILED; }
            ///<summary> The time-out has been reached while waiting. </summary>
        bool Timedout() const { return m_dwResult == WAIT_TIMEOUT; }

            ///<summary> Indicates one object is signalled (or all if we waited for all). </summary>
        bool AnySignalled() const { return m_dwResult >= WAIT_OBJECT_0 && m_dwResult < (DWORD) (WAIT_OBJECT_0 + m_cHandles); }
            ///<summary> Indicates all objects are signalled when we waited for all. </summary>
        bool AllSignalled() const { return m_bWaitedAll && AnySignalled(); }
            ///<summary> Tells the index of the *only* object signalled. </summary>
        int  IndexSignalled() const { return AnySignalled() && !m_bWaitedAll ? (int) (m_dwResult - WAIT_OBJECT_0) : -1; }

            ///<summary> Tells if one object has been abandoned. If we waited for all objects, then this
            ///  indicates one *or more* objects have been abandoned and the other objects are signalled. </summary>
        bool AnyAbandoned() const { return m_dwResult >= WAIT_ABANDONED_0 && m_dwResult < (DWORD) (WAIT_ABANDONED_0 + m_cHandles); }
            ///<summary> Tells if we waited for all objects, but one or more have been abandoned and the other objects are signalled. </summary>
        bool SomeAbandonedOthersSignalled() const { return AnyAbandoned() && m_bWaitedAll; }
            ///<summary> Tells the index of the *only* object abandoned. </summary>
        int  IndexAbandoned() const { return AnyAbandoned() && !m_bWaitedAll ? (int) (m_dwResult - WAIT_ABANDONED_0) : -1; }

            ///<summary> Tells new input of the type specified in the wait call's
            ///  <b>dwWakeMask</b> parameter is available in the thread's input queue. </summary>
        bool NewInput() const { return m_dwResult == (DWORD) (WAIT_OBJECT_0 + m_cHandles); }

        #ifdef WAIT_IO_COMPLETION
            ///<summary> Tells the wait was ended by one or more user-mode
            ///  asynchronous procedure calls (APC) queued to the thread. </summary>
        bool IoCompletion() const { return m_dwResult == WAIT_IO_COMPLETION; }
        #endif
};



class multiWait
{
    protected :
        Array<HANDLE, 8> m_aHandles;

    public :
        multiWait() { m_aHandles.SetBlockSize( 64 ); }
        multiWait( HANDLE *handles, int count ) { m_aHandles.SetBlockSize( 64 ); Add( handles, count ); }
            ///<summary> For the variable numbers of parameters: First argument is number of
            ///  handles, then the HANDLE values must follow one after each other. </summary>
        multiWait( int count, ... )
            { va_list args; va_start(args, count); m_aHandles.SetBlockSize( 64 );
              for( int i = 0 ; i < count ; ++i ) Add( va_arg(args, HANDLE) );
              va_end(args); }
        ~multiWait() {  }

    // retrieving content:
        int      count() const { return (int) m_aHandles.GetSize(); }
        HANDLE  *get()         { return m_aHandles.Ptr(); }
        operator HANDLE *()    { return m_aHandles.Ptr(); }

    // add handles:
        bool Add( HANDLE h ) { return Add( &h, 1 ); }
        bool Add( HANDLE *handles, int count )
        {
            if( count == 0 ) return true;
            if( !handles || count <= 0 ) return false;

            m_aHandles.Add( handles, count );

            return true;
        }
            ///<summary> For the variable numbers of parameters: First argument is number of
            ///  handles, then the HANDLE values must follow one after each other. </summary>
        bool Add( int count, ... )
        {
            if( count == 0 ) return true;
            if( count <= 0 ) return false;

            va_list args;
            va_start( args, count );

            for( int i = 0 ; i < count ; ++i )
                Add( va_arg(args, HANDLE) );

            va_end( args );
            return true;
        }

    // waiting calls:
            ///<summary> This is equivalent to calling <b>WaitForMultipleObjects()</b>. </summary>
        waitResult Wait( DWORD dwTimeout = INFINITE, bool bWaitAll = false )
        {
            HANDLE *ptr = get();
            if( count() <= 0 || !ptr ) return waitResult( WAIT_FAILED, 0, bWaitAll );

            DWORD ret = ::WaitForMultipleObjects( (DWORD) count(), ptr, bWaitAll, dwTimeout );

            return waitResult( ret, count(), bWaitAll );
        }

            ///<summary> This is equivalent to calling <b>MsgWaitForMultipleObjects()</b>. </summary>
            ///<param name="dwWakeMask"> Same flags as for <b>MsgWaitForMultipleObjects()</b>. </param>
        waitResult MsgWait( DWORD dwTimeout = INFINITE, bool bWaitAll = false, DWORD dwWakeMask = 0 )
        {
            HANDLE *ptr = get();
            if( count() <= 0 || !ptr ) return waitResult( WAIT_FAILED, 0, bWaitAll );

            DWORD ret = ::MsgWaitForMultipleObjects( (DWORD) count(), ptr, bWaitAll, dwTimeout, dwWakeMask );

            return waitResult( ret, count(), bWaitAll );
        }
};




}

#endif
