/*
    File:               argonThreadLocalStorage.h
    Description:        Handling thread local storage more easily.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       02.02.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonThreadLocalStorage_h
#define __argonThreadLocalStorage_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonString.h"
    #include "argonArray.h"
    #include "argonLocking.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{

/*
    The interresting classes and templates for the user:
        tlsPtr              holds a pointer to an object in each thread but does not delete it
        tlsautoPtr          holds a pointer to an object in each thread and deletes it automatically somewhen after the thread has ended
        tlsHandleMapper     maps handles to pointers to objects in each thread
*/




#define TLS_INVALID_INDEX               ( (DWORD) -1 )


// ---------------------------------------------------------------- //
//      thread local storage pointer

/*
    Holds a pointer in thread local storage.
    The object it points to will not be deleted automatically.

    Usage:
    Declare a specific #tlsPtr# object (e.g. in global space).
    It will not have allocated a pointer storage place yet
    (#get()# will return NULL).
    When you call #set()# the first time, a storage place will
    be allocated internally and the #tlsPtr# object from now
    on will access that pointer storage place.
    By calling #set()# again, you overwrite the stored pointer.

    After you have a storage place (so after calling #set()# the
    first time), you can assing the #tlsPtr# object to
    another #tlsPtr# object. The 2nd object will then access
    the same pointer storage place.

    A #tlsPtr# object even can be created as global variable
    if you want to.
*/

class Thread_Local_Storage;

template< class T >
class tlsPtr
{
    friend class Thread_Local_Storage;

    private :
        _idx_ m_Index;      // index into the array of allocated ThreadLocalPointers
        bool  m_hasIndex;   // because it is only one char, we can check and change this even outside LockModuleWide lock.

    public :
        tlsPtr() { m_Index = -1; m_hasIndex = false; }
        tlsPtr( const tlsPtr& src ) { m_Index = src.m_Index; m_hasIndex = src.m_hasIndex; }

        tlsPtr& operator =( const tlsPtr& src ) { m_Index = src.m_Index; m_hasIndex = src.m_hasIndex; return *this; }

        tlsPtr& operator =( T *ptr ) { set( ptr ); return *this; }
        tlsPtr& set( T *ptr );
        T      *get() const;

        operator T *() { return get(); }
        T& operator *() { return *get(); }
        T *operator ->() { return get(); }
        const T *operator ->() const { return get(); }

        bool isAllocated() const { return m_hasIndex; } // if storage place is allocated already
};

template< class T >
tlsPtr<T>& tlsPtr<T>::set( T *ptr )
{
    if( !m_hasIndex )
    {
        LockModuleWide lmw;

        if( !m_hasIndex )
        {
            Thread_Local_Storage::AllocThreadLocalPointer( *this );
            if( !m_hasIndex ) return *this;
        }
    }

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return *this;

    p->m_aThreadLocalPointers.SetAt( m_Index, ptr, true );

    return *this;
}

template< class T >
T *tlsPtr<T>::get() const
{
    if( !m_hasIndex ) return NULL;

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return NULL;

    return (T*) p->m_aThreadLocalPointers[ m_Index ];
}


// ----------------------------------------------------------------
//      thread local storage pointer with auto-delete-operation

/*
    Holds a pointer in thread local storage pointing to
    an object.
    Such an object will be deleted automatically at the end
    of the thread it belongs to, or somewhen later,
    but not later than when exiting the process.

    The further handling and usage of the #tlsautoPtr#
    class is the same as for #tlsPtr#.
    See remarks there.
*/

class ThreadLocalAutoBase
{
    public :
        ThreadLocalAutoBase() {  }
        virtual ~ThreadLocalAutoBase() {  }
};

template< class T >
class ThreadLocalAutoDeleter : public ThreadLocalAutoBase
{
    public :
        T *m_ptr;

        ThreadLocalAutoDeleter( T *ptr ) { m_ptr = ptr; }
        virtual ~ThreadLocalAutoDeleter() { if( m_ptr ) delete m_ptr; }
};

template< class T >
class tlsautoPtr
{
    friend class Thread_Local_Storage;

    private :
        _idx_ m_Index;      // index into the array of allocated ThreadLocalAutoObjectPointers
        bool  m_hasIndex;   // because it is only one char, we can check and change this even outside LockModuleWide lock.

    public :
        tlsautoPtr() { m_Index = -1; m_hasIndex = false; }
        tlsautoPtr( const tlsautoPtr& src ) { m_Index = src.m_Index; m_hasIndex = src.m_hasIndex; }

        tlsautoPtr& operator =( const tlsautoPtr& src ) { m_Index = src.m_Index; m_hasIndex = src.m_hasIndex; return *this; }

        tlsautoPtr& operator =( T *ptr ) { set( ptr ); return *this; }
        tlsautoPtr& set( T *ptr );
        T          *get() const;

        operator T *() { return get(); }
        T& operator *() { return *get(); }
        T *operator ->() { return get(); }
        const T *operator ->() const { return get(); }

        bool isAllocated() const { return m_hasIndex; } // if storage place is allocated already
};

///<summary> By calling this method, you pass the object pointed
///  to by <b>ptr</b> over to this <b>tlsautoPtr</b> object.
///  You must not delete the object pointed to by <b>ptr</b>
///  by yourself. The thread local storage handler will make this for
///  you. The object (<b>ptr</b>) will be deleted not earlier than
///  when the thread it belongs to has ended. When the thread has
///  ended, somewhen later the object will be deleted. This will
///  be done by another thread or the last thread. </summary>
template< class T >
tlsautoPtr<T>& tlsautoPtr<T>::set( T *ptr )
{
    if( !m_hasIndex )
    {
        LockModuleWide lmw;

        if( !m_hasIndex )
        {
            Thread_Local_Storage::AllocThreadLocalAutoObjectPointer( *this );
            if( !m_hasIndex ) return *this;
        }
    }

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return *this;

    if( p->m_aThreadLocalAutoObjectPointers[m_Index] )
        delete p->m_aThreadLocalAutoObjectPointers[m_Index];

    p->m_aThreadLocalAutoObjectPointers.SetAt( m_Index, new ThreadLocalAutoDeleter<T>(ptr), true );

    return *this;
}

template< class T >
T *tlsautoPtr<T>::get() const
{
    if( !m_hasIndex ) return NULL;

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return NULL;

    ThreadLocalAutoDeleter<T> *ptr = (ThreadLocalAutoDeleter<T>*) p->m_aThreadLocalAutoObjectPointers[ m_Index ];
    return ptr ? (T*) ptr->m_ptr : NULL;
}


// ----------------------------------------------------------------
//      thread local storage handle to pointer mapper

/*
    The #tlsHandleMapper# can be used to map a handle
    (e.g HWND) to a pointer (e.g. a C++ window object).

    Usage:
    Declare a #tlsHandleMapper# object specifying a type name.
    The name will be used to identify different maps.
    Then call #set()# and #get()# whenever you want. It will store the
    pointer related to the specified handle within each thread separately.

    In the ctor of #tlsHandleMapper#, you will have to specify
    the name of the handle-map. If you specify a different name, then
    the handles will be handled in a different table.

    A #tlsHandleMapper# object is mainly meant to be
    created as global variable. But you don't need to.
*/

struct TLHM_Handle
{
    INT_PTR m_id;
    HANDLE  m_handle;

    TLHM_Handle() {  }
    TLHM_Handle( const TLHM_Handle& h )
        { m_id = h.m_id; m_handle = h.m_handle; }

    TLHM_Handle& operator =( const TLHM_Handle& h )
        { m_id = h.m_id; m_handle = h.m_handle; return *this; }

    bool operator ==( const TLHM_Handle& h ) const
        { return m_id == h.m_id && m_handle == h.m_handle; }

    _hashval_t_ EvalHashKey() const
        { return (_hashval_t_) ((((UINT_PTR) m_handle) >> 3) + m_id); }
};

template< class Thandle, class Tobject >
class tlsHandleMapper
{
    private :
        String  m_strHandleTypeName; // name of the handle map to use
        INT_PTR m_id;

    public :
        tlsHandleMapper( PCTSTR pstrHandleTypeName );

        tlsHandleMapper& set( Thandle handle, Tobject *pointer );
        Tobject         *get( Thandle handle ) const;

        bool hasMappings() const;
};

template< class Thandle, class Tobject >
tlsHandleMapper<Thandle, Tobject>::tlsHandleMapper( PCTSTR pstrHandleTypeName )
{
    m_strHandleTypeName = pstrHandleTypeName;
    m_id                = -1;
}

template< class Thandle, class Tobject >
tlsHandleMapper<Thandle, Tobject>& tlsHandleMapper<Thandle, Tobject>::set( Thandle handle, Tobject *pointer )
{
    if( handle == NULL ) return *this;

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return *this;

    if( m_id < 0 )
    {
        m_id = p->AllocHandleNameID( m_strHandleTypeName );
        if( m_id < 0 ) return *this;
    }

    TLHM_Handle h;
    h.m_id     = m_id;
    h.m_handle = (HANDLE) handle;

    if( pointer ) p->m_htHandleMapper.SetAt( h, pointer );
    else          p->m_htHandleMapper.RemoveAt( h );

    return *this;
}

template< class Thandle, class Tobject >
Tobject *tlsHandleMapper<Thandle, Tobject>::get( Thandle handle ) const
{
    if( m_id < 0 ) return NULL;

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return NULL;

    TLHM_Handle h;
    h.m_id     = m_id;
    h.m_handle = (HANDLE) handle;

    return (Tobject*) p->m_htHandleMapper.GetAt( h );
}

///<summary> Returns TRUE if currently there are handles
///  being mapped in the table. </summary>
template< class Thandle, class Tobject >
bool tlsHandleMapper<Thandle, Tobject>::hasMappings() const
{
    if( m_id < 0 ) return false;

    Thread_Local_Storage *p = Thread_Local_Storage::GetStorage();
    if( !p )
        return false;

    return p->m_htHandleMapper.Count() > 0;
}


// ----------------------------------------------------------------
//      Thread local storage

///<summary>
///  This class is used for thread local storage.
///
///  Usage:
///  #Thread_Local_Storage::GetStorage()# will return a pointer to a
///  #Thread_Local_Storage# object that represents the current thread.
///
///  Each thread will have its own #Thread_Local_Storage# object.
///  It will be created only when #Thread_Local_Storage::GetStorage()#
///  is called the first time within a thread.
///</summary>
class Thread_Local_Storage
{
    template<class> friend class tlsPtr;
    template<class> friend class tlsautoPtr;
    template<class, class> friend class tlsHandleMapper;

    private :
        DWORD   m_idThrd;
        HANDLE  m_hInternalThrd; // an internal handle to the thread so that we
                                 // can internally check if the thread ended already

    public :
        Thread_Local_Storage();
        ~Thread_Local_Storage();
        // CAUTION: This object can be deleted even by another thread
        //  than the thread that created it or that it belongs to,
        //  whichever is the last thread running.

        DWORD GetThreadID() const { return m_idThrd; }
        bool  ThreadHasEnded() const;


    // Thread local storage internals
    private :
        struct Creator
        {
            Creator();
        };

        static Creator s_Creator;

        typedef Array<Thread_Local_Storage*, 16> TlsPtrArray;

        static bool s_bDeinited;
        static DWORD s_dwTlsIdx;
        static TlsPtrArray *s_paTlsPtrs;

    public :
        static Thread_Local_Storage *GetStorage();

    private :
        static void DeleteEndedThreadStorageObjs();
        static bool InitTls();
        static void DeinitTls();


    // thread local storage pointers
    private :
        // Array of pointers that the user can use by himself
        // for storing other things in thread local storage.
        Array<PVOID, 16> m_aThreadLocalPointers;
        // and the index counter
        static _idx_ s_ThreadLocalIndexCounter;
    private :
        template< class T >
        static void AllocThreadLocalPointer( tlsPtr<T>& tlsp );


    // thread local storage pointers with auto-delete-operation
    private :
        Array<ThreadLocalAutoBase*, 16> m_aThreadLocalAutoObjectPointers;
        // and the index counter
        static _idx_ s_ThreadLocalAutoObjectIndexCounter;
    private :
        template< class T >
        static void AllocThreadLocalAutoObjectPointer( tlsautoPtr<T>& tlsap );
        void DeleteThreadLocalAutoObjects();


    // thread local storage handle to pointer mapper
    private :
        HashTable<TLHM_Handle, PVOID> m_htHandleMapper;
        Array<String>                 m_aHandleNames;
    private :
        INT_PTR AllocHandleNameID( PCTSTR pstrHandleName );
};

template< class T >
void Thread_Local_Storage::AllocThreadLocalPointer( tlsPtr<T>& tlsp )
{
    LockModuleWide::Enter();
    _idx_ idx = s_ThreadLocalIndexCounter++;
    LockModuleWide::Leave();

    tlsp.m_Index    = idx;
    tlsp.m_hasIndex = true;
}

template< class T >
void Thread_Local_Storage::AllocThreadLocalAutoObjectPointer( tlsautoPtr<T>& tlsap )
{
    LockModuleWide::Enter();
    _idx_ idx = s_ThreadLocalAutoObjectIndexCounter++;
    LockModuleWide::Leave();

    tlsap.m_Index    = idx;
    tlsap.m_hasIndex = true;
}




}

#endif
