/*
    File:               argonTrayicon.cpp
    Description:        Class for easy tray icon handling.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         17.11.2003
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonTrayicon_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonTrayicon.h"

namespace ArgonLibrary
{




Trayicon::Trayicon( UINT idTrayicon )
{
    m_idTrayicon            = idTrayicon;
    m_bTrayiconExists       = false;
    m_hwndNotify            = NULL;

    memset( &m_nid, 0, sizeof(m_nid) );
    m_hIcon                 = NULL;
    m_autodelHicon          = false;

    m_autoPopupMenu         = false;
    m_pstrMenuname_or_resid = NULL;
    m_nSubmenu              = 0;
    m_idDefaultMenuitem     = 0;
    m_dwPopupmenuFlags      = traypopupflags_Default;
}



bool Trayicon::CreateIcon( HWND hwndNotify, UINT idIcon, UINT idMsg, PCTSTR pstrTip )
{
    if( m_bTrayiconExists ) return false;

    HICON hicon = (HICON) LoadImage( Get_Resource_Instance(),
                                     MAKEINTRESOURCE( idIcon ),
                                     IMAGE_ICON,
                                     GetSystemMetrics(SM_CXSMICON),
                                     GetSystemMetrics(SM_CYSMICON),
                                     LR_DEFAULTCOLOR );
    if( !hicon )
        return false;

    bool ok = CreateIcon( hwndNotify, hicon, idMsg, pstrTip );

    if( ok ) m_autodelHicon = true;
    else     DestroyIcon( hicon );

    return ok;
}

///<remarks> IMPORTANT: <b>hIcon</b> should exist as long as the icon is used.
///  Normally this would not be necessary. It is only necessary if you want
///  <b>RecreateIcon()</b> to work (e.g. when the taskbar is restarted). </remarks>
bool Trayicon::CreateIcon( HWND hwndNotify, HICON hIcon, UINT idMsg, PCTSTR pstrTip )
{
    if( m_bTrayiconExists ) return false;
    if( !hwndNotify || !hIcon ) return false;

    if( hIcon != m_hIcon )
        DelHicon();

    NOTIFYICONDATA nid;
    memset( &nid, 0, sizeof(nid) );
    nid.cbSize      = sizeof(nid);
    nid.hWnd        = hwndNotify;
    nid.uID         = m_idTrayicon;
    nid.uFlags      = NIF_MESSAGE | NIF_ICON | ( pstrTip ? NIF_TIP : 0 );
    nid.uCallbackMessage = idMsg;
    nid.hIcon       = hIcon;

    if( pstrTip )
    {
        #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
        _tcsncpy_s( nid.szTip, _countof(nid.szTip), pstrTip, _TRUNCATE );
        #else
        _tcsncpy( nid.szTip, pstrTip, _countof(nid.szTip) - 1 );  // null-terminator has been set already before above in #memset()#.
        #endif
    }

    bool ok = __Shell_NotifyIcon( NIM_ADD, &nid );

    if( ok )
    {
        m_bTrayiconExists = true;
        m_hwndNotify      = hwndNotify;

        m_nid             = nid;
        m_hIcon           = hIcon;
        m_autodelHicon    = false;
    }

    return ok;
}

bool Trayicon::DeleteIcon()
{
    if( !m_bTrayiconExists ) return false;

    NOTIFYICONDATA nid;
    memset( &nid, 0, sizeof(nid) );
    nid.cbSize      = sizeof(nid);
    nid.hWnd        = m_hwndNotify;
    nid.uID         = m_idTrayicon;

    __Shell_NotifyIcon( NIM_DELETE, &nid );

    m_bTrayiconExists = false;
    m_hwndNotify      = NULL;

    memset( &m_nid, 0, sizeof(m_nid) );

    m_autoPopupMenu   = false;

    return true;
}

bool Trayicon::SetIcon( UINT idIcon, PCTSTR pstrTip )
{
    if( !m_bTrayiconExists ) return false;

    HICON hicon = (HICON) LoadImage( Get_Resource_Instance(),
                                     MAKEINTRESOURCE( idIcon ),
                                     IMAGE_ICON,
                                     GetSystemMetrics(SM_CXSMICON),
                                     GetSystemMetrics(SM_CYSMICON),
                                     LR_DEFAULTCOLOR );
    if( !hicon )
        return false;

    bool ok = SetIcon( hicon, pstrTip );

    if( ok ) m_autodelHicon = true;
    else     DestroyIcon( hicon );

    return ok;
}

///<remarks> IMPORTANT: <b>hIcon</b> should exist as long as the icon is used.
///  Normally this would not be necessary. It is only necessary if you want
///  <b>RecreateIcon()</b> to work (e.g. when the taskbar is restarted). </remarks>
bool Trayicon::SetIcon( HICON hIcon, PCTSTR pstrTip )
{
    if( !m_bTrayiconExists ) return false;
    if( !hIcon ) return false;

    if( hIcon != m_hIcon )
        DelHicon();

    NOTIFYICONDATA nid;
    memset( &nid, 0, sizeof(nid) );
    nid.cbSize      = sizeof(nid);
    nid.hWnd        = m_hwndNotify;
    nid.uID         = m_idTrayicon;
    nid.uFlags      = NIF_ICON | ( pstrTip ? NIF_TIP : 0 );
    nid.hIcon       = hIcon;

    if( pstrTip )
    {
        #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
        _tcsncpy_s( nid.szTip, _countof(nid.szTip), pstrTip, _TRUNCATE );
        #else
        _tcsncpy( nid.szTip, pstrTip, _countof(nid.szTip) - 1 );  // null-terminator has been set already before above in #memset()#.
        #endif
    }

    bool ok = __Shell_NotifyIcon( NIM_MODIFY, &nid );

    if( ok )
    {
        m_nid.hWnd      = nid.hWnd;
        m_nid.uID       = nid.uID;
        m_nid.uFlags   |= nid.uFlags;
        m_nid.hIcon     = nid.hIcon;
        if( pstrTip )
            memcpy( m_nid.szTip, nid.szTip, sizeof(m_nid.szTip) );
        m_hIcon         = hIcon;
        m_autodelHicon  = false;
    }

    return ok;
}

bool Trayicon::SetTip( PCTSTR pstrTip )
{
    if( !m_bTrayiconExists ) return false;

    NOTIFYICONDATA nid;
    memset( &nid, 0, sizeof(nid) );
    nid.cbSize      = sizeof(nid);
    nid.hWnd        = m_hwndNotify;
    nid.uID         = m_idTrayicon;
    nid.uFlags      = NIF_TIP;
    nid.szTip[0]    = 0;

    if( pstrTip )
    {
        #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
        _tcsncpy_s( nid.szTip, _countof(nid.szTip), pstrTip, _TRUNCATE );
        #else
        _tcsncpy( nid.szTip, pstrTip, _countof(nid.szTip) - 1 );  // null-terminator has been set already before above in #memset()#.
        #endif
    }

    bool ok = __Shell_NotifyIcon( NIM_MODIFY, &nid );

    if( ok )
    {
        m_nid.hWnd      = nid.hWnd;
        m_nid.uID       = nid.uID;
        m_nid.uFlags   |= nid.uFlags;
        memcpy( m_nid.szTip, nid.szTip, sizeof(m_nid.szTip) );
    }

    return ok;
}

bool Trayicon::SetFocus()
{
    if( !m_bTrayiconExists ) return false;

    NOTIFYICONDATA nid;
    memset( &nid, 0, sizeof(nid) );
    nid.cbSize      = sizeof(nid);
    nid.hWnd        = m_hwndNotify;
    nid.uID         = m_idTrayicon;

    return __Shell_NotifyIcon( NIM_SETFOCUS, &nid );
}

bool Trayicon::RecreateIcon()
{
    if( !m_bTrayiconExists ) return true;

    __Shell_NotifyIcon( NIM_DELETE, &m_nid );
    return m_bTrayiconExists = __Shell_NotifyIcon( NIM_ADD, &m_nid );
}



///<remarks> IMPORTANT: If <b>pstrMenuname</b> points to a string rather
///  than it is a resource id, then you, the user, must ensure to
///  hold the string memory valid as long as this Trayicon ojbect
///  exists or until you set another popup menu. </remarks>
///<param name="idDefaultMenuitem"> If != 0 then this is the menuitem id
///  that shall be invoked automatically when the tray icon is double
///  clicked using left mouse button. <br/>
///  In addition, the menuitem will be displayed bold in the popup menu
///  when the tray icon is right clicked. <br/>
///  -1 = no default menuitem </param>
///<param name="flags"> combination of <b>trayPopupFlags_t</b> values </param>
void Trayicon::SetAutoPopupMenu( PCTSTR pstrMenuname, int nSubmenu, UINT idDefaultMenuitem, DWORD flags )
{
    m_autoPopupMenu         = pstrMenuname != NULL;
    m_pstrMenuname_or_resid = pstrMenuname;
    m_nSubmenu              = nSubmenu;
    m_idDefaultMenuitem     = idDefaultMenuitem;
    m_dwPopupmenuFlags      = flags;
}



void Trayicon::openPopupMenu()
{
    if( !m_bTrayiconExists || !m_hwndNotify || !m_autoPopupMenu || !m_pstrMenuname_or_resid )
        return;

    HWND hwnd = m_hwndNotify;

    if( m_dwPopupmenuFlags & traypopupflags_setForeground )
        SetForegroundWindow( hwnd );    // Is necessary so that the PopupMenu disappears when the mouse is clicked outside of the PopupMenu

    if( !(m_dwPopupmenuFlags & traypopupflags_allowDuringModalDlg) )
        if( !IsWindowEnabled( hwnd ) ) return;

    POINT m;
    GetCursorPos( &m );
    //DWORD dwM = GetMessagePos();
    //m.x = LOWORD(dwM);
    //m.y = HIWORD(dwM);

    // Now create PopupMenu (for SysTray Icon) and set default menu item (will be bold)
    HMENU hmenu = LoadMenu( Get_Resource_Instance(), m_pstrMenuname_or_resid );
    HMENU hPopupmenu = GetSubMenu( hmenu, m_nSubmenu );

    if( !hmenu || !hPopupmenu ) return;

    if( m_idDefaultMenuitem != 0 )
        SetMenuDefaultItem( hPopupmenu, m_idDefaultMenuitem, FALSE );

    TrackPopupMenuEx( hPopupmenu,
                      TPM_LEFTALIGN,
                      m.x,
                      m.y,
                      hwnd,
                      NULL );

    DestroyMenu( hmenu );

    if( m_dwPopupmenuFlags & traypopupflags_setFocusToTrayicon )
        SetFocus();

    PostMessage( hwnd, WM_NULL, 0, 0 ); // recommended by VC-help in the case, the popup menu is created from a tray icon
}



bool Trayicon::__Shell_NotifyIcon( DWORD dwMessage, PNOTIFYICONDATA pData )
{
    if( !pData ) return false;

    // If we have an older Shell32.dll version, we use
    // the old structure instead.

    DWORD dwPackedVersion = GetDllVersion( _T("Shell32.dll") );

    if( dwPackedVersion < PACKDLLVERSION(5, 0) )
        if( pData->cbSize > (DWORD)NOTIFYICONDATA_V1_SIZE )
        {
            pData->cbSize = (DWORD)NOTIFYICONDATA_V1_SIZE;
            pData->szTip[63] = 0;   // shorten tip text
        }

    #if( _WIN32_IE >= 0x0600 )
    if( dwPackedVersion < PACKDLLVERSION(6, 0) )
        if( pData->cbSize > (DWORD)NOTIFYICONDATA_V2_SIZE )
        {
            pData->cbSize = (DWORD)NOTIFYICONDATA_V2_SIZE;
        }
    #endif

    return Shell_NotifyIcon( dwMessage, pData ) != FALSE;
}

void Trayicon::DelHicon()
{
    if( m_autodelHicon && m_hIcon )
        { ::DestroyIcon(m_hIcon); m_hIcon = NULL; }
}



UINT Trayicon::s_wmTaskbarRestart = 0;

UINT Trayicon::getTaskbarRestartMessage()
{
    if( !s_wmTaskbarRestart )
        s_wmTaskbarRestart = RegisterWindowMessage( _T("TaskbarCreated") );

    return s_wmTaskbarRestart;
}



bool Trayicon::HandleWndMessage( LRESULT& lr, HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    if( hWnd != m_hwndNotify || !hWnd || !m_bTrayiconExists ) return false;

    // recreate icon if taskbar has restarted
    if( message == getTaskbarRestartMessage() )
    {
        RecreateIcon();
    }

    // handle auto-popup-menu
    if( message == m_nid.uCallbackMessage && m_autoPopupMenu )
    {
        if( (UINT) wParam == getTrayiconID() )
            switch( (UINT) lParam )
            {
                case WM_LBUTTONDBLCLK :
                {
                    if( m_idDefaultMenuitem == 0 ) break;

                    if( m_dwPopupmenuFlags & traypopupflags_setForeground )
                        SetForegroundWindow( hWnd );
                        // If we wouldnt bring it to foreground, a modal dlg
                        // would not be created in foreground but in background,
                        // So without bringing anything to foreground.

                    if( !(m_dwPopupmenuFlags & traypopupflags_allowDuringModalDlg) )
                        if( !IsWindowEnabled( hWnd ) ) break;   // First: If a modal dialog is present -> ignore the message.

                    // Double Click shall execute the default menu item.
                    PostMessage( hWnd, WM_COMMAND, MAKEWPARAM(m_idDefaultMenuitem, 0), 0 );
                    // Sending would also work, but I post it just for
                    // nothing else than to be on the safe side.
                    break;
                }

                case WM_CONTEXTMENU :
                case WM_RBUTTONUP   :
                {
                    openPopupMenu();
                    break;
                }
            }
    }

    switch( message )
    {
        case WM_DESTROY :
        {
            DeleteIcon();
            break;
        }
    }

    return false;
}




}
