/*
    File:               argonUser32Dll.h
    Description:        Extensions and emulations of parts of the
                        system's USER32.DLL
                        useful mainly to support system features that
                        are not or may be not present on all available
                        operating system versions out there.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         9.11.2003
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonUser32Dll_h
#define __argonUser32Dll_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




/*
    -----------------------------------------------------------------
    Usage:

    Simply call the functions #UpdateLayeredWindow()#, #SetLayeredWindowAttributes()#,
    #GetLayeredWindowAttributes()# etc. in your source code as you need it.

    All these functions defined in here are emulating the system behavior
    that way, that they load the functions during runtime from the system
    and only if this did succeed, they will be called.

    The first time you call any of these functions, the library will try
    to be loaded and the function pointers will try to be retrieved.

    The unloading of the library will be done automatically
    during the deleting of the global variables.
    If you included this file in a DLL project, you will not want this
    behavior to happen because it should be prevented to call FreeLibrary()
    when a DLL is detached from a process (where global variables will be
    deleted).
    So in the case of DLL project, it is recommended to export an own
    DLL-deinit-function which you should redirect to a call to
    #User32Dll::UnloadAPI()#. If the user then calls this deinit-function,
    the USER32.DLL will be unloaded and later on, when your DLL is detachted
    from a process, it is harmless to call #User32Dll::UnloadAPI()# again
    because it will not do anything any more.

    -----------------------------------------------------------------
    Additional, new functions:

    SetLayeredWindowExstyle
*/


#define USER32DLL_NAME              "USER32.DLL"


// ____ Layered Windows ___________________________________
//      UpdateLayeredWindow
//      SetLayeredWindowAttributes
//      GetLayeredWindowAttributes

#ifndef WS_EX_LAYERED
    #define WS_EX_LAYERED           0x00080000

    #ifndef AC_SRC_ALPHA
    #define AC_SRC_ALPHA            0x00000001
    #endif

    #define LWA_COLORKEY            0x00000001
    #define LWA_ALPHA               0x00000002

    #define ULW_COLORKEY            0x00000001
    #define ULW_ALPHA               0x00000002
    #define ULW_OPAQUE              0x00000004
#endif

// Note: The following functions I cannot import by
// using the extern command because this would require
// the DLL to contain the functions at program start
// and on some Windows versions this would fail and
// therefore refuse the application to execute.
// So we must import the functions during runtime.

/*
EXTERN_C WINUSERAPI BOOL WINAPI UpdateLayeredWindow(
    HWND hwnd,             // handle to layered window
    HDC hdcDst,            // handle to screen DC
    POINT *pptDst,         // new screen position
    SIZE *psize,           // new size of the layered window
    HDC hdcSrc,            // handle to surface DC
    POINT *pptSrc,         // layer position
    COLORREF crKey,        // color key
    BLENDFUNCTION *pblend, // blend function
    DWORD dwFlags          // options
);

EXTERN_C WINUSERAPI BOOL WINAPI SetLayeredWindowAttributes(
    HWND hwnd,             // handle to the layered window
    COLORREF crKey,        // specifies the color key
    BYTE bAlpha,           // value for the blend function
    DWORD dwFlags          // action
);

EXTERN_C WINUSERAPI BOOL WINAPI GetLayeredWindowAttributes(
    HWND hwnd,
    COLORREF *pcrKey,
    BYTE *pbAlpha,
    DWORD *pdwFlags
);
*/

#if defined (__cplusplus)
extern "C" {
#endif

typedef BOOL (WINAPI *TfnUpdateLayeredWindow)(
    HWND hwnd,             // handle to layered window
    HDC hdcDst,            // handle to screen DC
    POINT *pptDst,         // new screen position
    SIZE *psize,           // new size of the layered window
    HDC hdcSrc,            // handle to surface DC
    POINT *pptSrc,         // layer position
    COLORREF crKey,        // color key
    BLENDFUNCTION *pblend, // blend function
    DWORD dwFlags          // options
);

BOOL WINAPI argonUpdateLayeredWindow(
    HWND hwnd,             // handle to layered window
    HDC hdcDst,            // handle to screen DC
    POINT *pptDst,         // new screen position
    SIZE *psize,           // new size of the layered window
    HDC hdcSrc,            // handle to surface DC
    POINT *pptSrc,         // layer position
    COLORREF crKey,        // color key
    BLENDFUNCTION *pblend, // blend function
    DWORD dwFlags          // options
);

typedef BOOL (WINAPI *TfnSetLayeredWindowAttributes)(
    HWND hwnd,             // handle to the layered window
    COLORREF crKey,        // specifies the color key
    BYTE byAlpha,          // value for the blend function
    DWORD dwFlags          // action
);

BOOL WINAPI argonSetLayeredWindowAttributes(
    HWND hwnd,             // handle to the layered window
    COLORREF crKey,        // specifies the color key
    BYTE byAlpha,          // value for the blend function
    DWORD dwFlags          // action
);

typedef BOOL (WINAPI *TfnGetLayeredWindowAttributes)(
    HWND hwnd,
    COLORREF *pcrKey,
    BYTE *pbyAlpha,
    DWORD *pdwFlags
);

BOOL WINAPI argonGetLayeredWindowAttributes(
    HWND hwnd,
    COLORREF *pcrKey,
    BYTE *pbyAlpha,
    DWORD *pdwFlags
);

typedef DWORD (WINAPI *TfnGetGuiResources)(
    IN HANDLE hProcess,
    IN DWORD uiFlags
);

DWORD WINAPI argonGetGuiResources(
    IN HANDLE hProcess,
    IN DWORD uiFlags
);

#define UpdateLayeredWindow         ArgonLibrary::argonUpdateLayeredWindow
#define SetLayeredWindowAttributes  ArgonLibrary::argonSetLayeredWindowAttributes
#define GetLayeredWindowAttributes  ArgonLibrary::argonGetLayeredWindowAttributes
#define GetGuiResources             ArgonLibrary::argonGetGuiResources

// additional, new functions

///<summary> Sets the WS_EX_LAYERED ex-style bit, but only if layered windows
///  are supported on the system. </summary>
///<remarks> This should be called before calling <b>UpdateLayeredWindow()</b>
///  or <b>SetLayeredWindowAttributes()</b>. </remarks>
///<param name="bSet"> If TRUE -> sets the style bit, if FALSE -> clears the style bit. </param>
///<param name="dwSwitchAlso"> If you want to switch also additional ex-styles together
///  with the WS_EX_LAYERED (e.g. WS_EX_TRANSPARENT), then you can specify them here.
///  They will be switched on or off depending on <b>bSet</b>. </param>
BOOL SetLayeredWindowExstyle( HWND hwnd, BOOL bSet = TRUE, DWORD dwSwitchAlso = 0 );

#if defined (__cplusplus)
}
#endif


class User32Dll
{
    protected :
        class Unloader
        {
            public :
                Unloader() {  }
                ~Unloader() { if( User32Dll::s_bAutoUnload ) User32Dll::UnloadAPI(); }
        };

        static Unloader s_Unloader;

    public :
        static HMODULE s_hModule;
        static TfnUpdateLayeredWindow        s_pfnUpdateLayeredWindow;
        static TfnSetLayeredWindowAttributes s_pfnSetLayeredWindowAttributes;
        static TfnGetLayeredWindowAttributes s_pfnGetLayeredWindowAttributes;
        static TfnGetGuiResources            s_pfnGetGuiResources;
        static bool s_bAutoUnload;  // true = automatically unloads the Dll when global objects are destructed.

        User32Dll();

        static bool IsLayeredWindowsSupported();

        static void LoadAPI();
        static void UnloadAPI();

    protected :
        static void Nullify();
};




}

#endif
