/*
    File:               argonString.cpp
    Description:        Contains string handling extensions.
                        Global functions as well as extensions to the
                        ATL::CStringT<> template (and therefore derived
                        MFC classes CString, CStringA, CStringW).
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         5.10.2003
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonString_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonString.h"

#ifndef _argon_no_warning_disables
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
    #pragma warning( disable: 4996 )    // warning C4996: 'sscanf' was declared deprecated
    // Note: Usage of #sscanf()# and #swscanf()# are no security problem in here because
    // we do not use any "%s" sequences in the format strings.
    // Further, we use #_vsnprintf()# and #_vsnwprintf()# for dest buffer limitation handling
    // so in our case here, there is no security issue.
    #endif
#endif

namespace ArgonLibrary
{




namespace privateString
{
    #ifndef _argonString_no_Initer
    Initer g_Initer;
    #endif

    Initer::Initer()
    {
        // switch to system's default locale
        // using #_set_default_locale_()#.
        #ifndef _WIN32_WCE
        _set_default_locale_();
        #endif
    }
}




path_t g_PathType = PathT_Default;




#ifndef _WIN32_WCE

///<summary> Sets the C-runtime's locale to
///  the system's default user-locale, except for numbers. </summary>
///<remarks> The advantage using the system's default user-locale
///  is that functions like <b>toupper</b>, <b>tolower</b>,
///  <b>strupr()</b>, ... will also handle umlaute, french
///  chars etc. In addition, the system's default time format
///  will be used for C-runtime time-string output. And so on. </remarks>
void _set_default_locale_()
{
    // Now we switch to the system's default user-locale.
    // This is the locale that the system holds when you use the
    // LOCALE_USER_DEFAULT in #GetLocaleInfo()# or when you
    // call #GetUserDefaultLCID()#.
    _setlocale_( LC_ALL, _T("") );
    // Note: To switch back call:   _setlocale_( LC_ALL, _T("C") );

    // Note: If you have switched to "C" locale, I am not sure
    // what the operating system's equivalent define number is.
    // It could be LOCALE_NEUTRAL or LOCALE_INVARIANT, but I am
    // not sure.

    // For numeric values, we want the default C handling
    // so that the decimal point is always a '.' instead
    // of ',' (like in german areas) or whatever.
    _setlocale_( LC_NUMERIC, _T("C") );
    // Note: This does not influence the monetary/currency settings.
}

///<summary> Sets the C-runtime's locale to
///  the generic C handling. </summary>
void _set_C_locale_()
{
    _setlocale_( LC_NUMERIC, _T("C") );
}



StoreLocale::~StoreLocale()
{
    if( m_bRestoreOnEnd ) RestoreNow();

    if( m_pstrLocale )
        { free( m_pstrLocale ); m_pstrLocale = NULL; }
}



bool StoreLocale::StoreNow()
{
    if( m_pstrLocale )
        { free( m_pstrLocale ); m_pstrLocale = NULL; }

    PCTSTR pstr = _setlocale_( LC_ALL, (PCTSTR) NULL );
    if( _isstrempty_( pstr ) )
        return false;

    m_pstrLocale = _tcsdup( pstr );
    if( !m_pstrLocale )
        return false;

    return true;
}

bool StoreLocale::RestoreNow()
{
    if( !m_pstrLocale ) return false;

    _setlocale_( LC_ALL, m_pstrLocale );

    return true;
}

#endif



bool _startswithdigit_( PCSTR string, bool bAllowNegativeNumber )
{
    if( !string ) return false;
    if( bAllowNegativeNumber && *string == '-' ) ++string;
    return _isdigit_( *string );
}

bool _startswithdigit_( PCWSTR string, bool bAllowNegativeNumber )
{
    if( !string ) return false;
    if( bAllowNegativeNumber && *string == L'-' ) ++string;
    return _isdigit_( *string );
}



///<summary> If this char belongs to command word (a command word is a word similar
///  to a C/C++-command which may contain alpha chars or a '_' char). </summary>
bool _iscommand_( CHAR c )
{
    return( (c >= 'A' && c <= 'Z') ||
            (c >= 'a' && c <= 'z') ||
             c == '_' );
}

///<summary> If this char belongs to command word (a command word is a word similar
///  to a C/C++-command which may contain alpha chars or a '_' char). </summary>
bool _iscommand_( WCHAR c )
{
    return( (c >= L'A' && c <= L'Z') ||
            (c >= L'a' && c <= L'z') ||
             c == L'_' );
}

///<summary> This one checks if the char is a command char
///  (like in <b>_iscommand_()</b>) or a digit number. </summary>
bool _iscmdnum_( CHAR c )
{
    return( (c >= 'A' && c <= 'Z') ||
            (c >= 'a' && c <= 'z') ||
             c == '_' ||
            (c >= '0' && c <= '9') );
}

///<summary> This one checks if the char is a command char
///  (like in <b>_iscommand_()</b>) or a digit number. </summary>
bool _iscmdnum_( WCHAR c )
{
    return( (c >= L'A' && c <= L'Z') ||
            (c >= L'a' && c <= L'z') ||
             c == L'_' ||
            (c >= L'0' && c <= L'9') );
}

///<summary> If the char is the *first* char of a
///  decimal floating point value (even negative). </summary>
bool _isfloatvalue_( CHAR c )
{
    return( (c >= '0' && c <= '9') ||
             c == '-' || c == '.' );
}

///<summary> If the char is the *first* char of a
///  decimal floating point value (even negative). </summary>
bool _isfloatvalue_( WCHAR c )
{
    return( (c >= L'0' && c <= L'9') ||
             c == L'-' || c == L'.' );
}

///<summary> If the char is another than the *first*
///  char of a decimal floating point value (even negative). </summary>
bool _isstillfloatvalue_( CHAR c )
{
    return( (c >= '0' && c <= '9') ||
             c == 'e' || c == 'E' || c == '.' );
}

///<summary> If the char is another than the *first*
///  char of a decimal floating point value (even negative). </summary>
bool _isstillfloatvalue_( WCHAR c )
{
    return( (c >= L'0' && c <= L'9') ||
             c == L'e' || c == L'E' || c == L'.' );
}



double _stof_( PCSTR string )
{
    if( !string ) return 0.0;
    return atof( string );
}

double _stof_( PCWSTR string )
{
    if( !string ) return 0.0;

    #ifndef _WIN32_WCE
    return _wtof( string );
    #else
    // Note: In case of WindowsCE (at least 5.0), _wtof() does not
    // exist. So to emulate, we convert into ascii (and hope not
    // to have need for special unicode characters) and convert
    // that into float.
    char s[512];
    _strconv_( s, string, _countof(s) );
    return atof( s );
    #endif
}

int _stoi_( PCSTR string )
{
    if( !string ) return 0;
    return atoi( string );
}

int _stoi_( PCWSTR string )
{
    if( !string ) return 0;
    return _wtoi( string );
}

INT64 _stoi64_( PCSTR string )
{
    if( !string ) return 0;
    return _atoi64( string );
}

INT64 _stoi64_( PCWSTR string )
{
    if( !string ) return 0;
    return _wtoi64( string );
}

long _stol_( PCSTR string )
{
    if( !string ) return 0;
    return atol( string );
}

long _stol_( PCWSTR string )
{
    if( !string ) return 0;
    return _wtol( string );
}

///<summary> This will use sscanf to check for a
///  decimal, hexadecimal (C syntax) or octal (when beginning with 0) number. </summary>
int _stoI_( PCSTR string )
{
    if( !string ) return 0;
    int n;
    if( sscanf( string, "%i", &n ) != 1 ) return 0;
    return n;
}

int _stoI_( PCWSTR string )
{
    if( !string ) return 0;
    int n;
    if( swscanf( string, L"%i", &n ) != 1 ) return 0;
    return n;
}

///<summary> This will use sscanf to check for a
///  hexadecimal (C syntax or without prefix) number. </summary>
int _stox_( PCSTR string )
{
    if( !string ) return 0;
    int n;
    if( sscanf( string, "%x", &n ) != 1 ) return 0;
    return n;
}

int _stox_( PCWSTR string )
{
    if( !string ) return 0;
    int n;
    if( swscanf( string, L"%x", &n ) != 1 ) return 0;
    return n;
}



///<summary> Length of a string.
///  Is equivalent to C-runtime <b>strlen()</b>. </summary>
_size_t_ _strlen_( PCSTR string )
{
    if( !string ) return 0;
    _idx_ i = 0;
    while( *(string++) )
        i++;
    return (_size_t_) i;
}

///<summary> Length of a string.
///  Is equivalent to C-runtime <b>wcslen()</b>. </summary>
_size_t_ _strlen_( PCWSTR string )
{
    if( !string ) return 0;
    _idx_ i = 0;
    while( *(string++) )
        i++;
    return (_size_t_) i;
}

///<summary> Length of a string.
///  Is equivalent to C-runtime <b>strlen()</b>. </summary>
_size_t_ _strlen_unaligned_( const CHAR UNALIGNED *string )
{
    if( !string ) return 0;
    _idx_ i = 0;
    while( *(string++) )
        i++;
    return (_size_t_) i;
}

///<summary> Length of a string.
///  Is equivalent to C-runtime <b>wcslen()</b>. </summary>
_size_t_ _strlen_unaligned_( const WCHAR UNALIGNED *string )
{
    if( !string ) return 0;
    _idx_ i = 0;
    while( *(string++) )
        i++;
    return (_size_t_) i;
}

///<summary> Is similar to <b>_strlen_()</b>, but will count only till <b>max</b>
///  chars and if the string is *larger* than <b>max</b>, -1 will be returned.<br/>
///  In the case, <b>string</b> is exactly <b>max</b> chars large, then
///  <b>max</b> will be returned (and not -1). </summary>
///<remarks> This function will only access a maximum of <b>max + 1</b>
///  characters (including terminating null char) of the string <b>string</b>
///  if the param <b>dont_exceed</b> is false. <br/>
///  If <b>dont_exceed</b> is true, a maximum of <b>max</b> characters
///  will be scanned (and *never* more). In that case, -1 will *never*
///  be returned. Only a value greaterequal 0 and lessequal <b>max</b>
///  will be returned. </remarks>
_ssize_t_ _strlenmax_( PCSTR string, _size_t_ max, bool dont_exceed )
{
    if( !string ) return 0;
    _idx_ i = 0;
    if( dont_exceed ) {
        while( i < (_idx_)max ? *(string++) != (CHAR) 0 : false )
            i++;
    } else {
        while( i <= (_idx_)max ? *(string++) != (CHAR) 0 : false )
            i++;
        if( i > (_idx_)max ) i = -1;
    }
    return (_ssize_t_) i;
}

///<summary> Is similar to <b>_strlen_()</b>, but will count only till <b>max</b>
///  chars and if the string is *larger* than <b>max</b>, -1 will be returned.<br/>
///  In the case, <b>string</b> is exactly <b>max</b> chars large, then
///  <b>max</b> will be returned (and not -1). </summary>
///<remarks> This function will only access a maximum of <b>max + 1</b>
///  characters (including terminating null char) of the string <b>string</b>
///  if the param <b>dont_exceed</b> is false. <br/>
///  If <b>dont_exceed</b> is true, a maximum of <b>max</b> characters
///  will be scanned (and *never* more). In that case, -1 will *never*
///  be returned. Only a value greaterequal 0 and lessequal <b>max</b>
///  will be returned. </remarks>
_ssize_t_ _strlenmax_( PCWSTR string, _size_t_ max, bool dont_exceed )
{
    if( !string ) return 0;
    _idx_ i = 0;
    if( dont_exceed ) {
        while( i < (_idx_)max ? *(string++) != (WCHAR) 0 : false )
            i++;
    } else {
        while( i <= (_idx_)max ? *(string++) != (WCHAR) 0 : false )
            i++;
        if( i > (_idx_)max ) i = -1;
    }
    return (_ssize_t_) i;
}

///<summary> Is similar to <b>_strlen_()</b>, but will count only till <b>max</b>
///  chars and if the string is *larger* than <b>max</b>, -1 will be returned.<br/>
///  In the case, <b>string</b> is exactly <b>max</b> chars large, then
///  <b>max</b> will be returned (and not -1). </summary>
///<remarks> This function will only access a maximum of <b>max + 1</b>
///  characters (including terminating null char) of the string <b>string</b>
///  if the param <b>dont_exceed</b> is false. <br/>
///  If <b>dont_exceed</b> is true, a maximum of <b>max</b> characters
///  will be scanned (and *never* more). In that case, -1 will *never*
///  be returned. Only a value greaterequal 0 and lessequal <b>max</b>
///  will be returned. </remarks>
_ssize_t_ _strlenmax_unaligned_( const CHAR UNALIGNED *string, _size_t_ max, bool dont_exceed )
{
    if( !string ) return 0;
    _idx_ i = 0;
    if( dont_exceed ) {
        while( i < (_idx_)max ? *(string++) != (CHAR) 0 : false )
            i++;
    } else {
        while( i <= (_idx_)max ? *(string++) != (CHAR) 0 : false )
            i++;
        if( i > (_idx_)max ) i = -1;
    }
    return (_ssize_t_) i;
}

///<summary> Is similar to <b>_strlen_()</b>, but will count only till <b>max</b>
///  chars and if the string is *larger* than <b>max</b>, -1 will be returned.<br/>
///  In the case, <b>string</b> is exactly <b>max</b> chars large, then
///  <b>max</b> will be returned (and not -1). </summary>
///<remarks> This function will only access a maximum of <b>max + 1</b>
///  characters (including terminating null char) of the string <b>string</b>
///  if the param <b>dont_exceed</b> is false. <br/>
///  If <b>dont_exceed</b> is true, a maximum of <b>max</b> characters
///  will be scanned (and *never* more). In that case, -1 will *never*
///  be returned. Only a value greaterequal 0 and lessequal <b>max</b>
///  will be returned. </remarks>
_ssize_t_ _strlenmax_unaligned_( const WCHAR UNALIGNED *string, _size_t_ max, bool dont_exceed )
{
    if( !string ) return 0;
    _idx_ i = 0;
    if( dont_exceed ) {
        while( i < (_idx_)max ? *(string++) != (WCHAR) 0 : false )
            i++;
    } else {
        while( i <= (_idx_)max ? *(string++) != (WCHAR) 0 : false )
            i++;
        if( i > (_idx_)max ) i = -1;
    }
    return (_ssize_t_) i;
}

///<summary> like <b>strncpy()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in CHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
PSTR _strcpy_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size < 1 ) return dest;
    PCSTR s = srce;
    PSTR  d = dest;
    _size_t_ cnt = size - 1;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> like <b>wcsncpy()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in WCHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
PWSTR _strcpy_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size < 1 ) return dest;
    PCWSTR s = srce;
    PWSTR  d = dest;
    _size_t_ cnt = size - 1;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> like <b>strncpy()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in CHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
CHAR UNALIGNED *_strcpy_unaligned_( CHAR UNALIGNED *dest, const CHAR UNALIGNED *srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size < 1 ) return dest;
    const CHAR UNALIGNED *s = srce;
    CHAR UNALIGNED       *d = dest;
    _size_t_ cnt = size - 1;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> like <b>wcsncpy()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in WCHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
WCHAR UNALIGNED *_strcpy_unaligned_( WCHAR UNALIGNED *dest, const WCHAR UNALIGNED *srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size < 1 ) return dest;
    const WCHAR UNALIGNED *s = srce;
    WCHAR UNALIGNED       *d = dest;
    _size_t_ cnt = size - 1;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> like <b>strncat()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in CHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
PSTR _strcat_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size <= 1 ) return dest;
    PCSTR s = srce;
    PSTR  d = dest;
    _size_t_ cnt = size - 1, i;
    for( i = 0 ; (i < cnt) && *d ; ++i, ++d );
    if( i >= cnt ) return dest;
    for( ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> like <b>wcsncat()</b>, but <b>size</b> must be the size
///  of the <b>dest</b>-buffer in WCHARs and it will always be set a terminating 0. </summary>
///<param name="dest"> Destination string </param>
///<param name="srce"> Source string </param>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> returns <b>dest</b> </returns>
PWSTR _strcat_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    if( size <= 1 ) return dest;
    PCWSTR s = srce;
    PWSTR  d = dest;
    _size_t_ cnt = size - 1, i;
    for( i = 0 ; (i < cnt) && *d ; ++i, ++d );
    if( i >= cnt ) return dest;
    for( ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
            return dest;
    }
    *d = 0;
    return dest;
}

///<summary> Similar to <b>_strcat_()</b>, but
///  only adds one character to the <b>dest</b> string. </summary>
PSTR _straddch_( PSTR dest, CHAR ch, _size_t_ size )
{
    CHAR s[2] = { ch, 0 };
    return _strcat_( dest, s, size );
}

///<summary> Similar to <b>_strcat_()</b>, but
///  only adds one character to the <b>dest</b> string. </summary>
PWSTR _straddch_( PWSTR dest, WCHAR ch, _size_t_ size )
{
    WCHAR s[2] = { ch, 0 };
    return _strcat_( dest, s, size );
}

///<summary> The char <b>ch</b> will be appended multiple times
///  to the string <b>dest</b> till it has reached the
///  length specified by <b>length</b>. </summary>
///<remarks> NOTE: Of course the buffer pointed to by <b>dest</b>
///  must be <b>length + 1</b> CHARs large in size. </remarks>
PSTR _straddmultch_( PSTR dest, CHAR ch, _size_t_ length )
{
    if( !dest || !ch ) return dest;
    PSTR d = dest;
    if( length <= 0 ) return dest;
    for( ; (length > 0) && *d ; ++d, --length );
    for( ; length > 0 ; --length )
        *(d++) = ch;
    *d = 0;
    return dest;
}

///<summary> The char <b>ch</b> will be appended multiple times
///  to the string <b>dest</b> till it has reached the
///  length specified by <b>length</b>. </summary>
///<remarks> NOTE: Of course the buffer pointed to by <b>dest</b>
///  must be <b>length + 1</b> WCHARs large in size. </remarks>
PWSTR _straddmultch_( PWSTR dest, WCHAR ch, _size_t_ length )
{
    if( !dest || !ch ) return dest;
    PWSTR d = dest;
    if( length <= 0 ) return dest;
    for( ; (length > 0) && *d ; ++d, --length );
    for( ; length > 0 ; --length )
        *(d++) = ch;
    *d = 0;
    return dest;
}

///<summary> Returns the length of a string list buffer but *WITHOUT* any of the two finishing
///  '\0' chars included. So the complete size of that buffer in CHARs (including all '\0' chars)
///  would be the returned value added with 2. </summary>
///<remarks> NOTE ON STRING-LISTS:
///  A string list can contain one or more strings separated by single '\0' chars.
///  The end of a string list is defined by two '\0' chars.
///  This automatically includes the fact that a string of a string list cannot be an
///  empty string (because this would have to be stored as two following '\0' chars
///  which would indicate the end of the stringlist).
///  IMPORTANT: It would be theoretically possible for a string list to start
///   with one single '\0' char followed by strings. This case is illegal for a string list
///   and will be handled as if this first '\0' char were not included in the string list.
///   But of course, <b>_strlstlen_()</b> will return a length with that first single '\0' char
///   included. </remarks>
_size_t_ _strlstlen_( PCSTR stringlist )
{
    if( !stringlist ) return 0;
    _size_t_ i = 0;
    while( stringlist[i] || stringlist[i+1] )
        i++;
    return i;
}

///<summary> Returns the length of a string list buffer but *WITHOUT* any of the two finishing
///  '\0' chars included. So the complete size of that buffer in WCHARs (including all '\0' chars)
///  would be the returned value added with 2. </summary>
///<remarks> NOTE ON STRING-LISTS:
///  A string list can contain one or more strings separated by single '\0' chars.
///  The end of a string list is defined by two '\0' chars.
///  This automatically includes the fact that a string of a string list cannot be an
///  empty string (because this would have to be stored as two following '\0' chars
///  which would indicate the end of the stringlist).
///  IMPORTANT: It would be theoretically possible for a string list to start
///   with one single '\0' char followed by strings. This case is illegal for a string list
///   and will be handled as if this first '\0' char were not included in the string list.
///   But of course, <b>_strlstlen_()</b> will return a length with that first single '\0' char
///   included. </remarks>
_size_t_ _strlstlen_( PCWSTR stringlist )
{
    if( !stringlist ) return 0;
    _size_t_ i = 0;
    while( stringlist[i] || stringlist[i+1] )
        i++;
    return i;
}

///<summary> like <b>_strcpy_()</b> but for string lists. </summary>
///<param name="dest"> will contain a string list </param>
///<param name="srce"> must be a valid string list </param>
///<param name="size"> size of <b>dest</b> buffer in CHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: For explanations of a string list,
///  see remarks to <b>_strlstlen_()</b>. </remarks>
PSTR _strlstcpy_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    PCSTR s = srce;
    PSTR  d = dest;
    if( size < 2 ) return dest;
    _size_t_ cnt = size - 2;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
          if( !*s ) {
              *d = 0;
              return dest;
          }
    }
    d[0] = d[1] = 0;
    return dest;
}

///<summary> like <b>_strcpy_()</b> but for string lists. </summary>
///<param name="dest"> will contain a string list </param>
///<param name="srce"> must be a valid string list </param>
///<param name="size"> size of <b>dest</b> buffer in WCHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: For explanations on string lists,
///  see remarks to <b>_strlstlen_()</b>. </remarks>
PWSTR _strlstcpy_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    PCWSTR s = srce;
    PWSTR  d = dest;
    if( size < 2 ) return dest;
    _size_t_ cnt = size - 2;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) )
          if( !*s ) {
              *d = 0;
              return dest;
          }
    }
    d[0] = d[1] = 0;
    return dest;
}

///<summary> like <b>_strcat_()</b> but for string lists. </summary>
///<param name="dest"> must be a valid string list </param>
///<param name="srce"> must be a valid string list </param>
///<param name="size"> size of <b>dest</b> buffer in CHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: For explanations on string lists,
///  see remarks to <b>_strlstlen_()</b>. </remarks>
PSTR _strlstcat_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    _size_t_ len = _strlstlen_( dest );
    _size_t_ ofs = ( len == 0 )? 0 : len + 1;
    PSTR d = dest + ofs;
    _size_t_ cnt = size - (_size_t_)ofs;
    _strlstcpy_( d, srce, cnt );
    return dest;
}

///<summary> like <b>_strcat_()</b> but for string lists. </summary>
///<param name="dest"> must be a valid string list </param>
///<param name="srce"> must be a valid string list </param>
///<param name="size"> size of <b>dest</b> buffer in WCHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: For explanations on string lists,
///  see remarks to <b>_strlstlen_()</b>. </remarks>
PWSTR _strlstcat_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( dest == srce || !dest || !srce ) return dest;
    _size_t_ len = _strlstlen_( dest );
    _size_t_ ofs = ( len == 0 )? 0 : len + 1;
    PWSTR d = dest + ofs;
    _size_t_ cnt = size - (_size_t_)ofs;
    _strlstcpy_( d, srce, cnt );
    return dest;
}

///<summary> Returns the number of strings in a string list </summary>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
_size_t_ _cstrlststrs_( PCSTR stringlist )
{
    if( !stringlist ) return 0;
    _size_t_ cnt = 0;
    _idx_    i   = 0;
    // Note: As described in the info on string lists in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present. So it will be ignored in #_cstrlststrs_()#.
    while( stringlist[i] || stringlist[i+1] )
        if( !stringlist[++i] )
            cnt++;
    return cnt;
}

///<summary> Returns the number of strings in a string list </summary>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
_size_t_ _cstrlststrs_( PCWSTR stringlist )
{
    if( !stringlist ) return 0;
    _size_t_ cnt = 0;
    _idx_    i   = 0;
    // Note: As described in the info on string lists in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present. So it will be ignored in #_cstrlststrs_()#.
    while( stringlist[i] || stringlist[i+1] )
        if( !stringlist[++i] )
            cnt++;
    return cnt;
}

///<summary> Copyies a string from a string list into the destination buffer.
///  If the index indicates a string that cannot be found,
///  <b>deststr</b> will be set to an empty string. </summary>
///<param name="deststr"> will contain the final string </param>
///<param name="srcestrlst"> must be a valid string list </param>
///<param name="index"> index (beginning with 0) of the string to retrieve </param>
///<param name="size"> size of <b>deststr</b> buffer in CHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststr</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PSTR _getstrlststr_( PSTR deststr, PCSTR srcestrlst, _idx_ index, _size_t_ size )
{
    if( deststr == srcestrlst || !deststr || !srcestrlst ) return deststr;
    PCSTR s = srcestrlst;
    if( size < 1 ) return deststr;
    *deststr = 0;
    _idx_ i = 0;
    // Note: As described in the info on string lists in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present.
    while( *s || s[1] ) {
        if( *s && i == index )
            return _strcpy_( deststr, s, size );
        if( !*(++s) )
            i++;
    }
    return deststr;
}

///<summary> Copyies a string from a string list into the destination buffer.
///  If the index indicates a string that cannot be found,
///  <b>deststr</b> will be set to an empty string. </summary>
///<param name="deststr"> will contain the final string </param>
///<param name="srcestrlst"> must be a valid string list </param>
///<param name="index"> index (beginning with 0) of the string to retrieve </param>
///<param name="size"> size of <b>deststr</b> buffer in WCHARs (incl. terminating zeros).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststr</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PWSTR _getstrlststr_( PWSTR deststr, PCWSTR srcestrlst, _idx_ index, _size_t_ size )
{
    if( deststr == srcestrlst || !deststr || !srcestrlst ) return deststr;
    PCWSTR s = srcestrlst;
    if( size < 1 ) return deststr;
    *deststr = 0;
    _idx_ i = 0;
    // Note: As described in the info on string lists in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present.
    while( *s || s[1] ) {
        if( *s && i == index )
            return _strcpy_( deststr, s, size );
        if( !*(++s) )
            i++;
    }
    return deststr;
}

///<summary> Gets the pointer to a string in the string list.
///  If the index indicates a string that cannot be found, NULL will be returned. </summary>
///<param name="stringlist"> must be a valid string list </param>
///<param name="index"> index (beginning with 0) of the string to find </param>
///<returns> pointer of string found, or<br/>
///  NULL = the index indicates a string that cannot be found. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PCSTR _get_strlst_str_ptr_( PCSTR stringlist, _idx_ index )
{
    if( !stringlist ) return NULL;
    PCSTR s = stringlist;
    _idx_ i = 0;
    // Note: As described in the info on string list in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present.
    while( *s || s[1] ) {
        if( *s && i == index )
            return s;
        if( !*(++s) )
            i++;
    }
    return NULL;
}

///<summary> Gets the pointer to a string in the string list.
///  If the index indicates a string that cannot be found, NULL will be returned. </summary>
///<param name="stringlist"> must be a valid string list </param>
///<param name="index"> index (beginning with 0) of the string to find </param>
///<returns> pointer of string found, or<br/>
///  NULL = the index indicates a string that cannot be found. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PCWSTR _get_strlst_str_ptr_( PCWSTR stringlist, _idx_ index )
{
    if( !stringlist ) return NULL;
    PCWSTR s = stringlist;
    _idx_ i = 0;
    // Note: As described in the info on string list in #_strlstlen_()#,
    //  a string list that starts with a '\0' char but the next is not a '\0' char
    //  (which means that strings are following in the list) will be handled as
    //  if that char were not present.
    while( *s || s[1] ) {
        if( *s && i == index )
            return s;
        if( !*(++s) )
            i++;
    }
    return NULL;
}

///<returns> Returns NULL if end of string-list has been reached. </returns>
PCSTR _fwd_to_next_strlst_str_( PCSTR curr_strlst_string )
{
    if( !curr_strlst_string ) return NULL;

    PCSTR s = curr_strlst_string;

    // go to end of current string
    while( *s )
        ++s;
    // skip null terminator
    ++s;
    // check for end of string-list
    if( !*s )
        return NULL;

    return s;
}

///<returns> Returns NULL if end of string-list has been reached. </returns>
PCWSTR _fwd_to_next_strlst_str_( PCWSTR curr_strlst_string )
{
    if( !curr_strlst_string ) return NULL;

    PCWSTR s = curr_strlst_string;

    // go to end of current string
    while( *s )
        ++s;
    // skip null terminator
    ++s;
    // check for end of string-list
    if( !*s )
        return NULL;

    return s;
}

///<summary> Creates a string list which only contains the specified string by copying it
///  into the destination buffer. </summary>
///<param name="deststrlst"> will be a string list with only the specified string included. </param>
///<param name="srcestr"> must be a valid string<br/>
///  NOTE: <b>deststrlst</b> and <b>srcestr</b> can be equal pointers.
///   In that case, a string list will be created from the string
///   if <b>size</b> is large enough. </param>
///<param name="size"> size of <b>deststrlst</b> buffer in CHARs (incl. terminating zero).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststrlst</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PSTR _strlststrcreate_( PSTR deststrlst, PCSTR srcestr, _size_t_ size )
{
    if( size < 2 || !deststrlst || !srcestr ) return deststrlst;
    if( deststrlst == srcestr ) {
        int len = (int)_strlen_( srcestr );
        if( len <= (int)size - 2 ) deststrlst[len+1] = 0;
        return deststrlst;
    }
    PCSTR s = srcestr;
    PSTR  d = deststrlst;
    _size_t_ cnt = size - 2;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) ) {
            *d = 0;
            return deststrlst;
        }
    }
    d[0] = d[1] = 0;
    return deststrlst;
}

///<summary> Creates a string list which only contains the specified string by copying it
///  into the destination buffer. </summary>
///<param name="deststrlst"> will be a string list with only the specified string included. </param>
///<param name="srcestr"> must be a valid string<br/>
///  NOTE: <b>deststrlst</b> and <b>srcestr</b> can be equal pointers.
///   In that case, a string list will be created from the string
///   if <b>size</b> is large enough. </param>
///<param name="size"> size of <b>deststrlst</b> buffer in WCHARs (incl. terminating zero).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststrlst</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PWSTR _strlststrcreate_( PWSTR deststrlst, PCWSTR srcestr, _size_t_ size )
{
    if( size < 2 || !deststrlst || !srcestr ) return deststrlst;
    if( deststrlst == srcestr ) {
        int len = (int)_strlen_( srcestr );
        if( len <= (int)size - 2 ) deststrlst[len+1] = 0;
        return deststrlst;
    }
    PCWSTR s = srcestr;
    PWSTR  d = deststrlst;
    _size_t_ cnt = size - 2;
    for( _size_t_ i = 0 ; i < cnt ; i++ ) {
        if( !(*(d++) = *(s++)) ) {
            *d = 0;
            return deststrlst;
        }
    }
    d[0] = d[1] = 0;
    return deststrlst;
}

///<summary> Adds a string to an existing string list. The string will be added as new
///  string at the end of the list. </summary>
///<param name="deststrlst"> will be a string list with only the specified string included. </param>
///<param name="srcestr"> must be a valid string </param>
///<param name="size"> size of <b>deststrlst</b> buffer in CHARs (incl. terminating zero).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststrlst</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PSTR _strlststradd_( PSTR deststrlst, PCSTR srcestr, _size_t_ size )
{
    if( deststrlst == srcestr || !deststrlst || !srcestr ) return deststrlst;
    int len = (int)_strlstlen_( deststrlst );
    int ofs = ( len == 0 )? 0 : len + 1;
    PSTR d = deststrlst + ofs;
    int cnt = (int)size - ofs;
    _strlststrcreate_( d, srcestr, cnt );
    return deststrlst;
}

///<summary> Adds a string to an existing string list. The string will be added as new
///  string at the end of the list. </summary>
///<param name="deststrlst"> will be a string list with only the specified string included. </param>
///<param name="srcestr"> must be a valid string </param>
///<param name="size"> size of <b>deststrlst</b> buffer in WCHARs (incl. terminating zero).
///  Must not be smaller than 2 or else nothing will be done. </param>
///<returns> Returns <b>deststrlst</b>. </returns>
///<remarks> NOTE: For explanations of a string list, see remarks to <b>_strlstlen_()</b>. </remarks>
PWSTR _strlststradd_( PWSTR deststrlst, PCWSTR srcestr, _size_t_ size )
{
    if( deststrlst == srcestr || !deststrlst || !srcestr ) return deststrlst;
    int len = (int)_strlstlen_( deststrlst );
    int ofs = ( len == 0 )? 0 : len + 1;
    PWSTR d = deststrlst + ofs;
    int cnt = (int)size - ofs;
    _strlststrcreate_( d, srcestr, cnt );
    return deststrlst;
}

///<summary> This is like tries to find one of the strings
///  that is exactly <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-sensitive. </remarks>
_idx_ _findstrlststr_( PCSTR string, PCSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCSTR curr_str = strlst;

    while( curr_str )
    {
        if( _stris_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that is exactly <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-sensitive. </remarks>
_idx_ _findstrlststr_( PCWSTR string, PCWSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCWSTR curr_str = strlst;

    while( curr_str )
    {
        if( _stris_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that is exactly <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-INsensitive. </remarks>
_idx_ _findistrlststr_( PCSTR string, PCSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCSTR curr_str = strlst;

    while( curr_str )
    {
        if( _striis_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that is exactly <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-INsensitive. </remarks>
_idx_ _findistrlststr_( PCWSTR string, PCWSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCWSTR curr_str = strlst;

    while( curr_str )
    {
        if( _striis_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that begin right at the *start* of the <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-sensitive. </remarks>
_idx_ _findstrlststrhere_( PCSTR string, PCSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCSTR curr_str = strlst;

    while( curr_str )
    {
        if( _strishere_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that begin right at the *start* of the <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-sensitive. </remarks>
_idx_ _findstrlststrhere_( PCWSTR string, PCWSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCWSTR curr_str = strlst;

    while( curr_str )
    {
        if( _strishere_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that begin right at the *start* of the <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-INsensitive. </remarks>
_idx_ _findistrlststrhere_( PCSTR string, PCSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCSTR curr_str = strlst;

    while( curr_str )
    {
        if( _striishere_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}

///<summary> This is like tries to find one of the strings
///  that begin right at the *start* of the <b>string</b>. </summary>
///<returns> -1 if not found. </returns>
///<remarks> The compare is case-INsensitive. </remarks>
_idx_ _findistrlststrhere_( PCWSTR string, PCWSTR strlst )
{
    if( !strlst || !string ) return -1;

    _idx_ idx = 0;
    PCWSTR curr_str = strlst;

    while( curr_str )
    {
        if( _striishere_( string, curr_str ) )
            return idx;

        ++idx;
        curr_str = _fwd_to_next_strlst_str_( curr_str );
    }

    return -1;
}



///<summary> If the string pointed to by <b>string2</b> is exactly the string pointed to by
///  <b>string1</b>, true will be returned else false is returned.
///  The comparing of the characters is case-sensitive. </summary>
bool _stris_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return *string1 ? false : true;
}

///<summary> If the string pointed to by <b>string2</b> is exactly the string pointed to by
///  <b>string1</b>, true will be returned else false is returned.
///  The comparing of the characters is case-sensitive. </summary>
bool _stris_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return *string1 ? false : true;
}

///<summary> like <b>_stris_()</b> without separators but the comparing is case-insensitive. </summary>
bool _striis_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return *string1 ? false : true;
}

///<summary> like <b>_stris_()</b> without separators but the comparing is case-insensitive. </summary>
bool _striis_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return *string1 ? false : true;
}

///<summary> This function returns true if <b>string2</b> is exactly
///  the first string in the stringlist <b>string1</b> where the stringlist
///  is a list of strings separated by any of the specified separation chars. </summary>
bool _strsepis_( PCSTR string1, PCSTR string2, PCSTR pstr1_Separators )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return *string1 ? ( _strchr_( pstr1_Separators, *string1 ) != NULL ) : true;
}

///<summary> This function returns true if <b>string2</b> is exactly
///  the first string in the stringlist <b>string1</b> where the stringlist
///  is a list of strings separated by any of the specified separation chars. </summary>
bool _strsepis_( PCWSTR string1, PCWSTR string2, PCWSTR pstr1_Separators )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return *string1 ? ( _strchr_( pstr1_Separators, *string1 ) != NULL ) : true;
}

///<summary> like <b>_strsepis_()</b> with separators but the comparing is case-insensitive. </summary>
bool _strsepiis_( PCSTR string1, PCSTR string2, PCSTR pstr1_Separators )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return *string1 ? ( _strichr_( pstr1_Separators, *string1 ) != NULL ) : true;
}

///<summary> like <b>_strsepis_()</b> with separators but the comparing is case-insensitive. </summary>
bool _strsepiis_( PCWSTR string1, PCWSTR string2, PCWSTR pstr1_Separators )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return *string1 ? ( _strichr_( pstr1_Separators, *string1 ) != NULL ) : true;
}

///<summary> If the string pointed to by <b>string2</b> is exactly the beginning of the string pointed to by
///  <b>string1</b> or if both strings are exactly the same, true will be returned else false is returned.
///  The comparing of the characters is case-sensitive. </summary>
bool _strishere_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return true;
}

///<summary> If the string pointed to by <b>string2</b> is exactly the beginning of the string pointed to by
///  <b>string1</b> or if both strings are exactly the same, true will be returned else false is returned.
///  The comparing of the characters is case-sensitive. </summary>
bool _strishere_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( *string1 != *string2 )  )
            return false;
    }
    return true;
}

///<summary> like <b>_strishere_()</b> but the comparing is case-insensitive. </summary>
bool _striishere_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return true;
}

///<summary> like <b>_strishere_()</b> but the comparing is case-insensitive. </summary>
bool _striishere_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return false;
    for( ; *string2 ; string1++, string2++ )
    {
        if(  !*string1 ||
             ( _toupper_(*string1) != _toupper_(*string2) )  )
            return false;
    }
    return true;
}

///<summary> like <b>strcmp()</b> </summary>
int _strcmp_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return 0;
    for( ; *string1 && *string2 ; string1++, string2++ )
    {
        int d = (int)(UCHAR)*string1 - (int)(UCHAR)*string2;
        if( d ) return d;
    }
    return *string1 - *string2;
}

///<summary> like <b>wcscmp()</b> </summary>
int _strcmp_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return 0;
    for( ; *string1 && *string2 ; string1++, string2++ )
    {
        int d = (int)*string1 - (int)*string2;
        if( d ) return d;
    }
    return *string1 - *string2;
}

///<summary> like <b>_strcmp_()</b> but case-insensitive </summary>
int _stricmp_( PCSTR string1, PCSTR string2 )
{
    if( !string1 || !string2 ) return 0;
    for( ; *string1 && *string2 ; string1++, string2++ )
    {
        int d = (int)(UCHAR)_toupper_(*string1) - (int)(UCHAR)_toupper_(*string2);
        if( d ) return d;
    }
    return _toupper_(*string1) - _toupper_(*string2);
}

///<summary> like <b>_strcmp_()</b> but case-insensitive </summary>
int _stricmp_( PCWSTR string1, PCWSTR string2 )
{
    if( !string1 || !string2 ) return 0;
    for( ; *string1 && *string2 ; string1++, string2++ )
    {
        int d = (int)_toupper_(*string1) - (int)_toupper_(*string2);
        if( d ) return d;
    }
    return _toupper_(*string1) - _toupper_(*string2);
}

///<summary> Returns true if <b>pstrIn</b> ends (case sensitive)
///  with the string specified in <b>pstrEnding</b>. </summary>
bool _strendswith_( PCSTR pstrIn, PCSTR pstrEnding )
{
    if( !pstrIn || !pstrEnding ) return false;
    if( !*pstrEnding ) return true;

    _size_t_ lenIn     = _strlen_( pstrIn     );
    _size_t_ lenEnding = _strlen_( pstrEnding );

    if( lenIn < lenEnding ) return false;

    return _stris_( pstrIn + (lenIn - lenEnding), pstrEnding );
}

///<summary> Returns true if <b>pstrIn</b> ends (case sensitive)
///  with the string specified in <b>pstrEnding</b>. </summary>
bool _strendswith_( PCWSTR pstrIn, PCWSTR pstrEnding )
{
    if( !pstrIn || !pstrEnding ) return false;
    if( !*pstrEnding ) return true;

    _size_t_ lenIn     = _strlen_( pstrIn     );
    _size_t_ lenEnding = _strlen_( pstrEnding );

    if( lenIn < lenEnding ) return false;

    return _stris_( pstrIn + (lenIn - lenEnding), pstrEnding );
}

///<summary> Returns true if <b>pstrIn</b> ends (case INsensitive)
///  with the string specified in <b>pstrEnding</b>. </summary>
bool _striendswith_( PCSTR pstrIn, PCSTR pstrEnding )
{
    if( !pstrIn || !pstrEnding ) return false;
    if( !*pstrEnding ) return true;

    _size_t_ lenIn     = _strlen_( pstrIn     );
    _size_t_ lenEnding = _strlen_( pstrEnding );

    if( lenIn < lenEnding ) return false;

    return _striis_( pstrIn + (lenIn - lenEnding), pstrEnding );
}

///<summary> Returns true if <b>pstrIn</b> ends (case INsensitive)
///  with the string specified in <b>pstrEnding</b>. </summary>
bool _striendswith_( PCWSTR pstrIn, PCWSTR pstrEnding )
{
    if( !pstrIn || !pstrEnding ) return false;
    if( !*pstrEnding ) return true;

    _size_t_ lenIn     = _strlen_( pstrIn     );
    _size_t_ lenEnding = _strlen_( pstrEnding );

    if( lenIn < lenEnding ) return false;

    return _striis_( pstrIn + (lenIn - lenEnding), pstrEnding );
}



///<summary> like <b>strchr()</b> </summary>
PCSTR _strchr_( PCSTR string, CHAR c )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if( *string == c )
            return string;
    }
    if( *string == c )
        return string;
    return NULL;
}

///<summary> like <b>wcschr()</b> </summary>
PCWSTR _strchr_( PCWSTR string, WCHAR c )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if( *string == c )
            return string;
    }
    if( *string == c )
        return string;
    return NULL;
}

///<summary> like <b>strchr()</b> but case-insensitive </summary>
PCSTR _strichr_( PCSTR string, CHAR c )
{
    if( !string ) return NULL;
    c = _toupper_( c );
    for( ; *string ; string++ )
    {
        if( _toupper_(*string) == c )
            return string;
    }
    if( _toupper_(*string) == c )
        return string;
    return NULL;
}

///<summary> like <b>wcschr()</b> but case-insensitive </summary>
PCWSTR _strichr_( PCWSTR string, WCHAR c )
{
    if( !string ) return NULL;
    c = _toupper_( c );
    for( ; *string ; string++ )
    {
        if( _toupper_(*string) == c )
            return string;
    }
    if( _toupper_(*string) == c )
        return string;
    return NULL;
}

///<summary> like <b>strrchr()</b> </summary>
PCSTR _strrchr_( PCSTR string, CHAR c )
{
    if( !string ) return NULL;
    return strrchr( string, c );
}

///<summary> like <b>wcsrchr()</b> </summary>
PCWSTR _strrchr_( PCWSTR string, WCHAR c )
{
    if( !string ) return NULL;
    return wcsrchr( string, c );
}

///<summary> like <b>strchr()</b> but it does not search only for <b>c</b>
///  but also for spaces as defined with <b>_isspace_()</b> </summary>
PCSTR _strspcchr_( PCSTR string, CHAR c )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if(  *string == c  ||
             _isspace_( *string )  )
            return string;
    }
    if( *string == c )
        return string;
    return NULL;
}

///<summary> like <b>wcschr()</b> but it does not search only for <b>c</b>
///  but also for spaces as defined with <b>_isspace_()</b> </summary>
PCWSTR _strspcchr_( PCWSTR string, WCHAR c )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if(  *string == c  ||
             _isspace_( *string )  )
            return string;
    }
    if( *string == c )
        return string;
    return NULL;
}

///<summary> like <b>strchr()</b> but it does not search only for <b>c</b>
///  but also for spaces as defined with <b>_isspace_()</b> </summary>
PCSTR _strspcichr_( PCSTR string, CHAR c )
{
    if( !string ) return NULL;
    c = _toupper_( c );
    for( ; *string ; string++ )
    {
        if(  _toupper_(*string) == c  ||
             _isspace_( *string )  )
            return string;
    }
    if( _toupper_(*string) == c )
        return string;
    return NULL;
}

///<summary> like <b>wcschr()</b> but it does not search only for <b>c</b>
///  but also for spaces as defined with <b>_isspace_()</b> </summary>
PCWSTR _strspcichr_( PCWSTR string, WCHAR c )
{
    if( !string ) return NULL;
    c = _toupper_( c );
    for( ; *string ; string++ )
    {
        if(  _toupper_(*string) == c  ||
             _isspace_( *string )  )
            return string;
    }
    if( _toupper_(*string) == c )
        return string;
    return NULL;
}

///<summary> like <b>_strspcchr_()</b> but searches only for spaces
///  without an additional character </summary>
PCSTR _strspc_( PCSTR string )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if( _isspace_( *string ) )
            return string;
    }
    return NULL;
}

///<summary> like <b>_strspcchr_()</b> but searches only for spaces
///  without an additional character </summary>
PCWSTR _strspc_( PCWSTR string )
{
    if( !string ) return NULL;
    for( ; *string ; string++ )
    {
        if( _isspace_( *string ) )
            return string;
    }
    return NULL;
}

///<summary> makes string uppercase. </summary>
PSTR _strupr_( PSTR string )
{
    if( !string ) return string;
    for( _idx_ i = 0 ; string[i] ; i++ )
        string[i] = _toupper_( string[i] );
    return string;
}

///<summary> makes string uppercase. </summary>
PWSTR _strupr_( PWSTR string )
{
    if( !string ) return string;
    for( _idx_ i = 0 ; string[i] ; i++ )
        string[i] = _toupper_( string[i] );
    return string;
}

///<summary> makes string lowercase. </summary>
PSTR _strlwr_( PSTR string )
{
    if( !string ) return string;
    for( _idx_ i = 0 ; string[i] ; i++ )
        string[i] = _tolower_( string[i] );
    return string;
}

///<summary> makes string lowercase. </summary>
PWSTR _strlwr_( PWSTR string )
{
    if( !string ) return string;
    for( _idx_ i = 0 ; string[i] ; i++ )
        string[i] = _tolower_( string[i] );
    return string;
}

///<summary> makes first non-space char in the string uppercase. </summary>
PSTR _strupr_firstchar_( PSTR string )
{
    if( !string ) return string;

    while( _isspace_(*string) ) ++string;

    if( *string )
        *string = _toupper_( *string );

    return string;
}

///<summary> makes first non-space char in the string uppercase. </summary>
PWSTR _strupr_firstchar_( PWSTR string )
{
    if( !string ) return string;

    while( _isspace_(*string) ) ++string;

    if( *string )
        *string = _toupper_( *string );

    return string;
}

///<summary> makes first non-space char in the string lowercase. </summary>
PSTR _strlwr_firstchar_( PSTR string )
{
    if( !string ) return string;

    while( _isspace_(*string) ) ++string;

    if( *string )
        *string = _tolower_( *string );

    return string;
}

///<summary> makes first non-space char in the string lowercase. </summary>
PWSTR _strlwr_firstchar_( PWSTR string )
{
    if( !string ) return string;

    while( _isspace_(*string) ) ++string;

    if( *string )
        *string = _tolower_( *string );

    return string;
}

///<summary> Reverses the string. </summary>
///<returns> returns the string </returns>
PSTR _strrev_( PSTR string )
{
    _size_t_ l = _strlen_( string );
    if( l <= 1 ) return string;
    _size_t_ cnt = l >> 1;
    for( _idx_ i = 0 ; (_size_t_) i < cnt ; i++ ) {
        CHAR h          = string[ i ];
        string[ i ]     = string[ l-1-i ];
        string[ l-1-i ] = h;
    }
    return string;
}

///<summary> Reverses the string. </summary>
///<returns> returns the string </returns>
PWSTR _strrev_( PWSTR string )
{
    _size_t_ l = _strlen_( string );
    if( l <= 1 ) return string;
    _size_t_ cnt = l >> 1;
    for( _idx_ i = 0 ; (_size_t_) i < cnt ; i++ ) {
        WCHAR h         = string[ i ];
        string[ i ]     = string[ l-1-i ];
        string[ l-1-i ] = h;
    }
    return string;
}

///<summary> The same as <b>strstr()</b>. </summary>
PCSTR _strstr_( PCSTR string, PCSTR strfind )
{
    if( !string || !strfind ) return string;
    if( !*strfind ) return string;
    PCSTR p = string;
    while( *p ) {
        if( _strishere_( p, strfind ) )
            return p;
        p++;
    }
    return NULL;
}

///<summary> The same as <b>wcsstr()</b>. </summary>
PCWSTR _strstr_( PCWSTR string, PCWSTR strfind )
{
    if( !string || !strfind ) return string;
    if( !*strfind ) return string;
    PCWSTR p = string;
    while( *p ) {
        if( _strishere_( p, strfind ) )
            return p;
        p++;
    }
    return NULL;
}

///<summary> The same as <b>strstr()</b>, but with
///  case-insensitive search. </summary>
PCSTR _stristr_( PCSTR string, PCSTR strfind )
{
    if( !string || !strfind ) return string;
    if( !*strfind ) return string;
    PCSTR p = string;
    while( *p ) {
        if( _striishere_( p, strfind ) )
            return p;
        p++;
    }
    return NULL;
}

///<summary> The same as <b>wcsstr()</b>, but with
///  case-insensitive search. </summary>
PCWSTR _stristr_( PCWSTR string, PCWSTR strfind )
{
    if( !string || !strfind ) return string;
    if( !*strfind ) return string;
    PCWSTR p = string;
    while( *p ) {
        if( _striishere_( p, strfind ) )
            return p;
        p++;
    }
    return NULL;
}

///<summary> The same as <b>strpbrk()</b>. </summary>
PCSTR _strpbrk_( PCSTR string, PCSTR charset )
{
    if( !string || !charset ) return string;
    return strpbrk( string, charset );
}

///<summary> The same as <b>wcspbrk()</b>. </summary>
PCWSTR _strpbrk_( PCWSTR string, PCWSTR charset )
{
    if( !string || !charset ) return string;
    return wcspbrk( string, charset );
}

///<summary> Counts how often <b>c</b> is present in <b>string</b>.
///  The scanning is case-sensitive. </summary>
int _strcntchr_( PCSTR string, CHAR c )
{
    CHAR s[2] = { c, 0 };
    return _strcntchrs_( string, s );
}

///<summary> Counts how often <b>c</b> is present in <b>string</b>.
///  The scanning is case-sensitive. </summary>
int _strcntchr_( PCWSTR string, WCHAR c )
{
    WCHAR s[2] = { c, 0 };
    return _strcntchrs_( string, s );
}

///<summary> Counts how often <b>c</b> is present in <b>string</b>.
///  The scanning is case-insensitive. </summary>
int _stricntchr_( PCSTR string, CHAR c )
{
    CHAR s[2] = { c, 0 };
    return _stricntchrs_( string, s );
}

///<summary> Counts how often <b>c</b> is present in <b>string</b>.
///  The scanning is case-insensitive. </summary>
int _stricntchr_( PCWSTR string, WCHAR c )
{
    WCHAR s[2] = { c, 0 };
    return _stricntchrs_( string, s );
}

///<summary> Counts the sum of how often all chars
///  of <b>chars</b> are present in <b>string</b>.
///  The scanning is case-sensitive. </summary>
///<remarks> Note: If a char in <b>chars</b> is present
///  twice, then its occurrence in <b>string</b> will
///  also be counted twice. </remarks>
int _strcntchrs_( PCSTR string, PCSTR chars )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    _size_t_ cchars = _strlen_( chars );
    if( l == 0 || cchars == 0 )
        return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ )
        for( _idx_ j = 0 ; (_size_t_) j < cchars ; j++ )
            if( string[i] == chars[j] )
                cnt++;
    return (int) cnt;
}

///<summary> Counts the sum of how often all chars
///  of <b>chars</b> are present in <b>string</b>.
///  The scanning is case-sensitive. </summary>
///<remarks> Note: If a char in <b>chars</b> is present
///  twice, then its occurrence in <b>string</b> will
///  also be counted twice. </remarks>
int _strcntchrs_( PCWSTR string, PCWSTR chars )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    _size_t_ cchars = _strlen_( chars );
    if( l == 0 || cchars == 0 )
        return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ )
        for( _idx_ j = 0 ; (_size_t_) j < cchars ; j++ )
            if( string[i] == chars[j] )
                cnt++;
    return (int) cnt;
}

///<summary> Counts the sum of how often all chars
///  of <b>chars</b> are present in <b>string</b>.
///  The scanning is case-insensitive. </summary>
///<remarks> Note: If a char in <b>chars</b> is present
///  twice, then its occurrence in <b>string</b> will
///  also be counted twice. </remarks>
int _stricntchrs_( PCSTR string, PCSTR chars )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    _size_t_ cchars = _strlen_( chars );
    if( l == 0 || cchars == 0 )
        return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ ) {
        int c = (int)_toupper_( string[i] );
        for( _idx_ j = 0 ; (_size_t_) j < cchars ; j++ )
            if( c == (int)_toupper_( chars[j] ) )
                cnt++;
    }
    return (int) cnt;
}

///<summary> Counts the sum of how often all chars
///  of <b>chars</b> are present in <b>string</b>.
///  The scanning is case-insensitive. </summary>
///<remarks> Note: If a char in <b>chars</b> is present
///  twice, then its occurrence in <b>string</b> will
///  also be counted twice. </remarks>
int _stricntchrs_( PCWSTR string, PCWSTR chars )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    _size_t_ cchars = _strlen_( chars );
    if( l == 0 || cchars == 0 )
        return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ ) {
        int c = (int)_toupper_( string[i] );
        for( _idx_ j = 0 ; (_size_t_) j < cchars ; j++ )
            if( c == (int)_toupper_( chars[j] ) )
                cnt++;
    }
    return (int) cnt;
}

///<summary>
///  <b>_strcntspc_()</b> counts how many spacing chars
///  are present in the passed string.
///</summary>
int _strcntspc_( PCSTR string )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    if( l == 0 ) return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ )
        if( _isspace_( string[i] ) )
            cnt++;
    return (int) cnt;
}

///<summary>
///  <b>_strcntspc_()</b> counts how many spacing chars
///  are present in the passed string.
///</summary>
int _strcntspc_( PCWSTR string )
{
    _size_t_ cnt = 0;
    _size_t_ l = _strlen_( string );
    if( l == 0 ) return 0;
    for( _idx_ i = 0 ; (_size_t_) i < l ; i++ )
        if( _isspace_( string[i] ) )
            cnt++;
    return (int) cnt;
}

void _strHasWhichCharCases_( PCSTR string, bool& hasLowercaseChars, bool& hasUppercaseChars )
{
    hasLowercaseChars = hasUppercaseChars = false;
    if( !string ) return;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
        {
            if( *string == _tolower_(*string) ) hasLowercaseChars = true;
            if( *string == _toupper_(*string) ) hasUppercaseChars = true;

            if( hasLowercaseChars && hasUppercaseChars )
                break;
        }
}

void _strHasWhichCharCases_( PCWSTR string, bool& hasLowercaseChars, bool& hasUppercaseChars )
{
    hasLowercaseChars = hasUppercaseChars = false;
    if( !string ) return;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
        {
            if( *string == _tolower_(*string) ) hasLowercaseChars = true;
            if( *string == _toupper_(*string) ) hasUppercaseChars = true;

            if( hasLowercaseChars && hasUppercaseChars )
                break;
        }
}

bool _strHasLowercaseChars_( PCSTR string )
{
    if( !string ) return false;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
            if( *string == _tolower_(*string) )
                return true;

    return false;
}

bool _strHasLowercaseChars_( PCWSTR string )
{
    if( !string ) return false;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
            if( *string == _tolower_(*string) )
                return true;

    return false;
}

bool _strHasUppercaseChars_( PCSTR string )
{
    if( !string ) return false;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
            if( *string == _toupper_(*string) )
                return true;

    return false;
}

bool _strHasUppercaseChars_( PCWSTR string )
{
    if( !string ) return false;

    for( ; *string ; ++string )
        if( _isalpha_(*string) )
            if( *string == _toupper_(*string) )
                return true;

    return false;
}



///<summary> checks if <b>c</b> is a hexadecimal digit.
///  If it is so -> returns the digit value or
///  if it is not a hex digit -> returns -1.
///  The checking is case-insensitive. </summary>
int _chtohex_( CHAR c )
{
    if( c >= '0' && c <= '9' ) return( c - '0' );
    if( c >= 'A' && c <= 'F' ) return( c - 'A' + 10 );
    if( c >= 'a' && c <= 'f' ) return( c - 'a' + 10 );
    return -1;
}

///<summary> checks if <b>c</b> is a hexadecimal digit.
///  If it is so -> returns the digit value or
///  if it is not a hex digit -> returns -1.
///  The checking is case-insensitive. </summary>
int _chtohex_( WCHAR c )
{
    if( c >= L'0' && c <= L'9' ) return( c - L'0' );
    if( c >= L'A' && c <= L'F' ) return( c - L'A' + 10 );
    if( c >= L'a' && c <= L'f' ) return( c - L'a' + 10 );
    return -1;
}

///<summary> converts a hexadecimal value (which must be inbetween
///  0x0 - 0xf) into a textual lowercase hex-digit character. </summary>
///<returns> -1 = <b>hex</b> is not within 0x0 - 0xf </returns>
int _hextoch_( int hex )
{
    if( hex < 0 || hex > 0xf ) return -1;
    if( hex < 0xa ) return( hex + '0' );
    else return( hex - 10 + 'a' );
}

///<summary> converts a hexadecimal value (which must be inbetween
///  0x0 - 0xf) into a textual uppercase hex-digit character. </summary>
///<returns> -1 = <b>hex</b> is not within 0x0 - 0xf </returns>
int _hextoCH_( int hex )
{
    if( hex < 0 || hex > 0xf ) return -1;
    if( hex < 0xa ) return( hex + '0' );
    else return( hex - 10 + 'A' );
}

///<summary> checks if <b>c</b> is a decimal digit.
///  If it is so -> returns the digit value or
///  if it is not a dec digit -> returns -1.
///  The checking is case-insensitive. </summary>
int _chtodec_( CHAR c )
{
    if( c >= '0' && c <= '9' )
        return( c - '0' );
    return -1;
}

///<summary> checks if <b>c</b> is a decimal digit.
///  If it is so -> returns the digit value or
///  if it is not a dec digit -> returns -1.
///  The checking is case-insensitive. </summary>
int _chtodec_( WCHAR c )
{
    if( c >= L'0' && c <= L'9' )
        return( c - L'0' );
    return -1;
}

///<summary> converts a decimal value (which must be inbetween
///  0 - 9) into a textual decimal-digit character. </summary>
///<returns> -1 = <b>dec</b> is not within 0 - 9 </returns>
int _dectoch_( int dec )
{
    if( dec < 0 || dec > 9 )
        return -1;
    return( dec + '0' );
}

///<summary> Returns the next char of the string pointed to by <b>pstring</b>.
///  If in the string there is defined an escape sequence like for C or C++ strings, then
///  this function will expand the escape sequence. </summary>
///<param name="pstring"> IN: pointer to a string which can contain escape sequences<br/>
///  OUT: will point to the next character in the string, or to the end of the string
///  (which is the terminating 0-character). </param>
///<returns> returns the character or 0 if the end of the string has been reached.<br/>
///  LT 0 = (if the return value is less than 0) there also was a wrong escape sequence,
///   but a character has been detected successfully. To get that character, simply
///   negate the return value.<br/>
///   Such an error will occur e.g. when you specify an escape sequence like "\q"
///   which is invalid. The negated return value in that case will be a 'q'. </returns>
///<remarks> Note: The following escape-sequences are supported:
///  <ul>
///   <li> simple-escape-sequence: <br/>
///         \a \b \f \n \r \t \v \' \" \\ \? </li>
///   <li> octal-escape-sequence: <br/>
///         \ octal-digit <br/>
///         \ octal-digit octal-digit <br/>
///         \ octal-digit octal-digit octal-digit <br/>
///         In here, a maximum of 3 oct-digits are supported </li>
///   <li> hexadecimal-escape-sequence: <br/>
///         \x hexadecimal-digit <br/>
///         hexadecimal-escape-sequence hexadecimal-digit <br/>
///         In here, a maximum of 3 hex-digits are supported </li>
///  </ul>
///  Note: If a hexadecimal or octal sequence is too large to fit in only one character (so if
///   its value is larger than 255), the value will be returned nevertheless. </remarks>
int _esctoch_( PCSTR *pstring )
{
    if( !pstring ) return 0;
    if( !*pstring ) return 0;
    CHAR c = **pstring;
    if( !c ) return 0;
    ( *pstring )++;
    if( c != '\\' ) return (int)(UINT)(UCHAR)c;
    if( c = **pstring )
    {
        if( !c ) return 0;
        ( *pstring )++;
        // check for simple-escape-sequence
        switch( c )
        {
            case 'a'  : return (int)(UINT)(UCHAR)'\a';
            case 'b'  : return (int)(UINT)(UCHAR)'\b';
            case 'f'  : return (int)(UINT)(UCHAR)'\f';
            case 'n'  : return (int)(UINT)(UCHAR)'\n';
            case 'r'  : return (int)(UINT)(UCHAR)'\r';
            case 't'  : return (int)(UINT)(UCHAR)'\t';
            case 'v'  : return (int)(UINT)(UCHAR)'\v';
            case '\'' : return (int)(UINT)(UCHAR)'\'';
            case '\"' : return (int)(UINT)(UCHAR)'\"';
            case '\\' : return (int)(UINT)(UCHAR)'\\';
            case '?'  : return (int)(UINT)(UCHAR)'\?';
        }
        // check for hexadecimal-escape-sequence
        if( c == 'x' )
        {
            int hex = 0;
            int v = 0;
            for( int i = 0 ; i < 3 ; i++ )
            {
                if( !( c = **pstring ) )         return( i ? v : -(int)(UINT)(UCHAR)'x' );
                if( ( hex = _chtohex_(c) ) < 0 ) return( i ? v : -(int)(UINT)(UCHAR)'x' );
                ( *pstring )++;
                v *= 16;
                v += hex;
            }
            return v;
        }
        // check for octal-escape-sequence
        if( c >= '0' && c <= '7' )
        {
            int oct = (int)(UCHAR)( c - '0' );
            int v = oct;
            for( int i = 0 ; i < 2 ; i++ )
            {
                if( !( c = **pstring ) ) return v;
                if( c < '0' || c > '7' ) return v;
                ( *pstring )++;
                oct = (int)(UCHAR)( c - '0' );
                v *= 8;
                v += oct;
            }
            return v;
        }
        return -(int)(UINT)(UCHAR)c;
    }
    return -(int)(UINT)(UCHAR)'\\';
}

///<summary> Returns the next char of the string pointed to by <b>pstring</b>.
///  If in the string there is defined an escape sequence like for C or C++ strings, then
///  this function will expand the escape sequence. </summary>
///<param name="pstring"> IN: pointer to a string which can contain escape sequences<br/>
///  OUT: will point to the next character in the string, or to the end of the string
///  (which is the terminating 0-character). </param>
///<returns> returns the character or 0 if the end of the string has been reached.<br/>
///  LT 0 = (if the return value is less than 0) there also was a wrong escape sequence,
///   but a character has been detected successfully. To get that character, simply
///   negate the return value.<br/>
///   Such an error will occur e.g. when you specify an escape sequence like "\q"
///   which is invalid. The negated return value in that case will be a 'q'. </returns>
///<remarks> Note: The following escape-sequences are supported:
///  <ul>
///   <li> simple-escape-sequence: <br/>
///         \a \b \f \n \r \t \v \' \" \\ \? </li>
///   <li> octal-escape-sequence: <br/>
///         \ octal-digit <br/>
///         \ octal-digit octal-digit <br/>
///         \ octal-digit octal-digit octal-digit <br/>
///         In here, a maximum of 3 oct-digits are supported </li>
///   <li> hexadecimal-escape-sequence: <br/>
///         \x hexadecimal-digit <br/>
///         hexadecimal-escape-sequence hexadecimal-digit <br/>
///         In here, a maximum of 3 hex-digits are supported </li>
///  </ul>
///  Note: If a hexadecimal or octal sequence is too large to fit in only one character (so if
///   its value is larger than 255), the value will be returned nevertheless. </remarks>
int _esctoch_( PCWSTR *pstring )
{
    if( !pstring ) return 0;
    if( !*pstring ) return 0;
    WCHAR c = **pstring;
    if( !c ) return 0;
    ( *pstring )++;
    if( c != L'\\' ) return (int)(UINT)(USHORT)c;
    if( c = **pstring )
    {
        if( !c ) return 0;
        ( *pstring )++;
        // check for simple-escape-sequence
        switch( c )
        {
            case L'a'  : return (int)(UINT)(USHORT)L'\a';
            case L'b'  : return (int)(UINT)(USHORT)L'\b';
            case L'f'  : return (int)(UINT)(USHORT)L'\f';
            case L'n'  : return (int)(UINT)(USHORT)L'\n';
            case L'r'  : return (int)(UINT)(USHORT)L'\r';
            case L't'  : return (int)(UINT)(USHORT)L'\t';
            case L'v'  : return (int)(UINT)(USHORT)L'\v';
            case L'\'' : return (int)(UINT)(USHORT)L'\'';
            case L'\"' : return (int)(UINT)(USHORT)L'\"';
            case L'\\' : return (int)(UINT)(USHORT)L'\\';
            case L'?'  : return (int)(UINT)(USHORT)L'\?';
        }
        // check for hexadecimal-escape-sequence
        if( c == L'x' )
        {
            int hex = 0;
            int v = 0;
            for( int i = 0 ; i < 3 ; i++ )
            {
                if( !( c = **pstring ) )         return( i ? v : -(int)(UINT)(USHORT)L'x' );
                if( ( hex = _chtohex_(c) ) < 0 ) return( i ? v : -(int)(UINT)(USHORT)L'x' );
                ( *pstring )++;
                v *= 16;
                v += hex;
            }
            return v;
        }
        // check for octal-escape-sequence
        if( c >= L'0' && c <= L'7' )
        {
            int oct = (int)(USHORT)( c - L'0' );
            int v = oct;
            for( int i = 0 ; i < 2 ; i++ )
            {
                if( !( c = **pstring ) ) return v;
                if( c < L'0' || c > L'7' ) return v;
                ( *pstring )++;
                oct = (int)(USHORT)( c - L'0' );
                v *= 8;
                v += oct;
            }
            return v;
        }
        return -(int)(UINT)(USHORT)c;
    }
    return -(int)(UINT)(USHORT)L'\\';
}

///<summary> If <b>ch</b> is not a writeable character, it will be converted into an escape sequence
///  and copied into <b>dest</b>, else <b>ch</b> itself will be copied into <b>dest</b>.
///  If <b>size</b> is too small to hold the escape sequence or the character, NULL will
///  be returned, else <b>dest</b> will be returned. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
PSTR _esccpy_( PSTR dest, CHAR ch, _size_t_ size )
{
    if( !dest || size <= 0 ) return NULL;
    PSTR pstr = NULL;
    switch( ch )
    {
        case 0    : pstr = "\\000"; break;
        // Note: "\\0" is too short because if e.g. a '1' will follow this escape sequence, then
        // the string would look like "\01" which will be interpreted as one char but a
        // sequence like "\\0001" represents 2 chars: '\0' and '1'
        case '\a' : pstr = "\\a";  break;
        case '\b' : pstr = "\\b";  break;
        case '\f' : pstr = "\\f";  break;
        case '\n' : pstr = "\\n";  break;
        case '\r' : pstr = "\\r";  break;
        case '\t' : pstr = "\\t";  break;
        case '\v' : pstr = "\\v";  break;
        case '\'' : pstr = "\\\'"; break;
        case '\"' : pstr = "\\\""; break;
        case '\\' : pstr = "\\\\"; break;
        case '\?' : pstr = "\\?";  break;
    }
    CHAR str[10];
    if( !pstr )
    {
        if( (int)ch < 0x20 )   // Note: Because #ch# is of type #CHAR#, values within 0x80 - 0xff also are negative
        {
            str[0] = '\\';
            str[1] = 'x';
            str[2] = '0';
            str[3] = (CHAR)_hextoch_( (ch >> 4) & 0xf );
            str[4] = (CHAR)_hextoch_( ch & 0xf );
            str[5] = 0;
            pstr = str;
        }
        else
        {
            str[0] = ch;
            str[1] = 0;
            pstr = str;
        }
    }
    _size_t_ c = _strlen_( pstr ) + 1;
    if( size < c ) return NULL;
    return _strcpy_( dest, pstr, size );
}

///<summary> If <b>ch</b> is not a writeable character, it will be converted into an escape sequence
///  and copied into <b>dest</b>, else <b>ch</b> itself will be copied into <b>dest</b>.
///  If <b>size</b> is too small to hold the escape sequence or the character, NULL will
///  be returned, else <b>dest</b> will be returned. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<remarks> IMPORTANT NOTE: Only WCHARs whose value has the upper byte set
///  to zero can be converted into an escape sequence. </remarks>
PWSTR _esccpy_( PWSTR dest, WCHAR ch, _size_t_ size )
{
    if( !dest || size <= 0 ) return NULL;
    PWSTR pstr = NULL;
    switch( ch )
    {
        case 0    : pstr = L"\\000"; break;
        // Note: "\\0" is too short because if e.g. a '1' will follow this escape sequence, then
        // the string would look like "\01" which will be interpreted as one char but a
        // sequence like "\\0001" represents 2 chars: '\0' and '1'
        case L'\a' : pstr = L"\\a";  break;
        case L'\b' : pstr = L"\\b";  break;
        case L'\f' : pstr = L"\\f";  break;
        case L'\n' : pstr = L"\\n";  break;
        case L'\r' : pstr = L"\\r";  break;
        case L'\t' : pstr = L"\\t";  break;
        case L'\v' : pstr = L"\\v";  break;
        case L'\'' : pstr = L"\\\'"; break;
        case L'\"' : pstr = L"\\\""; break;
        case L'\\' : pstr = L"\\\\"; break;
        case L'\?' : pstr = L"\\?";  break;
    }
    WCHAR str[10];
    if( !pstr )
    {
        if(( HIBYTE(ch) == 0 )&&( (int)(CHAR)ch < 0x20 ))   // Note: Because #ch# is casted to #CHAR#, values within 0x80 - 0xff also are negative
        {
            str[0] = L'\\';
            str[1] = L'x';
            str[2] = L'0';
            str[3] = (WCHAR)_hextoch_( (ch >> 4) & 0xf );
            str[4] = (WCHAR)_hextoch_( ch & 0xf );
            str[5] = 0;
            pstr = str;
        }
        else
        {
            str[0] = ch;
            str[1] = 0;
            pstr = str;
        }
    }
    _size_t_ c = _strlen_( pstr ) + 1;
    if( size < c ) return NULL;
    return _strcpy_( dest, pstr, size );
}

///<summary> Equal to <b>_esccpy_()</b> except that the esc-sequence or
///  the char will not be copied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
PSTR _esccat_( PSTR dest, CHAR ch, _size_t_ size )
{
    if( !dest || size <= 0 ) return NULL;
    _size_t_ l = _strlen_( dest );
    return _esccpy_( dest + l, ch, size - l );
}

///<summary> Equal to <b>_esccpy_()</b> except that the esc-sequence or
///  the char will not be copied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
PWSTR _esccat_( PWSTR dest, WCHAR ch, _size_t_ size )
{
    if( !dest || size <= 0 ) return NULL;
    _size_t_ l = _strlen_( dest );
    return _esccpy_( dest + l, ch, size - l );
}

///<summary> Equal to <b>_esccpy_()</b> but this function will convert a whole string into escape
///  sequences (only for those characters which cannot be written) and then copy it to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> It will return the number of chars copied from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escscpy_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( !dest || !srce || size <= 0 ) return 0;
    int cnt = 0;
    PSTR d = dest;
    _size_t_ c = size;
    while( *srce && c > 0 )
    {
        if( !_esccpy_( d, *srce, c ) ) return cnt;
        _size_t_ l = _strlen_( d );
        d += l;
        c -= l;
        ++srce;
        ++cnt;
    }
    return cnt;
}

///<summary> Equal to <b>_esccpy_()</b> but this function will convert a whole string into escape
///  sequences (only for those characters which cannot be written) and then copy it to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> It will return the number of chars copied from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escscpy_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( !dest || !srce || size <= 0 ) return 0;
    int cnt = 0;
    PWSTR d = dest;
    _size_t_ c = size;
    while( *srce && c > 0 )
    {
        if( !_esccpy_( d, *srce, c ) ) return cnt;
        _size_t_ l = _strlen_( d );
        d += l;
        c -= l;
        ++srce;
        ++cnt;
    }
    return cnt;
}

///<summary> Equal to <b>_escscpy_()</b> but the converted
///  string will not be copyied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> It will return the number of chars appended from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escscat_( PSTR dest, PCSTR srce, _size_t_ size )
{
    if( !dest || !srce || size <= 0 ) return 0;
    _size_t_ l = _strlen_( dest );
    return _escscpy_( dest + l, srce, size - l );
}

///<summary> Equal to <b>_escscpy_()</b> but the converted
///  string will not be copyied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> It will return the number of chars appended from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escscat_( PWSTR dest, PCWSTR srce, _size_t_ size )
{
    if( !dest || !srce || size <= 0 ) return 0;
    _size_t_ l = _strlen_( dest );
    return _escscpy_( dest + l, srce, size - l );
}

///<summary> Moves one character (which can be an escape-sequence-char) forward in the <b>escstr</b> string
///  which must be a string which can contain escape-sequences. </summary>
///<param name="ipos"> current cursor position in the string </param>
///<returns> index of the following char. Can be the length of <b>escstr</b> which indicates that the end of string has been reached.
///  If <b>ipos</b> is negative, or if it is greater or equal than the <b>escstr</b> length, -1 will be returned. </returns>
///<remarks>
///  This function checks the length of the string <b>escstr</b> which
///  could be time consuming. The advantage is that we cannot exceed the
///  string limits. <br/>
///  If you need it faster (and more dangerous because you have to check the
///  string limit by yourself), you can use <b>_esctoch_()</b> instead.
///</remarks>
_idx_ _escfwdchr_( PCSTR escstr, _idx_ ipos )
{
    if( ipos < 0 ) return -1;
    _size_t_ len = _strlen_( escstr );
    if( (_size_t_) ipos >= len ) return -1;
    PCSTR ppos = escstr + ipos;
    _esctoch_( &ppos );
    if( ppos == escstr + ipos ) return -1;
    return (_idx_)( ppos - escstr );    // subtraction handles char-size correctly
}

///<summary> Moves one character (which can be an escape-sequence-char) forward in the <b>escstr</b> string
///  which must be a string which can contain escape-sequences. </summary>
///<param name="ipos"> current cursor position in the string </param>
///<returns> index of the following char. Can be the length of <b>escstr</b> which indicates that the end of string has been reached.
///  If <b>ipos</b> is negative, or if it is greater or equal than the <b>escstr</b> length, -1 will be returned. </returns>
///<remarks>
///  This function checks the length of the string <b>escstr</b> which
///  could be time consuming. The advantage is that we cannot exceed the
///  string limits. <br/>
///  If you need it faster (and more dangerous because you have to check the
///  string limit by yourself), you can use <b>_esctoch_()</b> instead.
///</remarks>
_idx_ _escfwdchr_( PCWSTR escstr, _idx_ ipos )
{
    if( ipos < 0 ) return -1;
    _size_t_ len = _strlen_( escstr );
    if( (_size_t_) ipos >= len ) return -1;
    PCWSTR ppos = escstr + ipos;
    _esctoch_( &ppos );
    if( ppos == escstr + ipos ) return -1;
    return (_idx_)( ppos - escstr );    // subtraction handles char-size correctly
}

///<summary> Moves one character (which can be an escape-sequence-char) backward in the <b>escstr</b> string
///  which must be a string which can contain escape-sequences. </summary>
///<param name="ipos"> current cursor position in the string (can also be the length of the string
///  to indicate a position after the last char in the string) </param>
///<returns> index of the previous char. Will be less than the length of <b>escstr</b>.
///  If <b>ipos</b> is negative or 0, or if it is greater than the <b>escstr</b> length, -1 will be returned. </returns>
_idx_ _escbwdchr_( PCSTR escstr, _idx_ ipos )
{
    if( ipos <= 0 ) return -1;
    _size_t_ len = _strlen_( escstr );
    if( (_size_t_) ipos > len ) return -1;
    _idx_ i = ipos - 1;
    bool escfound = false;
    for( ; i >= 0 ; --i )
        if( escstr[i] == '\\' )
        {
            if( i > 0 ) {
                if( escstr[i-1] != '\\' ) {
                    escfound = true;
                    break;
                }
            } else {
                escfound = true;
                break;
            }
        }
    if( escfound && i >= 0 )
    {
        PCSTR ppos = escstr + i;
        _esctoch_( &ppos );
        if( ppos == escstr + i ) return -1;
        _idx_ iafter = (_idx_)( ppos - escstr );    // subtraction handles char-size correctly
        if( iafter >= ipos ) return i;
    }
    return ipos - 1;
}

///<summary> Moves one character (which can be an escape-sequence-char) backward in the <b>escstr</b> string
///  which must be a string which can contain escape-sequences. </summary>
///<param name="ipos"> current cursor position in the string (can also be the length of the string
///  to indicate a position after the last char in the string) </param>
///<returns> index of the previous char. Will be less than the length of <b>escstr</b>.
///  If <b>ipos</b> is negative or 0, or if it is greater than the <b>escstr</b> length, -1 will be returned. </returns>
_idx_ _escbwdchr_( PCWSTR escstr, _idx_ ipos )
{
    if( ipos <= 0 ) return -1;
    _size_t_ len = _strlen_( escstr );
    if( (_size_t_) ipos > len ) return -1;
    _idx_ i = ipos - 1;
    bool escfound = false;
    for( ; i >= 0 ; --i )
        if( escstr[i] == L'\\' )
        {
            if( i > 0 ) {
                if( escstr[i-1] != L'\\' ) {
                    escfound = true;
                    break;
                }
            } else {
                escfound = true;
                break;
            }
        }
    if( escfound && i >= 0 )
    {
        PCWSTR ppos = escstr + i;
        _esctoch_( &ppos );
        if( ppos == escstr + i ) return -1;
        _idx_ iafter = (_idx_)( ppos - escstr );    // subtraction handles char-size correctly
        if( iafter >= ipos ) return i;
    }
    return ipos - 1;
}

///<summary> Like <b>_escfwdchr_()</b> except that char pointers instead
///  of indices are used. This allows faster handling since the string length
///  does not need to be evaluated (which can be very time consuming when
///  calling this function here again and again). </summary>
PCSTR _escfwdchr_ptr_( PCSTR escstr_ptr )
{
    if( !escstr_ptr ) return NULL;
    PCSTR old = escstr_ptr;
    _esctoch_( &escstr_ptr );
    if( escstr_ptr == old ) return NULL;
    return escstr_ptr;
}

///<summary> Like <b>_escfwdchr_()</b> except that char pointers instead
///  of indices are used. This allows faster handling since the string length
///  does not need to be evaluated (which can be very time consuming when
///  calling this function here again and again). </summary>
PCWSTR _escfwdchr_ptr_( PCWSTR escstr_ptr )
{
    if( !escstr_ptr ) return NULL;
    PCWSTR old = escstr_ptr;
    _esctoch_( &escstr_ptr );
    if( escstr_ptr == old ) return NULL;
    return escstr_ptr;
}

///<summary> Equal to <b>_escscpy_()</b> but this function will convert a whole string-list into escape
///  sequences (only for those characters which cannot be written) and copy it to <b>dest</b>
///  where <b>dest</b> will then be only a string (with one terminating 0 char). </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> It will return the number of chars copied from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
///<remarks> IMPORTANT: The <b>srcestrlist</b> ends with 2 0 chars. In the <b>dest</b> string *ONE* of these
///   2 0 chars (the first one) will also be converted to a string-escape-sequence and appended
///   to the <b>dest</b> string. </remarks>
int _escstrlstcpy_( PSTR dest, PCSTR srcestrlst, _size_t_ size )
{
    if( !dest || !srcestrlst || size <= 0 ) return 0;
    int cnt = 0;
    PSTR d = dest;
    _size_t_ c = size;
    while( ( srcestrlst[0] || srcestrlst[1] ) && c > 0 )
    {
        if( !_esccpy_( d, *srcestrlst, c ) ) return cnt;
        _size_t_ l = _strlen_( d );
        d += l;
        c -= l;
        ++srcestrlst;
        ++cnt;
    }
    if( c > 0 )
    {
        if( !_esccpy_( d, *srcestrlst, c ) ) return cnt;
        ++cnt;
    }
    return cnt;
}

///<summary> Equal to <b>_escscpy_()</b> but this function will convert a whole string-list into escape
///  sequences (only for those characters which cannot be written) and copy it to <b>dest</b>
///  where <b>dest</b> will then be only a string (with one terminating 0 char). </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> It will return the number of chars copied from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
///<remarks> IMPORTANT: The <b>srcestrlist</b> ends with 2 0 chars. In the <b>dest</b> string *ONE* of these
///   2 0 chars (the first one) will also be converted to a string-escape-sequence and appended
///   to the <b>dest</b> string. </remarks>
int _escstrlstcpy_( PWSTR dest, PCWSTR srcestrlst, _size_t_ size )
{
    if( !dest || !srcestrlst || size <= 0 ) return 0;
    int cnt = 0;
    PWSTR d = dest;
    _size_t_ c = size;
    while( ( srcestrlst[0] || srcestrlst[1] ) && c > 0 )
    {
        if( !_esccpy_( d, *srcestrlst, c ) ) return cnt;
        _size_t_ l = _strlen_( d );
        d += l;
        c -= l;
        ++srcestrlst;
        ++cnt;
    }
    if( c > 0 )
    {
        if( !_esccpy_( d, *srcestrlst, c ) ) return cnt;
        ++cnt;
    }
    return cnt;
}

///<summary> Equal to <b>_escstrlstcpy_()</b> but the converted
///  string will not be copied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> It will return the number of chars appended from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escstrlstcat_( PSTR dest, PCSTR srcestrlst, _size_t_ size )
{
    if( !dest || !srcestrlst || size <= 0 ) return 0;
    _size_t_ l = _strlen_(dest);
    return _escstrlstcpy_( dest + l, srcestrlst, size - l );
}

///<summary> Equal to <b>_escstrlstcpy_()</b> but the converted
///  string will not be copied but appended to <b>dest</b>. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> It will return the number of chars appended from <b>srce</b> to <b>dest</b> (note: each
///  escape-sequence also will be counted as only one char (do *not* change this behavior since it is expected)).
///  The terminating 0-byte is not included in that counting process. </returns>
int _escstrlstcat_( PWSTR dest, PCWSTR srcestrlst, _size_t_ size )
{
    if( !dest || !srcestrlst || size <= 0 ) return 0;
    _size_t_ l = _strlen_(dest);
    return _escstrlstcpy_( dest + l, srcestrlst, size - l );
}

///<summary> Finds all wildcard chars defined by <b>wild</b> in the string <b>string</b> and will replace
///  it with the number in <b>n</b>. </summary>
///<param name="n"> the number to insert. Only positive values are supported at the moment. </param>
///<param name="nbase"> the base of the number: <br/>
///   10 = <b>n</b> is a decimal value <br/>
///   16 = <b>n</b> is a hexadecimal value (the numbers A to F will be uppercase). <br/>
///  -16 = <b>n</b> is a hexadecimal value (the numbers a to f will be lowercase). <br/>
///    2 = <b>n</b> is a binary value <br/>
///  Also all other base values are supported <br/>
///  (as long as their absolute values are greater-equal 2). </param>
///<returns>
///   1 = ok <br/>
///  -1 = If in <b>string</b> there are not enough wildcard chars to hold the whole number,
///       but nevertheless it will be tried to replace as many wildcard chars as possible.
///       ( The lowest digits of <b>n</b> will be set. ) <br/>
///   0 = Error </returns>
///<remarks> Note: If there are too much wildcard chars so that the number is smaller, the remaining
///  leading wildcard chars will be filled with '0' chars. </remarks>
int _replace_nr_( PSTR string, CHAR wild, int n, int nbase )
{
    bool upper = ( nbase > 0 );
    if( nbase < 0 ) nbase = -nbase;
    if( !string || !wild || n < 0 || nbase < 2 ) return 0;
    _size_t_ l = _strlen_( string ), v;
    for( _idx_ i = l - 1 ; i >= 0 ; i-- )
        if( string[i] == wild )
        {
            v = (int)( ((UINT)n) % ((UINT)nbase) );
            string[i] = ( v > 9 )? (CHAR)( upper ? v - 10 + 'A' : v - 10 + 'a' ) : (CHAR)( v + '0' );
            n /= nbase;
        }
    return( n ? -1 : 1 );
}

///<summary> Finds all wildcard chars defined by <b>wild</b> in the string <b>string</b> and will replace
///  it with the number in <b>n</b>. </summary>
///<param name="n"> the number to insert. Only positive values are supported at the moment. </param>
///<param name="nbase"> the base of the number: <br/>
///   10 = <b>n</b> is a decimal value <br/>
///   16 = <b>n</b> is a hexadecimal value (the numbers A to F will be uppercase). <br/>
///  -16 = <b>n</b> is a hexadecimal value (the numbers a to f will be lowercase). <br/>
///    2 = <b>n</b> is a binary value <br/>
///  Also all other base values are supported <br/>
///  (as long as their absolute values are greater-equal 2). </param>
///<returns>
///   1 = ok <br/>
///  -1 = If in <b>string</b> there are not enough wildcard chars to hold the whole number,
///       but nevertheless it will be tried to replace as many wildcard chars as possible.
///       ( The lowest digits of <b>n</b> will be set. ) <br/>
///   0 = Error </returns>
///<remarks> Note: If there are too much wildcard chars so that the number is smaller, the remaining
///  leading wildcard chars will be filled with '0' chars. </remarks>
int _replace_nr_( PWSTR string, WCHAR wild, int n, int nbase )
{
    bool upper = ( nbase > 0 );
    if( nbase < 0 ) nbase = -nbase;
    if( !string || !wild || n < 0 || nbase < 2 ) return 0;
    _size_t_ l = (int)_strlen_( string ), v;
    for( _idx_ i = l - 1 ; i >= 0 ; i-- )
        if( string[i] == wild )
        {
            v = (int)( ((UINT)n) % ((UINT)nbase) );
            string[i] = ( v > 9 )? (WCHAR)( upper ? v - 10 + L'A' : v - 10 + L'a' ) : (WCHAR)( v + L'0' );
            n /= nbase;
        }
    return( n ? -1 : 1 );
}



///<summary> Searches for a path separator and returns the
///  string pointer of its position, or NULL if not found </summary>
PCSTR _findpathsep_( PCSTR path )
{
    if( !path ) return NULL;
    while( *path ) {
        if( _ispathsep_( *path ) )
            return path;
        ++path;
    }
    return NULL;
}

///<summary> Searches for a path separator and returns the
///  string pointer of its position, or NULL if not found </summary>
PCWSTR _findpathsep_( PCWSTR path )
{
    if( !path ) return NULL;
    while( *path ) {
        if( _ispathsep_( *path ) )
            return path;
        ++path;
    }
    return NULL;
}

///<summary> Returns the default path delimiter char </summary>
int _defpathsep_()
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return '/'  & 0xff;
        case PathT_WindowsPath  : return '\\' & 0xff;
        default                 : return '\\' & 0xff;
    }
}

///<summary> Searches for a path delimiter char and returns
///  this one. If none found, the default path delimiter char
///  will be returned. </summary>
CHAR _getpathsep_( PCSTR path )
{
    if( !path ) return (CHAR)_defpathsep_();
    for( _idx_ ix = 0 ; path[ix] && (ix <= _MAX_PATH_) ; ++ix )
        if( _ispathsep_( path[ix] ) )
            return path[ix];
    return (CHAR)_defpathsep_();
}

///<summary> Searches for a path delimiter char and returns
///  this one. If none found, the default path delimiter char
///  will be returned. </summary>
WCHAR _getpathsep_( PCWSTR path )
{
    if( !path ) return (WCHAR)_defpathsep_();
    for( _idx_ ix = 0 ; path[ix] && (ix <= _MAX_PATH_) ; ++ix )
        if( _ispathsep_( path[ix] ) )
            return path[ix];
    return (WCHAR)_defpathsep_();
}

///<summary>
///  Determines, if the passed string contains a (possible) path
///  (with or without a filename). It *cannot* determine a single
///  filename.
///</summary>
///<param name="strength">
///  How strong a text shall be checked for fitting
///  to a type of filepath. <br/>
///  The lower the value (even negative possible), the
///  more liberal it will be handled (as filepath).
///  <br/><br/>
///  <b> 0 ....              </b> Default <br/>
///  <b> less-equal -10 .... </b> spaces at the beginning and at the end of
///                               the string also will be allowed for filepaths. <br/>
///  <b> less-equal -20 .... </b> leading and trailing spaces - also inside the
///                               string but directly having a path-delimiter as
///                               neighbour - also will be allowed for filepaths.
///</param>
///<returns>
///  <b> 0 = </b> is not a filepath, this is sure. <br/>
///  <b> 1 = </b> seems to be a filepath, but this is not 100% sure. <br/>
///  <b> 2 = </b> seems to be a masked filepath, but this is not 100% sure.
///               This is similar to the return-value 1, but with the difference
///               that the <b>string</b> contains wildcard chars ('?', '*'), too. <br/>
///  <b>-1 = </b> can or cannot be a filepath, it is not determinable
///</returns>
///<remarks>
/// This function will handle both, \ and / chars as path delimiter.
///  <br/><br/>
/// If <b>string</b> is larger than _MAX_PATH_ chars, then 0 will be returned.
///  <br/><br/>
/// <b>The following</b> string sequences are some samples which
///  <u>will</u> be determined as filepath: <br/>
///    "A:", "C:\", "C:\path", "\", "\path", "\path\",
///    "\file.extension", "\\mypc\d_drive\", "path\path"
///    ".", "..", ".\", "..\", "C:\path\.\path\..\path",
///    etc. <br/>
///  <b>The following</b> string sequences are some samples which
///  will <u>not</u> be determined as filepath: <br/>
///    "A", "\\", "\\mypc\d_drive", "\\mypc\d_drive\path:",
///    "x\\mypc\d_drive", "\\\d_drive\", "C:\path\\path",
///    "path\\path", "...",
///    "XY:", "XY:\", "@:", "C::", "C:\\", "C:\path\\", "C:\path:"
///</remarks>
int _isfilepath_( PCSTR string, int strength )
{
    static CHAR notallowed[] = "\"<>|\r\n\t";
    static CHAR wildcards [] = "*?";
    _idx_ i;
    // Note: Not allowed chars:   \ / : * ? " < > |
    // But the \ and / and : must be allowed to support path delimiters.
    // * and ? must also be supported to allow masked filepaths to be supported.
    if( !string ) return 0;
    _ssize_t_ l = _strlenmax_( string, _MAX_PATH_, false );
    // Note: #_strlenmax_()# is used to prevent a time consuming scanning
    // if #string# is e.g. a whole large text file content or something.
    if( l <= 0 ) return -1;
    if( _strcntchrs_( string, notallowed ) > 0 ) return 0;
    int isvalid = ( _strcntchrs_( string, wildcards ) > 0 )? 2 : 1;
    if( l >= 2 ? ( _ispathsep_(string[0]) && _ispathsep_(string[1]) ) : false )
    {
        // check for network-drive (with or without paths)
        if( l < 6 ) return 0;
        int cslashes    = 2;
        int cheadspaces = 0;  // after a pathsep
        int ctailspaces = 0;  // before a pathsep
        int cendspaces  = 0;  // at end of string
        for( i = 2 ; (_ssize_t_) i < l ; i++ ) {
            if( _iscntrl_( string[i] ) ) return 0;
            if( string[i] == ':' ) return 0;
            if( _ispathsep_( string[i] ) ) {
                if( _ispathsep_( string[i-1] ) ) return 0;
                cslashes++;
            }
            if( _ispathsep_(string[i]) && string[i-1] == ' ' ) ctailspaces++;
            if( string[i] == ' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
            if( string[i] == ' ' ) if( string[i+1] == '\0' ) cendspaces++;
            if( string[i] == '.' && string[i-1] == '.' && string[i-2] == '.' ) return 0;
        }
        if( strength > -10 ) if( cheadspaces || ctailspaces || cendspaces ) return 0;
        if( strength > -20 ) if( cheadspaces || ctailspaces               ) return 0;
        if( cslashes < 4 ) return 0;
        return isvalid;
    }
    if( l >= 2 ? ( _isalpha_(string[0]) && string[1] == ':' ) : false )
    {
        // check for normal drive letter (with or without paths)
        int cheadspaces = 0;  // after a pathsep
        int ctailspaces = 0;  // before a pathsep
        int cendspaces  = 0;  // at end of string
        for( i = 2 ; (_ssize_t_) i < l ; i++ ) {
            if( _iscntrl_( string[i] ) ) return 0;
            if( string[i] == ':' ) return 0;
            if( _ispathsep_(string[i]) && _ispathsep_(string[i-1]) ) return 0;
            if( _ispathsep_(string[i]) && string[i-1] == ' ' ) ctailspaces++;
            if( string[i] == ' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
            if( string[i] == ' ' ) if( string[i+1] == '\0' ) cendspaces++;
            if( string[i] == '.' && string[i-1] == '.' && string[i-2] == '.' ) return 0;
        }
        if( strength > -10 ) if( cheadspaces || ctailspaces || cendspaces ) return 0;
        if( strength > -20 ) if( cheadspaces || ctailspaces               ) return 0;
        return isvalid;
    }
    // check for relative path
    int cslashes     = 0;
    int cspaces      = 0;
    int cheadspaces  = 0;   // after a pathsep
    int ctailspaces  = 0;   // before a pathsep
    int cbeginspaces = 0;   // at start of string
    int cendspaces   = 0;   // at end of string
    if( string[0] == ' ' ) cbeginspaces++;
    int nbracketdepth = 0, maxbracketdepth = 0, minbracketdepth = 0;
    int cplus = 0, cminus = 0, cmult = 0;
    int cdblops = 0;
    for( i = 0 ; (_ssize_t_) i < l ; i++ ) {
        if( _iscntrl_( string[i] ) ) return 0;
        if( string[i] == ':' ) return 0;
        if( string[i] == ' ' ) cspaces++;
        if( _ispathsep_( string[i] ) ) cslashes++;
        if( i >= 1 ) {
            if( _ispathsep_(string[i]) && _ispathsep_(string[i-1]) ) return 0;
            if( _ispathsep_(string[i]) && string[i-1] == ' ' ) ctailspaces++;
            if( string[i] == ' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
        }
        if( string[i] == ' ' ) if( string[i+1] == '\0' ) cendspaces++;
        if( i >= 2 ) if( string[i] == '.' && string[i-1] == '.' && string[i-2] == '.' ) return 0;
        // Now check if we can find some specific characterisics for mathematic expressions.
        // This can be useful e.g. for expression evaluators like the <b>ExpressionEvaluator</b> class.
        if( string[i] == '(' || string[i] == '[' || string[i] == '{' ) {
            if( ++nbracketdepth > maxbracketdepth ) maxbracketdepth = nbracketdepth;
        } else if( string[i] == ')' || string[i] == ']' || string[i] == '}' ) {
            if( --nbracketdepth < minbracketdepth ) minbracketdepth = nbracketdepth;
        }
        if( string[i] == '+' ) cplus++;
        if( string[i] == '-' ) cminus++;
        if( string[i] == '*' ) cmult++;
        if( string[i] == '*' ) if( string[i+1] == '*' ) return 0;
        if( string[i] == '+' || string[i] == '-' )
            if( string[i+1] == '+' || string[i+1] == '-' ) cdblops++;
    }
    if( nbracketdepth == 0 && minbracketdepth == 0 ) {
        if( maxbracketdepth > 1 ) return -1;
        if( cplus > 0 && ( cminus > 0 || cmult > 0 ) ) return -1;
        if( cdblops > 0 ) return -1;
    }
    // continue checking for relative path
    if( strength > -10 ) if( cbeginspaces || cheadspaces || ctailspaces || cendspaces ) return 0;
    if( strength > -20 ) if(                 cheadspaces || ctailspaces               ) return 0;
    if( _ispathsep_( string[0  ] ) ) return isvalid;
    if( _ispathsep_( string[l-1] ) ) return isvalid;
    if( cslashes < 2 || cspaces ) return -1;
    return isvalid;
}

///<summary>
///  Determines, if the passed string contains a (possible) path
///  (with or without a filename). It *cannot* determine a single
///  filename.
///</summary>
///<param name="strength">
///  How strong a text shall be checked for fitting
///  to a type of filepath. <br/>
///  The lower the value (even negative possible), the
///  more liberal it will be handled (as filepath).
///  <br/><br/>
///  <b> 0 ....              </b> Default <br/>
///  <b> less-equal -10 .... </b> spaces at the beginning and at the end of
///                               the string also will be allowed for filepaths. <br/>
///  <b> less-equal -20 .... </b> leading and trailing spaces - also inside the
///                               string but directly having a path-delimiter as
///                               neighbour - also will be allowed for filepaths.
///</param>
///<returns>
///  <b> 0 = </b> is not a filepath, this is sure. <br/>
///  <b> 1 = </b> seems to be a filepath, but this is not 100% sure. <br/>
///  <b> 2 = </b> seems to be a masked filepath, but this is not 100% sure.
///               This is similar to the return-value 1, but with the difference
///               that the <b>string</b> contains wildcard chars ('?', '*'), too. <br/>
///  <b>-1 = </b> can or cannot be a filepath, it is not determinable
///</returns>
///<remarks>
/// This function will handle both, \ and / chars as path delimiter.
///  <br/><br/>
/// If <b>string</b> is larger than _MAX_PATH_ chars, then 0 will be returned.
///  <br/><br/>
/// <b>The following</b> string sequences are some samples which
///  <u>will</u> be determined as filepath: <br/>
///    "A:", "C:\", "C:\path", "\", "\path", "\path\",
///    "\file.extension", "\\mypc\d_drive\", "path\path"
///    ".", "..", ".\", "..\", "C:\path\.\path\..\path",
///    etc. <br/>
///  <b>The following</b> string sequences are some samples which
///  will <u>not</u> be determined as filepath: <br/>
///    "A", "\\", "\\mypc\d_drive", "\\mypc\d_drive\path:",
///    "x\\mypc\d_drive", "\\\d_drive\", "C:\path\\path",
///    "path\\path", "...",
///    "XY:", "XY:\", "@:", "C::", "C:\\", "C:\path\\", "C:\path:"
///</remarks>
int _isfilepath_( PCWSTR string, int strength )
{
    static WCHAR notallowed[] = L"\"<>|\r\n\t";
    static WCHAR wildcards [] = L"*?";
    _idx_ i;
    // Note: Not allowed chars:   \ / : * ? " < > |
    // But the \ and / and : must be allowed to support path delimiters.
    // * and ? must also be supported to allow masked filepaths to be supported.
    if( !string ) return 0;
    _ssize_t_ l = _strlenmax_( string, _MAX_PATH_, false );
    // Note: #_strlenmax_()# is used to prevent a time consuming scanning
    // if #string# is e.g. a whole large text file content or something.
    if( l <= 0 ) return -1;
    if( _strcntchrs_( string, notallowed ) > 0 ) return 0;
    int isvalid = ( _strcntchrs_( string, wildcards ) > 0 )? 2 : 1;
    if( l >= 2 ? ( _ispathsep_(string[0]) && _ispathsep_(string[1]) ) : false )
    {
        // check for network-drive (with or without paths)
        if( l < 6 ) return 0;
        int cslashes    = 2;
        int cheadspaces = 0;  // after a pathsep
        int ctailspaces = 0;  // before a pathsep
        int cendspaces  = 0;  // at end of string
        for( i = 2 ; (_ssize_t_) i < l ; i++ ) {
            if( _iscntrl_( string[i] ) ) return 0;
            if( string[i] == L':' ) return 0;
            if( _ispathsep_( string[i] ) ) {
                if( _ispathsep_( string[i-1] ) ) return 0;
                cslashes++;
            }
            if( _ispathsep_(string[i]) && string[i-1] == L' ' ) ctailspaces++;
            if( string[i] == L' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
            if( string[i] == L' ' ) if( string[i+1] == L'\0' ) cendspaces++;
            if( string[i] == L'.' && string[i-1] == L'.' && string[i-2] == L'.' ) return 0;
        }
        if( strength > -10 ) if( cheadspaces || ctailspaces || cendspaces ) return 0;
        if( strength > -20 ) if( cheadspaces || ctailspaces               ) return 0;
        if( cslashes < 4 ) return 0;
        return isvalid;
    }
    if( l >= 2 ? ( _isalpha_(string[0]) && string[1] == L':' ) : false )
    {
        // check for normal drive letter (with or without paths)
        int cheadspaces = 0;  // after a pathsep
        int ctailspaces = 0;  // before a pathsep
        int cendspaces  = 0;  // at end of string
        for( i = 2 ; (_ssize_t_) i < l ; i++ ) {
            if( _iscntrl_( string[i] ) ) return 0;
            if( string[i] == L':' ) return 0;
            if( _ispathsep_(string[i]) && _ispathsep_(string[i-1]) ) return 0;
            if( _ispathsep_(string[i]) && string[i-1] == L' ' ) ctailspaces++;
            if( string[i] == L' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
            if( string[i] == L' ' ) if( string[i+1] == L'\0' ) cendspaces++;
            if( string[i] == L'.' && string[i-1] == L'.' && string[i-2] == L'.' ) return 0;
        }
        if( strength > -10 ) if( cheadspaces || ctailspaces || cendspaces ) return 0;
        if( strength > -20 ) if( cheadspaces || ctailspaces               ) return 0;
        return isvalid;
    }
    // check for relative path
    int cslashes     = 0;
    int cspaces      = 0;
    int cheadspaces  = 0;   // after a pathsep
    int ctailspaces  = 0;   // before a pathsep
    int cbeginspaces = 0;   // at start of string
    int cendspaces   = 0;   // at end of string
    if( string[0] == L' ' ) cbeginspaces++;
    int nbracketdepth = 0, maxbracketdepth = 0, minbracketdepth = 0;
    int cplus = 0, cminus = 0, cmult = 0;
    int cdblops = 0;
    for( i = 0 ; (_ssize_t_) i < l ; i++ ) {
        if( _iscntrl_( string[i] ) ) return 0;
        if( string[i] == L':' ) return 0;
        if( string[i] == L' ' ) cspaces++;
        if( _ispathsep_( string[i] ) ) cslashes++;
        if( i >= 1 ) {
            if( _ispathsep_(string[i]) && _ispathsep_(string[i-1]) ) return 0;
            if( _ispathsep_(string[i]) && string[i-1] == L' ' ) ctailspaces++;
            if( string[i] == L' ' && _ispathsep_(string[i-1]) ) cheadspaces++;
        }
        if( string[i] == L' ' ) if( string[i+1] == L'\0' ) cendspaces++;
        if( i >= 2 ) if( string[i] == L'.' && string[i-1] == L'.' && string[i-2] == L'.' ) return 0;
        // Now check if we can find some specific characterisics for mathematic expressions.
        // This can be useful e.g. for expression evaluators like the <b>ExpressionEvaluator</b> class.
        if( string[i] == L'(' || string[i] == L'[' || string[i] == L'{' ) {
            if( ++nbracketdepth > maxbracketdepth ) maxbracketdepth = nbracketdepth;
        } else if( string[i] == L')' || string[i] == L']' || string[i] == L'}' ) {
            if( --nbracketdepth < minbracketdepth ) minbracketdepth = nbracketdepth;
        }
        if( string[i] == L'+' ) cplus++;
        if( string[i] == L'-' ) cminus++;
        if( string[i] == L'*' ) cmult++;
        if( string[i] == L'*' ) if( string[i+1] == L'*' ) return 0;
        if( string[i] == L'+' || string[i] == L'-' )
            if( string[i+1] == L'+' || string[i+1] == L'-' ) cdblops++;
    }
    if( nbracketdepth == 0 && minbracketdepth == 0 ) {
        if( maxbracketdepth > 1 ) return -1;
        if( cplus > 0 && ( cminus > 0 || cmult > 0 ) ) return -1;
        if( cdblops > 0 ) return -1;
    }
    // continue checking for relative path
    if( strength > -10 ) if( cbeginspaces || cheadspaces || ctailspaces || cendspaces ) return 0;
    if( strength > -20 ) if(                 cheadspaces || ctailspaces               ) return 0;
    if( _ispathsep_( string[0  ] ) ) return isvalid;
    if( _ispathsep_( string[l-1] ) ) return isvalid;
    if( cslashes < 2 || cspaces ) return -1;
    return isvalid;
}

///<summary> Gets the length of the drive of an
///  absolute path or zero if no absolute path. </summary>
///<returns> 2 for e.g. C:, C:path\path <br/>
///          3 for e.g. C:\, C:\path\path <br/>
///          16 for e.g. \\wks\sharedpath <br/>
///          0 = not an absolute path <br/>
///          0 for e.g. \\wks </returns>
_size_t_ _getabsolutepathlen_( PCSTR path, _size_t_ *plength )
{
    if( plength ) *plength = 0;
    if( !path ) return 0;
    if( !*path ) return 0;

    //_size_t_ len = plength ? _strlen_( path ) :
    //                         (_size_t_) _strlenmax_( path, _MAX_PATH_, true );
    _size_t_ len = (_size_t_) _strlenmax_( path, _MAX_PATH_, true );
    if( plength ) *plength = len;
    if( len < 2 ) return 0;

    if( _isalpha_( path[0] ) )
        if( path[1] == ':' ) {
            if( len >= 3 && _ispathsep_( path[2] ) ) return 3;
            else return 2;
        }
    if( len < 3 ) return 0;

    if(  !_ispathsep_( path[0] ) ||
         !_ispathsep_( path[1] ) ||
         _ispathsep_( path[2] )  )
        return 0;

    PCSTR sep = _findpathsep_( path + 2 );
    if( !sep ) return 0;
    if( _ispathsep_( sep[1] ) ) return 0;

    PCSTR sep2 = _findpathsep_( sep + 1 );
    if( !sep2 ) return len;

    return (_size_t_)( sep2 - path ) + 1;   // subtraction handles char-size correctly
}

///<summary> Gets the length of the drive of an
///  absolute path or zero if no absolute path. </summary>
///<returns> 2 for e.g. C:, C:path\path <br/>
///          3 for e.g. C:\, C:\path\path <br/>
///          16 for e.g. \\wks\sharedpath <br/>
///          0 = not an absolute path <br/>
///          0 for e.g. \\wks </returns>
_size_t_ _getabsolutepathlen_( PCWSTR path, _size_t_ *plength )
{
    if( plength ) *plength = 0;
    if( !path ) return 0;
    if( !*path ) return 0;

    //_size_t_ len = plength ? _strlen_( path ) :
    //                         (_size_t_) _strlenmax_( path, _MAX_PATH_, true );
    _size_t_ len = (_size_t_) _strlenmax_( path, _MAX_PATH_, true );
    if( plength ) *plength = len;
    if( len < 2 ) return 0;

    if( _isalpha_( path[0] ) )
        if( path[1] == ':' ) {
            if( len >= 3 && _ispathsep_( path[2] ) ) return 3;
            else return 2;
        }
    if( len < 3 ) return 0;

    if(  !_ispathsep_( path[0] ) ||
         !_ispathsep_( path[1] ) ||
         _ispathsep_( path[2] )  )
        return 0;

    PCWSTR sep = _findpathsep_( path + 2 );
    if( !sep ) return 0;
    if( _ispathsep_( sep[1] ) ) return 0;

    PCWSTR sep2 = _findpathsep_( sep + 1 );
    if( !sep2 ) return len;

    return (_size_t_)( sep2 - path ) + 1;   // subtraction handles char-size correctly
}

///<summary> makes all backslashes and slashes to the path-separator
///  format applicable to the selected operating system specified
///  in <b>g_PathType</b>. </summary>
PSTR _ospath_( PSTR path )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _unixpath_   ( path );
        case PathT_WindowsPath  : return _windowspath_( path );
        default                 : return                path;
    }
}

///<summary> makes all backslashes and slashes to the path-separator
///  format applicable to the selected operating system specified
///  in <b>g_PathType</b>. </summary>
PWSTR _ospath_( PWSTR path )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _unixpath_   ( path );
        case PathT_WindowsPath  : return _windowspath_( path );
        default                 : return                path;
    }
}

///<summary> makes all backslashes to slashes in a path </summary>
PSTR _unixpath_( PSTR path )
{
    if( !path ) return path;
    PSTR p = path;
    for( ; *p ; ++p )
        if( *p == '\\' ) *p = '/';
    return path;
}

///<summary> makes all backslashes to slashes in a path </summary>
PWSTR _unixpath_( PWSTR path )
{
    if( !path ) return path;
    PWSTR p = path;
    for( ; *p ; ++p )
        if( *p == L'\\' ) *p = L'/';
    return path;
}

///<summary> makes all slashes to backslashes in a path </summary>
PSTR _windowspath_( PSTR path )
{
    if( !path ) return path;
    PSTR p = path;
    for( ; *p ; ++p )
        if( *p == '/' ) *p = '\\';
    return path;
}

///<summary> makes all slashes to backslashes in a path </summary>
PWSTR _windowspath_( PWSTR path )
{
    if( !path ) return path;
    PWSTR p = path;
    for( ; *p ; ++p )
        if( *p == L'/' ) *p = L'\\';
    return path;
}

///<summary> makes the parentpath of a path by truncating the last
///  pathname off if present else the path won't be changed. </summary>
///<param name="path"> INOUT: path to make a parentpath of
///  It may contain \ or / as well <br/>
///  It must not contain any "\.\" or "\..\" <br/>
///  E.g.<br/> d:\path\path -> d:\path <br/>
///           d:\path\path\ -> d:\path\ <br/>
///           d:\path -> d:\ <br/>
///           d:\path\ -> d:\ <br/>
///           d:path -> d: <br/>
///           d:path\ -> d: <br/>
///           \path -> \ <br/>
///           path ->  <br/>
///           d:, d:\, \, , \\mypc\d_drive won't be changed </param>
///<remarks> NOTE: As drive also strings like \\mypc\c_drive are handled correctly. In such
/// a case, if the result only contains the drive name and no path any more, the drive name
/// will *not* contain a trailing path delimiter. </remarks>
PSTR _parentpath_( PSTR path )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );
    path[len] = 0;

    if( len == 0 || len == abslen )
        return path;

    _idx_ ix = len - 1;
    bool bRootSep = _ispathsep_( path[0] ) && abslen == 0;
    bool bEndSep = _ispathsep_( path[ix] );
    if( bEndSep && len == 1 )   // do not make parent path from \ root path only
        return path;
    if( bEndSep )
        --ix;

    if( ix <= 0 ) {
        path[0] = 0;
        return path;
    }

    for( ; !_ispathsep_( path[ix] ) ; --ix )
        if( ix == 0 ) {
            if( (_size_t_) ix < abslen ) ix = (_idx_) abslen;
            if( bRootSep && ix == 0 ) ++ix;
            path[ ix ] = 0;
            return path;
        }

    if( bEndSep ) ++ix;
    if( (_size_t_) ix < abslen ) ix = (_idx_) abslen;
    if( bRootSep && ix == 0 ) ++ix;
    path[ ix ] = 0;

    return path;
}

///<summary> makes the parentpath of a path by truncating the last
///  pathname off if present else the path won't be changed. </summary>
///<param name="path"> INOUT: path to make a parentpath of
///  It may contain \ or / as well <br/>
///  It must not contain any "\.\" or "\..\" <br/>
///  E.g.<br/> d:\path\path -> d:\path <br/>
///           d:\path\path\ -> d:\path\ <br/>
///           d:\path -> d:\ <br/>
///           d:\path\ -> d:\ <br/>
///           d:path -> d: <br/>
///           d:path\ -> d: <br/>
///           \path -> \ <br/>
///           path ->  <br/>
///           d:, d:\, \, , \\mypc\d_drive won't be changed </param>
///<remarks> NOTE: As drive also strings like \\mypc\c_drive are handled correctly. In such
/// a case, if the result only contains the drive name and no path any more, the drive name
/// will *not* contain a trailing path delimiter. </remarks>
PWSTR _parentpath_( PWSTR path )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );
    path[len] = 0;

    if( len == 0 || len == abslen )
        return path;

    _idx_ ix = len - 1;
    bool bRootSep = _ispathsep_( path[0] ) && abslen == 0;
    bool bEndSep = _ispathsep_( path[ix] );
    if( bEndSep && len == 1 )   // do not make parent path from \ root path only
        return path;
    if( bEndSep )
        --ix;

    if( ix <= 0 ) {
        path[0] = 0;
        return path;
    }

    for( ; !_ispathsep_( path[ix] ) ; --ix )
        if( ix == 0 ) {
            if( (_size_t_) ix < abslen ) ix = (_idx_) abslen;
            if( bRootSep && ix == 0 ) ++ix;
            path[ ix ] = 0;
            return path;
        }

    if( bEndSep ) ++ix;
    if( (_size_t_) ix < abslen ) ix = (_idx_) abslen;
    if( bRootSep && ix == 0 ) ++ix;
    path[ ix ] = 0;

    return path;
}

///<summary> truncates a path so that only its root path is contained. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
PSTR _rootpath_( PSTR path, _size_t_ size )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    path[abslen] = 0;
    len = abslen;

    _size_t_ maxlen = size - 1;

    if( len <= maxlen - 1 )
        if( len > 0 ? !_ispathsep_( path[len-1] ) : true ) {
            path[len  ] = _getpathsep_( path );
            path[len+1] = 0;
        }

    return path;
}

///<summary> truncates a path so that only its root path is contained. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
PWSTR _rootpath_( PWSTR path, _size_t_ size )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    path[abslen] = 0;
    len = abslen;

    _size_t_ maxlen = size - 1;

    if( len <= maxlen - 1 )
        if( len > 0 ? !_ispathsep_( path[len-1] ) : true ) {
            path[len  ] = _getpathsep_( path );
            path[len+1] = 0;
        }

    return path;
}

_size_t_ _rootpathlen_( PCSTR path )
{
    if( !path ) return 0;

    _size_t_ abslen = _getabsolutepathlen_( path, NULL );

    if( abslen > 0 ? !_ispathsep_( path[abslen-1] ) : true ) {
        ++abslen;
    }

    return abslen;
}

_size_t_ _rootpathlen_( PCWSTR path )
{
    if( !path ) return 0;

    _size_t_ abslen = _getabsolutepathlen_( path, NULL );

    if( abslen > 0 ? !_ispathsep_( path[abslen-1] ) : true ) {
        ++abslen;
    }

    return abslen;
}

///<summary> Removes a trailing path separator.
///  But it will not be done if the path needs the trailing
///  path delimiter to be the same path as before. </summary>
///<remarks> This is the vice versa function to <b>_aspath_()</b>. </remarks>
PSTR _strippath_( PSTR path )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    // We do *not* remove a trailing path separator
    // if it is necessary to specify a correct root drive.
    if( len == abslen )
        return path;

    if( len > 0 )
        if( _ispathsep_( path[len - 1] ) )
            path[len - 1] = 0;

    return path;
}

///<summary> Removes a trailing path separator.
///  But it will not be done if the path needs the trailing
///  path delimiter to be the same path as before. </summary>
///<remarks> This is the vice versa function to <b>_aspath_()</b>. </remarks>
PWSTR _strippath_( PWSTR path )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    // We do *not* remove a trailing path separator
    // if it is necessary to specify a correct root drive.
    if( len == abslen )
        return path;

    if( len > 0 )
        if( _ispathsep_( path[len - 1] ) )
            path[len - 1] = 0;

    return path;
}

///<summary> Appends a trailing path separator if there is none. </summary>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<remarks> This is the vice versa function to <b>_strippath_()</b>. <br/>
///  Note: If <b>path</b> is an empty string, a path delimiter will
///   *not* be appended because an empty string as directory indicates
///   the current directory and setting a path delimiter will jump
///   to the root directory. This must not happen. </remarks>
PSTR _aspath_( PSTR path, _size_t_ size )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    _size_t_ maxlen = size - 1;

    // If we have something like C:, we do *not*
    // append a path separator, because this would
    // jump to the root directory, which should not happen.
    if( abslen == 2 && len == abslen )
        return path;

    if( len > 0 && len < maxlen )
        if( !_ispathsep_( path[len-1] ) ) {
            path[len  ] = _getpathsep_( path );
            path[len+1] = 0;
        }

    return path;
}

///<summary> Appends a trailing path separator if there is none. </summary>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<remarks> This is the vice versa function to <b>_strippath_()</b>. <br/>
///  Note: If <b>path</b> is an empty string, a path delimiter will
///   *not* be appended because an empty string as directory indicates
///   the current directory and setting a path delimiter will jump
///   to the root directory. This must not happen. </remarks>
PWSTR _aspath_( PWSTR path, _size_t_ size )
{
    if( !path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    _size_t_ maxlen = size - 1;

    // If we have something like C:, we do *not*
    // append a path separator, because this would
    // jump to the root directory, which should not happen.
    if( abslen == 2 && len == abslen )
        return path;

    if( len > 0 && len < maxlen )
        if( !_ispathsep_( path[len-1] ) ) {
            path[len  ] = _getpathsep_( path );
            path[len+1] = 0;
        }

    return path;
}

_size_t_ _aspathlen_( PCSTR path )
{
    if( !path ) return 0;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    // If we have something like C:, we do *not*
    // append a path separator, because this would
    // jump to the root directory, which should not happen.
    if( abslen == 2 && len == abslen )
        return len;

    if( len > 0 )
        if( !_ispathsep_( path[len-1] ) ) {
            ++len;
        }

    return len;
}

_size_t_ _aspathlen_( PCWSTR path )
{
    if( !path ) return 0;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    // If we have something like C:, we do *not*
    // append a path separator, because this would
    // jump to the root directory, which should not happen.
    if( abslen == 2 && len == abslen )
        return len;

    if( len > 0 )
        if( !_ispathsep_( path[len-1] ) ) {
            ++len;
        }

    return len;
}

///<summary> Checks if the string ends with a path delimiter. </summary>
///<remarks> It can be that a path does *not* have a trailing path
///  delimiter even after a call to <b>_aspath_()</b> due to the fact
///  that <b>_aspath_()</b> will append a path delimiter only if
///  it would not change the path (e.g. imagine something like
///  C: which includes C's current directory). </remarks>
bool _isendpathsep_( PCSTR path )
{
    if( !path ) return false;
    _size_t_ len = _strlen_( path );
    if( len == 0 ) return false;
    return _ispathsep_( path[len - 1] );
}

///<summary> Checks if the string ends with a path delimiter. </summary>
///<remarks> It can be that a path does *not* have a trailing path
///  delimiter even after a call to <b>_aspath_()</b> due to the fact
///  that <b>_aspath_()</b> will append a path delimiter only if
///  it would not change the path (e.g. imagine something like
///  C: which includes C's current directory). </remarks>
bool _isendpathsep_( PCWSTR path )
{
    if( !path ) return false;
    _size_t_ len = _strlen_( path );
    if( len == 0 ) return false;
    return _ispathsep_( path[len - 1] );
}

///<summary> Similar to <b>_makepath()</b>. <br/>
///  Builds a path name from components and store it to the array
///  pointed to by <b>dest</b>. The array <b>dest</b> should be
///  of size <b>size</b> in CHARs. </summary>
///<remarks> If <b>drive</b> is not NULL and does not point to
///   an empty string, the first character of the string (a drive
///   name) pointed to by <b>drive</b> followed by a colon is stored
///   to the array pointed to by <b>dest</b>. <br/>
///  If <b>dir</b> is not NULL and does not point to an empty
///   string, the string (a directory name) pointed to by <b>dir</b>
///   is copied to <b>dest</b>. If it does not end with a path
///   delimiter, <b>_makepath_()</b> appends a path delimiter. <br/>
///  If <b>ftitle</b> is not NULL, the string pointed to by
///   <b>ftitle</b> (a filetitle) is copied to <b>dest</b>. <br/>
///  If <b>ext</b> is not NULL and does not point to an empty
///   string, the string pointed to by <b>ext</b> (an extension)
///   is copied to <b>dest</b>. If the string pointed to by
///   <b>ext</b> does not start with a '.' character,
///   <b>_makepath_()</b> inserts a '.' in front of the string
///   pointed to by <b>ext</b>. <br/>
///  If the length of the resulting string exceeds <b>size - 1</b>
///   (in CHARs), the string is truncated to <b>size - 1</b>
///   characters (not including the terminating 0). <br/>
///  NOTE: As <b>drive</b> also strings like \\mypc\c_drive
///   are supported. </remarks>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> Returns <b>dest</b>. </returns>
PSTR _makepath_( PSTR dest, PCSTR drive, PCSTR dir,
                 PCSTR ftitle, PCSTR ext, _size_t_ size )
{
    if( !dest || size < 1 ) return dest;
    *dest = 0;

    if( !_isstrempty_( drive ) )
    {
        _size_t_ abslen = _getabsolutepathlen_( drive );
        if( abslen > size - 1 ) return dest;
        _strcat_( dest, drive, abslen + 1 );
    }

    if( !_isstrempty_( dir ) )
    {
        _aspath_( dest, size );
        // Now if #dest# already ends with a pathsep (from #_aspath_()#)
        // and #dir# starts with one, we must skip it.
        if( _isendpathsep_( dest ) && _ispathsep_( *dir ) )
            ++dir;
        _strcat_( dest, dir, size );
        _aspath_( dest, size );
    }

    if( ftitle )
    {
        _aspath_( dest, size );
        _strcat_( dest, ftitle, size );
    }

    if( !_isstrempty_( ext ) )
    {
        if( !ftitle ) _aspath_( dest, size );
        if( *ext != '.' )
            _strcat_( dest, ".", size );
        _strcat_( dest, ext, size );
    }

    return dest;
}

///<summary> Similar to <b>_makepath()</b>. <br/>
///  Builds a path name from components and store it to the array
///  pointed to by <b>dest</b>. The array <b>dest</b> should be
///  of size <b>size</b> in WCHARs. </summary>
///<remarks> If <b>drive</b> is not NULL and does not point to
///   an empty string, the first character of the string (a drive
///   name) pointed to by <b>drive</b> followed by a colon is stored
///   to the array pointed to by <b>dest</b>. <br/>
///  If <b>dir</b> is not NULL and does not point to an empty
///   string, the string (a directory name) pointed to by <b>dir</b>
///   is copied to <b>dest</b>. If it does not end with a path
///   delimiter, <b>_makepath_()</b> appends a path delimiter. <br/>
///  If <b>ftitle</b> is not NULL, the string pointed to by
///   <b>ftitle</b> (a filetitle) is copied to <b>dest</b>. <br/>
///  If <b>ext</b> is not NULL and does not point to an empty
///   string, the string pointed to by <b>ext</b> (an extension)
///   is copied to <b>dest</b>. If the string pointed to by
///   <b>ext</b> does not start with a '.' character,
///   <b>_makepath_()</b> inserts a '.' in front of the string
///   pointed to by <b>ext</b>. <br/>
///  If the length of the resulting string exceeds <b>size - 1</b>
///   (in WCHARs), the string is truncated to <b>size - 1</b>
///   characters (not including the terminating 0). <br/>
///  NOTE: As <b>drive</b> also strings like \\mypc\c_drive
///   are supported. </remarks>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> Returns <b>dest</b>. </returns>
PWSTR _makepath_( PWSTR dest, PCWSTR drive, PCWSTR dir,
                  PCWSTR ftitle, PCWSTR ext, _size_t_ size )
{
    if( !dest || size < 1 ) return dest;
    *dest = 0;

    if( !_isstrempty_( drive ) )
    {
        _size_t_ abslen = _getabsolutepathlen_( drive );
        if( abslen > size - 1 ) return dest;
        _strcat_( dest, drive, abslen + 1 );
    }

    if( !_isstrempty_( dir ) )
    {
        _aspath_( dest, size );
        // Now if #dest# already ends with a pathsep (from #_aspath_()#)
        // and #dir# starts with one, we must skip it.
        if( _isendpathsep_( dest ) && _ispathsep_( *dir ) )
            ++dir;
        _strcat_( dest, dir, size );
        _aspath_( dest, size );
    }

    if( ftitle )
    {
        _aspath_( dest, size );
        _strcat_( dest, ftitle, size );
    }

    if( !_isstrempty_( ext ) )
    {
        if( !ftitle ) _aspath_( dest, size );
        if( *ext != L'.' )
            _strcat_( dest, L".", size );
        _strcat_( dest, ext, size );
    }

    return dest;
}

///<summary> Similar to <b>_splitpath()</b>. <br/>
///  Splits the path name pointed to by <b>srce</b> into
///  its components. </summary>
///<remarks> The <i>drive name</i> (including the colon) is
///   stored in the string pointed to by <b>drive</b>.
///   A drive name also can be something like \\mypc\c_drive. <br/>
///  The <i>directory</i> (including a trailing path delimiter) will be
///   copied to the string pointed to by <b>dir</b>. <br/>
///  The <i>filename</i> without extension is copied to the string
///   pointed to by <b>ftitle</b>. <br/>
///  The <i>extension</i> (including the dot) is copied to the string
///   pointed to by <b>ext</b>. <br/>
///  NOTE: The 4 destintion buffers should be of size
///   <b>size_each</b> each (including terminating 0). <br/>
///   If one of the pointers (except for <b>srce</b>) is NULL,
///   then that component will not be stored. <br/>
///  REMARK: If the <b>srce</b> ends with *no* path delimiter,
///   then the last word-name will be interpreted as filename
///   (with a possible extension included). <br/>
///  NOTE: As <b>drive</b> also strings like \\mypc\c_drive
///   are supported. </remarks>
///<param name="size_each"> Size of each of the 4 destination
///  string buffers in CHARs (including terminating 0). </param>
void _splitpath_( PCSTR srce, PSTR drive, PSTR dir,
                  PSTR ftitle, PSTR ext, _size_t_ size_each )
{
    if( size_each <= 0 ) return;
    if( drive  ) *drive  = 0;
    if( dir    ) *dir    = 0;
    if( ftitle ) *ftitle = 0;
    if( ext    ) *ext    = 0;
    if( !srce ) return;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( srce, &len );

    if( drive )
    {
        if( abslen > size_each - 1 ) return;
        _strcpy_( drive, srce, abslen + 1 );
    }

    _idx_ ilastpathsep = -1;
    if( dir || ftitle || ext )
        for( _idx_ i = abslen ; srce[i] ; ++i )
            if( _ispathsep_( srce[i] ) )
                ilastpathsep = i;

    if( dir )
    {
        bool bRootDir = ( _ispathsep_( srce[abslen] ) ) ||
                        ( abslen > 0 ? _ispathsep_( srce[abslen - 1] ) : false );
        if( bRootDir )
            _straddch_( dir, _getpathsep_( srce ), size_each );  // set preceding root path delimiter if present
        _idx_ i = abslen;
        if( _ispathsep_( srce[abslen] ) )
            ++i;
        _size_t_ srcedirlen = ilastpathsep - i;
        if( ilastpathsep >= 0 )
            ++srcedirlen;   // include trailing path delimiter if present
        if( srcedirlen > 0 ) {
            if( _strlen_(dir) + srcedirlen > size_each - 1 ) return;
            _strcat_( dir, srce + i, srcedirlen + 1 );
        }
    }

    _idx_ iext = -1;
    if( ftitle || ext )
        for( _idx_ i = len - 1 ; i >= 0 && iext < 0 ; --i )
            if( srce[i] == '.' ) iext = i;
            else if( _ispathsep_( srce[i] ) ) break;

    if( ftitle )
    {
        _idx_ i = ilastpathsep + 1;
        _size_t_ ftitlelen = ( iext >= 0 ? iext : len ) - i;
        if( ftitlelen > 0 ) {
            if( ftitlelen > size_each - 1 ) return;
            _strcat_( ftitle, srce + i, ftitlelen + 1 );
        }
    }

    if( ext )
    {
        _size_t_ extlen = len - iext;
        if( iext >= 0 && extlen > 0 ) {
            if( extlen > size_each - 1 ) return;
            _strcat_( ext, srce + iext, extlen + 1 );
        }
    }
}

///<summary> Similar to <b>_splitpath()</b>. <br/>
///  Splits the path name pointed to by <b>srce</b> into
///  its components. </summary>
///<remarks> The <i>drive name</i> (including the colon) is
///   stored in the string pointed to by <b>drive</b>.
///   A drive name also can be something like \\mypc\c_drive. <br/>
///  The <i>directory</i> (including a trailing path delimiter) will be
///   copied to the string pointed to by <b>dir</b>. <br/>
///  The <i>filename</i> without extension is copied to the string
///   pointed to by <b>ftitle</b>. <br/>
///  The <i>extension</i> (including the dot) is copied to the string
///   pointed to by <b>ext</b>. <br/>
///  NOTE: The 4 destintion buffers should be of size
///   <b>size_each</b> each (including terminating 0). <br/>
///   If one of the pointers (except for <b>srce</b>) is NULL,
///   then that component will not be stored. <br/>
///  REMARK: If the <b>srce</b> ends with *no* path delimiter,
///   then the last word-name will be interpreted as filename
///   (with a possible extension included). <br/>
///  NOTE: As <b>drive</b> also strings like \\mypc\c_drive
///   are supported. </remarks>
///<param name="size_each"> Size of each of the 4 destination
///  string buffers in WCHARs (including terminating 0). </param>
void _splitpath_( PCWSTR srce, PWSTR drive, PWSTR dir,
                  PWSTR ftitle, PWSTR ext, _size_t_ size_each )
{
    if( size_each <= 0 ) return;
    if( drive  ) *drive  = 0;
    if( dir    ) *dir    = 0;
    if( ftitle ) *ftitle = 0;
    if( ext    ) *ext    = 0;
    if( !srce ) return;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( srce, &len );

    if( drive )
    {
        if( abslen > size_each - 1 ) return;
        _strcpy_( drive, srce, abslen + 1 );
    }

    _idx_ ilastpathsep = -1;
    if( dir || ftitle || ext )
        for( _idx_ i = abslen ; srce[i] ; ++i )
            if( _ispathsep_( srce[i] ) )
                ilastpathsep = i;

    if( dir )
    {
        bool bRootDir = ( _ispathsep_( srce[abslen] ) ) ||
                        ( abslen > 0 ? _ispathsep_( srce[abslen - 1] ) : false );
        if( bRootDir )
            _straddch_( dir, _getpathsep_( srce ), size_each );  // set preceding root path delimiter if present
        _idx_ i = abslen;
        if( _ispathsep_( srce[abslen] ) )
            ++i;
        _size_t_ srcedirlen = ilastpathsep - i;
        if( ilastpathsep >= 0 )
            ++srcedirlen;   // include trailing path delimiter if present
        if( srcedirlen > 0 ) {
            if( _strlen_(dir) + srcedirlen > size_each - 1 ) return;
            _strcat_( dir, srce + i, srcedirlen + 1 );
        }
    }

    _idx_ iext = -1;
    if( ftitle || ext )
        for( _idx_ i = len - 1 ; i >= 0 && iext < 0 ; --i )
            if( srce[i] == L'.' ) iext = i;
            else if( _ispathsep_( srce[i] ) ) break;

    if( ftitle )
    {
        _idx_ i = ilastpathsep + 1;
        _size_t_ ftitlelen = ( iext >= 0 ? iext : len ) - i;
        if( ftitlelen > 0 ) {
            if( ftitlelen > size_each - 1 ) return;
            _strcat_( ftitle, srce + i, ftitlelen + 1 );
        }
    }

    if( ext )
    {
        _size_t_ extlen = len - iext;
        if( iext >= 0 && extlen > 0 ) {
            if( extlen > size_each - 1 ) return;
            _strcat_( ext, srce + iext, extlen + 1 );
        }
    }
}

///<summary> Makes a path from 2 paths by concatenating them. </summary>
///<param name="dest"> path in return (this may not be a
///  fullpath if <b>pre</b> was not a fullpath). <br/>
///  The last character will be a path delimiter or it will be
///  not depending on the last character of the string in
///  <b>post</b>. If <b>post</b> is NULL, the ending depends on
///  the last character of the <b>pre</b> string. </param>
///<param name="pre"> main path with or without trailing path delimiter. <br/>
///  Samples: d:\path, \path\ <br/>
///  Normally this should be a fullpath. Only if it is something
///  like d:path\path, a wrong path could be returned because
///  no path delimiter follows the d:. </param>
///<param name="post"> path or only a filename being appended. <br/>
///  If this one ends with a path delimiter, the destination
///  path will have a trailing path delimiter. Otherwise the destination
///  path will also not have a trailing path delimiter. <br/>
///  Samples: file.name, \file.name, ../file.name, path\, .\path <br/>
///  Note: If <b>post</b> is already a fullpath -> <b>pre</b> will be ignored. </param>
///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: As drive for <b>pre</b> or <b>post</b> or <b>dest</b>
///   also strings like `\\mypc\c_drive` are supported. <br/>
///  NOTE: The buffer <b>dest</b> may be equal to one of the buffers
///   <b>pre</b> or <b>post</b>. <br/>
///  Note: If one of <b>pre</b> and <b>post</b> is NULL or is an
///   empty string -> the other one will be copied into <b>dest</b>. </remarks>
PSTR _joinpaths_( PSTR dest, PCSTR pre, PCSTR post, _size_t_ size )
{
    if( !dest || size <= 0 ) return dest;
    if( !pre  ) pre  = "";
    if( !post ) post = "";

    if( !*pre  ) return _strcpy_( dest, post, size );
    if( !*post ) return _strcpy_( dest, pre,  size );

    if( _isabsolutepath_( post ) )
        return _strcpy_( dest, post, size );

    PCSTR pstr2 = post;

    CHAR tmp[ _MAX_PATH_ * 2 + 1 ];
    _size_t_ tmpsize = _countof(tmp);
    _strcpy_( tmp, pre, tmpsize );
    for( ;; )
    {
        CHAR dirname[ _MAX_PATH_ + 1 ] = "";
        _size_t_ max_dirname_len = (_size_t_)_countof(dirname) - 1;

        PCSTR dirname_start = pstr2;
        while( *pstr2 && !_ispathsep_( *pstr2 ) )
            ++pstr2;
        PCSTR dirname_end = pstr2;

        if( _ispathsep_( *pstr2 ) )
            ++pstr2;
        bool bEndSep = ( pstr2 > dirname_end );
        if( dirname_start == dirname_end && dirname_start == post )
            dirname_end = pstr2; // if root path delimiter -> include it in dirname
        while( _ispathsep_( *pstr2 ) )
            ++pstr2;

        _size_t_ dirname_len = (_size_t_)( dirname_end - dirname_start );   // subtraction handles char-size correctly
        if( dirname_len > max_dirname_len )
            dirname_len = max_dirname_len;
        _strcpy_( dirname, dirname_start, dirname_len + 1 );

        if( _patheq_( dirname, "." ) ) {
            // nothing to do in here.
        } else if( _patheq_( dirname, ".." ) ) {
            _parentpath_( tmp );
        } else if( _ispathsep_( *dirname ) ) {
            _rootpath_( tmp, tmpsize );
        } else {
            _aspath_( tmp, tmpsize );
            _strcat_( tmp, dirname, tmpsize );
        }

        if( !*pstr2 ) {
            if( bEndSep )
                _aspath_( tmp, tmpsize );
            break;
        }
    }

    return _strcpy_( dest, tmp, size );
}

///<summary> Makes a path from 2 paths by concatenating them. </summary>
///<param name="dest"> path in return (this may not be a
///  fullpath if <b>pre</b> was not a fullpath). <br/>
///  The last character will be a path delimiter or it will be
///  not depending on the last character of the string in
///  <b>post</b>. If <b>post</b> is NULL, the ending depends on
///  the last character of the <b>pre</b> string. </param>
///<param name="pre"> main path with or without trailing path delimiter. <br/>
///  Samples: d:\path, \path\ <br/>
///  Normally this should be a fullpath. Only if it is something
///  like d:path\path, a wrong path could be returned because
///  no path delimiter follows the d:. </param>
///<param name="post"> path or only a filename being appended. <br/>
///  If this one ends with a path delimiter, the destination
///  path will have a trailing path delimiter. Otherwise the destination
///  path will also not have a trailing path delimiter. <br/>
///  Samples: file.name, \file.name, ../file.name, path\, .\path <br/>
///  Note: If <b>post</b> is already a fullpath -> <b>pre</b> will be ignored. </param>
///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
///<returns> Returns <b>dest</b>. </returns>
///<remarks> NOTE: As drive for <b>pre</b> or <b>post</b> or <b>dest</b>
///   also strings like `\\mypc\c_drive` are supported. <br/>
///  NOTE: The buffer <b>dest</b> may be equal to one of the buffers
///   <b>pre</b> or <b>post</b>. <br/>
///  Note: If one of <b>pre</b> and <b>post</b> is NULL or is an
///   empty string -> the other one will be copied into <b>dest</b>. </remarks>
PWSTR _joinpaths_( PWSTR dest, PCWSTR pre, PCWSTR post, _size_t_ size )
{
    if( !dest || size <= 0 ) return dest;
    if( !pre  ) pre  = L"";
    if( !post ) post = L"";

    if( !*pre  ) return _strcpy_( dest, post, size );
    if( !*post ) return _strcpy_( dest, pre,  size );

    if( _isabsolutepath_( post ) )
        return _strcpy_( dest, post, size );

    PCWSTR pstr2 = post;

    WCHAR tmp[ _MAX_PATH_ * 2 + 1 ];
    _size_t_ tmpsize = _countof(tmp);
    _strcpy_( tmp, pre, tmpsize );
    for( ;; )
    {
        WCHAR dirname[ _MAX_PATH_ + 1 ] = L"";
        _size_t_ max_dirname_len = (_size_t_)_countof(dirname) - 1;

        PCWSTR dirname_start = pstr2;
        while( *pstr2 && !_ispathsep_( *pstr2 ) )
            ++pstr2;
        PCWSTR dirname_end = pstr2;

        if( _ispathsep_( *pstr2 ) )
            ++pstr2;
        bool bEndSep = ( pstr2 > dirname_end );
        if( dirname_start == dirname_end && dirname_start == post )
            dirname_end = pstr2; // if root path delimiter -> include it in dirname
        while( _ispathsep_( *pstr2 ) )
            ++pstr2;

        _size_t_ dirname_len = (_size_t_)( dirname_end - dirname_start );   // subtraction handles char-size correctly
        if( dirname_len > max_dirname_len )
            dirname_len = max_dirname_len;
        _strcpy_( dirname, dirname_start, dirname_len + 1 );

        if( _patheq_( dirname, L"." ) ) {
            // nothing to do in here.
        } else if( _patheq_( dirname, L".." ) ) {
            _parentpath_( tmp );
        } else if( _ispathsep_( *dirname ) ) {
            _rootpath_( tmp, tmpsize );
        } else {
            _aspath_( tmp, tmpsize );
            _strcat_( tmp, dirname, tmpsize );
        }

        if( !*pstr2 ) {
            if( bEndSep )
                _aspath_( tmp, tmpsize );
            break;
        }
    }

    return _strcpy_( dest, tmp, size );
}

bool _driveof_( PSTR drive, PCSTR path, _size_t_ size )
{
    if( !drive || size <= 0 || !path ) return false;
    _size_t_ abslen = _getabsolutepathlen_( path );
    if( abslen == 0 ) return false;
    if( abslen >= size ) return false;
    _strcpy_( drive, path, abslen + 1 );
    return true;
}

bool _driveof_( PWSTR drive, PCWSTR path, _size_t_ size )
{
    if( !drive || size <= 0 || !path ) return false;
    _size_t_ abslen = _getabsolutepathlen_( path );
    if( abslen == 0 ) return false;
    if( abslen >= size ) return false;
    _strcpy_( drive, path, abslen + 1 );
    return true;
}

PSTR _trunc_to_dir_( PSTR path )
{
    if( !path ) return path;
    if( !*path ) return path;

    _size_t_ len = _strlen_( path );
    if( _ispathsep_( path[len-1] ) )
        return path;

    return _parentpath_( path );
}

PWSTR _trunc_to_dir_( PWSTR path )
{
    if( !path ) return path;
    if( !*path ) return path;

    _size_t_ len = _strlen_( path );
    if( _ispathsep_( path[len-1] ) )
        return path;

    return _parentpath_( path );
}

PSTR _directoryof_( PSTR dir, PCSTR path, _size_t_ size )
{
    if( !dir || !path ) return dir;
    CHAR tmp[ _MAX_PATH_ * 2 + 1 ];
    _strcpy_( tmp, path, _countof(tmp) );
    _trunc_to_dir_( tmp );
    return _strcpy_( dir, tmp, size );
}

PWSTR _directoryof_( PWSTR dir, PCWSTR path, _size_t_ size )
{
    if( !dir || !path ) return dir;
    WCHAR tmp[ _MAX_PATH_ * 2 + 1 ];
    _strcpy_( tmp, path, _countof(tmp) );
    _trunc_to_dir_( tmp );
    return _strcpy_( dir, tmp, size );
}

PCSTR _filenameptrof_( PCSTR path )
{
    if( !path ) return path;
    if( !*path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    _idx_ ix = (_idx_) len - 1;
    for( ; ix > (_idx_) abslen ; --ix )
        if( _ispathsep_( path[ix] ) )
            break;

    if( _ispathsep_( path[ix] ) )
        ix++;

    return path + ix;
}

PCWSTR _filenameptrof_( PCWSTR path )
{
    if( !path ) return path;
    if( !*path ) return path;

    _size_t_ len;
    _size_t_ abslen = _getabsolutepathlen_( path, &len );

    _idx_ ix = (_idx_) len - 1;
    for( ; ix > (_idx_) abslen ; --ix )
        if( _ispathsep_( path[ix] ) )
            break;

    if( _ispathsep_( path[ix] ) )
        ix++;

    return path + ix;
}

PSTR _filenameof_( PSTR filename, PCSTR path, _size_t_ size )
{
    CHAR ftitle[_MAX_PATH_ + 1], ext[_MAX_PATH_ + 1];
    _splitpath_( path, NULL, NULL, ftitle, ext, _MAX_PATH_ + 1 );
    return _makepath_( filename, NULL, NULL, ftitle, ext, size );
}

PWSTR _filenameof_( PWSTR filename, PCWSTR path, _size_t_ size )
{
    WCHAR ftitle[_MAX_PATH_ + 1], ext[_MAX_PATH_ + 1];
    _splitpath_( path, NULL, NULL, ftitle, ext, _MAX_PATH_ + 1 );
    return _makepath_( filename, NULL, NULL, ftitle, ext, size );
}

PSTR _filetitleof_( PSTR filetitle, PCSTR path, _size_t_ size )
{
    _splitpath_( path, NULL, NULL, filetitle, NULL, size );
    return filetitle;
}

PWSTR _filetitleof_( PWSTR filetitle, PCWSTR path, _size_t_ size )
{
    _splitpath_( path, NULL, NULL, filetitle, NULL, size );
    return filetitle;
}

PSTR _fileextof_( PSTR fileext, PCSTR path, _size_t_ size )
{
    _splitpath_( path, NULL, NULL, NULL, fileext, size );
    return fileext;
}

PWSTR _fileextof_( PWSTR fileext, PCWSTR path, _size_t_ size )
{
    _splitpath_( path, NULL, NULL, NULL, fileext, size );
    return fileext;
}

static int _s_pathcmp_( PCSTR path1, PCSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return 0;
    _idx_ l;
    for( l = 0 ; *path1 || *path2 ; path1++, path2++, l++ )
    {
        if( _ispathsep_(*path1) && _ispathsep_(*path2) ) continue;
        int d = bCaseSensitive
                ? (int)(UCHAR)*path1            - (int)(UCHAR)*path2
                : (int)(UCHAR)_toupper_(*path1) - (int)(UCHAR)_toupper_(*path2);
        if( d ) {
            // If we reached at least the end of one of the strings, and
            // if the strings only differ that way that one has a trailing
            // path separator while the other not, then treat the paths
            // as equal. But this is allowed only (1) when the last char is
            // not a ':' (where a following path delimiter makes a difference
            // between a current directory and a root directory)
            // and (2) when not just a root directory is specified (so that
            // one is an empty string while the other only contains a
            // path delimiter indicating a root directory).
            if( l > 0 && !*path1 && _ispathsep_(*path2) ) if( !path2[1] && path1[-1] != ':' ) return 0;
            if( l > 0 && !*path2 && _ispathsep_(*path1) ) if( !path1[1] && path2[-1] != ':' ) return 0;
            return d;
        }
    }
    return 0;
}

static int _s_pathcmp_( PCWSTR path1, PCWSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return 0;
    _idx_ l;
    for( l = 0 ; *path1 || *path2 ; path1++, path2++, l++ )
    {
        if( _ispathsep_(*path1) && _ispathsep_(*path2) ) continue;
        int d = bCaseSensitive
                ? (int)(USHORT)*path1            - (int)(USHORT)*path2
                : (int)(USHORT)_toupper_(*path1) - (int)(USHORT)_toupper_(*path2);
        if( d ) {
            // If we reached at least the end of one of the strings, and
            // if the strings only differ that way that one has a trailing
            // path separator while the other not, then treat the paths
            // as equal. But this is allowed only (1) when the last char is
            // not a ':' (where a following path delimiter makes a difference
            // between a current directory and a root directory)
            // and (2) when not just a root directory is specified (so that
            // one is an empty string while the other only contains a
            // path delimiter indicating a root directory).
            if( l > 0 && !*path1 && _ispathsep_(*path2) ) if( !path2[1] && path1[-1] != L':' ) return 0;
            if( l > 0 && !*path2 && _ispathsep_(*path1) ) if( !path1[1] && path2[-1] != L':' ) return 0;
            return d;
        }
    }
    return 0;
}

int _pathcmp_( PCSTR path1, PCSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_pathcmp_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_pathcmp_( path1, path2, false );
        default                 : return _s_pathcmp_( path1, path2, false );
    }
}

int _pathcmp_( PCWSTR path1, PCWSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_pathcmp_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_pathcmp_( path1, path2, false );
        default                 : return _s_pathcmp_( path1, path2, false );
    }
}

static bool _s_pathishere_( PCSTR path1, PCSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return false;
    _idx_ l;
    for( l = 0 ; *path2 ; path1++, path2++, l++ )
    {
        CHAR ch1 = *path1,
             ch2 = *path2;
        if( !bCaseSensitive ) {
            ch1 = _toupper_(ch1);
            ch2 = _toupper_(ch2);
        }
        if( _ispathsep_(ch1) ) ch1 = (CHAR)_defpathsep_();
        if( _ispathsep_(ch2) ) ch2 = (CHAR)_defpathsep_();
        if( !*path1 || ( ch1 != ch2 ) )
            return false;
    }
    CHAR ch1 = *path1;
    if( ch1 == 0 ) return true;
    CHAR ch2_ = ( l > 0 )? path2[-1] : (CHAR)0;
    if( _ispathsep_(ch2_) || ch2_ == ':' ) return true;
    if( _ispathsep_(ch1 ) || ch1  == ':' ) return true;
    return false;
}

static bool _s_pathishere_( PCWSTR path1, PCWSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return false;
    _idx_ l;
    for( l = 0 ; *path2 ; path1++, path2++, l++ )
    {
        WCHAR ch1 = *path1,
              ch2 = *path2;
        if( !bCaseSensitive ) {
            ch1 = _toupper_(ch1);
            ch2 = _toupper_(ch2);
        }
        if( _ispathsep_(ch1) ) ch1 = (WCHAR)_defpathsep_();
        if( _ispathsep_(ch2) ) ch2 = (WCHAR)_defpathsep_();
        if( !*path1 || ( ch1 != ch2 ) )
            return false;
    }
    WCHAR ch1 = *path1;
    if( ch1 == 0 ) return true;
    WCHAR ch2_ = ( l > 0 )? path2[-1] : (WCHAR)0;
    if( _ispathsep_(ch2_) || ch2_ == L':' ) return true;
    if( _ispathsep_(ch1 ) || ch1  == L':' ) return true;
    return false;
}

bool _pathishere_( PCSTR path1, PCSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_pathishere_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_pathishere_( path1, path2, false );
        default                 : return _s_pathishere_( path1, path2, false );
    }
}

bool _pathishere_( PCWSTR path1, PCWSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_pathishere_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_pathishere_( path1, path2, false );
        default                 : return _s_pathishere_( path1, path2, false );
    }
}

///<returns> -1 = the paths are completely different <br/>
///    0 = the paths are equal <br/>
///    1 = the <b>path1</b> is a subdirectory of <b>path2</b> <br/>
///    2 = the <b>path2</b> is a subdirectory of <b>path1</b> </returns>
static int _s_subpathcmp_( PCSTR path1, PCSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return 0;
    _idx_ l;
    for( l = 0 ; *path1 || *path2 ; path1++, path2++, l++ )
    {
        if( _ispathsep_(*path1) && _ispathsep_(*path2) ) continue;
        int d = bCaseSensitive
                ? (int)(UCHAR)*path1            - (int)(UCHAR)*path2
                : (int)(UCHAR)_toupper_(*path1) - (int)(UCHAR)_toupper_(*path2);
        if( d ) {
            // If we reached at least the end of one of the strings, and
            // if one has a path delimiter at the position where the
            // other one has the end-of-string mark, then we handle the
            // case as if they both have a path delimiter at that position.
            // But this is allowed only (1) when the char before the path
            // delimiter is not a ':' (where a following path delimiter
            // makes a difference between a current directory and a root
            // directory) and (2) when not just a root directory is
            // specified (so that one is an empty string while the other
            // starts with a path delimiter indicating a root directory).
            if( l > 0 && !*path1 && _ispathsep_(*path2) ) if( path1[-1] != ':' ) return path2[1] ? 2 : 0;
            if( l > 0 && !*path2 && _ispathsep_(*path1) ) if( path2[-1] != ':' ) return path1[1] ? 1 : 0;
            return -1;
        }
    }
    return 0;
}

///<returns> -1 = the paths are completely different <br/>
///    0 = the paths are equal <br/>
///    1 = the <b>path1</b> is a subdirectory of <b>path2</b> <br/>
///    2 = the <b>path2</b> is a subdirectory of <b>path1</b> </returns>
static int _s_subpathcmp_( PCWSTR path1, PCWSTR path2, bool bCaseSensitive )
{
    if( !path1 || !path2 ) return 0;
    _idx_ l;
    for( l = 0 ; *path1 || *path2 ; path1++, path2++, l++ )
    {
        if( _ispathsep_(*path1) && _ispathsep_(*path2) ) continue;
        int d = bCaseSensitive
                ? (int)(USHORT)*path1            - (int)(USHORT)*path2
                : (int)(USHORT)_toupper_(*path1) - (int)(USHORT)_toupper_(*path2);
        if( d ) {
            // If we reached at least the end of one of the strings, and
            // if one has a path delimiter at the position where the
            // other one has the end-of-string mark, then we handle the
            // case as if they both have a path delimiter at that position.
            // But this is allowed only (1) when the char before the path
            // delimiter is not a ':' (where a following path delimiter
            // makes a difference between a current directory and a root
            // directory) and (2) when not just a root directory is
            // specified (so that one is an empty string while the other
            // starts with a path delimiter indicating a root directory).
            if( l > 0 && !*path1 && _ispathsep_(*path2) ) if( path1[-1] != L':' ) return path2[1] ? 2 : 0;
            if( l > 0 && !*path2 && _ispathsep_(*path1) ) if( path2[-1] != L':' ) return path1[1] ? 1 : 0;
            return -1;
        }
    }
    return 0;
}

///<summary> Compares 2 paths: checks if the paths are
///  equal or if one is a subdirectory of the other. </summary>
///<remarks> path1, path2 = these must be fullpaths, or
///  if they are relative paths (without drive at the
///  front) then they must be interpreted relative to the same
///  directory, or the comparsion will return wrong results. </remarks>
///<returns> -1 = the paths are completely different <br/>
///    0 = the paths are equal <br/>
///    1 = the <b>path1</b> is a subdirectory of <b>path2</b> <br/>
///    2 = the <b>path2</b> is a subdirectory of <b>path1</b> </returns>
int _subpathcmp_( PCSTR path1, PCSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_subpathcmp_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_subpathcmp_( path1, path2, false );
        default                 : return _s_subpathcmp_( path1, path2, false );
    }
}

///<summary> Compares 2 paths: checks if the paths are
///  equal or if one is a subdirectory of the other. </summary>
///<remarks> path1, path2 = these must be fullpaths, or
///  if they are relative paths (without drive at the
///  front) then they must be interpreted relative to the same
///  directory, or the comparsion will return wrong results. </remarks>
///<returns> -1 = the paths are completely different <br/>
///    0 = the paths are equal <br/>
///    1 = the <b>path1</b> is a subdirectory of <b>path2</b> <br/>
///    2 = the <b>path2</b> is a subdirectory of <b>path1</b> </returns>
int _subpathcmp_( PCWSTR path1, PCWSTR path2 )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _s_subpathcmp_( path1, path2, true  );
        case PathT_WindowsPath  : return _s_subpathcmp_( path1, path2, false );
        default                 : return _s_subpathcmp_( path1, path2, false );
    }
}



///<summary> checks, if a filename matches the mask <br/>
///  This function will be called recursively. Mostly each time a '*' is in the mask, a new recursion will be called.
///  But many '*' following one after each other will be treated as if it was only one '*'. </summary>
///<param name="filename"> This must be only the filename, without drive or path </param>
///<param name="mask"> the mask, which may contain '?' and '*' and '[ ... ]' <br/><br/>
///           ?       represents exactly one character <br/><br/>
///           *       represents 0, 1 or more characters <br/><br/>
///           [ ... ] represents a list of chars where the character of the filename must be one of this list. <br/>
///                   <i>Examples:</i> <br/>
///                    [abDg]     = the character of the filename must be one out of 'a', 'b', 'D', 'g'. <br/>
///                                  If case-insensitive -> 'A', 'B', 'd', 'G' are also included in that list. <br/>
///                    [a-c]      = the character must be an 'a', 'b', 'c'. <br/>
///                                  If case-sensitive 'A', 'B', 'C' are also included in that list. <br/>
///                    [pd-Ha]    = in case-sensitive case: <br/>
///                                  the character must be a 'p' or 'a'. The sequence "d-H" is not valid,
///                                  because 'H' has a lower value than 'd'. <br/>
///                                 in case-insensitive case: <br/>
///                                  the character must be of the list
///                                  'p', 'P', 'd', 'D', 'e', 'E', 'f', 'F', 'g', 'G', 'h', 'H', 'a', 'A'. <br/>
///                    [?.--$1-4] = the character must be one of the list
///                                  '?', '.', '-', '$', '1', '2', '3', '4'. <br/>
///                                 Note: If you want to include a '-' within a "[ ... ]" listing but
///                                       the '-' should not have its special meaning but only the simple
///                                       '-' character, then write them twice within the list. <br/><br/>
///           .       Will be interpreted as separator between a file-name and its extension. <br/>
///                   If this is the last char in the <b>mask</b> and the <b>filename</b> string does
///                   not contain any '.' char -> a match *will* be returned as if the <b>filename</b>
///                   ends with a '.' char, because this would also be a file-name without extension.
///                   If you want force that the <b>filename</b> string *must* contain a '.', then specify the
///                   sequence "[.]" in the <b>mask</b> instead of only specifying a "." character. <br/><br/>
///           Many '*' following one after each other will be treated as if it was only one '*'. <br/>
///           WARNING: If the mask is only a "." -> this is NOT interpreted as the current directory
///                                                 (this behaviour is necessary for the recursion) </param>
///<param name="bCaseSensitive"> false = case-insensitive <br/>  true = case-sensitive </param>
///<param name="start"> will contain starting char (see description in the explained examples below) <br/>
///          NULL -> no starting char will be returned. </param>
///<param name="end"> will contain ending char (see description in the explained examples below) <br/>
///          NULL -> no ending char will be returned. </param>
///<param name="newmask"> if defined, it will contain the resulting mask which has been changed from
///              <b>mask</b> that way, that it will match for case-sensitive checks (which makes
///              sense if this function here has been called with <b>bCaseSensitive</b> set to false; but if
///              <b>bCaseSensitive</b> was set to true -> the <b>newmask</b> can
///              differ from <b>mask</b> nevertheless because
///              e.g. when the <b>mask</b> looks like this "***.*" then <b>newmask</b> will look like
///              "*.*"). <br/>
///             The resulting string will be valid only when a match did occur.
///             If this param is NULL -> no newmask calculation will be done. <br/><br/>
///            On return, contains the corrected mask (and also made case-sensitive),
///              but it is *only* valid, when a match did occur, otherwise this string
///              can contain anything (it could also have been changed somehow). </param>
///<param name="cnewmask"> size of the <b>newmask</b> string buffer in CHARs. <br/>
///             If this param is 0 -> no newmask calculation will be done. </param>
///<param name="pWildmaskItemArray"> If not NULL, will be filled with <b>WildmaskItem</b> struct items
///  that each indicates which sequence of the mask maps to which sequence of the filename. </param>
///<param name="cWildmaskItemArray"> the size of the <b>pWildmaskItemArray</b> array in <b>WildmaskItem</b> structs. </param>
///<param name="pcWildmaskItemArray"> if not NULL, it will contain the number of items filled into
///  the <b>pWildmaskItemArray</b> array. </param>
///<returns> false = no match, true = match <br/>
///    If a match did occur -> <b>start</b> will contain the index into <b>filename</b> where
///     the first char of the first match has been found.
///     And <b>end</b> will contain the index into <b>filename</b> of a char directly after the last
///     matching char. <br/>
///     Note: <b>start</b> will contain a valid and senseful value only if the <b>mask</b> starts with
///      a '*' wildcard.
///      In addition <b>end</b> will contain a valid and senseful value only if the <b>mask</b> ends
///      with a '*' wildcard. <br/>
///     <i>Example 1:</i> <br/>
///      Imagine <b>filename</b> contains the string "abc(X)cd(Y)12345". It has the length 16.
///      Imagine <b>mask</b> contains the string "*(?)*". <br/>
///      The function will return a match (the first one) with <b>start</b> and <b>end</b> containing: <br/>
///       <b>start</b> = 3   This indicates the index of the '(' within the <b>filename</b>, because
///                          the '(' is the char directly after the starting '*' in the <b>mask</b>. <br/>
///       <b>end</b>   = 6   This indicates the index of the char directly after the ')'
///                          within the <b>filename</b>, because the ')' is the char directly before
///                          the trailing '*' in the <b>mask</b>. <br/>
///     <i>Example 2:</i> <br/>
///       <b>filename</b> = "ab(XYZ)123456" <br/>
///       <b>mask</b>     = "*(*)*" <br/>
///      The result will be: <br/>
///       <b>start</b>    = 2 <br/>
///       <b>end</b>      = 7 <br/>
///    Note: If 0 (no match) is returned, the values in <b>start</b> and <b>end</b> could have been
///          set to values nevertheless, but in that case they should be ignored. </returns>
///<remarks> Additional note to masks: <br/>
///  According to the special characters in a mask, it is not easily possible to scan for
///  exactly those characters in the filename. Following is a list, how you can scan for
///  those chars by specifying a special sequence in the mask: <br/>
///      scan for:                   mask sequence: <br/>
///      '['                         "[[]" <br/>
///      ']'                         "]"             (so for this is no special sequence necessary because ']' is not a special character) <br/>
///      '?'                         "[?]" <br/>
///      '.'                         "[.]" <br/>
///      '*'                         "[*]" </remarks>
static bool _s_findmask_( PCSTR filename, PCSTR mask, bool bCaseSensitive, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask,
                          WildmaskItem *pWildmaskItemArray = NULL, _size_t_ cWildmaskItemArray = 0, _size_t_ *pcWildmaskItemArray = NULL )
{
    if( !filename || !mask ) return false;
    _size_t_ cCurrWildmaskItemArray = 0;
    bool start_set = false;      // if #start# already contains a value
    _idx_ f, m ;
    CHAR mchar, fchar;
    if( newmask && cnewmask > 0 ) newmask[0] = 0;
    for( f = 0, m = 0 ; filename[f] || mask[m] ; f++, m++ )     // || is correct
    {
        fchar = filename[f];
        mchar = mask[m];
        switch( mchar )
        {
            case '*' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, "*", cnewmask );
                _size_t_ nmlen = _strlen_( newmask );
                PCSTR mm = &mask[m + 1];
                while( *mm == '*' ) mm++;   // if more '*' follow one after each other -> jump to the last
                if( !*mm ) {
                    if( start && !start_set ) *start = f;
                    if( end ) *end = f;
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = (_idx_) _strlen_( filename );
                        ++cCurrWildmaskItemArray;
                    }
                    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                    return true;
                }
                _size_t_ cNewWildmaskItemArray;
                for( ; filename[f] ; f++ )
                {
                    cNewWildmaskItemArray = 0;
                    WildmaskItem *pItms = pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? pWildmaskItemArray + (cCurrWildmaskItemArray + 1) : NULL;
                    if( _s_findmask_( &filename[f], mm, bCaseSensitive, start_set ? NULL : start, end, newmask + nmlen, cnewmask - nmlen,
                                      pItms,
                                      pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? cWildmaskItemArray - (cCurrWildmaskItemArray + 1) : 0,
                                      &cNewWildmaskItemArray ) ) {
                        if( start && !start_set ) *start += f;
                        if( end ) *end += f;
                        if( pItms && cNewWildmaskItemArray > 0 )
                            for( _idx_ h = 0 ; h < (_idx_) cNewWildmaskItemArray ; ++h ) {
                                pItms[h].m_iMaskItemStart += mm - mask;
                                pItms[h].m_iMaskItemEnd   += mm - mask;
                                pItms[h].m_iFilenameStart += f;
                                pItms[h].m_iFilenameEnd   += f;
                            }
                        if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                            ++cCurrWildmaskItemArray;
                        }
                        cCurrWildmaskItemArray += cNewWildmaskItemArray;
                        if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                        return true;
                    } else if( filename[f] == '.' && *mm == '.' ) {
                        for( f++ ; filename[f] ; f++ )
                            if( filename[f] == '.' ) break;
                        if( filename[f] != '.' ) return false;
                        f--;
                    }
                }
                // and finally one test, when #&filename[f]# is a clear string
                cNewWildmaskItemArray = 0;
                WildmaskItem *pItms = pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? pWildmaskItemArray + (cCurrWildmaskItemArray + 1) : NULL;
                if( _s_findmask_( &filename[f], mm, bCaseSensitive, start_set ? NULL : start, end, newmask + nmlen, cnewmask - nmlen,
                                  pItms,
                                  pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? cWildmaskItemArray - (cCurrWildmaskItemArray + 1) : 0,
                                  &cNewWildmaskItemArray ) ) {
                    if( start && !start_set ) *start += f;
                    if( end ) *end += f;
                    if( pItms && cNewWildmaskItemArray > 0 )
                        for( _idx_ h = 0 ; h < (_idx_) cNewWildmaskItemArray ; ++h ) {
                            pItms[h].m_iMaskItemStart += mm - mask;
                            pItms[h].m_iMaskItemEnd   += mm - mask;
                            pItms[h].m_iFilenameStart += f;
                            pItms[h].m_iFilenameEnd   += f;
                        }
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                        ++cCurrWildmaskItemArray;
                    }
                    cCurrWildmaskItemArray += cNewWildmaskItemArray;
                    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                    return true;
                }
                else return false;
            }
            case '?' :
                if( start && !start_set ) { *start = f; start_set = true; }
                if( !fchar ) return false;
                if( newmask && cnewmask > 0 ) _strcat_( newmask, "?", cnewmask );
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Questionmark;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            case '.' :
                if( start && !start_set ) { *start = f; start_set = true; }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, ".", cnewmask );
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Dot;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                }
                if( fchar != mchar )
                {
                    // Now check: If the #mask# ends with a '.' or with a '.' followed by one or more
                    // '*' chars -> and if the #filename# has no more characters ( that is when #fchar#
                    // is a '\0' char ) -> that's a match. In all other cases -> it is no match and
                    // the scanning must be aborted and false must be returned.
                    PCSTR mm = &mask[m + 1];
                    if( *mm == '*' )
                        if( newmask && cnewmask > 0 ) _strcat_( newmask, "*", cnewmask );
                    while( *mm == '*' ) mm++;
                    // now if #*mm != 0# -> there have to be characters after the '.' in the filename
                    if( !*mm && !fchar )
                    {
                        if( end ) *end = f;
                        if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                            ++cCurrWildmaskItemArray;
                        }
                        if( mask[m + 1] == '*' )
                            if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m + 1;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = mm - mask;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd   = f;
                                ++cCurrWildmaskItemArray;
                            }
                        if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                        return true;
                    }
                    return false;
                }
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray )
                    ++cCurrWildmaskItemArray;
                break;
            case '[' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_oneCharOfMany;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                }
                if( start && !start_set ) { *start = f; start_set = true; }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, "[", cnewmask );
                bool found = false;
                m++;
                while( mask[m] && mask[m] != ']' )
                {
                    if( !found )
                    {
                        bool test_char_directly = true;
                        if( mask[m + 1] == '-' )      // e.g. "[a-d]" but also e.g. "[--]"
                        {
                            if( !mask[m + 2] ) return false;
                            if( mask[m + 2] != '-' ) {
                                UINT c0 = (UINT)(UCHAR)( bCaseSensitive ? mask[m]     : _toupper_(mask[m]    ) );
                                UINT c1 = (UINT)(UCHAR)( bCaseSensitive ? mask[m + 2] : _toupper_(mask[m + 2]) );
                                if( c0 > c1 ) { UINT ch = c0; c0 = c1; c1 = ch; }
                                UINT cf = (UINT)(UCHAR)( bCaseSensitive ? fchar : _toupper_(fchar) );
                                if( cf >= c0 && cf <= c1 ) found = true;
                                test_char_directly = false;
                            }
                            if( newmask && cnewmask > 0 ) {
                                _straddch_( newmask, mask[m],     cnewmask );
                                _straddch_( newmask, mask[m + 1], cnewmask );
                            }
                            m += 2;
                        }
                        if( test_char_directly )
                        {
                            if( bCaseSensitive ) {
                                if( fchar == mask[m] )
                                    found = true;
                            } else {
                                if( _toupper_(fchar) == _toupper_(mask[m]) )
                                    found = true;
                            }
                        }
                    }
                    if( newmask && cnewmask > 0 )
                        _straddch_( newmask, mask[m], cnewmask );
                    m++;
                }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, "]", cnewmask );
                if( !mask[m] ) return false;     // no corresponding ']' in mask found -> error -> return false for not found
                if( !found ) return false;
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            }
            case 0 :
                return false;
            default :
            {
                if( start && !start_set ) { *start = f; start_set = true; }
                if( bCaseSensitive ) {
                    if( fchar != mchar )
                        return false;
                } else {
                    if( _toupper_(fchar) != _toupper_(mchar) )
                        return false;
                }
                if( newmask && cnewmask > 0 )
                    _straddch_( newmask, fchar, cnewmask );
                // Now we check if the last WildmaskItem is already of type Other.
                // If so, we just extend this one, otherwise we have to add a new one.
                bool bCreateNewWildmaskItem = true;
                if( pWildmaskItemArray && cWildmaskItemArray > 0 && cCurrWildmaskItemArray > 0 && cCurrWildmaskItemArray <= cWildmaskItemArray )
                    if( pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_Type == wildmaskitem_Other ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iMaskItemEnd = m + 1;
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iFilenameEnd = f + 1;
                        bCreateNewWildmaskItem = false;
                    }
                if( bCreateNewWildmaskItem )
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Other;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd   = f + 1;
                        ++cCurrWildmaskItemArray;
                    }
                break;
            }
        }
    }
    if( start && !start_set ) *start = f;
    if( end ) *end = f;
    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
    return true;
}

///<summary> checks, if a filename matches the mask <br/>
///  This function will be called recursively. Mostly each time a '*' is in the mask, a new recursion will be called.
///  But many '*' following one after each other will be treated as if it was only one '*'. </summary>
///<param name="filename"> This must be only the filename, without drive or path </param>
///<param name="mask"> the mask, which may contain '?' and '*' and '[ ... ]' <br/><br/>
///           ?       represents exactly one character <br/><br/>
///           *       represents 0, 1 or more characters <br/><br/>
///           [ ... ] represents a list of chars where the character of the filename must be one of this list. <br/>
///                   <i>Examples:</i> <br/>
///                    [abDg]     = the character of the filename must be one out of 'a', 'b', 'D', 'g'. <br/>
///                                  If case-insensitive -> 'A', 'B', 'd', 'G' are also included in that list. <br/>
///                    [a-c]      = the character must be an 'a', 'b', 'c'. <br/>
///                                  If case-sensitive 'A', 'B', 'C' are also included in that list. <br/>
///                    [pd-Ha]    = in case-sensitive case: <br/>
///                                  the character must be a 'p' or 'a'. The sequence "d-H" is not valid,
///                                  because 'H' has a lower value than 'd'. <br/>
///                                 in case-insensitive case: <br/>
///                                  the character must be of the list
///                                  'p', 'P', 'd', 'D', 'e', 'E', 'f', 'F', 'g', 'G', 'h', 'H', 'a', 'A'. <br/>
///                    [?.--$1-4] = the character must be one of the list
///                                  '?', '.', '-', '$', '1', '2', '3', '4'. <br/>
///                                 Note: If you want to include a '-' within a "[ ... ]" listing but
///                                       the '-' should not have its special meaning but only the simple
///                                       '-' character, then write them twice within the list. <br/><br/>
///           .       Will be interpreted as separator between a file-name and its extension. <br/>
///                   If this is the last char in the <b>mask</b> and the <b>filename</b> string does
///                   not contain any '.' char -> a match *will* be returned as if the <b>filename</b>
///                   ends with a '.' char, because this would also be a file-name without extension.
///                   If you want force that the <b>filename</b> string *must* contain a '.', then specify the
///                   sequence "[.]" in the <b>mask</b> instead of only specifying a "." character. <br/><br/>
///           Many '*' following one after each other will be treated as if it was only one '*'. <br/>
///           WARNING: If the mask is only a "." -> this is NOT interpreted as the current directory
///                                                 (this behaviour is necessary for the recursion) </param>
///<param name="bCaseSensitive"> false = case-insensitive <br/>  true = case-sensitive </param>
///<param name="start"> will contain starting char (see description in the explained examples below) <br/>
///          NULL -> no starting char will be returned. </param>
///<param name="end"> will contain ending char (see description in the explained examples below) <br/>
///          NULL -> no ending char will be returned. </param>
///<param name="newmask"> if defined, it will contain the resulting mask which has been changed from
///              <b>mask</b> that way, that it will match for case-sensitive checks (which makes
///              sense if this function here has been called with <b>bCaseSensitive</b> set to false; but if
///              <b>bCaseSensitive</b> was set to true -> the <b>newmask</b> can
///              differ from <b>mask</b> nevertheless because
///              e.g. when the <b>mask</b> looks like this "***.*" then <b>newmask</b> will look like
///              "*.*"). <br/>
///             The resulting string will be valid only when a match did occur.
///             If this param is NULL -> no newmask calculation will be done. <br/><br/>
///            On return, contains the corrected mask (and also made case-sensitive),
///              but it is *only* valid, when a match did occur, otherwise this string
///              can contain anything (it could also have been changed somehow). </param>
///<param name="cnewmask"> size of the <b>newmask</b> string buffer in WCHARs. <br/>
///             If this param is 0 -> no newmask calculation will be done. </param>
///<param name="pWildmaskItemArray"> If not NULL, will be filled with <b>WildmaskItem</b> struct items
///  that each indicates which sequence of the mask maps to which sequence of the filename. </param>
///<param name="cWildmaskItemArray"> the size of the <b>pWildmaskItemArray</b> array in <b>WildmaskItem</b> structs. </param>
///<param name="pcWildmaskItemArray"> if not NULL, it will contain the number of items filled into
///  the <b>pWildmaskItemArray</b> array. </param>
///<returns> false = no match, true = match <br/>
///    If a match did occur -> <b>start</b> will contain the index into <b>filename</b> where
///     the first char of the first match has been found.
///     And <b>end</b> will contain the index into <b>filename</b> of a char directly after the last
///     matching char. <br/>
///     Note: <b>start</b> will contain a valid and senseful value only if the <b>mask</b> starts with
///      a '*' wildcard.
///      In addition <b>end</b> will contain a valid and senseful value only if the <b>mask</b> ends
///      with a '*' wildcard. <br/>
///     <i>Example 1:</i> <br/>
///      Imagine <b>filename</b> contains the string "abc(X)cd(Y)12345". It has the length 16.
///      Imagine <b>mask</b> contains the string "*(?)*". <br/>
///      The function will return a match (the first one) with <b>start</b> and <b>end</b> containing: <br/>
///       <b>start</b> = 3   This indicates the index of the '(' within the <b>filename</b>, because
///                          the '(' is the char directly after the starting '*' in the <b>mask</b>. <br/>
///       <b>end</b>   = 6   This indicates the index of the char directly after the ')'
///                          within the <b>filename</b>, because the ')' is the char directly before
///                          the trailing '*' in the <b>mask</b>. <br/>
///     <i>Example 2:</i> <br/>
///       <b>filename</b> = "ab(XYZ)123456" <br/>
///       <b>mask</b>     = "*(*)*" <br/>
///      The result will be: <br/>
///       <b>start</b>    = 2 <br/>
///       <b>end</b>      = 7 <br/>
///    Note: If 0 (no match) is returned, the values in <b>start</b> and <b>end</b> could have been
///          set to values nevertheless, but in that case they should be ignored. </returns>
///<remarks> Additional note to masks: <br/>
///  According to the special characters in a mask, it is not easily possible to scan for
///  exactly those characters in the filename. Following is a list, how you can scan for
///  those chars by specifying a special sequence in the mask: <br/>
///      scan for:                   mask sequence: <br/>
///      '['                         "[[]" <br/>
///      ']'                         "]"             (so for this is no special sequence necessary because ']' is not a special character) <br/>
///      '?'                         "[?]" <br/>
///      '.'                         "[.]" <br/>
///      '*'                         "[*]" </remarks>
static bool _s_findmask_( PCWSTR filename, PCWSTR mask, bool bCaseSensitive, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask,
                          WildmaskItem *pWildmaskItemArray = NULL, _size_t_ cWildmaskItemArray = 0, _size_t_ *pcWildmaskItemArray = NULL )
{
    if( !filename || !mask ) return false;
    _size_t_ cCurrWildmaskItemArray = 0;
    bool start_set = false;      // if #start# already contains a value
    _idx_ f, m ;
    WCHAR mchar, fchar;
    if( newmask && cnewmask > 0 ) newmask[0] = 0;
    for( f = 0, m = 0 ; filename[f] || mask[m] ; f++, m++ )     // || is correct
    {
        fchar = filename[f];
        mchar = mask[m];
        switch( mchar )
        {
            case L'*' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, L"*", cnewmask );
                _size_t_ nmlen = _strlen_( newmask );
                PCWSTR mm = &mask[m + 1];
                while( *mm == L'*' ) mm++;   // if more '*' follow one after each other -> jump to the last
                if( !*mm ) {
                    if( start && !start_set ) *start = f;
                    if( end ) *end = f;
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = (_idx_) _strlen_( filename );
                        ++cCurrWildmaskItemArray;
                    }
                    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                    return true;
                }
                _size_t_ cNewWildmaskItemArray;
                for( ; filename[f] ; f++ )
                {
                    cNewWildmaskItemArray = 0;
                    WildmaskItem *pItms = pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? pWildmaskItemArray + (cCurrWildmaskItemArray + 1) : NULL;
                    if( _s_findmask_( &filename[f], mm, bCaseSensitive, start_set ? NULL : start, end, newmask + nmlen, cnewmask - nmlen,
                                      pItms,
                                      pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? cWildmaskItemArray - (cCurrWildmaskItemArray + 1) : 0,
                                      &cNewWildmaskItemArray ) ) {
                        if( start && !start_set ) *start += f;
                        if( end ) *end += f;
                        if( pItms && cNewWildmaskItemArray > 0 )
                            for( _idx_ h = 0 ; h < (_idx_) cNewWildmaskItemArray ; ++h ) {
                                pItms[h].m_iMaskItemStart += mm - mask;
                                pItms[h].m_iMaskItemEnd   += mm - mask;
                                pItms[h].m_iFilenameStart += f;
                                pItms[h].m_iFilenameEnd   += f;
                            }
                        if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                            ++cCurrWildmaskItemArray;
                        }
                        cCurrWildmaskItemArray += cNewWildmaskItemArray;
                        if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                        return true;
                    } else if( filename[f] == L'.' && *mm == L'.' ) {
                        for( f++ ; filename[f] ; f++ )
                            if( filename[f] == L'.' ) break;
                        if( filename[f] != L'.' ) return false;
                        f--;
                    }
                }
                // and finally one test, when #&filename[f]# is a clear string
                cNewWildmaskItemArray = 0;
                WildmaskItem *pItms = pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? pWildmaskItemArray + (cCurrWildmaskItemArray + 1) : NULL;
                if( _s_findmask_( &filename[f], mm, bCaseSensitive, start_set ? NULL : start, end, newmask + nmlen, cnewmask - nmlen,
                                  pItms,
                                  pWildmaskItemArray && cWildmaskItemArray > (cCurrWildmaskItemArray + 1) ? cWildmaskItemArray - (cCurrWildmaskItemArray + 1) : 0,
                                  &cNewWildmaskItemArray ) ) {
                    if( start && !start_set ) *start += f;
                    if( end ) *end += f;
                    if( pItms && cNewWildmaskItemArray > 0 )
                        for( _idx_ h = 0 ; h < (_idx_) cNewWildmaskItemArray ; ++h ) {
                            pItms[h].m_iMaskItemStart += mm - mask;
                            pItms[h].m_iMaskItemEnd   += mm - mask;
                            pItms[h].m_iFilenameStart += f;
                            pItms[h].m_iFilenameEnd   += f;
                        }
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = mm - mask;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                        ++cCurrWildmaskItemArray;
                    }
                    cCurrWildmaskItemArray += cNewWildmaskItemArray;
                    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                    return true;
                }
                else return false;
            }
            case L'?' :
                if( start && !start_set ) { *start = f; start_set = true; }
                if( !fchar ) return false;
                if( newmask && cnewmask > 0 ) _strcat_( newmask, L"?", cnewmask );
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Questionmark;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            case L'.' :
                if( start && !start_set ) { *start = f; start_set = true; }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, L".", cnewmask );
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Dot;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                }
                if( fchar != mchar )
                {
                    // Now check: If the #mask# ends with a '.' or with a '.' followed by one or more
                    // '*' chars -> and if the #filename# has no more characters ( that is when #fchar#
                    // is a '\0' char ) -> that's a match. In all other cases -> it is no match and
                    // the scanning must be aborted and false must be returned.
                    PCWSTR mm = &mask[m + 1];
                    if( *mm == L'*' )
                        if( newmask && cnewmask > 0 ) _strcat_( newmask, L"*", cnewmask );
                    while( *mm == L'*' ) mm++;
                    // now if #*mm != 0# -> there have to be characters after the '.' in the filename
                    if( !*mm && !fchar )
                    {
                        if( end ) *end = f;
                        if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                            pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f;
                            ++cCurrWildmaskItemArray;
                        }
                        if( mask[m + 1] == L'*' )
                            if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m + 1;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = mm - mask;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                                pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd   = f;
                                ++cCurrWildmaskItemArray;
                            }
                        if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
                        return true;
                    }
                    return false;
                }
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray )
                    ++cCurrWildmaskItemArray;
                break;
            case L'[' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_oneCharOfMany;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                }
                if( start && !start_set ) { *start = f; start_set = true; }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, L"[", cnewmask );
                bool found = false;
                m++;
                while( mask[m] && mask[m] != L']' )
                {
                    if( !found )
                    {
                        bool test_char_directly = true;
                        if( mask[m + 1] == L'-' )      // e.g. "[a-d]" but also e.g. "[--]"
                        {
                            if( !mask[m + 2] ) return false;
                            if( mask[m + 2] != L'-' ) {
                                UINT c0 = (UINT)( bCaseSensitive ? mask[m]     : _toupper_(mask[m]    ) );
                                UINT c1 = (UINT)( bCaseSensitive ? mask[m + 2] : _toupper_(mask[m + 2]) );
                                if( c0 > c1 ) { UINT ch = c0; c0 = c1; c1 = ch; }
                                UINT cf = (UINT)( bCaseSensitive ? fchar : _toupper_(fchar) );
                                if( cf >= c0 && cf <= c1 ) found = true;
                                test_char_directly = false;
                            }
                            if( newmask && cnewmask > 0 ) {
                                _straddch_( newmask, mask[m],     cnewmask );
                                _straddch_( newmask, mask[m + 1], cnewmask );
                            }
                            m += 2;
                        }
                        if( test_char_directly )
                        {
                            if( bCaseSensitive ) {
                                if( fchar == mask[m] )
                                    found = true;
                            } else {
                                if( _toupper_(fchar) == _toupper_(mask[m]) )
                                    found = true;
                            }
                        }
                    }
                    if( newmask && cnewmask > 0 )
                        _straddch_( newmask, mask[m], cnewmask );
                    m++;
                }
                if( newmask && cnewmask > 0 ) _strcat_( newmask, L"]", cnewmask );
                if( !mask[m] ) return false;     // no corresponding ']' in mask found -> error -> return false for not found
                if( !found ) return false;
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd = f + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            }
            case 0 :
                return false;
            default :
            {
                if( start && !start_set ) { *start = f; start_set = true; }
                if( bCaseSensitive ) {
                    if( fchar != mchar )
                        return false;
                } else {
                    if( _toupper_(fchar) != _toupper_(mchar) )
                        return false;
                }
                if( newmask && cnewmask > 0 )
                    _straddch_( newmask, fchar, cnewmask );
                // Now we check if the last WildmaskItem is already of type Other.
                // If so, we just extend this one, otherwise we have to add a new one.
                bool bCreateNewWildmaskItem = true;
                if( pWildmaskItemArray && cWildmaskItemArray > 0 && cCurrWildmaskItemArray > 0 && cCurrWildmaskItemArray <= cWildmaskItemArray )
                    if( pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_Type == wildmaskitem_Other ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iMaskItemEnd = m + 1;
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iFilenameEnd = f + 1;
                        bCreateNewWildmaskItem = false;
                    }
                if( bCreateNewWildmaskItem )
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Other;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameStart = f;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iFilenameEnd   = f + 1;
                        ++cCurrWildmaskItemArray;
                    }
                break;
            }
        }
    }
    if( start && !start_set ) *start = f;
    if( end ) *end = f;
    if( pcWildmaskItemArray ) *pcWildmaskItemArray = cCurrWildmaskItemArray;
    return true;
}

///<summary> checks, if a filename matches the mask
///  the comparing will be case-sensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _matchmask_( PCSTR filename, PCSTR mask )
{
    return _s_findmask_( filename, mask, true, NULL, NULL, NULL, 0 ) ;
}

///<summary> checks, if a filename matches the mask
///  the comparing will be case-sensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _matchmask_( PCWSTR filename, PCWSTR mask )
{
    return _s_findmask_( filename, mask, true, NULL, NULL, NULL, 0 ) ;
}

///<summary> checks, if a filename matches the mask
///  the comparing will be case-insensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _matchimask_( PCSTR filename, PCSTR mask )
{
    return _s_findmask_( filename, mask, false, NULL, NULL, NULL, 0 );
}

///<summary> checks, if a filename matches the mask
///  the comparing will be case-insensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _matchimask_( PCWSTR filename, PCWSTR mask )
{
    return _s_findmask_( filename, mask, false, NULL, NULL, NULL, 0 );
}

bool _matchpathmask_( PCSTR filename, PCSTR mask )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _matchmask_ ( filename, mask );
        case PathT_WindowsPath  : return _matchimask_( filename, mask );
        default                 : return _matchimask_( filename, mask );
    }
}

bool _matchpathmask_( PCWSTR filename, PCWSTR mask )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _matchmask_ ( filename, mask );
        case PathT_WindowsPath  : return _matchimask_( filename, mask );
        default                 : return _matchimask_( filename, mask );
    }
}

///<summary> checks, if a filename matches the mask and
///  returns starting and ending matching position
///  the comparing will be case-sensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findmask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end )
{
    return _s_findmask_( filename, mask, true, start, end, NULL, 0 );
}

///<summary> checks, if a filename matches the mask and
///  returns starting and ending matching position
///  the comparing will be case-sensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end )
{
    return _s_findmask_( filename, mask, true, start, end, NULL, 0 );
}

///<summary> checks, if a filename matches the mask and
///  returns starting and ending matching position
///  the comparing will be case-insensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findimask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end )
{
    return _s_findmask_( filename, mask, false, start, end, NULL, 0 );
}

///<summary> checks, if a filename matches the mask and
///  returns starting and ending matching position
///  the comparing will be case-insensitive
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findimask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end )
{
    return _s_findmask_( filename, mask, false, start, end, NULL, 0 );
}

bool _findpathmask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findmask_ ( filename, mask, start, end );
        case PathT_WindowsPath  : return _findimask_( filename, mask, start, end );
        default                 : return _findimask_( filename, mask, start, end );
    }
}

bool _findpathmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findmask_ ( filename, mask, start, end );
        case PathT_WindowsPath  : return _findimask_( filename, mask, start, end );
        default                 : return _findimask_( filename, mask, start, end );
    }
}

///<summary> like <b>_findmask_()</b> but also returns the new
///  mask which is corrected and made case-sensitive.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewmask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask )
{
    return _s_findmask_( filename, mask, true, start, end, newmask, cnewmask );
}

///<summary> like <b>_findmask_()</b> but also returns the new
///  mask which is corrected and made case-sensitive.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask )
{
    return _s_findmask_( filename, mask, true, start, end, newmask, cnewmask );
}

///<summary> like <b>_findimask_()</b> but also returns the new
///  mask which is corrected and made case-sensitive.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewimask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask )
{
    return _s_findmask_( filename, mask, false, start, end, newmask, cnewmask );
}

///<summary> like <b>_findimask_()</b> but also returns the new
///  mask which is corrected and made case-sensitive.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewimask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask )
{
    return _s_findmask_( filename, mask, false, start, end, newmask, cnewmask );
}

bool _findnewpathmask_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findnewmask_ ( filename, mask, start, end, newmask, cnewmask );
        case PathT_WindowsPath  : return _findnewimask_( filename, mask, start, end, newmask, cnewmask );
        default                 : return _findnewimask_( filename, mask, start, end, newmask, cnewmask );
    }
}

bool _findnewpathmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findnewmask_ ( filename, mask, start, end, newmask, cnewmask );
        case PathT_WindowsPath  : return _findnewimask_( filename, mask, start, end, newmask, cnewmask );
        default                 : return _findnewimask_( filename, mask, start, end, newmask, cnewmask );
    }
}

///<summary> like <b>_findnewmask_()</b> but also returns the array
///  of sequences matching between filename and mask.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewmask2_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    return _s_findmask_( filename, mask, true, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
}

///<summary> like <b>_findnewmask_()</b> but also returns the array
///  of sequences matching between filename and mask.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewmask2_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    return _s_findmask_( filename, mask, true, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
}

///<summary> like <b>_findnewimask_()</b> but also returns the array
///  of sequences matching between filename and mask.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewimask2_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    return _s_findmask_( filename, mask, false, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
}

///<summary> like <b>_findnewimask_()</b> but also returns the array
///  of sequences matching between filename and mask.
///  For parameter description see <b>_s_findmask_()</b> above. </summary>
bool _findnewimask2_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    return _s_findmask_( filename, mask, false, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
}

bool _findnewpathmask2_( PCSTR filename, PCSTR mask, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findnewmask2_ ( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
        case PathT_WindowsPath  : return _findnewimask2_( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
        default                 : return _findnewimask2_( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
    }
}

bool _findnewpathmask2_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findnewmask2_ ( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
        case PathT_WindowsPath  : return _findnewimask2_( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
        default                 : return _findnewimask2_( filename, mask, start, end, newmask, cnewmask, pWildmaskItemArray, cWildmaskItemArray, pcWildmaskItemArray );
    }
}

///<summary> if a filename matches several masks (separated by ';') </summary>
///<param name="filename"> Filename (without paths) </param>
///<param name="masklist"> a list of masks (one or more, separated by ';') </param>
///<param name="bCaseSensitive"> false = case-insensitive <br/>  true = case-sensitive </param>
///<param name="start"> when a match has been found, <b>start</b> will be an index into the <b>masklist</b>
///   string indicating the start of the mask string which caused the match.
///   <b>end</b> will point the char directly after that mask string (so it will
///   point normally on a ';' or '\0' character). <br/>
///   If one of the pointers is NULL -> this value will not be returned. <br/><br/>
///  On return: <b>start</b>, <b>end</b> = will contain values as described above. They will have been only set when
///   a match did occur. If no match did occur, <b>start</b> and <b>end</b> will not
///   have been changed. </param>
///<param name="end"> see remarks to <b>start</b> param. </param>
///<param name="imask"> if <b>filename</b> is not NULL, then this index will be ignored.
///  But if <b>filename</b> is NULL -> this index will indicate a mask of <b>masklist</b>
///  which will then be returned in <b>newmask</b>. For this, <b>newmask</b> and <b>cnewmask</b>
///  must be defined. As well, <b>start</b> and <b>end</b> will be filled if not NULL. </param>
///<param name="newmask"> if defined, it will contain the resulting mask which has been changed from
///  <b>mask</b> that way, that it will match for case-sensitive checks (which makes
///  sense if this function here has been called with <b>bCaseSensitive</b> set to false; but if
///  <b>bCaseSensitive</b> was set to true -> the <b>newmask</b> can differ from matching mask nevertheless because
///  e.g. when the matching mask looks like this "***.*" then <b>newmask</b> will look like
///  "*.*"). <br/>
///  The resulting string will be valid only when a match did occur.
///  If this param is NULL -> no newmask calculation will be done. </param>
///<param name="cnewmask"> size of the <b>newmask</b> string buffer <br/>
///  If this param is 0 -> no newmask calculation will be done. </param>
///<param name="pMaskIndex"> If not NULL, will contain the index of the n'th mask within
///  the masklist that caused the match. If false is returned, this value remains unchanged. </param>
///<returns> false = no match, true = one of the masks did match </returns>
///<remarks> Additional not to masks;
///  In the description to <b>_s_findmask_()</b>, a list of special characters is written above.
///  Now a now special char is available and below the sequence is written, which allows to
///  scan the <b>filename</b> exactly for that sepcial character: <br/>
///      scan for:                   mask sequence: <br/>
///      ';'                         "[;]" </remarks>
static bool _s_matchmasks_( PCSTR filename, PCSTR masklist, bool bCaseSensitive, _idx_ *start, _idx_ *end, _idx_ imask, PSTR newmask, _size_t_ cnewmask,
                            _idx_ *pMaskIndex = NULL )
{
    if( newmask && cnewmask > 0 ) newmask[0] = 0;
    if( !filename && !( newmask && cnewmask > 0 && imask >= 0 ) ) return false;
    if( ! masklist ) return false; // this *must* be false
    if( !*masklist ) return false; // this *must* be false
    if( filename ) if( !*filename ) return false;
    _idx_ nth_mask = 0;
    PCSTR origmasklist = masklist;
    while( *masklist )
    {
        CHAR msk[_MAX_PATH_ + 1];
        bool within_charlist = false;
        while( *masklist == ';' ) { masklist++; imask--; ++nth_mask; }   // overjump following ';' chars
        _idx_ istart = (_idx_)( masklist - origmasklist );   // subtraction handles char-size correctly
        _idx_ i;
        for( i = 0 ; ( i < _countof(msk) - 1 ) && ( *masklist && *masklist != ';' ) ; i++, masklist++ )
            msk[i] = *masklist;
        msk[i] = 0;
        _idx_ iend = istart + i;
        if( *msk )          // if #msk# is a clear string -> this does not mean to be a match
            if( filename )
            {
                if( _s_findmask_( filename, msk, bCaseSensitive, NULL, NULL, newmask, cnewmask ) ) {
                    if( start ) *start = istart;
                    if( end ) *end = iend;
                    if( pMaskIndex ) *pMaskIndex = nth_mask;
                    return true;        // one match found
                }
            }
            else if( newmask && cnewmask > 0 )
            {
                if( imask < 0 ) return false;
                if( imask == 0 ) {
                    _strcpy_( newmask, msk, cnewmask );
                    if( start ) *start = istart;
                    if( end ) *end = iend;
                    if( pMaskIndex ) *pMaskIndex = nth_mask;
                    return true;        // mask (at index #imask#) found
                }
            }
        // Now find the next ';' but only that which is not within a "[ ... ]" sequence.
        while( *masklist )
        {
            if( within_charlist ) {
                if( *masklist == ']' ) within_charlist = false;
            } else {
                if( *masklist == '[' ) within_charlist = true;
                else if( *masklist == ';' ) break;
            }
            masklist++;
        }
    }
    return false;
}

///<summary> if a filename matches several masks (separated by ';') </summary>
///<param name="filename"> Filename (without paths) </param>
///<param name="masklist"> a list of masks (one or more, separated by ';') </param>
///<param name="bCaseSensitive"> false = case-insensitive <br/>  true = case-sensitive </param>
///<param name="start"> when a match has been found, <b>start</b> will be an index into the <b>masklist</b>
///   string indicating the start of the mask string which caused the match.
///   <b>end</b> will point the char directly after that mask string (so it will
///   point normally on a ';' or '\0' character). <br/>
///   If one of the pointers is NULL -> this value will not be returned. <br/><br/>
///  On return: <b>start</b>, <b>end</b> = will contain values as described above. They will have been only set when
///   a match did occur. If no match did occur, <b>start</b> and <b>end</b> will not
///   have been changed. </param>
///<param name="end"> see remarks to <b>start</b> param. </param>
///<param name="imask"> if <b>filename</b> is not NULL, then this index will be ignored.
///  But if <b>filename</b> is NULL -> this index will indicate a mask of <b>masklist</b>
///  which will then be returned in <b>newmask</b>. For this, <b>newmask</b> and <b>cnewmask</b>
///  must be defined. As well, <b>start</b> and <b>end</b> will be filled if not NULL. </param>
///<param name="newmask"> if defined, it will contain the resulting mask which has been changed from
///  <b>mask</b> that way, that it will match for case-sensitive checks (which makes
///  sense if this function here has been called with <b>bCaseSensitive</b> set to false; but if
///  <b>bCaseSensitive</b> was set to true -> the <b>newmask</b> can differ from matching mask nevertheless because
///  e.g. when the matching mask looks like this "***.*" then <b>newmask</b> will look like
///  "*.*"). <br/>
///  The resulting string will be valid only when a match did occur.
///  If this param is NULL -> no newmask calculation will be done. </param>
///<param name="cnewmask"> size of the <b>newmask</b> string buffer <br/>
///  If this param is 0 -> no newmask calculation will be done. </param>
///<param name="pMaskIndex"> If not NULL, will contain the index of the n'th mask within
///  the masklist that caused the match. If false is returned, this value remains unchanged. </param>
///<returns> false = no match, true = one of the masks did match </returns>
///<remarks> Additional not to masks;
///  In the description to <b>_s_findmask_()</b>, a list of special characters is written above.
///  Now a now special char is available and below the sequence is written, which allows to
///  scan the <b>filename</b> exactly for that sepcial character: <br/>
///      scan for:                   mask sequence: <br/>
///      ';'                         "[;]" </remarks>
static bool _s_matchmasks_( PCWSTR filename, PCWSTR masklist, bool bCaseSensitive, _idx_ *start, _idx_ *end, _idx_ imask, PWSTR newmask, _size_t_ cnewmask,
                            _idx_ *pMaskIndex = NULL )
{
    if( newmask && cnewmask > 0 ) newmask[0] = 0;
    if( !filename && !( newmask && cnewmask > 0 && imask >= 0 ) ) return false;
    if( ! masklist ) return false; // this *must* be false
    if( !*masklist ) return false; // this *must* be false
    if( filename ) if( !*filename ) return false;
    _idx_ nth_mask = 0;
    PCWSTR origmasklist = masklist;
    while( *masklist )
    {
        WCHAR msk[_MAX_PATH_ + 1];
        bool within_charlist = false;
        while( *masklist == L';' ) { masklist++; imask--; ++nth_mask; }  // overjump following ';' chars
        _idx_ istart = (_idx_)( masklist - origmasklist );   // subtraction handles char-size correctly
        _idx_ i;
        for( i = 0 ; ( i < _countof(msk) - 1 ) && ( *masklist && *masklist != L';' ) ; i++, masklist++ )
            msk[i] = *masklist;
        msk[i] = 0;
        _idx_ iend = istart + i;
        if( *msk )          // if #msk# is a clear string -> this does not mean to be a match
            if( filename )
            {
                if( _s_findmask_( filename, msk, bCaseSensitive, NULL, NULL, newmask, cnewmask ) ) {
                    if( start ) *start = istart;
                    if( end ) *end = iend;
                    if( pMaskIndex ) *pMaskIndex = nth_mask;
                    return true;        // one match found
                }
            }
            else if( newmask && cnewmask > 0 )
            {
                if( imask < 0 ) return false;
                if( imask == 0 ) {
                    _strcpy_( newmask, msk, cnewmask );
                    if( start ) *start = istart;
                    if( end ) *end = iend;
                    if( pMaskIndex ) *pMaskIndex = nth_mask;
                    return true;        // mask (at index #imask#) found
                }
            }
        // Now find the next ';' but only that which is not within a "[ ... ]" sequence.
        while( *masklist )
        {
            if( within_charlist ) {
                if( *masklist == L']' ) within_charlist = false;
            } else {
                if( *masklist == L'[' ) within_charlist = true;
                else if( *masklist == L';' ) break;
            }
            masklist++;
        }
    }
    return false;
}

///<summary> if a filename matches several masks (separated by ';')
///  comparing will be case-sensitive. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _matchmasks_( PCSTR filename, PCSTR masklist )
{
    return _s_matchmasks_( filename, masklist, true, NULL, NULL, -1, NULL, 0 );
}

///<summary> if a filename matches several masks (separated by ';')
///  comparing will be case-sensitive. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _matchmasks_( PCWSTR filename, PCWSTR masklist )
{
    return _s_matchmasks_( filename, masklist, true, NULL, NULL, -1, NULL, 0 );
}

///<summary> if a filename matches several masks (separated by ';')
///  comparing will be case-insensitive. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _matchimasks_( PCSTR filename, PCSTR masklist )
{
    return _s_matchmasks_( filename, masklist, false, NULL, NULL, -1, NULL, 0 );
}

///<summary> if a filename matches several masks (separated by ';')
///  comparing will be case-insensitive. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _matchimasks_( PCWSTR filename, PCWSTR masklist )
{
    return _s_matchmasks_( filename, masklist, false, NULL, NULL, -1, NULL, 0 );
}

bool _matchpathmasks_( PCSTR filename, PCSTR masklist )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _matchmasks_ ( filename, masklist );
        case PathT_WindowsPath  : return _matchimasks_( filename, masklist );
        default                 : return _matchimasks_( filename, masklist );
    }
}

bool _matchpathmasks_( PCWSTR filename, PCWSTR masklist )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _matchmasks_ ( filename, masklist );
        case PathT_WindowsPath  : return _matchimasks_( filename, masklist );
        default                 : return _matchimasks_( filename, masklist );
    }
}

///<summary> Similar to <b>_matchmasks_()</b>, the only difference is the return value.
///  For parameter description see <b>_s_matchmasks_()</b> above,
///  except for the return value. </summary>
///<returns> index number of the mask which did cause the match. <br/>
///    -1 = no matching occured. </returns>
_idx_ _findmatchmasks_( PCSTR filename, PCSTR masklist )
{
    _idx_ iMask = 0;
    if( !_s_matchmasks_( filename, masklist, true, NULL, NULL, -1, NULL, 0, &iMask ) ) return -1;
    if( !masklist ) return 0;
    return iMask;
}

///<summary> Similar to <b>_matchmasks_()</b>, the only difference is the return value.
///  For parameter description see <b>_s_matchmasks_()</b> above,
///  except for the return value. </summary>
///<returns> index number of the mask which did cause the match. <br/>
///    -1 = no matching occured. </returns>
_idx_ _findmatchmasks_( PCWSTR filename, PCWSTR masklist )
{
    _idx_ iMask = 0;
    if( !_s_matchmasks_( filename, masklist, true, NULL, NULL, -1, NULL, 0, &iMask ) ) return -1;
    if( !masklist ) return 0;
    return iMask;
}

///<summary> Similar to <b>_matchimasks_()</b>, the only difference is the return value.
///  For parameter description see <b>_s_matchmasks_()</b> above,
///  except for the return value. </summary>
///<returns> index number of the mask which did cause the match. <br/>
///    -1 = no matching occured. </returns>
_idx_ _findmatchimasks_( PCSTR filename, PCSTR masklist )
{
    _idx_ iMask = 0;
    if( !_s_matchmasks_( filename, masklist, false, NULL, NULL, -1, NULL, 0, &iMask ) ) return -1;
    if( !masklist ) return 0;
    return iMask;
}

///<summary> Similar to <b>_matchimasks_()</b>, the only difference is the return value.
///  For parameter description see <b>_s_matchmasks_()</b> above,
///  except for the return value. </summary>
///<returns> index number of the mask which did cause the match. <br/>
///    -1 = no matching occured. </returns>
_idx_ _findmatchimasks_( PCWSTR filename, PCWSTR masklist )
{
    _idx_ iMask = 0;
    if( !_s_matchmasks_( filename, masklist, false, NULL, NULL, -1, NULL, 0, &iMask ) ) return -1;
    if( !masklist ) return 0;
    return iMask;
}

_idx_ _findmatchpathmasks_( PCSTR filename, PCSTR masklist )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findmatchmasks_ ( filename, masklist );
        case PathT_WindowsPath  : return _findmatchimasks_( filename, masklist );
        default                 : return _findmatchimasks_( filename, masklist );
    }
}

_idx_ _findmatchpathmasks_( PCWSTR filename, PCWSTR masklist )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _findmatchmasks_ ( filename, masklist );
        case PathT_WindowsPath  : return _findmatchimasks_( filename, masklist );
        default                 : return _findmatchimasks_( filename, masklist );
    }
}

///<summary> Same as <b>_matchmasks_()</b>, but it also will return the char-index of
///  that one mask in the masks-string which did cause the matching. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _getmatchmasks_( PCSTR filename, PCSTR masklist, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    return _s_matchmasks_( filename, masklist, true, start, end, -1, newmask, cnewmask, pMaskIndex );
}

///<summary> Same as <b>_matchmasks_()</b>, but it also will return the char-index of
///  that one mask in the masks-string which did cause the matching. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _getmatchmasks_( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    return _s_matchmasks_( filename, masklist, true, start, end, -1, newmask, cnewmask, pMaskIndex );
}

///<summary> Same as <b>_matchimasks_()</b>, but it also will return the char-index of
///  that one mask in the masks-string which did cause the matching. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _getmatchimasks_( PCSTR filename, PCSTR masklist, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    return _s_matchmasks_( filename, masklist, false, start, end, -1, newmask, cnewmask, pMaskIndex );
}

///<summary> Same as <b>_matchimasks_()</b>, but it also will return the char-index of
///  that one mask in the masks-string which did cause the matching. <br/>
///  For parameter description see <b>_s_matchmasks_()</b> above. </summary>
bool _getmatchimasks_( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    return _s_matchmasks_( filename, masklist, false, start, end, -1, newmask, cnewmask, pMaskIndex );
}

bool _getmatchpathmasks_( PCSTR filename, PCSTR masklist, _idx_ *start, _idx_ *end, PSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _getmatchmasks_ ( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
        case PathT_WindowsPath  : return _getmatchimasks_( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
        default                 : return _getmatchimasks_( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
    }
}

bool _getmatchpathmasks_( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex )
{
    switch( g_PathType )
    {
        case PathT_UnixPath     : return _getmatchmasks_ ( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
        case PathT_WindowsPath  : return _getmatchimasks_( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
        default                 : return _getmatchimasks_( filename, masklist, start, end, newmask, cnewmask, pMaskIndex );
    }
}

///<summary> This function will return the mask of <b>masklist</b> with the index <b>imask</b>.
///  It will return it in <b>mask</b> which is a string buffer of length <b>cmask</b> (including
///  terminating 0). (The mask will be the same as it was in <b>masklist</b>, this
///  means e.g. "***.*" will stay the same and not be converted into "*.*"). </summary>
///<returns> true = the mask has been found <br/>  false = not found </returns>
bool _getmask_( PSTR mask, _size_t_ cmask, PCSTR masklist, _idx_ imask )
{
    return _s_matchmasks_( NULL, masklist, false, NULL, NULL, imask, mask, cmask );
}

///<summary> This function will return the mask of <b>masklist</b> with the index <b>imask</b>.
///  It will return it in <b>mask</b> which is a string buffer of length <b>cmask</b> (including
///  terminating 0). (The mask will be the same as it was in <b>masklist</b>, this
///  means e.g. "***.*" will stay the same and not be converted into "*.*"). </summary>
///<returns> true = the mask has been found <br/>  false = not found </returns>
bool _getmask_( PWSTR mask, _size_t_ cmask, PCWSTR masklist, _idx_ imask )
{
    return _s_matchmasks_( NULL, masklist, false, NULL, NULL, imask, mask, cmask );
}

///<summary> same like <b>_getmask_()</b> </summary>
///<param name="start"> see description to <b>_s_matchmasks_()</b> </param>
///<param name="end"> see description to <b>_s_matchmasks_()</b> </param>
bool _getmask2_( PSTR mask, _size_t_ cmask, PCSTR masklist, _idx_ imask, _idx_ *start, _idx_ *end )
{
    return _s_matchmasks_( NULL, masklist, false, start, end, imask, mask, cmask );
}

///<summary> same like <b>_getmask_()</b> </summary>
///<param name="start"> see description to <b>_s_matchmasks_()</b> </param>
///<param name="end"> see description to <b>_s_matchmasks_()</b> </param>
bool _getmask2_( PWSTR mask, _size_t_ cmask, PCWSTR masklist, _idx_ imask, _idx_ *start, _idx_ *end )
{
    return _s_matchmasks_( NULL, masklist, false, start, end, imask, mask, cmask );
}

///<summary>
///  <b>_getmaskitems_()</b> gets an array of <b>WildmaskItem</b> structs
///  where each item represents a sequence of the mask.
///</summary>
///<param name="mask">  </param>
///<param name="pWildmaskItemArray"> Will be filled with <b>WildmaskItem</b> struct items
///  that each indicates a sequence of the mask. </param>
///<param name="cWildmaskItemArray"> the size of the <b>pWildmaskItemArray</b> array in <b>WildmaskItem</b> structs. </param>
///<param name="bAutoCorrectWildcards"> If true (the default), consecutive '*' wildcard chars will
///  be interpreted as one (so as one sequence) and only one item will be created for them in
///  the returned array. <br/>
///  If false, consecutive '*' wildcard chars will be represented as independent
///  items in the array. </param>
///<returns> the number of items filled into the <b>pWildmaskItemArray</b> array. </returns>
_size_t_ _getmaskitems_( PCSTR mask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, bool bAutoCorrectWildcards )
{
    if( !mask ) return 0;
    _size_t_ cCurrWildmaskItemArray = 0;
    _idx_ m ;
    CHAR mchar;
    for( m = 0 ; mask[m] ; m++ )
    {
        mchar = mask[m];
        switch( mchar )
        {
            case '*' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                }
                ++m;
                if( bAutoCorrectWildcards )
                    while( mask[m] == '*' ) m++;   // if more '*' follow one after each other -> jump to the last
                --m;
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case '?' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Questionmark;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case '.' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Dot;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case '[' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_oneCharOfMany;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                }
                m++;
                while( mask[m] && mask[m] != ']' )
                {
                    if( mask[m + 1] == '-' )      // e.g. "[a-d]" but also e.g. "[--]"
                    {
                        if( !mask[m + 2] ) return 0;    // error
                        m += 2;
                    }
                    m++;
                }
                if( !mask[m] ) return 0;     // no corresponding ']' in mask found -> error
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            }

            case 0 :
                return 0;

            default :
            {
                // Now we check if the last WildmaskItem is already of type Other.
                // If so, we just extend this one, otherwise we have to add a new one.
                bool bCreateNewWildmaskItem = true;
                if( pWildmaskItemArray && cWildmaskItemArray > 0 && cCurrWildmaskItemArray > 0 && cCurrWildmaskItemArray <= cWildmaskItemArray )
                    if( pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_Type == wildmaskitem_Other ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iMaskItemEnd = m + 1;
                        bCreateNewWildmaskItem = false;
                    }
                if( bCreateNewWildmaskItem )
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Other;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                        ++cCurrWildmaskItemArray;
                    }
                break;
            }
        }
    }

    return cCurrWildmaskItemArray;
}

///<summary>
///  <b>_getmaskitems_()</b> gets an array of <b>WildmaskItem</b> structs
///  where each item represents a sequence of the mask.
///</summary>
///<param name="mask">  </param>
///<param name="pWildmaskItemArray"> Will be filled with <b>WildmaskItem</b> struct items
///  that each indicates a sequence of the mask. </param>
///<param name="cWildmaskItemArray"> the size of the <b>pWildmaskItemArray</b> array in <b>WildmaskItem</b> structs. </param>
///<param name="bAutoCorrectWildcards"> If true (the default), consecutive '*' wildcard chars will
///  be interpreted as one (so as one sequence) and only one item will be created for them in
///  the returned array. <br/>
///  If false, consecutive '*' wildcard chars will be represented as independent
///  items in the array. </param>
///<returns> the number of items filled into the <b>pWildmaskItemArray</b> array. </returns>
_size_t_ _getmaskitems_( PCWSTR mask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, bool bAutoCorrectWildcards )
{
    if( !mask ) return 0;
    _size_t_ cCurrWildmaskItemArray = 0;
    _idx_ m ;
    WCHAR mchar;
    for( m = 0 ; mask[m] ; m++ )
    {
        mchar = mask[m];
        switch( mchar )
        {
            case L'*' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Asterisk;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                }
                ++m;
                if( bAutoCorrectWildcards )
                    while( mask[m] == L'*' ) m++;   // if more '*' follow one after each other -> jump to the last
                --m;
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case L'?' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Questionmark;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case L'.' :
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Dot;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;

            case L'[' :
            {
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_oneCharOfMany;
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                }
                m++;
                while( mask[m] && mask[m] != L']' )
                {
                    if( mask[m + 1] == L'-' )      // e.g. "[a-d]" but also e.g. "[--]"
                    {
                        if( !mask[m + 2] ) return 0;    // error
                        m += 2;
                    }
                    m++;
                }
                if( !mask[m] ) return 0;     // no corresponding ']' in mask found -> error
                if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                    pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd = m + 1;
                    ++cCurrWildmaskItemArray;
                }
                break;
            }

            case 0 :
                return 0;

            default :
            {
                // Now we check if the last WildmaskItem is already of type Other.
                // If so, we just extend this one, otherwise we have to add a new one.
                bool bCreateNewWildmaskItem = true;
                if( pWildmaskItemArray && cWildmaskItemArray > 0 && cCurrWildmaskItemArray > 0 && cCurrWildmaskItemArray <= cWildmaskItemArray )
                    if( pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_Type == wildmaskitem_Other ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray - 1].m_iMaskItemEnd = m + 1;
                        bCreateNewWildmaskItem = false;
                    }
                if( bCreateNewWildmaskItem )
                    if( pWildmaskItemArray && cCurrWildmaskItemArray < cWildmaskItemArray ) {
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_Type = wildmaskitem_Other;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemStart = m;
                        pWildmaskItemArray[cCurrWildmaskItemArray].m_iMaskItemEnd   = m + 1;
                        ++cCurrWildmaskItemArray;
                    }
                break;
            }
        }
    }

    return cCurrWildmaskItemArray;
}

_size_t_ _countmasks_( PCSTR  masklist )
{
    if( !masklist ) return 0;
    if( !*masklist ) return 0;
    bool in_charlist = false;
    _idx_ cnt = 1;
    for( _idx_ i = 0 ; masklist[i] ; i++ ) {
        if( masklist[i] == '[' ) {
            in_charlist = true;
            if( masklist[i+1] ) i++; // and overjump a possible following ']' char
        } else if( masklist[i] == ']' ) {
            in_charlist = false;
        } else if( masklist[i] == ';' && !in_charlist )
            cnt++;
    }
    return cnt;
}

_size_t_ _countmasks_( PCWSTR masklist )
{
    if( !masklist ) return 0;
    if( !*masklist ) return 0;
    bool in_charlist = false;
    _idx_ cnt = 1;
    for( _idx_ i = 0 ; masklist[i] ; i++ ) {
        if( masklist[i] == L'[' ) {
            in_charlist = true;
            if( masklist[i+1] ) i++; // and overjump a possible following ']' char
        } else if( masklist[i] == L']' ) {
            in_charlist = false;
        } else if( masklist[i] == L';' && !in_charlist )
            cnt++;
    }
    return cnt;
}



///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
int _sprintf_( PSTR buffer, _size_t_ size, PCSTR format, ... )
{
    va_list v;
    va_start( v, format );
    int ret = _vsprintf_( buffer, size, format, v );
    va_end( v );
    return ret;
}

///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
int _sprintf_( PWSTR buffer, _size_t_ size, PCWSTR format, ... )
{
    va_list v;
    va_start( v, format );
    int ret = _vsprintf_( buffer, size, format, v );
    va_end( v );
    return ret;
}

///<param name="size"> Size of the destination string buffer
///  in CHARs (including terminating 0). </param>
int _vsprintf_( PSTR buffer, _size_t_ size, PCSTR format, va_list argptr )
{
    if( !buffer || size <= 0 ) return -1;
    *buffer = 0;
    if( !format || size == 1 ) return 0;
    int ret = _vsnprintf( buffer, size - 1, format, argptr );
    buffer[size - 1] = 0;
    return ret;
}

///<param name="size"> Size of the destination string buffer
///  in WCHARs (including terminating 0). </param>
int _vsprintf_( PWSTR buffer, _size_t_ size, PCWSTR format, va_list argptr )
{
    if( !buffer || size <= 0 ) return -1;
    *buffer = 0;
    if( !format || size == 1 ) return 0;
    int ret = _vsnwprintf( buffer, size - 1, format, argptr );
    buffer[size - 1] = 0;
    return ret;
}

#ifndef _WIN32_WCE

int _scprintf_( PCSTR format, ... )
{
    va_list v;
    va_start( v, format );
    int ret = _vscprintf_( format, v );
    va_end( v );
    return ret;
}

int _scprintf_( PCWSTR format, ... )
{
    va_list v;
    va_start( v, format );
    int ret = _vscprintf_( format, v );
    va_end( v );
    return ret;
}

int _vscprintf_( PCSTR format, va_list argptr )
{
    if( !format ) return 0;
    return _vscprintf( format, argptr );
}

int _vscprintf_( PCWSTR format, va_list argptr )
{
    if( !format ) return 0;
    return _vscwprintf( format, argptr );
}

#endif



#ifndef _WIN32_WCE

_size_t_ _strftime_( PSTR dest, _size_t_ zlen, PCSTR format, const struct tm *timeptr )
{
    if( !dest || zlen <= 0 || !format ) return 0;
    return (_size_t_) strftime( dest, zlen, format, timeptr );
}

_size_t_ _strftime_( PWSTR dest, _size_t_ zlen, PCWSTR format, const struct tm *timeptr )
{
    if( !dest || zlen <= 0 || !format ) return 0;
    return (_size_t_) wcsftime( dest, zlen, format, timeptr );
}

#endif



///<param name="zlenSrce">
///  Length of the source string in characters. This length may or may not
///  include the terminating null char. If you include it, it will be converted
///  and copied into the destination string, too. Otherwise not.
///  <br/>
///  If this param is -1, then the source string is imagined to be
///  null terminated and it will be converted including the null terminator.
///</param>
///<remarks>
///  This function will *not* append a null-terminator if it does not
///  have one in the range of <b>zlenSrce</b>.
///  But if in the range of <b>zlenSrce</b> is a null-terminator, or if
///  <b>zlenSrce</b> is -1, then the null-terminator will be included
///  in the conversion.
///</remarks>
bool _strconv_( PWSTR pstrDest, PCSTR pstrSrce, _ssize_t_ zlenSrce )
{
    if( !pstrDest || !pstrSrce || (void*) pstrDest == (void*) pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    if( zlenSrce < 0 )
        zlenSrce = (int) ::strlen( pstrSrce ) + 1;
    if( zlenSrce <= 0 ) return true;

    int ret = MultiByteToWideChar( CP_ACP, 0, pstrSrce, (int) zlenSrce, pstrDest, (int) zlenSrce );
    if( ret == 0 ) return false;

    return true;
}

///<param name="zlenSrce">
///  Length of the source string in characters. This length may or may not
///  include the terminating null char. If you include it, it will be converted
///  and copied into the destination string, too. Otherwise not.
///  <br/>
///  If this param is -1, then the source string is imagined to be
///  null terminated and it will be converted including the null terminator.
///</param>
///<remarks>
///  This function will *not* append a null-terminator if it does not
///  have one in the range of <b>zlenSrce</b>.
///  But if in the range of <b>zlenSrce</b> is a null-terminator, or if
///  <b>zlenSrce</b> is -1, then the null-terminator will be included
///  in the conversion.
///</remarks>
bool _strconv_( PSTR pstrDest, PCWSTR pstrSrce, _ssize_t_ zlenSrce )
{
    if( !pstrDest || !pstrSrce || (void*) pstrDest == (void*) pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    if( zlenSrce < 0 )
        zlenSrce = (int) ::wcslen( pstrSrce ) + 1;
    if( zlenSrce <= 0 ) return true;

    int ret = WideCharToMultiByte( CP_ACP, 0, pstrSrce, (int) zlenSrce, pstrDest, (int) zlenSrce, NULL, NULL );
    if( ret == 0 ) return false;

    return true;
}

///<remarks>
///  This function will *not* append a null-terminator if it does not
///  have one in the range of <b>zlenSrce</b>.
///  But if in the range of <b>zlenSrce</b> is a null-terminator, or if
///  <b>zlenSrce</b> is -1, then the null-terminator will be included
///  in the conversion.
///</remarks>
bool _strconv_( PSTR pstrDest, PCSTR pstrSrce, _ssize_t_ zlenSrce )
{
    if( !pstrDest || !pstrSrce || (void*) pstrDest == (void*) pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    memcpy( pstrDest, pstrSrce,
            (zlenSrce < 0 ? ::strlen(pstrSrce) + 1 : zlenSrce) * sizeof(CHAR) );

    return true;
}

///<remarks>
///  This function will *not* append a null-terminator if it does not
///  have one in the range of <b>zlenSrce</b>.
///  But if in the range of <b>zlenSrce</b> is a null-terminator, or if
///  <b>zlenSrce</b> is -1, then the null-terminator will be included
///  in the conversion.
///</remarks>
bool _strconv_( PWSTR pstrDest, PCWSTR pstrSrce, _ssize_t_ zlenSrce )
{
    if( !pstrDest || !pstrSrce || (void*) pstrDest == (void*) pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    memcpy( pstrDest, pstrSrce,
            (zlenSrce < 0 ? ::wcslen(pstrSrce) + 1 : zlenSrce) * sizeof(WCHAR) );

    return true;
}



#ifndef _WIN32_WCE

#ifndef __argonTime_h
// Note: "argonTime.h" includes the definitions of the following time functions,
// so we do not define them in here a second time.

struct tm *_localtime_( struct tm *ptm, const time_t *ptime )
{
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)

    if( localtime_s( ptm, ptime ) )
        return NULL;

    return ptm;

    #else

    if( !ptm ) return NULL;
    if( !ptime )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    struct tm *ret = localtime( ptime );
    if( !ret )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    *ptm = *ret;
    return ptm;

    #endif
}

struct tm *_gmtime_( struct tm *ptm, const time_t *ptime )
{
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)

    if( gmtime_s( ptm, ptime ) )
        return NULL;

    return ptm;

    #else

    if( !ptm ) return NULL;
    if( !ptime )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    struct tm *ret = gmtime( ptime );
    if( !ret )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    *ptm = *ret;
    return ptm;

    #endif
}

#endif

#endif




}
