/*
    File:               argonString.h
    Description:        Contains string handling extensions.
                        Global functions as well as extensions to the
                        ATL::CStringT<> template (and therefore derived
                        MFC classes CString, CStringA, CStringW).
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         5.10.2003
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    IMPORTANT:
    To be able to use CStrings (so CStringT<>, CString, CStringA, CStringW), either
    include this file into a project using the MFC or a non-MFC project.
    For a non-MFC project,
        #include <atlstr.h>
    will be done for you.

    This file will work with VC++ 7.0 or higher. (For MFC usage, this means that
    MFC v7.0 or higher is supported).


    _argon_no_CString       If you #define this name before including "argonString.h",
                            then this header file will not include <atlstr.h> and not
                            use CStringT<>, CString, CStringA, CStringW.
                            Instead, it will use an intneral memory handling.
                            So in your source files you need to include "argonMemory.h"
                            before including "argonString.h" or "argonErrors.h" or any
                            other include file that includes one of these.

    _argonString_no_Initer  If you #define this name before including "argonString.h"
                            in your "stdafx.h", then the root string initializations
                            that would have been done during global object
                            initialization will *not* be done.
                            The root inits would include:
                             -  setting the default locale (by calling #_set_default_locale_()#)


    String Usage:
    ----------------

    The main important class is the template class StringT<>.
    In there, you see all the public method declarations (below the ctor) you can use to
    manipulate strings.
    Further below, at the definitions of these methods (with their bodies) you usually
    will find the remarks to those methods. They show important informations and
    how to handle the methods and their parameters.

    In your code, you mainly will use String, StringA and StringW which are
    typedefs of StringT<> (declared at the bottom of this file).

    Using the _argon_no_CString define, you can specify if StringT<> shall be
    based on CString or on an own, independent base implementation.

    global string manipulation functions usage:
    ---------------------------------------------------

    More near the upper part of this file, you will find the declarations
    of global string manipulation functions that are based on and similar
    to CRT's string manipulation functions. But the versions in here
    allow automatic handling of ASCII and Unicode (wide) string pointers.
    Plus they add many helpful functionalities to common string handling life.
    Many of these functions are used by the StringT<> template.

    The functions's names start with a _ char and end with a _ char.

    The descriptions and remarks to these functions you will usually find
    in the "argonString.cpp" file.
*/

#ifndef __argonString_h
#define __argonString_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #ifndef _argon_no_CString
    #ifndef _MFC_VER
    // Do this include for a non-MFC project.
    #include <atlstr.h>
    #endif
    #endif

    #include <shlobj.h>
    #include <lmcons.h>     // for UNLEN
    #include <locale.h>
    #include <stdlib.h>
    #include <stdio.h>
    #include <math.h>
    #include <time.h>
    // Note: If your compiler yields problems with the #tm# structure,
    // try including <time.h> in your "stdafx.h" file.
    #include <objbase.h>

    #include "argonInstance.h"
    //#include "argonMemory.h"
    //  Note: "argonMemory.h" you need to include on your own if you want
    //  StringT<> to use an own memory implementation as base class instead
    //  of CString. (See #define _argon_no_CString).
    //  If you forget to include it before "argonString.h" and "argonErrors.h"
    //  an error message will be displayed by the compiler.
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4239 )
    // Disables a minor warning of nonstandard extension usage. This warning
    // normally occurs when passing a temporary object (no variable)
    // as parameter to a reference to the class.
    // Since in VC++ this construction *is* allowed as an extension and
    // we often use microsoft specific constellations, it is not harmful
    // at all to also allow using this microsoft specific feature
    // since it increases code writing and readability.

    #pragma warning( disable: 4706 )    // assignment within conditional expression

    #pragma warning( disable: 4100 )    // unreferenced parameter

    // The following warning disabling will prevent the compiler from
    // yawning floating point overflow errors (e.g. helpful when
    // using DirectX).
    #pragma warning( disable: 4056 )    // overflow in floating point constant arithmetic
    #pragma warning( disable: 4756 )    // overflow in constant arithmetic

    #pragma warning( disable: 4127 )    // conditional expression is constant
    // This would occur e.g. in something like  if( sizeof(TCHAR) == sizeof(CHAR) )

    #pragma warning( disable: 4201 )    // nonstandard extension used : nameless struct/union
    // Will be necessary e.g. when using   #include <mmsystem.h>

    #pragma warning( disable: 4267 )    // 'argument' : conversion from 'size_t' to 'int', possible loss of data

    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




//  initialization
namespace privateString
{
    ///<summary> This Initer class will just make some root
    ///  string initializations during global objects init time. </summary>
    class Initer
    {
        public :
            Initer();
    };
}

// ---------------------------------------------------------------- //


#ifndef _countof
#define _countof(array) (sizeof(array) / sizeof((array)[0]))
#endif

typedef INT_PTR     _idx_;
typedef size_t      _size_t_;
typedef INT_PTR     _ssize_t_;

typedef _size_t_    _hashval_t_;    // hashkey value type (see #HashTable<>#) (*must* be an unsigned type)


// ---------------------------------------------------------------- //
//  string and path and filesystem related

#define __W(x)          L ## x
#define _W(x)           __W(x)
// Note: To make #_W# macro work, we must do the detour using
//  the #__W# macro. This is necessary if you pass the #_W# macro
//  a macro-name instead of a string directly, so that #__W# is called
//  with the *expanded* macro-name, expanded to its defined string.


#define _MAX_PATH_  511

#ifdef MAX_PATH
    #if( MAX_PATH > _MAX_PATH_ )
        #undef  _MAX_PATH_
        #define _MAX_PATH_      (MAX_PATH)
    #endif
#endif

// path types
enum path_t
{
    PathT_UnixPath      = 1,
    PathT_WindowsPath,
    PathT_Default       = PathT_WindowsPath,

    _pad_path_t         = 0x7fffffff
};

extern "C" path_t g_PathType;
// This contains the global setting of the path type to use.
// You can set it whenever you need to.
// Its sense is used e.g. by several path handling string functions
// of "Std.cpp" to know if rather a \ or a / char shall be
// used as path seperator when forming path strings.



#ifndef __WFILE__
#define __WFILE__   _W(__FILE__)
#endif




// ---------------------------------------------------------------- //
//  root, extended global functions

#ifndef _WIN32_WCE
inline PSTR  _setlocale_( int category, PCSTR  locale ) { return   setlocale( category, locale ); }
inline PWSTR _setlocale_( int category, PCWSTR locale ) { return _wsetlocale( category, locale ); }

void _set_default_locale_();
void _set_C_locale_();

class StoreLocale
{
    private :
        PTSTR m_pstrLocale;
        bool  m_bRestoreOnEnd;

    public :
        StoreLocale()   { m_pstrLocale = NULL; m_bRestoreOnEnd = true; StoreNow(); }
        ~StoreLocale();

        bool StoreNow();
        bool RestoreNow();

        void Dismiss () { m_bRestoreOnEnd = false; } // call this to prevent the locale being restored when #this# object is deleted
        void Reenable() { m_bRestoreOnEnd = true;  }
};
#endif

inline CHAR  _toupper_ ( CHAR  c ) { return (CHAR)toupper( (UCHAR)c ); }
inline WCHAR _toupper_ ( WCHAR c ) { return (WCHAR)towupper( c ); }
inline CHAR  _tolower_ ( CHAR  c ) { return (CHAR)tolower( (UCHAR)c ); }
inline WCHAR _tolower_ ( WCHAR c ) { return (WCHAR)towlower( c ); }
inline bool  _isspace_ ( CHAR  c ) { return isspace( (UCHAR)c ) != 0; }
inline bool  _isspace_ ( WCHAR c ) { return iswspace( c ) != 0; }
inline bool  _isdigit_ ( CHAR  c ) { return isdigit( (UCHAR)c ) != 0; }
inline bool  _isdigit_ ( WCHAR c ) { return iswdigit( c ) != 0; }
inline bool  _isxdigit_( CHAR  c ) { return isxdigit( (UCHAR)c ) != 0; }
inline bool  _isxdigit_( WCHAR c ) { return iswxdigit( c ) != 0; }
bool _startswithdigit_( PCSTR  string, bool bAllowNegativeNumber = true );
bool _startswithdigit_( PCWSTR string, bool bAllowNegativeNumber = true );
inline bool  _isalpha_ ( CHAR  c ) { return isalpha( (UCHAR)c ) != 0; }
inline bool  _isalpha_ ( WCHAR c ) { return iswalpha( c ) != 0; }
inline bool  _isalnum_ ( CHAR  c ) { return isalnum( (UCHAR)c ) != 0; }
inline bool  _isalnum_ ( WCHAR c ) { return iswalnum( c ) != 0; }
inline bool  _iscntrl_ ( CHAR  c ) { return iscntrl( (UCHAR)c ) != 0; }
inline bool  _iscntrl_ ( WCHAR c ) { return iswcntrl( c ) != 0; }
bool _iscommand_        ( CHAR  c );
bool _iscommand_        ( WCHAR c );
bool _iscmdnum_         ( CHAR  c );
bool _iscmdnum_         ( WCHAR c );
bool _isfloatvalue_     ( CHAR  c );
bool _isfloatvalue_     ( WCHAR c );
bool _isstillfloatvalue_( CHAR  c );
bool _isstillfloatvalue_( WCHAR c );

double _stof_  ( PCSTR  string );
double _stof_  ( PCWSTR string );
int    _stoi_  ( PCSTR  string );
int    _stoi_  ( PCWSTR string );
INT64  _stoi64_( PCSTR  string );
INT64  _stoi64_( PCWSTR string );
long   _stol_  ( PCSTR  string );
long   _stol_  ( PCWSTR string );
int    _stoI_  ( PCSTR  string );
int    _stoI_  ( PCWSTR string );
int    _stox_  ( PCSTR  string );
int    _stox_  ( PCWSTR string );

_size_t_ _strlen_   ( PCSTR  string );
_size_t_ _strlen_   ( PCWSTR string );
_size_t_ _strlen_unaligned_( const CHAR  UNALIGNED *string );
_size_t_ _strlen_unaligned_( const WCHAR UNALIGNED *string );
_ssize_t_ _strlenmax_( PCSTR  string, _size_t_ max, bool dont_exceed );
_ssize_t_ _strlenmax_( PCWSTR string, _size_t_ max, bool dont_exceed );
_ssize_t_ _strlenmax_unaligned_( const CHAR  UNALIGNED *string, _size_t_ max, bool dont_exceed );
_ssize_t_ _strlenmax_unaligned_( const WCHAR UNALIGNED *string, _size_t_ max, bool dont_exceed );
inline bool _isstrempty_( PCSTR  string ) { return string ? (*string == ( CHAR)0) : true; }
inline bool _isstrempty_( PCWSTR string ) { return string ? (*string == (WCHAR)0) : true; }
PSTR  _strcpy_      ( PSTR  dest, PCSTR  srce, _size_t_ size );
PWSTR _strcpy_      ( PWSTR dest, PCWSTR srce, _size_t_ size );
template< int size > PSTR  _strcpy_( CHAR  (&dest)[size], PCSTR  srce ) { return _strcpy_( dest, srce, _countof(dest) ); }
template< int size > PWSTR _strcpy_( WCHAR (&dest)[size], PCWSTR srce ) { return _strcpy_( dest, srce, _countof(dest) ); }
CHAR  UNALIGNED *_strcpy_unaligned_( CHAR  UNALIGNED *dest, const CHAR  UNALIGNED *srce, _size_t_ size );
WCHAR UNALIGNED *_strcpy_unaligned_( WCHAR UNALIGNED *dest, const WCHAR UNALIGNED *srce, _size_t_ size );
PSTR  _strcat_      ( PSTR  dest, PCSTR  srce, _size_t_ size );
PWSTR _strcat_      ( PWSTR dest, PCWSTR srce, _size_t_ size );
template< int size > PSTR  _strcat_( CHAR  (&dest)[size], PCSTR  srce ) { return _strcat_( dest, srce, _countof(dest) ); }
template< int size > PWSTR _strcat_( WCHAR (&dest)[size], PCWSTR srce ) { return _strcat_( dest, srce, _countof(dest) ); }
PSTR  _straddch_    ( PSTR  dest, CHAR  ch, _size_t_ size );
PWSTR _straddch_    ( PWSTR dest, WCHAR ch, _size_t_ size );
PSTR  _straddmultch_( PSTR  dest, CHAR  ch, _size_t_ length );
PWSTR _straddmultch_( PWSTR dest, WCHAR ch, _size_t_ length );
_size_t_ _strlstlen_( PCSTR  stringlist );
_size_t_ _strlstlen_( PCWSTR stringlist );
PSTR   _strlstcpy_             ( PSTR  dest, PCSTR  srce, _size_t_ size );
PWSTR  _strlstcpy_             ( PWSTR dest, PCWSTR srce, _size_t_ size );
PSTR   _strlstcat_             ( PSTR  dest, PCSTR  srce, _size_t_ size );
PWSTR  _strlstcat_             ( PWSTR dest, PCWSTR srce, _size_t_ size );
_size_t_ _cstrlststrs_         ( PCSTR  stringlist );
_size_t_ _cstrlststrs_         ( PCWSTR stringlist );
PSTR   _getstrlststr_          ( PSTR  deststr, PCSTR  srcestrlst, _idx_ index, _size_t_ size );
PWSTR  _getstrlststr_          ( PWSTR deststr, PCWSTR srcestrlst, _idx_ index, _size_t_ size );
PCSTR  _get_strlst_str_ptr_    ( PCSTR  stringlist, _idx_ index );
PCWSTR _get_strlst_str_ptr_    ( PCWSTR stringlist, _idx_ index );
PCSTR  _fwd_to_next_strlst_str_( PCSTR  curr_strlst_string );
PCWSTR _fwd_to_next_strlst_str_( PCWSTR curr_strlst_string );
PSTR   _strlststrcreate_       ( PSTR  deststrlst, PCSTR  srcestr, _size_t_ size );
PWSTR  _strlststrcreate_       ( PWSTR deststrlst, PCWSTR srcestr, _size_t_ size );
PSTR   _strlststradd_          ( PSTR  deststrlst, PCSTR  srcestr, _size_t_ size );
PWSTR  _strlststradd_          ( PWSTR deststrlst, PCWSTR srcestr, _size_t_ size );
_idx_  _findstrlststr_         ( PCSTR  string, PCSTR  strlst );
_idx_  _findstrlststr_         ( PCWSTR string, PCWSTR strlst );
_idx_  _findistrlststr_        ( PCSTR  string, PCSTR  strlst );
_idx_  _findistrlststr_        ( PCWSTR string, PCWSTR strlst );
_idx_  _findstrlststrhere_     ( PCSTR  string, PCSTR  strlst );
_idx_  _findstrlststrhere_     ( PCWSTR string, PCWSTR strlst );
_idx_  _findistrlststrhere_    ( PCSTR  string, PCSTR  strlst );
_idx_  _findistrlststrhere_    ( PCWSTR string, PCWSTR strlst );

bool _stris_       ( PCSTR  string1, PCSTR  string2 );
bool _stris_       ( PCWSTR string1, PCWSTR string2 );
bool _striis_      ( PCSTR  string1, PCSTR  string2 );
bool _striis_      ( PCWSTR string1, PCWSTR string2 );
bool _strsepis_    ( PCSTR  string1, PCSTR  string2, PCSTR  pstr1_Separators );
bool _strsepis_    ( PCWSTR string1, PCWSTR string2, PCWSTR pstr1_Separators );
bool _strsepiis_   ( PCSTR  string1, PCSTR  string2, PCSTR  pstr1_Separators );
bool _strsepiis_   ( PCWSTR string1, PCWSTR string2, PCWSTR pstr1_Separators );
bool _strishere_   ( PCSTR  string1, PCSTR  string2 );
bool _strishere_   ( PCWSTR string1, PCWSTR string2 );
bool _striishere_  ( PCSTR  string1, PCSTR  string2 );
bool _striishere_  ( PCWSTR string1, PCWSTR string2 );
int _strcmp_       ( PCSTR  string1, PCSTR  string2 );
int _strcmp_       ( PCWSTR string1, PCWSTR string2 );
int _stricmp_      ( PCSTR  string1, PCSTR  string2 );
int _stricmp_      ( PCWSTR string1, PCWSTR string2 );
bool _strendswith_ ( PCSTR  pstrIn, PCSTR  pstrEnding );
bool _strendswith_ ( PCWSTR pstrIn, PCWSTR pstrEnding );
bool _striendswith_( PCSTR  pstrIn, PCSTR  pstrEnding );
bool _striendswith_( PCWSTR pstrIn, PCWSTR pstrEnding );

PCSTR  _strchr_    ( PCSTR  string, CHAR  c );
PCWSTR _strchr_    ( PCWSTR string, WCHAR c );
PCSTR  _strichr_   ( PCSTR  string, CHAR  c );
PCWSTR _strichr_   ( PCWSTR string, WCHAR c );
PCSTR  _strrchr_   ( PCSTR  string, CHAR  c );
PCWSTR _strrchr_   ( PCWSTR string, WCHAR c );
PCSTR  _strspcchr_ ( PCSTR  string, CHAR  c );
PCWSTR _strspcchr_ ( PCWSTR string, WCHAR c );
PCSTR  _strspcichr_( PCSTR  string, CHAR  c );
PCWSTR _strspcichr_( PCWSTR string, WCHAR c );
PCSTR  _strspc_    ( PCSTR  string );
PCWSTR _strspc_    ( PCWSTR string );
PSTR  _strupr_( PSTR  string );
PWSTR _strupr_( PWSTR string );
PSTR  _strlwr_( PSTR  string );
PWSTR _strlwr_( PWSTR string );
PSTR  _strupr_firstchar_( PSTR  string );
PWSTR _strupr_firstchar_( PWSTR string );
PSTR  _strlwr_firstchar_( PSTR  string );
PWSTR _strlwr_firstchar_( PWSTR string );
PSTR  _strrev_( PSTR  string );
PWSTR _strrev_( PWSTR string );
PCSTR  _strstr_ ( PCSTR  string, PCSTR  strfind );
PCWSTR _strstr_ ( PCWSTR string, PCWSTR strfind );
PCSTR  _stristr_( PCSTR  string, PCSTR  strfind );
PCWSTR _stristr_( PCWSTR string, PCWSTR strfind );
PCSTR  _strpbrk_( PCSTR  string, PCSTR  charset );
PCWSTR _strpbrk_( PCWSTR string, PCWSTR charset );
int _strcntchr_  ( PCSTR  string, CHAR  c );
int _strcntchr_  ( PCWSTR string, WCHAR c );
int _stricntchr_ ( PCSTR  string, CHAR  c );
int _stricntchr_ ( PCWSTR string, WCHAR c );
int _strcntchrs_ ( PCSTR  string, PCSTR  chars );
int _strcntchrs_ ( PCWSTR string, PCWSTR chars );
int _stricntchrs_( PCSTR  string, PCSTR  chars );
int _stricntchrs_( PCWSTR string, PCWSTR chars );
int _strcntspc_  ( PCSTR  string );
int _strcntspc_  ( PCWSTR string );
void _strHasWhichCharCases_( PCSTR  string, bool& hasLowercaseChars, bool& hasUppercaseChars );
void _strHasWhichCharCases_( PCWSTR string, bool& hasLowercaseChars, bool& hasUppercaseChars );
bool _strHasLowercaseChars_( PCSTR  string );
bool _strHasLowercaseChars_( PCWSTR string );
bool _strHasUppercaseChars_( PCSTR  string );
bool _strHasUppercaseChars_( PCWSTR string );

int _chtohex_( CHAR  c );
int _chtohex_( WCHAR c );
int _hextoch_( int hex );
int _hextoCH_( int hex );
int _chtodec_( CHAR  c );
int _chtodec_( WCHAR c );
int _dectoch_( int dec );
int _esctoch_( PCSTR  *pstring );
int _esctoch_( PCWSTR *pstring );
PSTR  _esccpy_( PSTR  dest, CHAR  ch, _size_t_ size );
PWSTR _esccpy_( PWSTR dest, WCHAR ch, _size_t_ size );
PSTR  _esccat_( PSTR  dest, CHAR  ch, _size_t_ size );
PWSTR _esccat_( PWSTR dest, WCHAR ch, _size_t_ size );
int _escscpy_( PSTR  dest, PCSTR  srce, _size_t_ size );
int _escscpy_( PWSTR dest, PCWSTR srce, _size_t_ size );
int _escscat_( PSTR  dest, PCSTR  srce, _size_t_ size );
int _escscat_( PWSTR dest, PCWSTR srce, _size_t_ size );
_idx_ _escfwdchr_( PCSTR  escstr, _idx_ ipos );
_idx_ _escfwdchr_( PCWSTR escstr, _idx_ ipos );
_idx_ _escbwdchr_( PCSTR  escstr, _idx_ ipos );
_idx_ _escbwdchr_( PCWSTR escstr, _idx_ ipos );
PCSTR  _escfwdchr_ptr_( PCSTR  escstr_ptr );
PCWSTR _escfwdchr_ptr_( PCWSTR escstr_ptr );
int _escstrlstcpy_( PSTR  dest, PCSTR  srcestrlst, _size_t_ size );
int _escstrlstcpy_( PWSTR dest, PCWSTR srcestrlst, _size_t_ size );
int _escstrlstcat_( PSTR  dest, PCSTR  srcestrlst, _size_t_ size );
int _escstrlstcat_( PWSTR dest, PCWSTR srcestrlst, _size_t_ size );
int _replace_nr_( PSTR  string, CHAR  wild, int n, int nbase );
int _replace_nr_( PWSTR string, WCHAR wild, int n, int nbase );

inline bool _ispathsep_( CHAR  ch ) { return( ch ==  '/' || ch ==  '\\' ); }
inline bool _ispathsep_( WCHAR ch ) { return( ch == L'/' || ch == L'\\' ); }
PCSTR  _findpathsep_( PCSTR  path );
PCWSTR _findpathsep_( PCWSTR path );
int    _defpathsep_();
CHAR   _getpathsep_( PCSTR  path );
WCHAR  _getpathsep_( PCWSTR path );
inline bool _are_paths_case_sensitive_() { return g_PathType == PathT_UnixPath; }
int    _isfilepath_( PCSTR  string, int strength );
int    _isfilepath_( PCWSTR string, int strength );
_size_t_ _getabsolutepathlen_( PCSTR  path, _size_t_ *plength = NULL );
_size_t_ _getabsolutepathlen_( PCWSTR path, _size_t_ *plength = NULL );
inline bool _isabsolutepath_ ( PCSTR  path ) { return( _getabsolutepathlen_( path ) > (_size_t_)0 ); }
inline bool _isabsolutepath_ ( PCWSTR path ) { return( _getabsolutepathlen_( path ) > (_size_t_)0 ); }
PSTR     _ospath_      ( PSTR  path );
PWSTR    _ospath_      ( PWSTR path );
PSTR     _unixpath_    ( PSTR  path );
PWSTR    _unixpath_    ( PWSTR path );
PSTR     _windowspath_ ( PSTR  path );
PWSTR    _windowspath_ ( PWSTR path );
PSTR     _parentpath_  ( PSTR  path );
PWSTR    _parentpath_  ( PWSTR path );
PSTR     _rootpath_    ( PSTR  path, _size_t_ size );
PWSTR    _rootpath_    ( PWSTR path, _size_t_ size );
_size_t_ _rootpathlen_ ( PCSTR  path );
_size_t_ _rootpathlen_ ( PCWSTR path );
PSTR     _strippath_   ( PSTR  path );
PWSTR    _strippath_   ( PWSTR path );
PSTR     _aspath_      ( PSTR  path, _size_t_ size );
PWSTR    _aspath_      ( PWSTR path, _size_t_ size );
_size_t_ _aspathlen_   ( PCSTR  path );
_size_t_ _aspathlen_   ( PCWSTR path );
bool     _isendpathsep_( PCSTR  path );
bool     _isendpathsep_( PCWSTR path );
PSTR  _makepath_ ( PSTR  dest, PCSTR  drive, PCSTR  dir,
                   PCSTR  ftitle, PCSTR  ext, _size_t_ size );
PWSTR _makepath_ ( PWSTR dest, PCWSTR drive, PCWSTR dir,
                   PCWSTR ftitle, PCWSTR ext, _size_t_ size );
void  _splitpath_( PCSTR  srce, PSTR  drive, PSTR  dir,
                   PSTR  ftitle, PSTR  ext, _size_t_ size_each );
void  _splitpath_( PCWSTR srce, PWSTR drive, PWSTR dir,
                   PWSTR ftitle, PWSTR ext, _size_t_ size_each );
PSTR  _joinpaths_( PSTR  dest, PCSTR  pre, PCSTR  post, _size_t_ size );
PWSTR _joinpaths_( PWSTR dest, PCWSTR pre, PCWSTR post, _size_t_ size );
bool  _driveof_       ( PSTR  drive, PCSTR  path, _size_t_ size );
bool  _driveof_       ( PWSTR drive, PCWSTR path, _size_t_ size );
PSTR  _trunc_to_dir_  ( PSTR  path );
PWSTR _trunc_to_dir_  ( PWSTR path );
PSTR  _directoryof_   ( PSTR  dir, PCSTR  path, _size_t_ size );
PWSTR _directoryof_   ( PWSTR dir, PCWSTR path, _size_t_ size );
PCSTR  _filenameptrof_( PCSTR  path );
PCWSTR _filenameptrof_( PCWSTR path );
PSTR  _filenameof_    ( PSTR  filename, PCSTR  path, _size_t_ size );
PWSTR _filenameof_    ( PWSTR filename, PCWSTR path, _size_t_ size );
PSTR  _filetitleof_   ( PSTR  filetitle, PCSTR  path, _size_t_ size );
PWSTR _filetitleof_   ( PWSTR filetitle, PCWSTR path, _size_t_ size );
PSTR  _fileextof_     ( PSTR  fileext, PCSTR  path, _size_t_ size );
PWSTR _fileextof_     ( PWSTR fileext, PCWSTR path, _size_t_ size );
int _pathcmp_       ( PCSTR  path1, PCSTR  path2 );
int _pathcmp_       ( PCWSTR path1, PCWSTR path2 );
inline bool _patheq_( PCSTR  path1, PCSTR  path2 ) { return 0 == _pathcmp_( path1, path2 ); }
inline bool _patheq_( PCWSTR path1, PCWSTR path2 ) { return 0 == _pathcmp_( path1, path2 ); }
bool _pathishere_   ( PCSTR  path1, PCSTR  path2 );
bool _pathishere_   ( PCWSTR path1, PCWSTR path2 );
int  _subpathcmp_   ( PCSTR  path1, PCSTR  path2 );
int  _subpathcmp_   ( PCWSTR path1, PCWSTR path2 );

enum eWildmaskItem
{
    wildmaskitem_unknown = 0,
    wildmaskitem_Other,         // just normal (non-wildcarded) characters
    wildmaskitem_Asterisk,      // '*' or a connected sequence of '*' chars
    wildmaskitem_Questionmark,  // '?'
    wildmaskitem_Dot,           // '.'
    wildmaskitem_oneCharOfMany  // '[...]'
};

struct WildmaskItem
{
    eWildmaskItem   m_Type;
    _idx_           m_iMaskItemStart;   // start char index of the mask item sequence
    _idx_           m_iMaskItemEnd;     // and (after) its end char (so length = end - start)
    _idx_           m_iFilenameStart;   // start and
    _idx_           m_iFilenameEnd;     // (after) end char index of the according sequence in the filename
    // Note: The Filename-sequence length can be 0 in that case of e.g. a '.' is specified
    // at the end of the mask but the filename does not have a '.' at its end but
    // is regarded as valid match. Or this can also happen in case the mask ends with '*' char
    // and there are no more chars in the filename (which will be regarded as valid match).
    // And some more...
};

bool _matchmask_ ( PCSTR  filename, PCSTR  mask );
bool _matchmask_ ( PCWSTR filename, PCWSTR mask );
bool _matchimask_( PCSTR  filename, PCSTR  mask );
bool _matchimask_( PCWSTR filename, PCWSTR mask );
bool _matchpathmask_( PCSTR  filename, PCSTR  mask );
bool _matchpathmask_( PCWSTR filename, PCWSTR mask );
bool _findmask_ ( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end );
bool _findmask_ ( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end );
bool _findimask_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end );
bool _findimask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end );
bool _findpathmask_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end );
bool _findpathmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end );
bool _findnewmask_ ( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask );
bool _findnewmask_ ( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask );
bool _findnewimask_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask );
bool _findnewimask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask );
bool _findnewpathmask_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask );
bool _findnewpathmask_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask );
bool _findnewmask2_ ( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _findnewmask2_ ( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _findnewimask2_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _findnewimask2_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _findnewpathmask2_( PCSTR  filename, PCSTR  mask, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _findnewpathmask2_( PCWSTR filename, PCWSTR mask, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, _size_t_ *pcWildmaskItemArray );
bool _matchmasks_ ( PCSTR  filename, PCSTR  masklist );
bool _matchmasks_ ( PCWSTR filename, PCWSTR masklist );
bool _matchimasks_( PCSTR  filename, PCSTR  masklist );
bool _matchimasks_( PCWSTR filename, PCWSTR masklist );
bool _matchpathmasks_( PCSTR  filename, PCSTR  masklist );
bool _matchpathmasks_( PCWSTR filename, PCWSTR masklist );
_idx_ _findmatchmasks_ ( PCSTR  filename, PCSTR  masklist );
_idx_ _findmatchmasks_ ( PCWSTR filename, PCWSTR masklist );
_idx_ _findmatchimasks_( PCSTR  filename, PCSTR  masklist );
_idx_ _findmatchimasks_( PCWSTR filename, PCWSTR masklist );
_idx_ _findmatchpathmasks_( PCSTR  filename, PCSTR  masklist );
_idx_ _findmatchpathmasks_( PCWSTR filename, PCWSTR masklist );
bool _getmatchmasks_ ( PCSTR  filename, PCSTR  masklist, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmatchmasks_ ( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmatchimasks_( PCSTR  filename, PCSTR  masklist, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmatchimasks_( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmatchpathmasks_( PCSTR  filename, PCSTR  masklist, _idx_ *start, _idx_ *end, PSTR  newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmatchpathmasks_( PCWSTR filename, PCWSTR masklist, _idx_ *start, _idx_ *end, PWSTR newmask, _size_t_ cnewmask, _idx_ *pMaskIndex = NULL );
bool _getmask_ ( PSTR  mask, _size_t_ cmask, PCSTR  masklist, _idx_ imask );
bool _getmask_ ( PWSTR mask, _size_t_ cmask, PCWSTR masklist, _idx_ imask );
bool _getmask2_( PSTR  mask, _size_t_ cmask, PCSTR  masklist, _idx_ imask, _idx_ *start, _idx_ *end );
bool _getmask2_( PWSTR mask, _size_t_ cmask, PCWSTR masklist, _idx_ imask, _idx_ *start, _idx_ *end );
_size_t_ _getmaskitems_( PCSTR  mask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, bool bAutoCorrectWildcards = true );
_size_t_ _getmaskitems_( PCWSTR mask, WildmaskItem *pWildmaskItemArray, _size_t_ cWildmaskItemArray, bool bAutoCorrectWildcards = true );
_size_t_ _countmasks_( PCSTR  masklist );
_size_t_ _countmasks_( PCWSTR masklist );

int _sprintf_  ( PSTR  buffer, _size_t_ size, PCSTR  format, ... );
int _sprintf_  ( PWSTR buffer, _size_t_ size, PCWSTR format, ... );
template< int size > int _sprintf_  ( CHAR  (&buffer)[size], PCSTR  format, ... ) { va_list v; va_start( v, format ); int ret = _vsprintf_( buffer, _countof(buffer), format, v ); va_end( v ); return ret; }
template< int size > int _sprintf_  ( WCHAR (&buffer)[size], PCWSTR format, ... ) { va_list v; va_start( v, format ); int ret = _vsprintf_( buffer, _countof(buffer), format, v ); va_end( v ); return ret; }
int _vsprintf_ ( PSTR  buffer, _size_t_ size, PCSTR  format, va_list argptr );
int _vsprintf_ ( PWSTR buffer, _size_t_ size, PCWSTR format, va_list argptr );
template< int size > int _vsprintf_ ( CHAR  (&buffer)[size], PCSTR  format, va_list argptr ) { int ret = _vsprintf_( buffer, _countof(buffer), format, argptr ); return ret; }
template< int size > int _vsprintf_ ( WCHAR (&buffer)[size], PCWSTR format, va_list argptr ) { int ret = _vsprintf_( buffer, _countof(buffer), format, argptr ); return ret; }
#ifndef _WIN32_WCE
int _scprintf_ ( PCSTR  format, ... );
int _scprintf_ ( PCWSTR format, ... );
int _vscprintf_( PCSTR  format, va_list argptr );
int _vscprintf_( PCWSTR format, va_list argptr );
#endif

#ifndef _WIN32_WCE
_size_t_ _strftime_( PSTR  dest, _size_t_ zlen, PCSTR  format, const struct tm *timeptr );
_size_t_ _strftime_( PWSTR dest, _size_t_ zlen, PCWSTR format, const struct tm *timeptr );
#endif

bool _strconv_( PWSTR pstrDest, PCSTR  pstrSrce, _ssize_t_ zlenSrce );
bool _strconv_( PSTR  pstrDest, PCWSTR pstrSrce, _ssize_t_ zlenSrce );
bool _strconv_( PSTR  pstrDest, PCSTR  pstrSrce, _ssize_t_ zlenSrce );
bool _strconv_( PWSTR pstrDest, PCWSTR pstrSrce, _ssize_t_ zlenSrce );

// The following functions are not string dependent. They are implemented
// in here nevertheless because we need them for the StringT<> template to
// prevent the deprecated warnings that VC++ 2005 yields for some CRT functions.

#ifndef _WIN32_WCE
struct tm *_localtime_( struct tm *ptm, const time_t *ptime );
struct tm *_gmtime_	  ( struct tm *ptm, const time_t *ptime );
#endif




// ---------------------------------------------------------------- //
//  StringT template and String class

namespace privateString
{
    template< class Tchar = char >
    class char_T
    {
        public :
            typedef wchar_t otherType;

            enum { isWide = false };

            static Tchar getChar( Tchar ch ) { return ch; }
            static Tchar getChar( otherType ch ) { Tchar ret = (Tchar) 0; _strconv_(&ret, &ch, 1); return ret; }

        // static functions that map to global A and W functions.
            static LRESULT __SendMessage( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
                { return SendMessageA( hwnd, msg, wParam, lParam ); }
            static LRESULT __SendMessageTimeout( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam, UINT fuFlags, UINT uTimeout, PDWORD_PTR pdwResult )
                { return SendMessageTimeoutA( hWnd, msg, wParam, lParam, fuFlags, uTimeout, pdwResult ); }

            static LONG __RegOpenKeyEx( HKEY hKey, const Tchar *pSubKey, DWORD ulOptions,
                                        REGSAM samDesired, PHKEY phkResult )
                { return RegOpenKeyExA( hKey, pSubKey, ulOptions, samDesired, phkResult ); }
            static LONG __RegQueryValueEx( HKEY hKey, const Tchar *pValueName, PDWORD pReserved,
                                        PDWORD pdwType, PBYTE pbyData, PDWORD pcbyData )
                { return RegQueryValueExA( hKey, pValueName, pReserved, pdwType, pbyData, pcbyData ); }

            static int __LoadString( HINSTANCE hInstance, UINT uID, Tchar *pBuffer, int nBufferMax )
                { return LoadStringA( hInstance, uID, pBuffer, nBufferMax ); }
            static int __GetWindowText( HWND hwnd, Tchar *pstrDest, int zlenDest )
                { return GetWindowTextA( hwnd, pstrDest, zlenDest ); }
            static int __GetClassName( HWND hwnd, Tchar *pstrDest, int zlenDest )
                { return GetClassNameA( hwnd, pstrDest, zlenDest ); }

            #ifndef _WIN32_WCE
            static DWORD __GetCurrentDirectory( DWORD zlenDest, Tchar *pstrDest )
                { return GetCurrentDirectoryA( zlenDest, pstrDest ); }
            #endif
            static DWORD __GetModuleFileName( HMODULE hModule, Tchar *pstrDest, DWORD zlenDest )
                { return GetModuleFileNameA( hModule, pstrDest, zlenDest ); }
            static DWORD __GetFullPathName( const Tchar *pstrFileName, DWORD dwBufferLength, Tchar *pstrBuffer, Tchar **ppstrFilePart )
                { return GetFullPathNameA( pstrFileName, dwBufferLength, pstrBuffer, ppstrFilePart ); }
            #ifndef _WIN32_WCE
            static BOOL __GetComputerName( Tchar *pstrDest, PDWORD pzlenDest )
                { return GetComputerNameA( pstrDest, pzlenDest ); }
            #endif
            static BOOL __GetUserName( Tchar *pstrDest, PDWORD pzlenDest )
                #ifndef _WIN32_WCE
                { return GetUserNameA( pstrDest, pzlenDest ); }
                #else
                { return GetUserNameExA( NameWindowsCeLocal, pstrDest, pzlenDest ); }
                #endif
            #ifndef _WIN32_WCE
            static UINT __GetWindowsDirectory( Tchar *pstrDest, UINT zlenDest )
                { return GetWindowsDirectoryA( pstrDest, zlenDest ); }
            static UINT __GetSystemDirectory( Tchar *pstrDest, UINT zlenDest )
                { return GetSystemDirectoryA( pstrDest, zlenDest ); }
            static BOOL __SHGetSpecialFolderPath( HWND hwndOwner, Tchar *pstrDest, int csidl, BOOL fCreate )
                { return SHGetSpecialFolderPathA( hwndOwner, pstrDest, csidl, fCreate ); }
            #endif
            static DWORD __GetTempPath( DWORD zlenDest, Tchar *pstrDest )
                { return GetTempPathA( zlenDest, pstrDest ); }

            static DWORD __FormatMessage( DWORD dwFlags, LPCVOID lpSource, DWORD dwMessageId, DWORD dwLanguageId, Tchar *lpBuffer, DWORD nSize, va_list *Arguments )
                { return FormatMessageA( dwFlags, lpSource, dwMessageId, dwLanguageId, lpBuffer, nSize, Arguments ); }
            static int __GetLocaleInfo( LCID Locale, LCTYPE LCType, Tchar *lpLCData, int cchData )
                { return GetLocaleInfoA( Locale, LCType, lpLCData, cchData ); }
    };

    template<>
    class char_T< wchar_t >
    {
        public :
            typedef wchar_t Tchar;
            typedef char otherType;

            enum { isWide = true };

            static Tchar getChar( Tchar ch ) { return ch; }
            static Tchar getChar( otherType ch ) { Tchar ret = (Tchar) 0; _strconv_(&ret, &ch, 1); return ret; }

        // static functions that map to global A and W functions.
            static LRESULT __SendMessage( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
                { return SendMessageW( hwnd, msg, wParam, lParam ); }
            static LRESULT __SendMessageTimeout( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam, UINT fuFlags, UINT uTimeout, PDWORD_PTR pdwResult )
                #ifndef _WIN32_WCE
                { return SendMessageTimeoutW( hWnd, msg, wParam, lParam, fuFlags, uTimeout, pdwResult ); }
                #else
                { return SendMessageTimeout( hWnd, msg, wParam, lParam, fuFlags, uTimeout, pdwResult ); }
                #endif

            static LONG __RegOpenKeyEx( HKEY hKey, const Tchar *pSubKey, DWORD ulOptions,
                                        REGSAM samDesired, PHKEY phkResult )
                { return RegOpenKeyExW( hKey, pSubKey, ulOptions, samDesired, phkResult ); }
            static LONG __RegQueryValueEx( HKEY hKey, const Tchar *pValueName, PDWORD pReserved,
                                        PDWORD pdwType, PBYTE pbyData, PDWORD pcbyData )
                { return RegQueryValueExW( hKey, pValueName, pReserved, pdwType, pbyData, pcbyData ); }

            static int __LoadString( HINSTANCE hInstance, UINT uID, Tchar *pBuffer, int nBufferMax )
                { return LoadStringW( hInstance, uID, pBuffer, nBufferMax ); }
            static int __GetWindowText( HWND hwnd, Tchar *pstrDest, int zlenDest )
                { return GetWindowTextW( hwnd, pstrDest, zlenDest ); }
            static int __GetClassName( HWND hwnd, Tchar *pstrDest, int zlenDest )
                { return GetClassNameW( hwnd, pstrDest, zlenDest ); }

            #ifndef _WIN32_WCE
            static DWORD __GetCurrentDirectory( DWORD zlenDest, Tchar *pstrDest )
                { return GetCurrentDirectoryW( zlenDest, pstrDest ); }
            #endif
            static DWORD __GetModuleFileName( HMODULE hModule, Tchar *pstrDest, DWORD zlenDest )
                { return GetModuleFileNameW( hModule, pstrDest, zlenDest ); }
            static DWORD __GetFullPathName( const Tchar *pstrFileName, DWORD dwBufferLength, Tchar *pstrBuffer, Tchar **ppstrFilePart )
                { return GetFullPathNameW( pstrFileName, dwBufferLength, pstrBuffer, ppstrFilePart ); }
            #ifndef _WIN32_WCE
            static BOOL __GetComputerName( Tchar *pstrDest, PDWORD pzlenDest )
                { return GetComputerNameW( pstrDest, pzlenDest ); }
            #endif
            static BOOL __GetUserName( Tchar *pstrDest, PDWORD pzlenDest )
                #ifndef _WIN32_WCE
                { return GetUserNameW( pstrDest, pzlenDest ); }
                #else
                { return GetUserNameExW( NameWindowsCeLocal, pstrDest, pzlenDest ); }
                #endif
            #ifndef _WIN32_WCE
            static UINT __GetWindowsDirectory( Tchar *pstrDest, UINT zlenDest )
                { return GetWindowsDirectoryW( pstrDest, zlenDest ); }
            static UINT __GetSystemDirectory( Tchar *pstrDest, UINT zlenDest )
                { return GetSystemDirectoryW( pstrDest, zlenDest ); }
            static BOOL __SHGetSpecialFolderPath( HWND hwndOwner, Tchar *pstrDest, int csidl, BOOL fCreate )
                { return SHGetSpecialFolderPathW( hwndOwner, pstrDest, csidl, fCreate ); }
            #endif
            static DWORD __GetTempPath( DWORD zlenDest, Tchar *pstrDest )
                { return GetTempPathW( zlenDest, pstrDest ); }

            static DWORD __FormatMessage( DWORD dwFlags, LPCVOID lpSource, DWORD dwMessageId, DWORD dwLanguageId, Tchar *lpBuffer, DWORD nSize, va_list *Arguments )
                { return FormatMessageW( dwFlags, lpSource, dwMessageId, dwLanguageId, lpBuffer, nSize, Arguments ); }
            static int __GetLocaleInfo( LCID Locale, LCTYPE LCType, Tchar *lpLCData, int cchData )
                { return GetLocaleInfoW( Locale, LCType, lpLCData, cchData ); }
    };

    template< class T > T tmin( T a, T b ) { return ( a <= b )? a : b; }

    template< class T > class type        { public : static T     round( T     x ) { return (T) floor ( (double) x + 0.5  ); } };
    template<         > class type<float> { public : static float round( float x ) { return     floorf(          x + 0.5f ); } };

    #define _Tstr(str)          ( isWide ? (const Tchar*) L##str : (const Tchar*) str )
    #define _Tchar(ch)          ( isWide ? (Tchar)        L##ch  : (Tchar)        ch  )
    #define _Tstr2(str, Class)  ( Class::isWide ? (const Class::Tchar*) L##str : (const Class::Tchar*) str )
    #define _Tchar2(ch, Class)  ( Class::isWide ? (Class::Tchar)        L##ch  : (Class::Tchar)        ch  )
    // Note: The macros _Tstr/2 and _Tchar/2 are not ideal because they
    // make a runtime if-statement although the type is known already at compile time.
    // But I did not find a better/easier/quicker way to pack the functionality into macros.
    // Let's hope the compiler is smart enough to remove unnecessary code.

    // ---- BaseStringShared as baseString for StringT<> using memory sharing -------------------

    #ifdef __argonMemory_h

    template< class Tchar >
    class BaseStringShared
    {
        friend class BaseStringShared;

        protected :
            typedef typename char_T<Tchar>::otherType TotherChar;

        private :
            MemoryShared m_Mem;

        public :
            BaseStringShared() {  }
            BaseStringShared( const BaseStringShared& src ) { _ShareFrom( src ); }
            BaseStringShared( const Tchar *pstrStr, mem_t memtype )
                { if( memtype == MemT_External ) SetExternalStr( pstrStr ); else _Strcpy( pstrStr, memtype, -1, false ); }
            BaseStringShared( const TotherChar *pstrStr, mem_t memtype )
                { _Strcpy( pstrStr, memtype == MemT_External ? MemT_Default : memtype, -1, false ); }

        // String related methods
            operator const Tchar *() const { static Tchar z = (Tchar) 0; const Tchar *p = (const Tchar*) m_Mem.ConstPtr(); return p ? p : &z; }

            Tchar   *GetBuffer() { return GetBuffer( 0 ); }
            Tchar   *GetBuffer( _size_t_ minLength, bool bClearBuffer = false );
            reterr_t ReleaseBuffer();

        // Memory related methods
            mem_t    GetMemType() const          { return m_Mem.GetMemType(); }
                ///<summary> the memory type this object will use when it has to create completely new memory </summary>
            mem_t    GetMemTypeCreate() const    { return m_Mem.GetMemTypeCreate(); }
            reterr_t SetMemType( mem_t memtype ) { return m_Mem.SetMemType( memtype ); }
            // Note: You can set a memory type, but a memory name like
            // in #MemorySpace# is not supported in here.

            _size_t_ GetBlockSize() const                   { return m_Mem.GetBlockSize(); }
            void     SetBlockSize( _size_t_ blocksize = 1 ) { m_Mem.SetBlockSize( blocksize ); }

            Tchar   *Lock  ( _size_t_ zlenBufferMin );
            reterr_t Unlock();
            Tchar   *ResizeLocked( _size_t_ zlenBufferMin );
            bool     IsLocked() const { return m_Mem.IsReallocLocked(); }

            reterr_t ClearPadding() { return m_Mem.ClearPadding(); }

            reterr_t CopyPropertiesFrom( const BaseStringShared& src )
                { return m_Mem.CopyPropertiesFrom( src.m_Mem ); }

            bool     IsShared() const { return m_Mem.IsShared(); }
            reterr_t Isolate ()       { return m_Mem.Isolate (); } // forces the string object to have its own non-shared memory.

            reterr_t Exchange( BaseStringShared& str ) { return m_Mem.Exchange( str.m_Mem ); }

            ///<summary> Use this method to set a static buffer. That is a buffer
            ///  which will not relocate or free its memory nor change its contents
            ///  as long as <b>this</b> object stores the pointer to that memory.
            ///  Internally, the passed memory will be accessed directly. A copy
            ///  will not be made. Use this if you have a large string and want to
            ///  do some StringT based read-operations on it without allocating another
            ///  memory. </summary>
            ///<remarks> When specifying a resource-string-id, an empty string will be returned. </remarks>
            BaseStringShared& SetExternalStr( const Tchar *pstrStaticStr )
            {
                if( !pstrStaticStr || IS_INTRESOURCE(pstrStaticStr) )
                    { Empty(); return *this; }

                _size_t_ len = _strlen_( pstrStaticStr );
                if( len <= 0 ) { Empty(); return *this; }

                m_Mem.SetExternalMem( pstrStaticStr, (len + 1) * sizeof(Tchar) );

                return *this;
            }

        protected :
            BaseStringShared& operator = ( const Tchar      *string ) { _AttachOrStrcpy( string ); return *this; }
            BaseStringShared& operator = ( const TotherChar *string ) {         _Strcpy( string, -1, false ); return *this; }
            BaseStringShared& operator = ( const Tchar       ch     ) { Tchar      s[2] = { ch, (Tchar)      0 }; _Strcpy( s, 2, false ); return *this; }
            BaseStringShared& operator = ( const TotherChar  ch     ) { TotherChar s[2] = { ch, (TotherChar) 0 }; _Strcpy( s, 2, false ); return *this; }
            BaseStringShared& operator = ( const BaseStringShared& src ) { _ShareFrom( src ); return *this; }

            BaseStringShared& operator +=( const Tchar      *string ) { _AttachOrStrcat( string ); return *this; }
            BaseStringShared& operator +=( const TotherChar *string ) {         _Strcat( string, -1, false ); return *this; }
            BaseStringShared& operator +=( const Tchar       ch     ) { Tchar      s[2] = { ch, (Tchar)      0 }; _Strcat( s, 1, false ); return *this; }
            BaseStringShared& operator +=( const TotherChar  ch     ) { TotherChar s[2] = { ch, (TotherChar) 0 }; _Strcat( s, 1, false ); return *this; }
            BaseStringShared& operator +=( const BaseStringShared& src )
                { if( GetLength() == 0 && !IsLocked() ) _ShareFrom( src ); else _Strcat( (const Tchar*) src, src.GetLength(), src.IsLocked() ); return *this; }

            friend BaseStringShared operator +( const BaseStringShared& str1, const BaseStringShared& str2 ) { BaseStringShared s; return _Concat( s, str1, str2 ); }
            friend BaseStringShared operator +( const BaseStringShared& str1, const Tchar *psz2 )            { BaseStringShared s; return _Concat( s, (const Tchar*) str1, str1.GetLength(), psz2, -1, str1.GetMemType() ); }
            friend BaseStringShared operator +( const Tchar *psz1, const BaseStringShared& str2 )            { BaseStringShared s; return _Concat( s, psz1, -1, (const Tchar*) str2, str2.GetLength(), str2.GetMemType() ); }
            friend BaseStringShared operator +( const BaseStringShared& str1, TotherChar ch2 )               { BaseStringShared s; return _Concat( s, (const Tchar*) str1, str1.GetLength(), (const Tchar*) BaseStringShared(ch2), 1, str1.GetMemType() ); }
            friend BaseStringShared operator +( const BaseStringShared& str1, Tchar ch2 )                    { BaseStringShared s; return _Concat( s, (const Tchar*) str1, str1.GetLength(), &ch2, 1, str1.GetMemType() ); }
            friend BaseStringShared operator +( TotherChar ch1, const BaseStringShared& str2 )               { BaseStringShared s; return _Concat( s, (const Tchar*) BaseStringShared(ch1), 1, (const Tchar*) str2, str2.GetLength(), str2.GetMemType() ); }
            friend BaseStringShared operator +( Tchar ch1, const BaseStringShared& str2 )                    { BaseStringShared s; return _Concat( s, &ch1, 1, (const Tchar*) str2, str2.GetLength(), str2.GetMemType() ); }

            bool     IsEmpty() const { return m_Mem.GetSize() <= (_size_t_) sizeof(Tchar); }
            _size_t_ GetLength() const { return tmax( m_Mem.GetSize() / (_size_t_) sizeof(Tchar) - 1, (_size_t_) 0 ); }

            void Empty() { m_Mem.Free(); }  // will fail if memory is locked

        protected :
            void _ShareFrom( const BaseStringShared& str )
            {
                if( IsLocked() || str.IsLocked() )
                {
                    // Since the source string is locked, we cannot
                    // share its data. We create a copy of its current
                    // string. But beware that the contents may be
                    // inaccurate, since the user may possibly edit
                    // the contents at the moment.
                    // Therefore, we must add the null-terminator by hand.
                    // #_Strcpy()# will do this for us.

                    _Strcpy( (const Tchar*) str.m_Mem.ConstPtr(), str.m_Mem.GetSize() / sizeof(Tchar), true );

                    return;
                }

                m_Mem.ShareFrom( str.m_Mem );
            }

            // Note: The names #StrCpy/Cat/A/W# are macros defined in "shlwapi.h".
            // That's why we use lower 'c' in here for #_Strcpy()# and #_Strcat()#
            // (although their names now start with a '_' char which has changed).

            ///<param name="zlenStr">
            ///  Length of the source string. The terminating null char *must* be included.
            ///  The last char of the speicified length *must* be the null terminator itself. <br/>
            ///  -1 = automatic
            ///</param>
            ///<param name="bForceCheckSourceString">
            ///  If <b>zlenStr</b> is positive value, this param will take effect.
            ///  If TRUE, the source string will be parsed if the specified length
            ///   really is the length of the source string or if the string is
            ///   shorter (a null-terminator in between). <br/>
            ///  If FALSE, it is imagined, that the source string does not have
            ///   any null-terminators in between. At most there may be a null-terminator
            ///   at the end of the specified length (or even no null-terminator
            ///   may be present at all).
            ///</param>
            template< typename srce_char_t >
            void _Strcpy( const srce_char_t *pstrStr, _ssize_t_ zlenStr, bool bForceCheckSourceString )
            { _Strcpy( pstrStr, m_Mem.GetMemTypeCreate(), zlenStr, bForceCheckSourceString ); }

            ///<param name="zlenStr">
            ///  Length of the source string. The terminating null char *must* be included.
            ///  The last char of the speicified length *must* be the null terminator itself. <br/>
            ///  -1 = automatic
            ///</param>
            ///<param name="bForceCheckSourceString">
            ///  If <b>zlenStr</b> is positive, this param will take effect.
            ///  If TRUE, the source string will be parsed if the specified length
            ///   really is the length of the source string or if the string is
            ///   shorter (a null-terminator in between). <br/>
            ///  If FALSE, it is imagined, that the source string does not have
            ///   any null-terminators in between. At most there may be a null-terminator
            ///   at the end of the specified length (or even no null-terminator
            ///   may be present at all).
            ///</param>
            template< typename srce_char_t >
            void _Strcpy( const srce_char_t *pstrStr, mem_t memtype, _ssize_t_ zlenStr, bool bForceCheckSourceString )
            {
                if( !pstrStr || IS_INTRESOURCE(pstrStr) )   // Note: We do not load from string resource in here. Will be done in StringT<>.
                    { Empty(); SetMemType( memtype ); return; }

                if( (_ssize_t_) zlenStr < 0 ) {
                    zlenStr = _strlen_( pstrStr ) + 1;
                } else if( (_ssize_t_) zlenStr > 0 ) {
                    if( bForceCheckSourceString )
                        zlenStr = _strlenmax_( pstrStr, zlenStr - 1, true ) + 1;
                    // Beware: In this case here, the last character in range
                    // #zlenStr# can be a non-null-terminator (so a normal char)
                    // if the actual length is longer than the specified maximum.
                }

                Tchar *ptr = (Tchar*) (PBYTE) m_Mem.WritePtr_Discard( zlenStr * sizeof(Tchar), memtype );
                if( !ptr ) return;

                _strconv_( ptr, pstrStr, zlenStr );

                // Because of the case described above, we must set the
                // null-terminator by hand.
                if( ptr )
                    ptr[zlenStr - 1] = (Tchar) 0;
            }

            void _AttachOrStrcpy( const Tchar *pstrStr )
            {
                if( IS_INTRESOURCE(pstrStr) ) return;   // resource strings are handled in StringT<> separately

                MemDat_Basic basic;
                if( !MemoryShared::IsAttachablePointer( pstrStr, &basic ) )
                    return _Strcpy( pstrStr, -1, false );

                // The string length must fit.
                // We must do this because it could be that the pointer is attachable, but it is
                // not only a memory string but a struct which has a string as first member (the pointer
                // address we have) followed by some possible other data.
                if( (_strlen_(pstrStr) + 1) * (_size_t_) sizeof(Tchar) != basic.m_cMemory )
                    { _Strcpy( pstrStr, -1, false ); return; }

                if( !m_Mem.AttachPointer( pstrStr ) )
                    { _Strcpy( pstrStr, -1, false ); return; }
            }

            ///<param name="lenStr">
            ///  Length of the source string *without* terminating null char. <br/>
            ///  -1 = automatic
            ///</param>
            ///<param name="bForceCheckSourceString">
            ///  If <b>lenStr</b> is positive value, this param will take effect.
            ///  If TRUE, the source string will be parsed if the specified length
            ///   really is the length of the source string or if the string is
            ///   shorter (a null-terminator in between). <br/>
            ///  If FALSE, it is imagined, that the source string does not have
            ///   any null-terminators in between, not even as last char
            ///   in the specified length.
            ///</param>
            template< typename srce_char_t >
            void _Strcat( const srce_char_t *pstrStr, _ssize_t_ lenStr, bool bForceCheckSourceString )
            {
                if( !pstrStr || lenStr == 0 ) return;
                if( IS_INTRESOURCE(pstrStr) ) return;   // resource strings are handled in StringT<> separately

                _size_t_ len = GetLength();

                if( len == 0 )
                    return _Strcpy( pstrStr, (_ssize_t_) lenStr < 0 ? lenStr : lenStr + 1, bForceCheckSourceString );

                if( ( (PBYTE) pstrStr >= m_Mem.ConstPtr() &&
                    (PBYTE) pstrStr < (m_Mem.ConstPtr() + m_Mem.GetSize()) ) ||
                    IS_INTRESOURCE( pstrStr ) )
                {
                    BaseStringShared<srce_char_t> str( pstrStr, m_Mem.GetMemTypeCreate() );
                    _Strcat( (const srce_char_t*) str, str.GetLength(), false );
                    return;
                }

                if( (_ssize_t_) lenStr < 0 ) {
                    lenStr = _strlen_( pstrStr );
                } else if( (_ssize_t_) lenStr > 0 ) {
                    if( bForceCheckSourceString )
                        lenStr = _strlenmax_( pstrStr, lenStr, true );
                }
                if( lenStr == 0 ) return;

                Tchar *ptr = (Tchar*) (PBYTE) m_Mem.WritePtr_Retain( (len + lenStr + 1) * sizeof(Tchar) );
                if( !ptr ) return;

                _strconv_( ptr + len, pstrStr, lenStr );

                // the terminating null char we handle separately
                ptr[len + lenStr] = (Tchar) 0;
            }

            void _AttachOrStrcat( const Tchar *pstrStr )
            {
                if( GetLength() > 0 )
                    return _Strcat( pstrStr, -1, false );

                return _AttachOrStrcpy( pstrStr );
            }

            static BaseStringShared& _Concat( BaseStringShared& strDest, const BaseStringShared& str1, const BaseStringShared& str2 )
            {
                // We take over the memory type if one of the source strings
                // is using a memory type capable of very large allocation sizes.
                mem_t memtype = strDest.GetMemTypeCreate();
                if(      IsVeryLargeMemType( str1.GetMemTypeCreate() ) ) memtype = str1.GetMemTypeCreate();
                else if( IsVeryLargeMemType( str2.GetMemTypeCreate() ) ) memtype = str2.GetMemTypeCreate();

                return _Concat( strDest, (const Tchar*) str1, str1.GetLength(), (const Tchar*) str2, str2.GetLength(), memtype );
            }

            ///<remarks> <b>lenstr1</b> and <b>lenstr2</b> can be -1 to let calculate the
            ///  string lengths automatically. If one is positive, it must be really its
            ///  string length. If a length is a positive value, it is allowed not to include
            ///  any terminating null-char. </remarks>
            template< typename srce_char_t >
            static BaseStringShared& _Concat( BaseStringShared& strDest, const srce_char_t *pstr1, _ssize_t_ lenstr1, const srce_char_t *pstr2, _ssize_t_ lenstr2, mem_t destmemtype )
            {
                if( destmemtype == MemT_External ) destmemtype = MemT_Dont_Change_Dynamic;

                strDest.Empty();
                strDest.SetMemType( destmemtype );

                if( !pstr1 || !pstr2 ) return strDest;

                if( (_ssize_t_) lenstr1 < 0 ) lenstr1 = _strlen_( pstr1 );
                if( (_ssize_t_) lenstr2 < 0 ) lenstr2 = _strlen_( pstr2 );

                _size_t_ len = lenstr1 + lenstr2;

                Tchar *ptr = (Tchar*) (PBYTE) strDest.m_Mem.WritePtr_Discard( (len + 1) * sizeof(Tchar) );
                if( !ptr ) return strDest;

                _strconv_( ptr          , pstr1, lenstr1 );
                _strconv_( ptr + lenstr1, pstr2, lenstr2 );

                // For savety reasons, we set the null-terminator by hand.
                ptr[len] = (Tchar) 0;

                return strDest;
            }
    };

    ///<param name="minLength"> Minimum length for the buffer (*excluding* the null terminator) </param>
    template< class Tchar >
    Tchar *BaseStringShared<Tchar>::GetBuffer( _size_t_ minLength, bool bClearBuffer )
    {
        if( IsLocked() ) return NULL;

        if( bClearBuffer ) m_Mem.Clear();

        return (Tchar*) (PBYTE) m_Mem.WritePtr_RetainMinimum( (minLength + 1) * (_size_t_) sizeof(Tchar) );
    }

    template< class Tchar >
    reterr_t BaseStringShared<Tchar>::ReleaseBuffer()
    {
        // We must check the locked state here by hand because
        // if we would let #m_Mem.SetSize()# handle this, then
        // we would have changed #ptr[idx]# already, which
        // is not ok.
        if( IsLocked() ) return false;

        _idx_ idx = (m_Mem.GetSize() / sizeof(Tchar)) - 1;
        if( idx <= 0 ) { Empty(); return true; }

        _retPByte_t_ ret = m_Mem.WritePtr_Retain();
        if( !ret ) return ret;

        Tchar *ptr = (Tchar*) (PBYTE) ret;
        if( !ptr ) return interr (errFailed);   // should not occur

        if( ptr && idx > 0 ) {
            ptr[idx] = (Tchar) 0;
            _size_t_ len = _strlen_( ptr );
            if( len == 0 ) return m_Mem.Free();
            if( len >= (_size_t_) idx ) return true;
            else ptr[len] = (Tchar) 0;
            return m_Mem.SetSize( (len + 1) * sizeof(Tchar) );
        } else {
            return m_Mem.Free();
        }
    }

    ///<param name="zlenBufferMin"> desired minimum size of the buffer in characters </param>
    ///<returns> Returns NULL if the string is already locked, or on memory allocation error. </returns>
    ///<remarks> Any newly allocated memory will <u>not</u> be cleared in here. </remarks>
    template< class Tchar >
    Tchar *BaseStringShared<Tchar>::Lock( _size_t_ zlenBufferMin )
    {
        if( zlenBufferMin <= 0 ) return NULL;
        if( IsLocked() ) return NULL;

        Tchar *ptr = (Tchar*) (PBYTE) m_Mem.WritePtr_RetainMinimum( zlenBufferMin * (_size_t_) sizeof(Tchar) );
        if( !ptr ) return NULL;

        if( !m_Mem.LockRealloc() )
            return NULL;
        // While locking, the memory could have been reallocated.
        // So we need to get the write pointer again.

        ptr = (Tchar*) (PBYTE) m_Mem.WritePtr();
        if( !ptr )
            { m_dat.m_Mem.UnlockRealloc(); return NULL; }

        return ptr;
    }

    ///<remarks> <b>this</b> string object can be successfully
    ///  deleted even if it is locked. </remarks>
    template< class Tchar >
    reterr_t BaseStringShared<Tchar>::Unlock()
    {
        if( !IsLocked() ) return interr (errFailed);

        m_Mem.UnlockRealloc();

        if( !IsLocked() )
            return ReleaseBuffer();

        return true;
    }

    ///<summary> This works only if you previously have called <b>Lock()</b> already.
    ///  Using this method, you can resize the desired minimum buffer size. <br/>
    ///  As you know, for each <b>Lock()</b> you will have to call <b>Unlock()</b>,
    ///  but this is <u>not</u> meant for calls to <b>ResizeLocked()</b>. </summary>
    template< class Tchar >
    Tchar *BaseStringShared<Tchar>::ResizeLocked( _size_t_ zlenBufferMin )
    {
        if( zlenBufferMin <= 0 ) return NULL;
        if( !IsLocked() ) return NULL;

        return (Tchar*) (PBYTE) m_Mem.WritePtr_RetainMinimum( zlenBufferMin * (_size_t_) sizeof(Tchar) );
    }

    #endif
}

// Flags for #StringT<>::PlaceholderCharsToString()#, #StringT<>::PlaceholderCharsToNumber()# and #StringT<>::OverwriteDecNumber()#.
enum strplaceholder_t {
    // number type

    StrNumT_Decimal                     = 0x00000000,
    StrNumT_Hexadecimal,
    StrNumT_Octal,

    StrNumT__Mask                       = 0x0000000f,

    // flags

    StrDigNumF_PadWithZeros             = 0x00000010,

    StrPlaceholderF_LeftAlign           = 0x00000000,
    StrPlaceholderF_RightAlign          = 0x00010000,
    StrPlaceholderF_SkipExceedingChars  = 0x00000020,   // If the string to insert needs more characters than can be
                                                        // replaced in the StringT<> string, then this flag will cause
                                                        // the upper or lower chars to be skipped. Otherwise they
                                                        // will be pre- or appended at the last char replaced.

    StrDigNumF__Default                 = ( StrNumT_Decimal | StrPlaceholderF_RightAlign | StrDigNumF_PadWithZeros ),

    _pad_strplaceholder_t = 0x7fffffff
};

#define strplaceholder_v DWORD

// Flags for #StringT<>::Findnextword()# and #StringT<>::Findword()#.
enum findword_t {
    findword_SeparatorChars             = 0x00000000,   // The specified chars for param #pstrChars# in #StringT<>::Findnextword()#
                                                        // are all separator chars between words.
    findword_WordChars                  = 0x00000001,   // The specified chars for param #pstrChars# in #StringT<>::Findnextword()#
                                                        // are all the chars that make up a word.
                                                        // All other chars are interpreted as separators.

    findword_inclSpaces                 = 0x00000002,   // If spaces shall be automatically included for param #pstrChars# in #StringT<>::Findnextword()#.
                                                        // If the spaces are separators or word-chars depends on
                                                        // #findword_SeparatorChars# and #findword_WordChars#.

    _pad_findword_t = 0x7fffffff
};

template< class baseString, typename Tchar >
class StringT : public baseString
{
    protected :
        class Proxy
        {
            private :
                StringT<baseString, Tchar> *m_str;
                _idx_                       m_idx;

            public :
                Proxy( StringT<baseString, Tchar>& s, _idx_ i ) : m_str(&s), m_idx(i) {  }
                //Proxy( const Proxy& rhs ) : m_str(rhs.m_str), m_idx(rhs.m_idx) {  }
                //  Note: We let the compiler do the copy-ctor and operator= automatically.

                // l-value uses (manipulate string if necessary)
                //Proxy& operator =( const Proxy& rhs ) { m_str = rhs.m_str; m_idx = rhs.m_idx; return *this; }
                //  Note: We let the compiler do the copy-ctor and operator= automatically.
                Proxy& operator =( Tchar ch ) { Tchar *p = m_str->WritePtrAt(m_idx); if( p ) { *p = ch; m_str->ReleaseBuffer(); } return *this; }

                // r-value uses (string does not change) (automatic conversion to char)
                //operator Tchar() const { const Tchar *p = m_str->ReadPtrAt(m_idx); return p ? *p : (Tchar) 0; }
                operator const Tchar& () const
                    { static Tchar c = (Tchar) 0; c = (Tchar) 0; const Tchar *p = m_str->ReadPtrAt(m_idx); return p ? *p : c; }
                    // Note: We use the Tchar&() operator instead of the Tchar() operator so that
                    //  the user can take the address from the returned Tchar and will receive
                    //  the index into the string memory. The idea is that the user now can
                    //  increment or decrement the address to get other characters.
                    //  This is to reflect the normal zero-terminated string handling as close as possible.

                const Tchar *operator &() const { return m_str->ReadPtrAt(m_idx); }
                    // Note: The operator &() is the address-of operator. It will be used if
                    //  you call e.g. &str[0] as r-value when str is a non-const StringT<> object.
                    //  Because when str is non-const, the Proxy will be used and the
                    //  operator &() will be used then.

                bool operator ==( const Proxy& rhs ) { const Tchar *p1 = m_str->ReadPtrAt(m_idx); const Tchar *p2 = rhs.m_str->ReadPtrAt(rhs.m_idx); return p1 && p2 ? *p1 == *p2 : false; }
                bool operator ==( Tchar ch ) { const Tchar *p = m_str->ReadPtrAt(m_idx); return p ? *p == ch : false; }

                bool operator !=( const Proxy& rhs ) { const Tchar *p1 = m_str->ReadPtrAt(m_idx); const Tchar *p2 = rhs.m_str->ReadPtrAt(rhs.m_idx); return p1 && p2 ? *p1 != *p2 : true; }
                bool operator !=( Tchar ch ) { const Tchar *p = m_str->ReadPtrAt(m_idx); return p ? *p != ch : true; }
        };
        // Note: We need the #Proxy# for the #StringT::operator[]# operator
        // to allow correct l-value and r-value handling at compile time.
        // Idea taken from:
        //  http://www.mpi-sb.mpg.de/~kettner/courses/lib_design_03/notes/advanced.html#Handle
        // But I did extend it. (the #operator &()# idea is mine as well as
        // using #operator Tchar& ()# with Tchar& instead of Tchar).

    public :
        typedef typename Tchar Tchar;
        typedef typename privateString::char_T<Tchar>::otherType TotherChar;
        typedef typename Tchar *Ptstr;
        typedef const typename Tchar *Pctstr;
        enum { isWide = privateString::char_T<Tchar>::isWide };

    // Initialization
        StringT() {  }
        StringT( const StringT& src ) : baseString( reinterpret_cast<const baseString&>(src) ) {  }
        StringT( const baseString& src ) : baseString( src ) {  }
        StringT( const Tchar *pstring_or_resid ) : baseString() { *this = pstring_or_resid; }
        StringT( const TotherChar *pstring_or_resid ) : baseString() { *this = pstring_or_resid; }
        StringT( Tchar ch, _size_t_ count = 1 ) : baseString() { Floodfill( ch, count ); }
        StringT( TotherChar ch, _size_t_ count = 1 ) : baseString() { Floodfill( ch, count ); }
        StringT( HINSTANCE hInstance, UINT resid ) { Loadstring( hInstance, resid ); }
        StringT( HINSTANCE hInstance, UINT resid, LANGID language ) { Loadstring( hInstance, resid, language ); }

    // String access
        _size_t_ length() const { __if_exists( baseString::GetLength ) { return (_size_t_) baseString::GetLength(); } __if_not_exists( baseString::GetLength ) { return _strlen_(*this); } }
        _size_t_ lengthInBytes() const { return length() * sizeof(Tchar); }
        _size_t_ sizeInBytes() const { return (length() + 1) * sizeof(Tchar); }

        void   Empty() { baseString::Empty(); }
        bool   IsEmpty() const { return baseString::IsEmpty(); }

            ///<summary> <b>Isolate()</b> will ensure that this string will not
            ///  have shared its memory with other string objects. </summary>
        void   Isolate() { __if_exists( baseString::Isolate ) { baseString::Isolate(); } __if_not_exists( baseString::Isolate ) { Lockbuffer(); Unlockbuffer(); } }

    // Character access
            // Note: We return a Tchar& from operator[]const instead of a Tchar so that
            //  the user can take the address from the returned Tchar and will receive
            //  the index into the string memory. The idea is that the user now can
            //  increment or decrement the address to get other characters.
            //  This is to reflect the normal zero-terminated string handling as close as possible.
        const Tchar& operator[]( _idx_ idx ) const { static Tchar c = (Tchar) 0; c = (Tchar) 0; const Tchar *p = ReadPtrAt(idx); return p ? *p : c; }
        Proxy        operator[]( _idx_ idx )       { return Proxy( *this, idx ); }
        Tchar GetAt( _idx_ idx ) const     { const Tchar *p = ReadPtrAt(idx); return p ? *p : (Tchar) 0; }
        bool  SetAt( _idx_ idx, Tchar ch ) { Tchar *p = WritePtrAt(idx); if( p ) { *p = ch; ReleaseBuffer(); return true; } return false; }

    // Assignment and operations
        StringT& operator =( const StringT& src )           { baseString::operator =( reinterpret_cast<const baseString&>(src) ); return *this; }
        StringT& operator =( const baseString& src )        { baseString::operator =( src ); return *this; }
        StringT& operator =( const Tchar *pstring )         { if( pstring && IS_INTRESOURCE(pstring) ) Loadstring((UINT) reinterpret_cast<DWORD_PTR>(pstring)); else baseString::operator =( pstring ); return *this; }
        StringT& operator =( const TotherChar *pstring )    { if( pstring && IS_INTRESOURCE(pstring) ) Loadstring((UINT) reinterpret_cast<DWORD_PTR>(pstring)); else baseString::operator =( pstring ); return *this; }
        StringT& operator =( Tchar ch )                     { Tchar      s[2] = { ch, (Tchar) 0 }; baseString::operator =( s ); return *this; }
        StringT& operator =( TotherChar ch )                { TotherChar s[2] = { ch, (Tchar) 0 }; baseString::operator =( s ); return *this; }

        StringT& operator +=( const StringT& src )          { baseString::operator +=( src ); return *this; }
        StringT& operator +=( const baseString& src )       { baseString::operator +=( src ); return *this; }
        StringT& operator +=( const Tchar *pstring )        { if( pstring && IS_INTRESOURCE(pstring) ) Loadappendstring((UINT) reinterpret_cast<DWORD_PTR>(pstring)); else baseString::operator +=( pstring ); return *this; }
        StringT& operator +=( const TotherChar *pstring )   { if( pstring && IS_INTRESOURCE(pstring) ) Loadappendstring((UINT) reinterpret_cast<DWORD_PTR>(pstring)); else baseString::operator +=( pstring ); return *this; }
        StringT& operator +=( Tchar ch )                    { Tchar      s[2] = { ch, (Tchar) 0 }; baseString::operator +=( s ); return *this; }
        StringT& operator +=( TotherChar ch )               { TotherChar s[2] = { ch, (Tchar) 0 }; baseString::operator +=( s ); return *this; }

        StringT& operator |=( const StringT& src )          { operator |=( (const Tchar*) src ); return *this; }
        StringT& operator |=( const baseString& src )       { operator |=( (const Tchar*) src ); return *this; }
        StringT& operator |=( const Tchar *pstring )        { Tchar s[_MAX_PATH_ + 1]; if( pstring && IS_INTRESOURCE(pstring) ) _joinpaths_(s, *this, StringT(pstring), _countof(s)); else _joinpaths_(s, *this, pstring, _countof(s)); *this = s; return *this; }
        StringT& operator |=( const TotherChar *pstring )   { StringT str(pstring); operator |=( (const Tchar*) str ); return *this; }
        StringT& operator |=( Tchar ch )                    { if( _ispathsep_(ch) ) AsPath(); else operator +=( ch ); return *this; }
        StringT& operator |=( TotherChar ch )               { if( _ispathsep_(ch) ) AsPath(); else operator +=( ch ); return *this; }

        friend StringT operator +( const StringT& str1, const StringT& str2 )     { return reinterpret_cast<const baseString&>(str1) + reinterpret_cast<const baseString&>(str2); }
        friend StringT operator +( const StringT& str1, const baseString& str2 )  { return reinterpret_cast<const baseString&>(str1) + str2; }
        friend StringT operator +( const baseString& str1, const StringT& str2 )  { return str1 + reinterpret_cast<const baseString&>(str2); }
        friend StringT operator +( const StringT& str1, const Tchar *psz2 )       { return ( psz2 && IS_INTRESOURCE(psz2) )? str1 + StringT(psz2) : reinterpret_cast<const baseString&>(str1) + psz2; }
        friend StringT operator +( const Tchar *psz1, const StringT& str2 )       { return ( psz1 && IS_INTRESOURCE(psz1) )? StringT(psz1) + str2 : psz1 + reinterpret_cast<const baseString&>(str2); }
        friend StringT operator +( const StringT& str1, wchar_t ch2 )             { return reinterpret_cast<const baseString&>(str1) + ch2; }
        friend StringT operator +( const StringT& str1, char ch2 )                { return reinterpret_cast<const baseString&>(str1) + ch2; }
        friend StringT operator +( wchar_t ch1, const StringT& str2 )             { return ch1 + reinterpret_cast<const baseString&>(str2); }
        friend StringT operator +( char ch1, const StringT& str2 )                { return ch1 + reinterpret_cast<const baseString&>(str2); }

        friend StringT operator |( const StringT& str1, const StringT& str2 )     { Tchar s[_MAX_PATH_ + 1]; _joinpaths_( s, str1, str2, _countof(s) ); return s; }
        friend StringT operator |( const StringT& str1, const baseString& str2 )  { Tchar s[_MAX_PATH_ + 1]; _joinpaths_( s, str1, str2, _countof(s) ); return s; }
        friend StringT operator |( const baseString& str1, const StringT& str2 )  { Tchar s[_MAX_PATH_ + 1]; _joinpaths_( s, str1, str2, _countof(s) ); return s; }
        friend StringT operator |( const StringT& str1, const Tchar *psz2 )       { Tchar s[_MAX_PATH_ + 1]; if( psz2 && IS_INTRESOURCE(psz2) ) _joinpaths_( s, str1, StringT(psz2), _countof(s) ); else _joinpaths_( s, str1, psz2, _countof(s) ); return s; }
        friend StringT operator |( const Tchar *psz1, const StringT& str2 )       { Tchar s[_MAX_PATH_ + 1]; if( psz1 && IS_INTRESOURCE(psz1) ) _joinpaths_( s, StringT(psz1), str2, _countof(s) ); else _joinpaths_( s, psz1, str2, _countof(s) ); return s; }
        friend StringT operator |( const StringT& str1, wchar_t ch2 )             { if( _ispathsep_(ch2) ) { StringT s = str1; s.AsPath(); return s; } return str1 + ch2; }
        friend StringT operator |( const StringT& str1, char ch2 )                { if( _ispathsep_(ch2) ) { StringT s = str1; s.AsPath(); return s; } return str1 + ch2; }

    // Comparison
        friend bool operator ==( const StringT& str1, const StringT& str2 )     { return _stris_( str1, str2 ); }
        friend bool operator ==( const StringT& str1, const baseString& str2 )  { return _stris_( str1, str2 ); }
        friend bool operator ==( const baseString& str1, const StringT& str2 )  { return _stris_( str1, str2 ); }
        friend bool operator ==( const StringT& str1, const Tchar *psz2 )       { return _stris_( str1, psz2 ); }
        friend bool operator ==( const Tchar *psz1, const StringT& str2 )       { return _stris_( psz1, str2 ); }
        friend bool operator ==( Tchar ch1, const StringT& str2 )               { Tchar s1[2] = { ch1, (Tchar) 0 }; return _stris_( s1, str2 ); }
        friend bool operator ==( const StringT& str1, Tchar ch2 )               { Tchar s2[2] = { ch2, (Tchar) 0 }; return _stris_( str1, s2 ); }

        friend bool operator !=( const StringT& str1, const StringT& str2 )     { return !_stris_( str1, str2 ); }
        friend bool operator !=( const StringT& str1, const baseString& str2 )  { return !_stris_( str1, str2 ); }
        friend bool operator !=( const baseString& str1, const StringT& str2 )  { return !_stris_( str1, str2 ); }
        friend bool operator !=( const StringT& str1, const Tchar *psz2 )       { return !_stris_( str1, psz2 ); }
        friend bool operator !=( const Tchar *psz1, const StringT& str2 )       { return !_stris_( psz1, str2 ); }
        friend bool operator !=( Tchar ch1, const StringT& str2 )               { Tchar s1[2] = { ch1, (Tchar) 0 }; return !_stris_( s1, str2 ); }
        friend bool operator !=( const StringT& str1, Tchar ch2 )               { Tchar s2[2] = { ch2, (Tchar) 0 }; return !_stris_( str1, s2 ); }

        friend bool operator <( const StringT& str1, const StringT& str2 )      { return _strcmp_( str1, str2 ) < 0; }
        friend bool operator <( const StringT& str1, const baseString& str2 )   { return _strcmp_( str1, str2 ) < 0; }
        friend bool operator <( const baseString& str1, const StringT& str2 )   { return _strcmp_( str1, str2 ) < 0; }
        friend bool operator <( const StringT& str1, const Tchar *psz2 )        { return _strcmp_( str1, psz2 ) < 0; }
        friend bool operator <( const Tchar *psz1, const StringT& str2 )        { return _strcmp_( psz1, str2 ) < 0; }

        friend bool operator >( const StringT& str1, const StringT& str2 )      { return _strcmp_( str1, str2 ) > 0; }
        friend bool operator >( const StringT& str1, const baseString& str2 )   { return _strcmp_( str1, str2 ) > 0; }
        friend bool operator >( const baseString& str1, const StringT& str2 )   { return _strcmp_( str1, str2 ) > 0; }
        friend bool operator >( const StringT& str1, const Tchar *psz2 )        { return _strcmp_( str1, psz2 ) > 0; }
        friend bool operator >( const Tchar *psz1, const StringT& str2 )        { return _strcmp_( psz1, str2 ) > 0; }

        friend bool operator <=( const StringT& str1, const StringT& str2 )     { return _strcmp_( str1, str2 ) <= 0; }
        friend bool operator <=( const StringT& str1, const baseString& str2 )  { return _strcmp_( str1, str2 ) <= 0; }
        friend bool operator <=( const baseString& str1, const StringT& str2 )  { return _strcmp_( str1, str2 ) <= 0; }
        friend bool operator <=( const StringT& str1, const Tchar *psz2 )       { return _strcmp_( str1, psz2 ) <= 0; }
        friend bool operator <=( const Tchar *psz1, const StringT& str2 )       { return _strcmp_( psz1, str2 ) <= 0; }

        friend bool operator >=( const StringT& str1, const StringT& str2 )     { return _strcmp_( str1, str2 ) >= 0; }
        friend bool operator >=( const StringT& str1, const baseString& str2 )  { return _strcmp_( str1, str2 ) >= 0; }
        friend bool operator >=( const baseString& str1, const StringT& str2 )  { return _strcmp_( str1, str2 ) >= 0; }
        friend bool operator >=( const StringT& str1, const Tchar *psz2 )       { return _strcmp_( str1, psz2 ) >= 0; }
        friend bool operator >=( const Tchar *psz1, const StringT& str2 )       { return _strcmp_( psz1, str2 ) >= 0; }

        int  Strcmp    ( const Tchar *pstr2 ) const { return _strcmp_ ( *this, pstr2 ); }
        int  Stricmp   ( const Tchar *pstr2 ) const { return _stricmp_( *this, pstr2 ); }
        bool Stris     ( const Tchar *pstr2 ) const { return _stris_  ( *this, pstr2 ); }
        bool Striis    ( const Tchar *pstr2 ) const { return _striis_ ( *this, pstr2 ); }
        bool Strishere ( const Tchar *pstr2 ) const { return _strishere_ ( *this, pstr2 ); }
        bool Striishere( const Tchar *pstr2 ) const { return _striishere_( *this, pstr2 ); }
        bool Strsepis  ( const Tchar *pstr2, const Tchar *pstr1_Separators ) const { return _strsepis_ ( *this, pstr2, pstr1_Separators ); }
        bool Strsepiis ( const Tchar *pstr2, const Tchar *pstr1_Separators ) const { return _strsepiis_( *this, pstr2, pstr1_Separators ); }

        bool StartsWith ( const Tchar *pstr2 ) const { return _strishere_   ( *this, pstr2 ); }
        bool StartsWithi( const Tchar *pstr2 ) const { return _striishere_  ( *this, pstr2 ); }
        bool EndsWith   ( const Tchar *pstr2 ) const { return _strendswith_ ( *this, pstr2 ); }
        bool EndsWithi  ( const Tchar *pstr2 ) const { return _striendswith_( *this, pstr2 ); }

    // String manipulation
        StringT Left ( _size_t_ count ) const { return Mid( 0, count ); }
        StringT Right( _size_t_ count ) const { return Mid( length() - count, count ); }
        StringT Mid  ( _idx_ first, _size_t_ count ) const;
            ///<summary> <b>Inflate()</b> takes an inner part of the string by cutting
            ///  <b>fromleft</b> chars from the left side and <b>fromright_neg</b> chars
            ///  from the right side. <b>fromright_neg</b> has to be a negative value. </summary>
        StringT Inflate( _idx_ fromleft, _idx_ fromright_neg ) const { return Mid( fromleft, length() + fromright_neg - fromleft ); }

        StringT& Floodfill( Tchar      ch, _size_t_ count = 1 );
        StringT& Floodfill( TotherChar ch, _size_t_ count = 1 ) { return Floodfill( privateString::char_T<Tchar>::getChar(ch), count ); }

        StringT& Strcpy_unaligned( const Tchar UNALIGNED *pstrString );
        StringT& Strcpy_unaligned( const Tchar UNALIGNED *pstrString, _size_t_ maxChars );

        StringT& Strcat( const Tchar      *pstrString ) { operator +=( pstrString ); return *this; }
        StringT& Strcat( const TotherChar *pstrString ) { operator +=( pstrString ); return *this; }
        StringT& Strcat( const Tchar      *pstrExternal, _size_t_ count );  // appends a string of maximum #count# chars excl. null-terminator
        StringT& Strcat( const TotherChar *pstrExternal, _size_t_ count );  // appends a string of maximum #count# chars excl. null-terminator

        StringT& Sprintf ( const Tchar *pstrFormat, ... ) { va_list v; va_start(v, pstrFormat); Vsprintf(pstrFormat, v); va_end(v); return *this; }
        StringT& Sprintf ( UINT         idsFormat,  ... ) { va_list v; va_start(v, idsFormat ); Vsprintf(idsFormat,  v); va_end(v); return *this; }
        StringT& Vsprintf( const Tchar *pstrFormat, va_list argptr );
        StringT& Vsprintf( UINT         idsFormat,  va_list argptr ) { return Vsprintf( (const Tchar*) (ULONG_PTR) MAKEINTRESOURCE(idsFormat), argptr ); }
        StringT& UnSformat() { return DuplicateChar( _Tchar('%') ); }

        StringT& DuplicateChar( Tchar chDup );

        StringT& AccomplishSegments( const Tchar *pstrSegmentStart = _Tstr("{"),
                                     const Tchar *pstrSegmentEnd   = _Tstr("}") );

        StringT EscToStr( Tchar chEndOfStr = (Tchar) 0 );
        StringT EscToStr( TotherChar chEndOfStr ) { return EscToStr( privateString::char_T<Tchar>::getChar(chEndOfStr) ); }
        StringT StrToEsc();

        Tchar    *toStrlst  ( Tchar chSep = _Tchar('|') );
        StringT&  fromStrlst( const Tchar *pstrStrlst, Tchar chSep = _Tchar('|') );
        _size_t_  fromOpenfilenameStrlst( const Tchar *pstrStrlst, DWORD dwOfnFlags, Tchar chSep = _Tchar('|') );

        StringT& Insert ( _idx_ idx, Tchar ch, _size_t_ count = 1 );
        StringT& Insert ( _idx_ idx, const Tchar *pstr );
        StringT& Delete ( _idx_ idx, _size_t_ count = 1 );
        StringT& Remove ( Tchar ch ) { Tchar s[2] = { ch, 0 }; return Remove( s ); }
        StringT& Remove ( const Tchar *pstrCharset );
        StringT& Removei( Tchar ch ) { Tchar s[2] = { ch, 0 }; return Removei( s ); }
        StringT& Removei( const Tchar *pstrCharset );
        StringT& RemoveSpaces();
        StringT& Remain ( Tchar ch ) { Tchar s[2] = { ch, 0 }; return Remain( s ); }
        StringT& Remain ( const Tchar *pstrCharset );
        StringT& Remaini( Tchar ch ) { Tchar s[2] = { ch, 0 }; return Remaini( s ); }
        StringT& Remaini( const Tchar *pstrCharset );

        StringT& Trim      ();
        StringT& Trim      ( const Tchar *pstrCharsToTrim );
        StringT& Trimi     ( const Tchar *pstrCharsToTrim );
        StringT& TrimLeft  ();
        StringT& TrimLeft  ( const Tchar *pstrCharsToTrim );
        StringT& TrimLefti ( const Tchar *pstrCharsToTrim );
        StringT& TrimRight ();
        StringT& TrimRight ( const Tchar *pstrCharsToTrim );
        StringT& TrimRighti( const Tchar *pstrCharsToTrim );
        StringT& TrimStart () { TrimLeft(); return *this; }
        StringT& TrimStart ( const Tchar *pstrCharsToTrim ) { TrimLeft ( pstrCharsToTrim ); return *this; }
        StringT& TrimStarti( const Tchar *pstrCharsToTrim ) { TrimLefti( pstrCharsToTrim ); return *this; }
        StringT& TrimEnd   () { TrimRight(); return *this; }
        StringT& TrimEnd   ( const Tchar *pstrCharsToTrim ) { TrimRight ( pstrCharsToTrim ); return *this; }
        StringT& TrimEndi  ( const Tchar *pstrCharsToTrim ) { TrimRighti( pstrCharsToTrim ); return *this; }

        StringT& Uppercase();
        StringT& Lowercase();
        StringT& UppercaseFirstChar();
        StringT& LowercaseFirstChar();
        StringT& Reverse();

        _idx_ Strchr (            Tchar ch                 ) const { return Strchr( 0, ch ); }
        _idx_ Strchr ( _idx_ idx, Tchar ch                 ) const;
        _idx_ Strichr(            Tchar ch                 ) const { return Strichr( 0, ch ); }
        _idx_ Strichr( _idx_ idx, Tchar ch                 ) const;
        _idx_ Strrchr(            Tchar ch                 ) const;
        _idx_ Strstr (            const Tchar *pstrSubstr  ) const { return Strstr( 0, pstrSubstr ); }
        _idx_ Strstr ( _idx_ idx, const Tchar *pstrSubstr  ) const;
        _idx_ Stristr(            const Tchar *pstrSubstr  ) const { return Stristr( 0, pstrSubstr ); }
        _idx_ Stristr( _idx_ idx, const Tchar *pstrSubstr  ) const;
        _idx_ Strpbrk(            const Tchar *pstrCharset ) const { return Strpbrk( 0, pstrCharset ); }
        _idx_ Strpbrk( _idx_ idx, const Tchar *pstrCharset ) const;

        bool  Matchmask         ( const Tchar *pstrMask     ) const { return _matchmask_         ( *this, pstrMask     ); }
        bool  Matchimask        ( const Tchar *pstrMask     ) const { return _matchimask_        ( *this, pstrMask     ); }
        bool  Matchpathmask     ( const Tchar *pstrMask     ) const { return _matchpathmask_     ( *this, pstrMask     ); }
        bool  Matchmasks        ( const Tchar *pstrMasklist ) const { return _matchmasks_        ( *this, pstrMasklist ); }
        bool  Matchimasks       ( const Tchar *pstrMasklist ) const { return _matchimasks_       ( *this, pstrMasklist ); }
        bool  Matchpathmasks    ( const Tchar *pstrMasklist ) const { return _matchpathmasks_    ( *this, pstrMasklist ); }
        _idx_ Findmatchmasks    ( const Tchar *pstrMasklist ) const { return _findmatchmasks_    ( *this, pstrMasklist ); } // returns mask index, -1 = not found
        _idx_ Findmatchimasks   ( const Tchar *pstrMasklist ) const { return _findmatchimasks_   ( *this, pstrMasklist ); } // returns mask index, -1 = not found
        _idx_ Findmatchpathmasks( const Tchar *pstrMasklist ) const { return _findmatchpathmasks_( *this, pstrMasklist ); } // returns mask index, -1 = not found

        StringT& Replace          ( Tchar        chFind,          Tchar        chReplace          );
        StringT& Replace          ( const Tchar *pstrFindCharset, Tchar        chReplace          );
        StringT& Replace          ( const Tchar *pstrFindCharset, const Tchar *pstrReplaceCharset );
        StringT& Replacei         ( Tchar        chFind,          Tchar        chReplace          );
        StringT& Replacei         ( const Tchar *pstrFindCharset, Tchar        chReplace          );
        StringT& Replacei         ( const Tchar *pstrFindCharset, const Tchar *pstrReplaceCharset );
        StringT& Replacestring    ( const Tchar *pstrFindString,  const Tchar *pstrReplaceString  ) { ReplacestringRet ( pstrFindString, pstrReplaceString ); return *this; }
        StringT& Replaceistring   ( const Tchar *pstrFindString,  const Tchar *pstrReplaceString  ) { ReplaceistringRet( pstrFindString, pstrReplaceString ); return *this; }
            ///<returns> number of replaces done </returns>
        _size_t_ ReplacestringRet ( const Tchar *pstrFindString,  const Tchar *pstrReplaceString  );
        _size_t_ ReplaceistringRet( const Tchar *pstrFindString,  const Tchar *pstrReplaceString  );

        StringT& PlaceholderCharsToString( const Tchar *pstrPlaceholderChars, const Tchar *pstrString, strplaceholder_v flags = 0                   );
        StringT& PlaceholderCharsToNumber( const Tchar *pstrPlaceholderChars, DWORD dwNum,             strplaceholder_v flags = StrDigNumF__Default );
            ///<summary> Changes all decimal number digits to the new value <b>dwNum</b>. <br/>
            ///  E.g. useful for numbering a filename. </summary>
        StringT& OverwriteDecNumber( DWORD dwNum, strplaceholder_v flags = StrDigNumF__Default ) { return PlaceholderCharsToNumber( _Tstr("0123456789"), dwNum, flags ); }

            ///<summary> <b>Findword()</b> and <b>Findnextword()</b> will find a word like usually used in
            ///  Find dialogs: for this, <b>nWord</b> is interpreted in a special way.
            ///  See description to <b>Findnextword()</b>. </summary>
        _idx_   Findword    ( _idx_& idxStart, _idx_& idxEnd, _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const { return Findnextword( idxStart, idxEnd, 0, nWord, flags, pstrChars ); }
        _idx_   Findnextword( _idx_& idxStart, _idx_& idxEnd, _idx_ idxStartParsing, _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const;
        _idx_   Findword    ( StringT& strFound, _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const { return Findnextword( strFound, 0, nWord, flags, pstrChars ); }
        _idx_   Findnextword( StringT& strFound, _idx_ idxStartParsing, _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const
            { _idx_ iS, iE; _idx_ i = Findnextword( iS, iE, idxStartParsing, nWord, flags, pstrChars ); strFound = Mid(iS, iE - iS); return i; }
        StringT Findword    ( _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const { StringT s; Findnextword( s, 0, nWord, flags, pstrChars ); return s; }
        StringT Findnextword( _idx_ idxStartParsing, _idx_ nWord, DWORD flags = findword_inclSpaces, const Tchar *pstrChars = NULL ) const { StringT s; Findnextword( s, idxStartParsing, nWord, flags, pstrChars ); return s; }
        static bool isFindwordChar( Tchar ch, DWORD flags, const Tchar *pstrChars = NULL );

        StringT& MakeCRLF();
        StringT& MakeLF();
        StringT& MakeCR();

        _size_t_ GetNumRows();
        _idx_    GetRowindexFromChar( _idx_ idxChar );
        _idx_    GetCharindexFromRow( _idx_ idxRow );
        _idx_    GetCharindexAtStartofrow( _idx_ idxChar );
        _idx_    GetCharindexAfterEndofrow( _idx_ idxChar );
        _idx_    GetStartcharindexOffsetRow( _idx_ idxChar, int nRowOffset ) { return GetStartcharindexOffsetRow( idxChar, nRowOffset, length() ); }

        StringT& Indent  ( _size_t_ cColumns, _size_t_ cTab, bool bAllowTabs, _idx_ iStartChar = 0, _idx_ iEndChar = -1 );
        StringT& Unindent( _size_t_ cColumns, _size_t_ cTab, bool bAllowTabs, _idx_ iStartChar = 0, _idx_ iEndChar = -1 );

        #define BeautyF_TrimComma           0x00000001  // if after the comma, there are only "0" digits, they *and* the comma will be removed
        #define BeautyF_AutoAddCommaZero    0x00000002  // if the value has no comma (so is an integer value), a ".0" will be appended
        #define BeautyF_CommaIsDot          0x00000004  // set this if the float value has/shall have a "." as comma char. Otherwise the current language's comma char will be used.
        #define BeautyF_2DigitsBeforeComma  0x00000008  // the value should have minimum 2 digits before the comma. If this is not the case, leading zeros will be inserted.
        StringT& BeautifyFloatValue( DWORD flags = BeautyF_AutoAddCommaZero );
        StringT& BeautifyVersionNr( int nMinDisplayNums = 2, int nMaxDisplayNums = 3, int nVersionNrNums = 4, const Tchar *pstrNewSep = _Tstr(".") );
        StringT  GetBuildNr() const { return Findnextword( length(), -1, findword_WordChars, _Tstr("0123456789") ); }

        bool  isAbsolutePath() const { return _isabsolutepath_( *this ); }
        StringT& ParentPath();
        StringT& RootPath();
        StringT& StripPath();
        StringT& AsPath();
        StringT  DriveOf    () const { Tchar s[_MAX_PATH_ + 1]; return _driveof_(s, *this, _countof(s)) ? s : (const Tchar*) NULL; }
        StringT  DirectoryOf() const { Tchar s[_MAX_PATH_ + 1]; return _directoryof_( s, *this, _countof(s) ); }
        StringT  FilenameOf () const { Tchar s[_MAX_PATH_ + 1]; return _filenameof_ ( s, *this, _countof(s) ); }
        StringT  FileTitleOf() const { Tchar s[_MAX_PATH_ + 1]; return _filetitleof_( s, *this, _countof(s) ); }
        StringT  FileExtOf  () const { Tchar s[_MAX_PATH_ + 1]; return _fileextof_  ( s, *this, _countof(s) ); } // preceding dot is included
        StringT& OsPath();
        StringT& UnixPath();
        StringT& WindowsPath();
        StringT  GetSubpathRelativeTo( const Tchar *pstrRelPath, bool *pbSucceeded = NULL );

        StringT& GetSystemerrorMessage( DWORD errorcode ) { GetSystemerrorMessage( errorcode, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) ); return *this; } // uses default language
        StringT& GetSystemerrorMessage( DWORD errorcode, DWORD langid );

        // Note: Names like #LoadString#, #GetWindowText# etc. are macros defined
        // in "winuser.h" and others. That's why we use slightly different names
        // for the methods in here.
        StringT& Loadstring( UINT resid );
        StringT& Loadstring( UINT resid, LANGID language ) { Loadstring( Get_Resource_Instance(), resid, language ); return *this; }
        StringT& Loadstring( HINSTANCE hInstance, UINT resid );
        StringT& Loadstring( HINSTANCE hInstance, UINT resid, LANGID language );
        StringT& Loadappendstring( UINT resid ) { operator +=( StringT( (const Tchar*) (ULONG_PTR) MAKEINTRESOURCE(resid) ) ); return *this; }
        StringT& Loadappendstring( UINT resid, LANGID language ) { operator +=( StringT().Loadstring(resid, language) ); return *this; }
        StringT& Loadappendstring( HINSTANCE hInstance, UINT resid ) { operator +=( StringT().Loadstring(hInstance, resid) ); return *this; }
        StringT& Loadappendstring( HINSTANCE hInstance, UINT resid, LANGID language ) { operator +=( StringT().Loadstring(hInstance, resid, language) ); return *this; }
        StringT& GetWindowtext( HWND hwnd );
        StringT& Wm_Gettext( HWND hwnd );  // Uses WM_GETTEXT. This allows getting text from windows in another process.
        bool     Wm_Gettext_timeout( HWND hwnd, UINT timeout, UINT fuFlags = SMTO_NORMAL | SMTO_ABORTIFHUNG ); // Uses WM_GETTEXT. This allows getting text from windows in another process. This uses also SendMessageTimeout to check if the destination process reacts within a minimum time delay (in ms).
        StringT& GetClassname( HWND hwnd );
        StringT& GetDlgitemtext( HWND hwndParent, UINT idCtrl ) { return GetWindowtext( ::GetDlgItem( hwndParent, idCtrl ) ); }
        StringT& GetListboxItemtext( HWND hwnd, int iItem );
        StringT& GetListboxCurrItemtext( HWND hwnd )
            { return GetListboxItemtext( hwnd, (int) privateString::char_T<Tchar>::__SendMessage(hwnd, LB_GETCURSEL, 0, 0) ); }
        StringT& GetComboboxLBItemtext( HWND hwnd, int iItem );
        StringT& GetComboboxCurrLBItemtext( HWND hwnd )
            { return GetComboboxLBItemtext( hwnd, (int) privateString::char_T<Tchar>::__SendMessage(hwnd, CB_GETCURSEL, 0, 0) ); }
        StringT& GetTreeviewItemtext( HWND hwndTV, HTREEITEM htiItem );
        #ifndef _WIN32_WCE
        StringT& GetCurrentdirectory();
        #endif
        StringT& GetModulefilename() { return GetModulefilename( Get_Module_Instance() ); }
        StringT& GetModulefilename( HMODULE hModule );
        StringT& GetExecutablefilename() { return GetModulefilename( NULL ); }
        StringT& GetExecutabledirectory() { *this = GetExecutablefilename().DirectoryOf(); return *this; }
        StringT& GetFullpathname() { return GetFullpathname( *this ); }
        StringT& GetFullpathname( const Tchar *pstrRelFile );
        StringT& GetUsername();
        #ifndef _WIN32_WCE
        StringT& GetComputername();
        StringT& GetWindowsdirectory();
        StringT& GetSystemdirectory();
        StringT& GetSpecialFolderPath( HWND hwndOwner, int nFolderCSIDL, bool bCreate = false );
        StringT& GetProgramFilesFolder();
        #endif
        StringT& GetTempDir();
        StringT& RegQueryStringVal( HKEY hkey, const Tchar *pstrValue, const Tchar *pstrSubkey = NULL );
        StringT& GetLocaleinfo( LCID locale, LCTYPE lctype );
        StringT& GetLocaleinfoLanguagename   ( LCID locale ) { return GetLocaleinfo( locale, LOCALE_SLANGUAGE    ); }  // gets the name of the locale-language written in the current language
        StringT& GetLocaleinfoEngLanguagename( LCID locale ) { return GetLocaleinfo( locale, LOCALE_SENGLANGUAGE ); }  // gets the name of the locale-language written in english

        #ifndef _WIN32_WCE
            ///<param name="format">
            ///  Can be a normal formatting string as passed
            ///  to <b>strftime()</b>, but also additional format commands with
            ///  syntax "%(...)" are supported. Inside the brackets you can
            ///  specify one of the following commands: <br/>
            ///      "hmt"       Hour-Minute-Time.
            ///                  That is 24 hours or 12 hours with AM/PM marker
            ///                  depending on the current locale. <br/>
            ///      "cwn"       the Calendar Week Number (starting with 1)
            ///                  related to the current locale. <br/>
            ///      ".10s", ".100s", ".ms", ".us", ".ns"
            ///                  Takes the tenths of secs, hundredths of secs, millisecs,
            ///                  microsecs, or nanosecs from the <b>nanos</b> param and
            ///                  displays it in form of a 1, 2, 3, 6 or 9-digit decimal number
            ///                  preceding with zeros if necessary to make all digits. <br/>
            ///      "ms", "us", "ns"
            ///                  Takes the millisecs, microsecs, or nanosecs from the
            ///                  <b>nanos</b> param and displays it in form of a normal
            ///                  integer number. <br/>
            ///</param>
            ///<param name="Locale">
            ///  Which value to use for <b>Locale</b> param is hard to say.
            ///  For more info on this, see all the remarks inside the
            ///  <b>_set_default_locale_()</b> function body.
            ///  <br/><br/>
            ///  But what can be said is, if you have switched to the locale
            ///  using <b>_set_default_locale_()</b>, then you can use
            ///  LOCALE_USER_DEFAULT for <b>Locale</b>. <br/>
            ///  But if you have switched to "C" locale (e.g. using
            ///  <b>_set_C_locale_()</b>), I am not sure what the correct
            ///  define is to use then. I guess its either LOCALE_NEUTRAL or
            ///  LOCALE_INVARIANT (as I explained in the remarks inside
            ///  the body of the <b>_set_default_locale_()</b> function).
            ///</param>
        StringT& StrftimeEx( const Tchar *format, const struct tm *ptm, int nanos, LCID Locale );
        // formatting using C-runtime's #strftime()#
        static StringT timeFmt     ( time_t t, const Tchar *format = _Tstr("%c") );  // Forms a time string in local time. (default = %c = date and time)
        static StringT timeFmtGmt  ( time_t t, const Tchar *format = _Tstr("%c") );  // Forms a time string in UTC time. (default = %c = date and time)
        // formatting using C-runtime's #strftime()# and some additional formattings
        static StringT timeFmtEx   ( time_t t, int nanos = 0, const Tchar *format = _Tstr("%c") , LCID Locale = LOCALE_USER_DEFAULT ) { struct tm _tm; return StringT().StrftimeEx( format, _localtime_(&_tm, &t), nanos, Locale ); }
        static StringT timeFmtExGmt( time_t t, int nanos = 0, const Tchar *format = _Tstr("%c") , LCID Locale = LOCALE_USER_DEFAULT ) { struct tm _tm; return StringT().StrftimeEx( format, _gmtime_   (&_tm, &t), nanos, Locale ); }
        // full date representation
        static StringT getFullDate            ( time_t t ) { return timeFmt   ( t, _Tstr("%#x") ); }
        static StringT getFullDateGmt         ( time_t t ) { return timeFmtGmt( t, _Tstr("%#x") ); }
        // special date/time related names
        static StringT getWeekdayFullName     ( time_t t ) { return timeFmt   ( t, _Tstr("%A") ); }
        static StringT getWeekdayFullNameGmt  ( time_t t ) { return timeFmtGmt( t, _Tstr("%A") ); }
        static StringT getWeekdayAbbrevName   ( time_t t ) { return timeFmt   ( t, _Tstr("%a") ); }
        static StringT getWeekdayAbbrevNameGmt( time_t t ) { return timeFmtGmt( t, _Tstr("%a") ); }
        static StringT getMonthFullName       ( time_t t ) { return timeFmt   ( t, _Tstr("%B") ); }
        static StringT getMonthFullNameGmt    ( time_t t ) { return timeFmtGmt( t, _Tstr("%B") ); }
        static StringT getMonthAbbrevName     ( time_t t ) { return timeFmt   ( t, _Tstr("%b") ); }
        static StringT getMonthAbbrevNameGmt  ( time_t t ) { return timeFmtGmt( t, _Tstr("%b") ); }
        #endif
        static StringT getWeekdayFullName     ( int ix );  // 0, 7 = sunday, 1 = monday, ...
        static StringT getWeekdayAbbrevName   ( int ix );  // 0, 7 = sunday, 1 = monday, ...
        static StringT getMonthFullName       ( int ix );  // 1 = january, 2 = february, ...
        static StringT getMonthAbbrevName     ( int ix );  // 1 = january, 2 = february, ...

        static StringT fromInt  ( int  val, const Tchar *pstrFormat = _Tstr("%d") ) { StringT s; s.Sprintf(pstrFormat, val); return s; }
        static StringT fromInt  ( int  val, strplaceholder_v flags );
        static StringT fromUint ( UINT val, const Tchar *pstrFormat = _Tstr("%u") ) { StringT s; s.Sprintf(pstrFormat, val); return s; }
        static StringT fromUint ( UINT val, strplaceholder_v flags );
        static StringT fromGUID ( const GUID& guid );
        static StringT fromGUID2( const GUID& guid );
        static StringT fromCLSID( const CLSID& clsid );
        static StringT fromIID  ( const IID& iid );

    // misc
        _hashval_t_ EvalHashKey() const;    // Hashkey value calculation for #HashTable<>#.

    // internal
    protected :
            ///<remarks> If in <b>WritePtrAt()</b> a non-NULL string pointer is returned,
            ///  <b>Releasebuffer()</b> must be called afterwards. </remarks>
        Tchar       *WritePtrAt( _idx_ idx )       { if( idx < 0 || (_size_t_) idx >= length() ) return NULL; Tchar *p = GetBuffer(); return p ? p + idx : p; }
        const Tchar *ReadPtrAt ( _idx_ idx ) const { if( idx < 0 || (_size_t_) idx >  length() ) return NULL; return ((const Tchar*) *this) + idx; }

            ///<returns> false when resulting range is zero or less. </returns>
        bool CheckIndexRange( _idx_& idx, _size_t_& count ) const
            { return CheckIndexRange( idx, count, length(), 0 ); }

            ///<returns> false when resulting range is zero or less. </returns>
        static bool CheckIndexRange( _idx_& idx, _size_t_& count, _size_t_ len, _idx_ left_limit )
        {
            if( idx < left_limit ) {
                count += idx - left_limit;
                idx = left_limit;
            }
            if( idx + (_idx_) count > (_idx_) len )
                count = len - idx;

            return (_idx_) count > 0;
        }

            ///<remarks> If <b>idxEnd</b> is -1, then it will be interpreted
            ///  as pointing after the last char of the string. <br/>
            ///  <b>idxEnd</b> is *exclusive*. </remarks>
            ///<returns> false when resulting range is zero or less. </returns>
        bool CheckIndexRange_Special( _idx_& idxStart, _idx_& idxEnd ) const
            { return CheckIndexRange_Special( idxStart, idxEnd, length(), 0 ); }

            ///<remarks> If <b>idxEnd</b> is -1, then it will be interpreted
            ///  as pointing after the last char of the string. <br/>
            ///  <b>idxEnd</b> is *exclusive*. </remarks>
            ///<returns> FALSE when resulting range is zero or less. </returns>
        static bool CheckIndexRange_Special( _idx_& idxStart, _idx_& idxEnd, _size_t_ len, _idx_ left_limit )
        {
            if( idxStart < left_limit )
                idxStart = left_limit;

            if( idxEnd > (_idx_) len || idxEnd == (_idx_) -1 )
                idxEnd = len;

            return idxEnd > idxStart;
        }

        _idx_ GetStartcharindexOffsetRow( _idx_ idxChar, int nRowOffset, _size_t_ overrideLength );

        #ifndef _WIN32_WCE
        static StringT GetTimeStringFromSpecialFormat( const Tchar *command, const struct tm *ptm, int nanos, LCID Locale )
        {
            if( !command || !ptm ) return StringT();

            _ssize_t_     len   = _strlenmax_( command, 40, true );
            const Tchar *pclose = _strchr_   ( command, _Tchar(')') );
            _idx_        iclose = (_idx_)    ( pclose - command );
            if( iclose < (_idx_) len ) len = iclose;

            Tchar thiscmd[40 + 1];
            _strcpy_( thiscmd, command, len + 1 );

            if( _stris_(thiscmd, _Tstr("hmt")) )
            {
                // Get Hour-Minute-Time.
                // That is 24 hours or 12 hours with AM/PM marker
                // depending on the current locale.

                Tchar sloc[2];
                GetLocaleInfo( Locale, LOCALE_ITIME, sloc, _countof(sloc) );
                int nHourFormat = _stoi_( sloc ); // 0 = AM/PM 12 hour format, 1 = 24 hour format

                if( nHourFormat == 0 )
                {
                    Tchar s[80];
                    _strftime_( s, _countof(s), _Tstr("%I:%M "), ptm );
                    _size_t_ len = _strlen_(s);
                    Tchar *p = s + len;

                    _strftime_( p, _countof(s) - len, _Tstr("%p"), ptm );
                    if( _isstrempty_(p) ) {
                        if( ptm->tm_hour <= 11 ) _strcpy_( p, _Tstr("AM"), _countof(s) - len );
                        else                     _strcpy_( p, _Tstr("PM"), _countof(s) - len );
                    }

                    return s;
                }
                else
                {
                    Tchar s[80];
                    _strftime_( s, _countof(s), _Tstr("%H:%M"), ptm );

                    return s;
                }
            }
            else
            if( _stris_(thiscmd, _Tstr("cwn")) )
            {
                // Get the Calendar Week Number (starting with 1)
                // related to the current locale.

                Tchar sloc[2];
                GetLocaleInfo( Locale, LOCALE_IFIRSTDAYOFWEEK, sloc, _countof(sloc) );
                int nFirstDayOfWeek = _stoi_( sloc ); // 0 = monday, 1= tuesday, ... 6 = sunday

                Tchar s[80];
                if( !_strftime_( s, _countof(s), nFirstDayOfWeek == 0 ? _Tstr("%W") : _Tstr("%U"), ptm ) )
                    s[0] = 0;
                if( _isstrempty_( s ) )
                    return s;

                // Week nr begins with 0. But we want it to begin with 1.
                int nWeekNr = _stoi_( s ) + 1;
                _sprintf_( s, _Tstr("%d"), nWeekNr );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr(".10s")) )
            {
                // From the #nanos# parameter, takes just the tenths of secs part
                // and writes it in form of a 1-digit decimal number.

                int tens = (nanos / 100000000) % 10;
                Tchar s[10];
                _sprintf_( s, _Tstr("%1d"), tens );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr(".100s")) )
            {
                // From the #nanos# parameter, takes just the hundredths of secs part
                // and writes it in form of a 2-digit decimal number preceding
                // with zeros if necessary to make 2 digits.

                int hnds = (nanos / 10000000) % 100;
                Tchar s[10];
                _sprintf_( s, _Tstr("%02d"), hnds );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr(".ms")) )
            {
                // From the #nanos# parameter, takes just the milliseconds part
                // and writes it in form of a 3-digit decimal number preceding
                // with zeros if necessary to make 3 digits.

                int ms = (nanos / 1000000) % 1000;
                Tchar s[10];
                _sprintf_( s, _Tstr("%03d"), ms );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr(".us")) )
            {
                // From the #nanos# parameter, takes just the microseconds part
                // and writes it in form of a 3-digit decimal number preceding
                // with zeros if necessary to make 3 digits.

                int us = (nanos / 1000) % 1000000;
                Tchar s[10];
                _sprintf_( s, _Tstr("%06d"), us );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr(".ns")) )
            {
                // From the #nanos# parameter, takes just the nanoseconds part
                // and writes it in form of a 3-digit decimal number preceding
                // with zeros if necessary to make 3 digits.

                int ns = nanos % 1000000000;
                Tchar s[16];
                _sprintf_( s, _Tstr("%09d"), ns );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr("ms")) )
            {
                // From the #nanos# parameter, takes just the milliseconds part
                // and writes it in form of a normal integer number.

                int ms = (nanos / 1000000) % 1000;
                Tchar s[10];
                _sprintf_( s, _Tstr("%d"), ms );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr("us")) )
            {
                // From the #nanos# parameter, takes just the microseconds part
                // and writes it in form of a normal integer number.

                int us = (nanos / 1000) % 1000000;
                Tchar s[10];
                _sprintf_( s, _Tstr("%d"), us );

                return s;
            }
            else
            if( _stris_(thiscmd, _Tstr("ns")) )
            {
                // From the #nanos# parameter, takes just the nanoseconds part
                // and writes it in form of a normal integer number.

                int ns = nanos % 1000000000;
                Tchar s[16];
                _sprintf_( s, _Tstr("%d"), ns );

                return s;
            }

            return StringT();
        }
        #endif
};

template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::Mid( _idx_ first, _size_t_ count ) const
{
    StringT str;

    if( !CheckIndexRange( first, count ) )
        return str;

    const Tchar *pstr = *this;

    Tchar *pdest = str.GetBuffer( count );
    if( !pdest ) return str;

    _strcpy_( pdest, pstr + first, count + 1 );
    pdest[count] = (Tchar) 0;

    str.ReleaseBuffer();
    return str;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Floodfill( Tchar ch, _size_t_ count )
{
    if( ch == (Tchar) 0 || count <= 0 )
        { Empty(); return *this; }

    Tchar *p = GetBuffer( count );
    if( !p ) return *this;

    for( _idx_ i = 0 ; (_size_t_) i < count ; ++i )
        *(p++) = ch;
    *p = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Strcpy_unaligned( const Tchar UNALIGNED *pstrString )
{
    if( !pstrString ) return *this;
    if( IS_INTRESOURCE( pstrString ) ) return operator =( (const Tchar *) pstrString );
    Empty();
    if( !*pstrString ) return *this;

    _size_t_ count = _strlen_unaligned_( pstrString );
    if( count == 0 ) return *this;

    Tchar *p = GetBuffer( count );
    if( !p ) return *this;

    _strcpy_unaligned_( p, pstrString, count + 1 );

    p[count] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Strcpy_unaligned( const Tchar UNALIGNED *pstrString, _size_t_ maxChars )
{
    if( !pstrString ) return *this;
    if( IS_INTRESOURCE( pstrString ) ) return operator =( (const Tchar *) pstrString );
    Empty();
    if( !*pstrString ) return *this;

    _size_t_ count = (_size_t_) _strlenmax_unaligned_( pstrString, maxChars, true );
    if( count <= 0 ) return *this;

    Tchar *p = GetBuffer( count );
    if( !p ) return *this;

    _strcpy_unaligned_( p, pstrString, count + 1 );

    p[count] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Strcat( const Tchar *pstrExternal, _size_t_ count )
{
    if( !pstrExternal || count == 0 ) return *this;
    if( IS_INTRESOURCE( pstrExternal ) ) return Strcat( pstrExternal );
    if( !*pstrExternal ) return *this;

    _size_t_ len = length();

    count = (_size_t_) _strlenmax_( pstrExternal, count, true );
    if( count == 0 ) return *this;

    Tchar *p = GetBuffer( len + count );
    if( !p ) return *this;

    p += len;

    _strconv_( p, pstrExternal, count );

    p[count] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Strcat( const TotherChar *pstrExternal, _size_t_ count )
{
    if( !pstrExternal || count == 0 ) return *this;
    if( IS_INTRESOURCE( pstrExternal ) ) return Strcat( pstrExternal );
    if( !*pstrExternal ) return *this;

    _size_t_ len = length();

    count = (_size_t_) _strlenmax_( pstrExternal, count, true );
    if( count == 0 ) return *this;

    Tchar *p = GetBuffer( len + count );
    if( !p ) return *this;

    p += len;

    _strconv_( p, pstrExternal, count );

    p[count] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Vsprintf( const Tchar *pstrFormat, va_list argptr )
{
    if( IS_INTRESOURCE( pstrFormat ) )
    {
        StringT str( pstrFormat );
        return Vsprintf( (const Tchar*) str, argptr );
    }

    #ifndef _WIN32_WCE
    _size_t_ count = (_size_t_) _vscprintf_( pstrFormat, argptr );
    #else
    _size_t_ count = _strlen_(pstrFormat) * 2 + (4 * 1024);
    // Note: In case of WindowsCE (at least 5.0) there is no _vscprintf() available
    // so in that case we make the buffer large and hope the result fits inside.
    // If it does not, then it is not causing a buffer overrun because
    // _vsprintf_() will ensure it.
    #endif
    if( count <= 0 )
        { Empty(); return *this; }

    Tchar *p = GetBuffer( count );
    if( !p ) return *this;

    _vsprintf_( p, count + 1, pstrFormat, argptr );

    ReleaseBuffer();
    return *this;
}

///<summary> For every <b>chDup</b> character found in the current string,
///  this method will insert a second one. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::DuplicateChar( Tchar chDup )
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *p = *this;

    for( _idx_ i = 0 ; (_size_t_) i < len ; ++i )
        if( p[i] == chDup )
        {
            Insert( ++i, chDup, 1 );

            len = length();
            p   = *this;

            if( !p )
                return *this;
        }

    return *this;
}

///<summary>
///  Performs and executes commands that are present inside segments
///  in the string.
///</summary>
///<remarks>
///  A <u>segment</u> is indicated by its start- and end-marker-string as
///  passed to this method. Default is "{" for the start and "}" for the end.
///   <br/><br/>
///  After the start-marker, one or more <u>commands</u> can follow. <br/>
///  The commands are separated by , characters. The commands will be
///  executed in the order they are specified. <br/>
///  The end of the command list within a segment must be indicated
///  by a ; character. <br/>
///  After the ; character (but still inside the segment), any characters
///  may follow which I will refer to as the <u>parameter-string</u> now.
///   <br/><br/>
///  The parameter-string's meaning depends on the command(s) specified
///  in the segment. Either it could be a number or a floating point
///  value or a textual string.
///   <br/><br/>
///  Now following is a list of possible <u>commands</u>: <br/>
///    <ul>
///      <li><b> -.0 </b><br/>
///        With this command, the <u>parameter-string</u> is interpreted
///        as a floating point value. <br/>
///        This command will cause the floating point value's trailing
///        zeros after the decimal point to be truncated. <br/>
///        The floating point value's decimal point depends on the current
///        locale set by using C-runtime <b>setlocale()</b> and will be
///        queried by using C-runtime <b>localeconv()</b>. <br/>
///        CAUTION: Since monetary quantity values could have a different
///        decimal point character than normal floating point values,
///        the parameter-string cannot handle monetary quantity values
///        correctly in this command. So for this command, use normal
///        floating point values only.
///      </li>
///      <li><b> by </b> or<br/>
///          <b> by+ </b> or<br/>
///          <b> by + </b><br/>
///        With this command, the <u>parameter-string</u> is interpreted
///        as an integer value indicating a file size in bytes. <br/>
///        This command will cause the integer value to be converted into
///        kilo-bytes or mega-bytes or giga-bytes automatically. <br/>
///        In case of <b>by+</b>, then after the (converted)
///        number a byte-marker will be inserted.
///        A byte-marker can be B for bytes, kB for kilo-bytes, MB for
///        mega-bytes or GB for giga-bytes.
///        In case of <b>by +</b>, a space char between the number and
///        the byte-marker will be inserted.
///      </li>
///    </ul>
///  Samples: <br/>
///    <samp> "some text {-.0;1234.56000}, some file size {by+;123456}" </samp>
///</remarks>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::AccomplishSegments( const Tchar *pstrSegmentStart, const Tchar *pstrSegmentEnd )
{
    enum segmentcmd_t {
        SegmentCmd_TruncFltZeros            = 0,
        SegmentCmd_ByteSize,
        SegmentCmd_ByteSizePlus,
        SegmentCmd_ByteSize_Plus,
        SegmentCmd_BeautifyFloatValue,
        SegmentCmd_BeautifyVersionNr,
        SegmentCmd_BuildNr,
        SegmentCmd_Reverse,
        SegmentCmd_Upper,
        SegmentCmd_Lower
    };

    static const Tchar *s_apstrSegmentCmds[] =
        { _Tstr("-.0"), _Tstr("by"), _Tstr("by+"), _Tstr("by +"),
          _Tstr("~f"), _Tstr("~ver#"), _Tstr("build#"),
          _Tstr("<-"), _Tstr("^"), _Tstr("_"), NULL };

    if( _isstrempty_(pstrSegmentStart) || _isstrempty_(pstrSegmentEnd) )
        return *this;

    _size_t_ lenSStart = _strlen_( pstrSegmentStart );
    _size_t_ lenSEnd   = _strlen_( pstrSegmentEnd   );

    const Tchar *pStr = *this;
    const Tchar *pstrScan = pStr;

    for( ;; )
    {
        // We search for the start of a segment.
        const Tchar *pstrCurrStart = _strstr_( pstrScan, pstrSegmentStart );
        if( !pstrCurrStart )
            return *this; // no more segments to find in the string

        const Tchar *pstrCurrSegment = pstrCurrStart + lenSStart;
        const Tchar *pstrCurrCmd     = pstrCurrSegment;

        // Now find end of segment. (and allow nested segments)
        const Tchar *pstrCurrEnd, *pstrNextStart, *pstrSearchSegmentEnd = pstrCurrSegment;
        int depth = 0;
        for( ;; )
        {
            pstrCurrEnd   = _strstr_( pstrSearchSegmentEnd, pstrSegmentEnd   );
            pstrNextStart = _strstr_( pstrSearchSegmentEnd, pstrSegmentStart );

            if( !pstrCurrEnd ) return *this; // unexpected end of data reached

            if( pstrNextStart && pstrNextStart < pstrCurrEnd )
                { ++depth; pstrSearchSegmentEnd = pstrNextStart + lenSStart; continue; }

            if( --depth < 0 ) break;
            pstrSearchSegmentEnd = pstrCurrEnd + lenSEnd;
        }
        // Now #pstrCurrEnd# indicates the position of our segment's end.

        // Now we try to find the ; in the segment.
        const Tchar *pstrCmdListEnd = _strchr_( pstrCurrSegment, _Tchar(';') );
        if( !pstrCmdListEnd || pstrCmdListEnd >= pstrCurrEnd )
            return *this; // ; char must be inside of segment

        pstrScan = pstrCurrEnd + lenSEnd;

        _idx_ idxBeforeSegmentStart = (_idx_) (pstrCurrStart - pStr);
        _idx_ idxAfterSegmentEnd    = (_idx_) (pstrCurrEnd - pStr) + lenSEnd;
        //_size_t_ cSegmentInnerLength = (_size_t_) (pstrCurrEnd - pstrCurrSegment);
        _size_t_ cSegmentOuterLength = idxAfterSegmentEnd - idxBeforeSegmentStart;
        _idx_ idxSegmentParamStart  = (_idx_) (pstrCmdListEnd - pStr) + 1;
        _idx_ idxSegmentParamEnd    = (_idx_) (pstrCurrEnd - pStr);

        StringT strParam = Mid( idxSegmentParamStart, idxSegmentParamEnd - idxSegmentParamStart );

        // We go through the list of commands in the segment.
        while( pstrCurrCmd && pstrCurrCmd < pstrCmdListEnd )
        {
            // Now we try to identify the current command.
            segmentcmd_t nCmd = (segmentcmd_t) -1;
            for( int iCmd = 0 ; s_apstrSegmentCmds[iCmd] ; ++iCmd )
            {
                if( !_strishere_(pstrCurrCmd, s_apstrSegmentCmds[iCmd]) )
                    continue;
                _size_t_ lenCmd = _strlen_( s_apstrSegmentCmds[iCmd] );
                if( pstrCurrCmd[lenCmd] != _Tchar(',') && pstrCurrCmd[lenCmd] != _Tchar(';') )
                    continue;
                nCmd = (segmentcmd_t) iCmd;
                pstrCurrCmd += lenCmd + 1;
                break;
            }

            if( nCmd < 0 ) {
                // Command not found.
                // We skip the current command and go on to the next.
                pstrCurrCmd = _strpbrk_(pstrCurrCmd, _Tstr(",;"));
                if( pstrCurrCmd ) ++pstrCurrCmd;
                continue;
            }

            // Now execute the command.
            switch( nCmd )
            {
                case SegmentCmd_TruncFltZeros :
                {
                    // We truncate trailing zeros after the
                    // decimal point in a floating point value.

                    const lconv *lc = localeconv();
                    Tchar chDot = lc->decimal_point ?
                                    privateString::char_T<Tchar>::getChar(*lc->decimal_point) :
                                    _Tchar('.');

                    _idx_ idxDot = strParam.Strrchr( chDot );
                    if( idxDot >= 0 )
                    {
                        strParam.TrimRight( _Tstr("0") );

                        if( (_idx_) strParam.length() - 1 == idxDot )
                            strParam.Delete( idxDot );

                        if( strParam.IsEmpty() )
                            strParam += _Tchar('0');
                    }

                    break;
                }

                case SegmentCmd_ByteSize :
                case SegmentCmd_ByteSizePlus :
                case SegmentCmd_ByteSize_Plus :
                {
                    // We convert the integer value in the parameter
                    // to either B (byte), kB (kilo-byte),
                    // MB (mega-byte), GB (giga-byte) or TB (tera-byte)
                    // and eventually append the byte-marker if requested.

                    UINT64 nBytesize = (UINT64) _stoi64_( strParam );
                    double dBytesize = (double) nBytesize;
                    int    nBytemarker = 4;

                    for( ; nBytemarker > 0 ; --nBytemarker )
                    {
                        UINT64 nMarkerVal = ((UINT64)1) << (10 * nBytemarker);
                        if( nBytesize >= nMarkerVal )
                            { dBytesize /= (double) nMarkerVal; break; }
                    }

                    dBytesize  = privateString::type<double>::round( dBytesize * 100.0 ) / 100.0;

                    if( nBytemarker > 0 ) strParam.Sprintf( _Tstr("%.10g"), dBytesize );
                    else                  strParam.Sprintf( _Tstr("%I64u"), nBytesize );

                    static const Tchar *s_apstrBytemarkers[] =
                        { _Tstr(" B"), _Tstr(" kB"), _Tstr(" MB"), _Tstr(" GB"), _Tstr(" TB") };
                    if( nCmd == SegmentCmd_ByteSizePlus )
                        strParam += s_apstrBytemarkers[nBytemarker] + 1;
                    if( nCmd == SegmentCmd_ByteSize_Plus )
                        strParam += s_apstrBytemarkers[nBytemarker];

                    break;
                }

                case SegmentCmd_BeautifyFloatValue :
                {
                    strParam.BeautifyFloatValue();
                    break;
                }

                case SegmentCmd_BeautifyVersionNr :
                {
                    strParam.BeautifyVersionNr();
                    break;
                }

                case SegmentCmd_BuildNr :
                {
                    strParam = strParam.GetBuildNr();
                    break;
                }

                case SegmentCmd_Reverse :
                {
                    strParam.Reverse();
                    break;
                }

                case SegmentCmd_Upper :
                {
                    strParam.Uppercase();
                    break;
                }

                case SegmentCmd_Lower :
                {
                    strParam.Lowercase();
                    break;
                }
            }
        }

        // Now we replace the complete segment in the original
        // string with the (adopted) parameter.

        Delete( idxBeforeSegmentStart, cSegmentOuterLength );
        Insert( idxBeforeSegmentStart, strParam );

        pStr = *this;
        if( !pStr ) return *this; // should never occur

        // Note: It would make sense to continue parsing for
        //  segments after the end of the last edited parameter string.
        //  But what we really do is to continue parsing of the
        //  edited parameter string itself so that we can handle nasted commands.
        //pstrScan = pStr + (idxBeforeSegmentStart + strParam.length());
        pstrScan = pStr + idxBeforeSegmentStart;
    }
}

///<summary> Converts a string which also contains escape sequences (like C/C++-strings)
///  into a normal string. </summary>
///<remarks> Notice, if the source-escape-sequence contains '\0' chars, they will
///  be converted to the specified char in <b>chEndOfStr</b>. <br/>
///  If <b>chEndOfStr</b> is zero, each '\0' char will be removed. </remarks>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::EscToStr( Tchar chEndOfStr )
{
    StringT str;

    _size_t_ len = length();
    if( len <= 0 ) return str;

    const Tchar *ptr = *this;

    while( *ptr )
    {
        int c = _esctoch_( &ptr );

        if( c < 0 ) c = -c;
        if( c == 0 ) c = chEndOfStr;
        if( c == 0 || c > 255 ) continue;

        str += (Tchar) c;
    }

    return str;
}

///<summary> Converts a normal string into a string with escape sequences (like C/C++-strings) </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::StrToEsc()
{
    StringT str;

    _size_t_ len = length();
    if( len <= 0 ) return str;

    const Tchar *ptr = *this;

    Tchar s[128];
    s[0] = (Tchar) 0;
    Tchar *end = &s[_countof(s) - 20];
    Tchar *p = s;

    while( *ptr )
    {
        while( *ptr && p < end )
        {
            _esccpy_( p, *(ptr++), 20 );
            p += _strlen_( p );
        }

        str += s;
        p = s;
    }

    return str;
}

///<summary>
///  Converts the current pseudo-string-list (in <b>this</b> object)
///  into a real string-list. (returning a malloc() allocated pointer).
///</summary>
///<remarks>
///  A <u>real string-list</u> consists of several null-terminated strings
///   one after each other so that the null-terminator separates a
///   string from another. <br/>
///   The end of the string-list is marked by 2 null-terminator chars.
///   <br/><br/>
///  A <u>pseudo-string-list</u> is a normal C string (null-terminated).
///   It contains several separator characters (of kind <b>chSep</b>) which
///   are meant as separators between 2 internal strings. <br/>
///   The end will be marked by 2 separator chars (but it need not when you
///   use <b>ToStrlst()</b>).
///</remarks>
///<returns> A pointer to the new real string list (allocated using
///  <b>malloc()</b>). If NULL -> error or invalid parameter. </returns>
template< class baseString, typename Tchar >
Tchar *StringT<baseString, Tchar>::toStrlst( Tchar chSep )
{
    if( chSep == 0 ) return NULL;

    const Tchar *pstrlst = *this;   // the pseudo strlst
    _size_t_ lstlen  = length();    // we are not yet sure if #this# string has 2 #chSep# chars at the end.

    if( _isstrempty_(pstrlst) || lstlen < 0 )
        lstlen = 0;

    Tchar *retptr = (Tchar*) malloc( (lstlen + 2) * (_size_t_) sizeof(Tchar) );
    if( !retptr ) return NULL;

    Tchar *pdest = retptr;

    _idx_ i;
    for( i = 0 ; (_size_t_) i < lstlen ; ++i, ++pstrlst )
    {
        if( *pstrlst == chSep )
            if( (i < (_idx_) lstlen - 1) ? (pstrlst[1] == chSep) : true )
                break;

        *(pdest++) = (*pstrlst == chSep) ? (Tchar) 0 : *pstrlst;
    }

    *(pdest++) = (Tchar) 0;
    *(pdest++) = (Tchar) 0;

    return retptr;
}

///<summary>
///  Converts a real string-list (see remarks for end-of-strlst marker)
///  into a pseudo-string-list (in <b>this</b> object).
///</summary>
///<remarks>
///  A <u>real string-list</u> consists of several null-terminated strings
///   one after each other so that the null-terminator separates a
///   string from another. <br/>
///   The end of the string-list is marked by 2 null-terminator chars.
///   <br/><br/>
///  A <u>pseudo-string-list</u> is a normal C string (null-terminated).
///   It contains several separator characters (of kind <b>chSep</b>) which
///   are meant as separators between 2 internal strings. <br/>
///   The end will be marked by 2 separator chars (but it need not when you
///   use <b>ToStrlst()</b>).
///</remarks>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::fromStrlst( const Tchar *pstrStrlst, Tchar chSep )
{
    if( chSep == 0 ) return *this;

    const Tchar *pstrlst = pstrStrlst;
    _size_t_ lstlen  = _strlstlen_( pstrStrlst );

    if( (pstrlst && lstlen > 0) ? (*pstrlst == 0) : false )
    {
        // Check for null-terminators at start of string-list.
        ++pstrlst;
        --lstlen;
        if( lstlen > 0 ) if( *pstrlst == 0 ) lstlen = 0;
    }

    if( lstlen <= 0 || !pstrlst )
    {
        Tchar s[3] = { chSep, chSep, (Tchar) 0 };
        *this = s;
        return *this;
    }

    lstlen += 2;

    Tchar *pdest = GetBuffer( lstlen + 2 );    // including 2 terminator chars + a C string null terminator
    if( !pdest ) return *this;

    for( _idx_ i = 0 ; (_size_t_) i < lstlen ; ++i, ++pstrlst )
    {
        if( *pstrlst == 0 )
            if( (i < (_idx_) lstlen - 1) ? (pstrlst[1] == 0) : true )
                break;

        *(pdest++) = (*pstrlst == 0) ? chSep : *pstrlst;
    }

    *(pdest++) = chSep;
    *(pdest++) = chSep;
    *(pdest++) = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

#ifndef OFN_ALLOWMULTISELECT
#define OFN_ALLOWMULTISELECT         0x00000200
#endif
#ifndef OFN_EXPLORER
#define OFN_EXPLORER                 0x00080000
#endif

///<summary>
///  Converts a specially formatted string-list or space-separated string
///  into a pseudo-string-list (in <b>this</b> object), or just interprets
///  the input string as normal string. <br/><br/>
///
///  The special format comes from the results of a <b>GetOpenFileName()</b>
///  or <b>GetSaveFileName()</b> when mutliselect was allowed. <br/><br/>
///
///  The special format is used when the OFN_ALLOWMULTISELECT flag is set: <br/>
///  The format is then like this: The first string in the list is the current directory
///  followed by the filenames of the selected files. <br/>
///  If the OFN_EXPLORER flag is specified, then the input string-list is interpreted
///  as real string-list. <br/>
///  If the OFN_EXPLORER flag is not specified, then the input string-list is interpreted
///  as a normal C string containing sub-strings separated by space characters. (In that
///  case, file- and directory-names cannot contain spaces which the system ensures
///  by returning the short (8.3) filenames). <br/><br/>
///
///  But if the OFN_ALLOWMULTISELECT flag is not specified, then the input string
///  will be interpreted just as a normal C string containing a filename with full path.
///
///  In all of these cases, this method will return either one string without any
///  separator chars included which contains one filename with with full.
///  Or it will return a pseudo-string-list with just filenames which each has a full path.
///</summary>
///<param name="pstrStrlst"> either a real string-list or a space-separated string,
///  depending on the <b>dwOfnFlags</b> parameter. </param>
///<param name="dwOfnFlags"> pass here the <b>.Flags</b> parameter of your <b>OPENFILENAME</b> struct. <br/>
///  If the OFN_ALLOWMULTISELECT is specified, the OFN_EXPLORER flag will be interpreted: <br/>
///  If the OFN_EXPLORER flag is set, then <b>pstrStrlst</b> will be interpreted as
///  real string-list. <br/>
///  If the OFN_EXPLORER flag is no set, <b>pstrStrlst</b> will be interpreted as
///  a normal string that contains sub-strings separated by space characters. <br/>
///  But if the OFN_ALLOWMULTISELECT is not specified, then <b>pstrStrlst</b> will be
///  interpreted just as one normal string containing a filename (or subdir) with full path. </param>
///<returns> the number of filenames (and subdirs) included in the returned string-list. </returns>
///<remarks>
///  A <u>real string-list</u> consists of several null-terminated strings
///   one after each other so that the null-terminator separates a
///   string from another. <br/>
///   The end of the string-list is marked by 2 null-terminator chars.
///   <br/><br/>
///  A <u>pseudo-string-list</u> is a normal C string (null-terminated).
///   It contains several separator characters (of kind <b>chSep</b>) which
///   are meant as separators between 2 internal strings. <br/>
///   The end will be marked by 2 separator chars (but it need not when you
///   use <b>ToStrlst()</b>).
///</remarks>
template< class baseString, typename Tchar >
_size_t_ StringT<baseString, Tchar>::fromOpenfilenameStrlst( const Tchar *pstrStrlst, DWORD dwOfnFlags, Tchar chSep )
{
    if( !(dwOfnFlags & OFN_ALLOWMULTISELECT) )
    {
        *this = pstrStrlst;
        return _isstrempty_(pstrStrlst) ? 0 : 1;
    }

    if( chSep == 0 ) return 0;

    if( !(dwOfnFlags & OFN_EXPLORER) )
    {
        // In the case of space-separated sub-strings, we convert them
        // into a real string-list and then call our method again but
        // with the OFN_EXPLORER flag included.

        String tmp = pstrStrlst;
        Tchar *pstrlstReal = tmp.toStrlst( _Tchar(' ') );

        _size_t_ cnt = fromOpenfilenameStrlst( pstrlstReal, dwOfnFlags | OFN_EXPLORER, chSep );

        if( pstrlstReal ) free( pstrlstReal );
        return cnt;
    }

    // Now we handle an explorer-style string-list (a real string-list)
    // returned from a GetOpen/SaveFileName() multiselection mode dialog.

    const Tchar *pstrlst = pstrStrlst;
    const Tchar *pstrDir = pstrlst;
    pstrlst = _fwd_to_next_strlst_str_( pstrlst );

    // If the stringlist contains only one string, then this is a filename with full path.
    if( !pstrlst && !_isstrempty_(pstrDir) )
        { *this = pstrDir; return 1; }

    // if a NULL stringlist, we return the empty pseudo-string-list
    if( !pstrlst )
        { fromStrlst( NULL, chSep ); return 0; }

    Empty();
    Tchar s[_MAX_PATH_ + 2];
    Tchar szsep[2] = { chSep, (Tchar) 0 };
    _size_t_ cnt = 0;

    do
    {
        _joinpaths_( s, pstrDir, pstrlst, _countof(s) - 1 );    // -1 because we need to ensure to have space to append a separator char.
        _strcat_( s, szsep, _countof(s) );
        *this += s;
        ++cnt;
    } while( pstrlst = _fwd_to_next_strlst_str_( pstrlst ) );

    *this += szsep;

    return cnt;
}

///<summary> Insert the character <b>ch</b> <b>count</b> times
///  beginning at character index <b>idx</b> </summary
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Insert( _idx_ idx, Tchar ch, _size_t_ count )
{
    if( ch == (Tchar) 0 || count <= 0 ) return *this;

    _size_t_ len = length();
    if( idx < 0 || (_size_t_) idx > len ) return *this;

    Tchar *p = GetBuffer( len + count );
    if( !p ) return *this;

    memmove( p + idx + count, p + idx, (len - idx + 1) * sizeof(Tchar) );
    p += idx;

    for( _idx_ i = 0 ; (_size_t_) i < count ; ++i )
        *(p++) = ch;

    ReleaseBuffer();
    return *this;
}

///<summary> Insert the specified string at character index <b>idx</b> </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Insert( _idx_ idx, const Tchar *pstr )
{
    _size_t_ count = _strlen_( pstr );
    if( count <= 0 ) return *this;

    _size_t_ len = length();
    if( idx < 0 || (_size_t_) idx > len ) return *this;

    Tchar *p = GetBuffer( len + count );
    if( !p ) return *this;

    memmove( p + idx + count, p + idx, (len - idx + 1) * sizeof(Tchar) );
    p += idx;

    memcpy( p, pstr, count * sizeof(Tchar) );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Delete( _idx_ idx, _size_t_ count )
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    if( !CheckIndexRange( idx, count, len, 0 ) )
        return *this;

    _idx_ iend = idx + count;

    if( idx == 0 && (_size_t_) iend >= len )
        { Empty(); return *this; }

    Tchar *p = GetBuffer();
    if( !p ) return *this;

    memmove( p + idx, p + iend, (len - iend + 1) * sizeof(Tchar) );

    ReleaseBuffer();
    return *this;
}

///<summary> Remove any of the characters of the specified character set
///  found inside the current string </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Remove( const Tchar *pstrCharset )
{
    if( ! pstrCharset ) return *this;
    if( !*pstrCharset ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _idx_ i = 0;
    while( ptr[i] )
    {
        _idx_ i2 = i;
        while( ptr[i2] && _strchr_( pstrCharset, ptr[i2] ) ) ++i2;

        if( i2 > i ) {
            _size_t_ cmove = len + 1 - i2;
            if( cmove > 0 ) memmove( &ptr[i], &ptr[i2], cmove * sizeof(Tchar) );
            len -= i2 - i;
            if( len <= 0 ) break;
        } else
            ++i;
    }

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Removei( const Tchar *pstrCharset )
{
    if( ! pstrCharset ) return *this;
    if( !*pstrCharset ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _idx_ i = 0;
    while( ptr[i] )
    {
        _idx_ i2 = i;
        while( ptr[i2] && _strichr_( pstrCharset, ptr[i2] ) ) ++i2;

        if( i2 > i ) {
            _size_t_ cmove = len + 1 - i2;
            if( cmove > 0 ) memmove( &ptr[i], &ptr[i2], cmove * sizeof(Tchar) );
            len -= i2 - i;
            if( len <= 0 ) break;
        } else
            ++i;
    }

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::RemoveSpaces()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _idx_ i = 0;
    while( ptr[i] )
    {
        _idx_ i2 = i;
        while( _isspace_( ptr[i2] ) ) ++i2;

        if( i2 > i ) {
            _size_t_ cmove = len + 1 - i2;
            if( cmove > 0 ) memmove( &ptr[i], &ptr[i2], cmove * sizeof(Tchar) );
            len -= i2 - i;
            if( len <= 0 ) break;
        } else
            ++i;
    }

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Remain( const Tchar *pstrCharset )
{
    if( ! pstrCharset ) return *this;
    if( !*pstrCharset ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _idx_ i = 0;
    while( ptr[i] )
    {
        _idx_ i2 = i;
        while( ptr[i2] && !_strchr_( pstrCharset, ptr[i2] ) ) ++i2;

        if( i2 > i ) {
            _size_t_ cmove = len + 1 - i2;
            if( cmove > 0 ) memmove( &ptr[i], &ptr[i2], cmove * sizeof(Tchar) );
            len -= i2 - i;
            if( len <= 0 ) break;
        } else
            ++i;
    }

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Remaini( const Tchar *pstrCharset )
{
    if( ! pstrCharset ) return *this;
    if( !*pstrCharset ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _idx_ i = 0;
    while( ptr[i] )
    {
        _idx_ i2 = i;
        while( ptr[i2] && !_strichr_( pstrCharset, ptr[i2] ) ) ++i2;

        if( i2 > i ) {
            _size_t_ cmove = len + 1 - i2;
            if( cmove > 0 ) memmove( &ptr[i], &ptr[i2], cmove * sizeof(Tchar) );
            len -= i2 - i;
            if( len <= 0 ) break;
        } else
            ++i;
    }

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Trim()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx1 = 0;
    while( _isspace_( ptr[idx1] ) )
        ++idx1;

    if( (_size_t_) idx1 >= len )
        { Empty(); return *this; }

    _idx_ idx2 = len - 1;
    while( idx2 >= idx1 ? _isspace_( ptr[idx2] ) : false )
        --idx2;

    _size_t_ newlen = idx2 + 1 - idx1;
    if( newlen <= 0 )
        { Empty(); return *this; }

    if( newlen == len )
        return *this;

    Tchar *ptr2 = GetBuffer();
    if( !ptr2 ) return *this;

    memmove( ptr2, ptr2 + idx1, newlen * sizeof(Tchar) );
    ptr2[newlen] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Trim( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx1 = 0;
    while( ptr[idx1] && _strchr_( pstrCharsToTrim, ptr[idx1] ) )
        ++idx1;

    if( idx1 >= len )
        { Empty(); return *this; }

    _idx_ idx2 = len - 1;
    while( idx2 >= idx1 ? _strchr_( pstrCharsToTrim, ptr[idx2] ) : false )
        --idx2;

    _size_t_ newlen = idx2 + 1 - idx1;
    if( newlen <= 0 )
        { Empty(); return *this; }

    if( newlen == len )
        return *this;

    Tchar *ptr2 = GetBuffer();
    if( !ptr2 ) return *this;

    memmove( ptr2, ptr2 + idx1, newlen * sizeof(Tchar) );
    ptr2[newlen] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Trimi( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx1 = 0;
    while( ptr[idx1] && _strichr_( pstrCharsToTrim, ptr[idx1] ) )
        ++idx1;

    if( idx1 >= len )
        { Empty(); return *this; }

    _idx_ idx2 = len - 1;
    while( idx2 >= idx1 ? _strichr_( pstrCharsToTrim, ptr[idx2] ) : FALSE )
        --idx2;

    _size_t_ newlen = idx2 + 1 - idx1;
    if( newlen <= 0 )
        { Empty(); return *this; }

    if( newlen == len )
        return *this;

    Tchar *ptr2 = GetBuffer();
    if( !ptr2 ) return *this;

    memmove( ptr2, ptr2 + idx1, newlen * sizeof(Tchar) );
    ptr2[newlen] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimLeft()
{
    if( IsEmpty() ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = 0;
    while( _isspace_( ptr[idx] ) )
        ++idx;

    return Delete( 0, idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimLeft( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    if( IsEmpty() ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = 0;
    while( ptr[idx] && _strchr_( pstrCharsToTrim, ptr[idx] ) )
        ++idx;

    return Delete( 0, idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimLefti( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    if( IsEmpty() ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = 0;
    while( ptr[idx] && _strichr_( pstrCharsToTrim, ptr[idx] ) )
        ++idx;

    return Delete( 0, idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimRight()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = len - 1;
    while( idx >= 0 ? _isspace_( ptr[idx] ) : false )
        --idx;

    ++idx;

    return Delete( idx, len - idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimRight( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = len - 1;
    while( idx >= 0 ? _strchr_( pstrCharsToTrim, ptr[idx] ) : false )
        --idx;

    ++idx;

    return Delete( idx, len - idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::TrimRighti( const Tchar *pstrCharsToTrim )
{
    if( ! pstrCharsToTrim ) return *this;
    if( !*pstrCharsToTrim ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *ptr = *this;

    _idx_ idx = len - 1;
    while( idx >= 0 ? _strichr_( pstrCharsToTrim, ptr[idx] ) : false )
        --idx;

    ++idx;

    return Delete( idx, len - idx );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Uppercase()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _strupr_( ptr );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Lowercase()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _strlwr_( ptr );

    ReleaseBuffer();
    return *this;
}

///<summary> makes only the first non-space char in the string uppercase. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::UppercaseFirstChar()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _strupr_firstchar_( ptr );

    ReleaseBuffer();
    return *this;
}

///<summary> makes only the first non-space char in the string lowercase. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::LowercaseFirstChar()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _strlwr_firstchar_( ptr );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Reverse()
{
    _size_t_ len = length();
    if( len <= 0 ) return *this;

    Tchar *ptr = GetBuffer();
    if( !ptr ) return *this;

    _strrev_( ptr );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Strchr( _idx_ idx, Tchar ch ) const
{
    if( ch == 0 ) return -1;
    if( idx < 0 || (_size_t_) idx >= length() ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _strchr_( p + idx, ch );
    if( !p2 ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Strichr( _idx_ idx, Tchar ch ) const
{
    if( ch == 0 ) return -1;
    if( idx < 0 || (_size_t_) idx >= length() ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _strichr_( p + idx, ch );
    if( !p2 ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Strrchr( Tchar ch ) const
{
    if( ch == 0 ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _strrchr_( p, ch );
    if( !p2 || p2 < p ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Strstr( _idx_ idx, const Tchar *pstrSubstr ) const
{
    if( ! pstrSubstr ) return -1;
    if( !*pstrSubstr ) return -1;
    if( idx < 0 || (_size_t_) idx >= length() ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _strstr_( p + idx, pstrSubstr );
    if( !p2 ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Stristr( _idx_ idx, const Tchar *pstrSubstr ) const
{
    if( ! pstrSubstr ) return -1;
    if( !*pstrSubstr ) return -1;
    if( idx < 0 || (_size_t_) idx >= length() ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _stristr_( p + idx, pstrSubstr );
    if( !p2 ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Strpbrk( _idx_ idx, const Tchar *pstrCharset ) const
{
    if( ! pstrCharset ) return -1;
    if( !*pstrCharset ) return -1;
    if( idx < 0 || (_size_t_) idx >= length() ) return -1;

    const Tchar *p = *this;
    if( !*p ) return -1;

    const Tchar *p2 = _strpbrk_( p + idx, pstrCharset );
    if( !p2 ) return -1;

    return (_idx_)( p2 - p );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replace( Tchar chFind, Tchar chReplace )
{
    if( chFind == 0 || chReplace == 0 ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;

    while( *p )
    {
        if( chFind == *p )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                if( chFind == *p2 )
                    *p2 = chReplace;
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replace( const Tchar *pstrFindCharset, Tchar chReplace )
{
    if( ! pstrFindCharset || chReplace == 0 ) return *this;
    if( !*pstrFindCharset ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;

    while( *p )
    {
        if( _strchr_(pstrFindCharset, *p) )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                if( _strchr_(pstrFindCharset, *p2) )
                    *p2 = chReplace;
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replace( const Tchar *pstrFindCharset, const Tchar *pstrReplaceCharset )
{
    if( !pstrFindCharset || !pstrReplaceCharset ) return *this;

    _size_t_ lfind = _strlen_( pstrFindCharset    );
    _size_t_ lrep  = _strlen_( pstrReplaceCharset );
    _size_t_ l     = privateString::tmin( lfind, lrep );
    if( l <= 0 ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;
    _idx_ ix;

    while( *p )
    {
        if( _strchr_(pstrFindCharset, *p) )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                p = _strchr_(pstrFindCharset, *p2);
                if( p )
                    if( ( ix = (_idx_) (p - pstrFindCharset) ) < (_idx_) l )
                        *p2 = pstrReplaceCharset[ ix ];
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replacei( Tchar chFind, Tchar chReplace )
{
    if( chFind == 0 || chReplace == 0 ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;
    chFind = _toupper_( chFind );

    while( *p )
    {
        if( chFind == _toupper_(*p) )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                if( chFind == _toupper_(*p2) )
                    *p2 = chReplace;
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replacei( const Tchar *pstrFindCharset, Tchar chReplace )
{
    if( ! pstrFindCharset || chReplace == 0 ) return *this;
    if( !*pstrFindCharset ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;

    while( *p )
    {
        if( _strichr_(pstrFindCharset, *p) )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                if( _strichr_(pstrFindCharset, *p2) )
                    *p2 = chReplace;
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Replacei( const Tchar *pstrFindCharset, const Tchar *pstrReplaceCharset )
{
    if( !pstrFindCharset || !pstrReplaceCharset ) return *this;

    _size_t_ lfind = _strlen_( pstrFindCharset    );
    _size_t_ lrep  = _strlen_( pstrReplaceCharset );
    _size_t_ l     = privateString::tmin( lfind, lrep );
    if( l <= 0 ) return *this;

    const Tchar *p0 = *this;
    const Tchar *p = p0;
    _idx_ ix;

    while( *p )
    {
        if( _strichr_(pstrFindCharset, *p) )
        {
            Tchar *p2 = GetBuffer();
            if( !p2 ) return *this;

            p2 += p - p0;

            while( *p2 )
            {
                p = _strichr_(pstrFindCharset, *p2);
                if( p )
                    if( ( ix = (_idx_) (p - pstrFindCharset) ) < (_idx_) l )
                        *p2 = pstrReplaceCharset[ ix ];
                ++p2;
            }

            ReleaseBuffer();
            return *this;
        }

        ++p;
    }

    return *this;
}

template< class baseString, typename Tchar >
_size_t_ StringT<baseString, Tchar>::ReplacestringRet( const Tchar *pstrFindString, const Tchar *pstrReplaceString )
{
    if( !pstrFindString || !pstrReplaceString ) return 0;

    _size_t_ lfind = _strlen_( pstrFindString    );
    _size_t_ lrep  = _strlen_( pstrReplaceString );
    if( lfind <= 0 ) return 0;

    const Tchar *p0 = *this;
    const Tchar *p = p0;

    bool gotbuffer = false;
    _size_t_ len = length();
    _size_t_ cnt = 0;

    while( p = _strstr_( p, pstrFindString ) )
    {
        _idx_ char_ofs = (_idx_) (p - p0);
        Tchar *pw0, *pw;

        if( lrep > lfind )
        {
            pw = pw0 = GetBuffer( len + (lrep - lfind) );
            if( !pw0 ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
            memmove( pw + lrep, pw + lfind, (len + 1 - (char_ofs + lfind)) * sizeof(Tchar) );
            len += lrep - lfind;
        }
        else if( lrep < lfind )
        {
            pw = pw0 = GetBuffer();
            if( !pw ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
            memmove( pw + lrep, pw + lfind, (len + 1 - (char_ofs + lfind)) * sizeof(Tchar) );
            len -= lfind - lrep;
        }
        else
        {
            pw = pw0 = GetBuffer();
            if( !pw ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
        }

        memcpy( pw, pstrReplaceString, lrep * sizeof(Tchar) );

        ++cnt;
        p0 = pw0;
        p  = pw + lrep;
    }

    if( gotbuffer )
        ReleaseBuffer();

    return cnt;
}

template< class baseString, typename Tchar >
_size_t_ StringT<baseString, Tchar>::ReplaceistringRet( const Tchar *pstrFindString, const Tchar *pstrReplaceString )
{
    if( !pstrFindString || !pstrReplaceString ) return 0;

    _size_t_ lfind = _strlen_( pstrFindString    );
    _size_t_ lrep  = _strlen_( pstrReplaceString );
    if( lfind <= 0 ) return 0;

    const Tchar *p0 = *this;
    const Tchar *p = p0;

    bool gotbuffer = false;
    _size_t_ len = length();
    _size_t_ cnt = 0;

    while( p = _stristr_( p, pstrFindString ) )
    {
        _idx_ char_ofs = (_idx_) (p - p0);
        Tchar *pw0, *pw;

        if( lrep > lfind )
        {
            pw = pw0 = GetBuffer( len + (lrep - lfind) );
            if( !pw0 ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
            memmove( pw + lrep, pw + lfind, (len + 1 - (char_ofs + lfind)) * sizeof(Tchar) );
            len += lrep - lfind;
        }
        else if( lrep < lfind )
        {
            pw = pw0 = GetBuffer();
            if( !pw ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
            memmove( pw + lrep, pw + lfind, (len + 1 - (char_ofs + lfind)) * sizeof(Tchar) );
            len -= lfind - lrep;
        }
        else
        {
            pw = pw0 = GetBuffer();
            if( !pw ) { if( gotbuffer ) ReleaseBuffer(); return cnt; }
            gotbuffer = true;
            pw += char_ofs;
        }

        memcpy( pw, pstrReplaceString, lrep * sizeof(Tchar) );

        ++cnt;
        p0 = pw0;
        p  = pw + lrep;
    }

    if( gotbuffer )
        ReleaseBuffer();

    return cnt;
}

///<summary> Replaces the placeholder chars by the chars of <b>pstrString</b> </summary>
///<param name="flags"> a combination of values from <b>strplaceholder_t</b> </param>
///<remarks> If <b>pstrString</b> is too short to fill all placeholder chars, then those
///  placeholder chars which cannot be filled will be *removed* from <b>this</b> string. </remarks>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::PlaceholderCharsToString( const Tchar *pstrPlaceholderChars, const Tchar *pstrString, strplaceholder_v flags )
{
    if( ! pstrPlaceholderChars || !pstrString ) return *this;
    if( !*pstrPlaceholderChars                ) return *this;

    _size_t_ len = length();
    if( len <= 0 ) return *this;

    const Tchar *p_r = *this;
    Tchar *p_w = NULL;

    bool gotbuffer = false;

    if( flags & StrPlaceholderF_RightAlign )
    {
        _size_t_ len_str = _strlen_( pstrString );
        _idx_    i_str   = len_str - 1;
        _idx_    i_last_replace = -1;

        for( _idx_ i = len - 1 ; i >= 0 ; --i )
            if( _strchr_( pstrPlaceholderChars, p_r[i] ) )
            {
                if( !gotbuffer ) {
                    if( !( p_r = p_w = GetBuffer() ) ) return *this;
                    gotbuffer = true;
                }

                if( i_str >= 0 ) {
                    p_w[i] = pstrString[ i_str-- ];
                    i_last_replace = i;
                } else {
                    memmove( p_w + i, p_w + i + 1, (len + 1 - (i + 1)) * sizeof(Tchar) );
                    --len;
                    --i_last_replace;
                }
            }

        if( !(flags & StrPlaceholderF_SkipExceedingChars ) )
        {
            // Now we check, if all of the chars have been replaced.
            // If not -> we insert the remaining chars right before the
            // last known char replacement.
            if( i_str >= 0 )
            {
                if( i_last_replace < 0 ) i_last_replace = 0;
                _size_t_ c_insert_chars = i_str + 1;

                if( !( p_r = p_w = GetBuffer( len + c_insert_chars ) ) )
                    { if( gotbuffer ) ReleaseBuffer(); return *this; }
                gotbuffer = true;

                memmove( p_w + i_last_replace + c_insert_chars, p_w + i_last_replace, (len + 1 - i_last_replace) * sizeof(Tchar) );
                memcpy( p_w + i_last_replace, pstrString, c_insert_chars * sizeof(Tchar) );
                len += c_insert_chars;
            }
        }
    }
    else
    {
        _size_t_ len_str = _strlen_( pstrString );
        _idx_    i_str   = 0;
        _idx_    i_last_replace = -1;

        for( _idx_ i = 0 ; (_size_t_) i < len ; ++i )
            if( _strchr_( pstrPlaceholderChars, p_r[i] ) )
            {
                if( !gotbuffer ) {
                    if( !( p_r = p_w = GetBuffer() ) ) return *this;
                    gotbuffer = true;
                }

                if( (_size_t_) i_str < len_str ) {
                    p_w[i] = pstrString[ i_str++ ];
                    i_last_replace = i;
                } else {
                    memmove( p_w + i, p_w + i + 1, (len + 1 - (i + 1)) * sizeof(Tchar) );
                    --len;
                    --i;
                }
            }

        if( !(flags & StrPlaceholderF_SkipExceedingChars ) )
        {
            // Now we check, if all of the chars have been replaced.
            // If not -> we insert the remaining chars right after the
            // last known char replacement.
            if( (_size_t_) i_str < len_str )
            {
                if( ++i_last_replace == 0 ) i_last_replace = len;
                _size_t_ c_insert_chars = len_str - i_str;

                if( !( p_r = p_w = GetBuffer( len + c_insert_chars ) ) )
                    { if( gotbuffer ) ReleaseBuffer(); return *this; }
                gotbuffer = true;

                memmove( p_w + i_last_replace + c_insert_chars, p_w + i_last_replace, (len + 1 - i_last_replace) * sizeof(Tchar) );
                memcpy( p_w + i_last_replace, pstrString + i_str, c_insert_chars * sizeof(Tchar) );
                len += c_insert_chars;
            }
        }
    }

    if( gotbuffer )
        ReleaseBuffer();

    return *this;
}

///<summary> All chars of <b>pstrPlaceholderChars</b> found in <b>this</b> string
///  will be replaced by digit chars taken from number value <b>nNum</b> interpreted
///  as numeric value of a type from selected in <b>flags</b> (a StrNumT_* value). </summary>
///<remarks> Sample: <br/>
///  Imagine <b>this</b> string is "This?is?a?sample???". <br/>
///  Imagine <b>flags</b> has <i>StrPlaceholderF_RightAlign</i> set. <br/>
///    Then for flag StrDigNumF_PadWithZeros:          set                   or not set:              <br/>
///    if #dwNum# ==      35 -> new string will be     "This0is0a0sample035" or "Thisisasample35"     <br/>
///    if #dwNum# ==   79235 -> new string will be     "This0is7a9sample235" or "Thisis7a9sample235"  <br/>
///    if #dwNum# == 8179235 -> new string will be     "This1is7a9sample235" or "This1is7a9sample235" <br/>
///       and the '8' will be skipped because there is no placeholder. <br/>
///  Now imagine <b>flags</b> has <i>StrPlaceholderF_LeftAlign</i> set instead right:                 <br/>
///    if #dwNum# ==      35 -> new string will be     "This3is5a0sample000" or "This3is5asample"     <br/>
///    if #dwNum# ==   79235 -> new string will be     "This7is9a2sample350" or "This7is9a2sample35"  <br/>
///    if #dwNum# == 8179235 -> new string will be     "This8is1a7sample923" or "This8is1a7sample923" <br/>
///       and the '5' will be skipped because there is no placeholder. </remarks>
///<param name="flags"> a StrNumT_* value and a combination of StrDigNumF_* and StrPlaceholderF_* values </param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::PlaceholderCharsToNumber( const Tchar *pstrPlaceholderChars, DWORD dwNum, strplaceholder_v flags )
{
    if( ! pstrPlaceholderChars ) return *this;
    if( !*pstrPlaceholderChars ) return *this;

    StringT str;
    DWORD   numtype = flags & StrNumT__Mask;

    if( flags & StrDigNumF_PadWithZeros )
    {
        int cplaceholders = 0;

        const Tchar *p = *this;

        while( *p )
            if( _strchr_( pstrPlaceholderChars, *(p++) ) )
                ++cplaceholders;

        if( cplaceholders == 0 )
            return *this;

        switch( numtype )
        {
            case StrNumT_Decimal     :  str.Sprintf( _Tstr("%0*u"), cplaceholders, (UINT) dwNum );  break;
            case StrNumT_Hexadecimal :  str.Sprintf( _Tstr("%0*x"), cplaceholders, (UINT) dwNum );  break;
            case StrNumT_Octal       :  str.Sprintf( _Tstr("%0*o"), cplaceholders, (UINT) dwNum );  break;
            default : return *this;
        }
    }
    else
    {
        switch( numtype )
        {
            case StrNumT_Decimal     :  str.Sprintf( _Tstr("%u"), (UINT) dwNum );  break;
            case StrNumT_Hexadecimal :  str.Sprintf( _Tstr("%x"), (UINT) dwNum );  break;
            case StrNumT_Octal       :  str.Sprintf( _Tstr("%o"), (UINT) dwNum );  break;
            default : return *this;
        }
    }

    return PlaceholderCharsToString( pstrPlaceholderChars, str, flags );
}

///<param name="nWord">
///  Which word to find.
///  <br/>
///  If 0 -> the word will be found where <b>idxStartParsing</b>
///  points in, or if <b>idxStartParsing</b> points on a separator character,
///  no word will be found (-1 will be returned then).
///  <br/>
///  If positive -> the n-th word that can be found when scanning forward will be returned.
///  E.g. <b>1</b> will return the current one, if <b>idxStartParsing</b> points to the *start* of a word, or
///  otherwise the next found word will be returned.
///  <br/>
///  If negative -> the n-th word that can be found when scanning backward will be returned.
///  E.g. <b>-1</b> will return the previous one.
///</param>
///<returns>
///  When scanning forward (<b>nWord</b> >= 0), returns the character index after the *end* of the found word. <br/>
///  When scanning backward (<b>nWord</b> < 0), returns the character index of the *start* of the found word. <br/>
///  -1 = word not found. <br/>
///  This allows easy iteration: <br/>
///  E.g. if you want to scan each word, call Findnextword with <b>nWord</b> set to 1 for forward scanning
///  or to -1 for backward scanning and set <b>idxStartParsing</b> to the return value of the previous call to Findnextword.
///</returns>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::Findnextword( _idx_& idxStart, _idx_& idxEnd, _idx_ idxStartParsing, _idx_ nWord, DWORD flags, const Tchar *pstrChars ) const
{
    idxStart = idxEnd = -1;

    _idx_    idx = idxStartParsing;
    _size_t_ len = length();
    if( idx < 0 || (_size_t_) idx > len ) return -1;

    const Tchar *ptr = *this;

    if( nWord == 0 )
        if( !isFindwordChar(ptr[idx], flags, pstrChars) )
            return -1;

    if( nWord > 0 )
    {
        bool isWordChar    = isFindwordChar(ptr[idx], flags, pstrChars);
        bool isStartOfWord = isWordChar &&
                             ( idx > 0 ? !isFindwordChar(ptr[idx - 1], flags, pstrChars) : true );
        if( isStartOfWord || !isWordChar )
            --nWord;
    }

    while( nWord >= 0 )
    {
        // Jump over words and separators till the beginning
        // of the desired word.
        // If #idxStartParsing# already pointed somewhere inside the desired
        // word, we must scan backwards to the beginning of the word.

        // Jump over separators.
        while( (_size_t_) idx < len ? !isFindwordChar(ptr[idx], flags, pstrChars) : false )
            ++idx;

        if( (_size_t_) idx >= len ) return -1;

        if( nWord == 0 )
        {
            // We have the desired word.
            // Now we must find the start of it.
            _idx_ idx_start = idx;
            while( idx_start > 0 ? isFindwordChar(ptr[idx_start - 1], flags, pstrChars) : false )
                --idx_start;

            // And its end.
            _idx_ idx_end = idx;
            while( (_size_t_) idx_end < len ? isFindwordChar(ptr[idx_end], flags, pstrChars) : false )
                ++idx_end;

            idxStart = idx_start;
            idxEnd   = idx_end;

            return idx_end;
        }

        // Jump over undesired word.
        while( (_size_t_) idx < len ? isFindwordChar(ptr[idx], flags, pstrChars) : false )
            ++idx;

        --nWord;
    }

    if( !isFindwordChar(ptr[idx], flags, pstrChars) )
        ++nWord;

    while( nWord <= 0 )
    {
        // Jump over words and separators till the beginning
        // of the desired word.
        // If #idxStartParsing# already pointed somewhere inside the desired
        // word, we must scan backwards to the beginning of the word.

        // Jump over separators.
        while( idx >= 0 ? !isFindwordChar(ptr[idx], flags, pstrChars) : false )
            --idx;

        if( idx < 0 ) return -1;

        if( nWord == 0 )
        {
            // We have the desired word.
            // Now we must find the start of it.
            _idx_ idx_start = idx;
            while( idx_start > 0 ? isFindwordChar(ptr[idx_start - 1], flags, pstrChars) : false )
                --idx_start;

            // And its end.
            _idx_ idx_end = idx;
            while( (_size_t_) idx_end < len ? isFindwordChar(ptr[idx_end], flags, pstrChars) : false )
                ++idx_end;

            idxStart = idx_start;
            idxEnd   = idx_end;

            return idx_start;
        }

        // Jump over undesired word.
        while( idx >= 0 ? isFindwordChar(ptr[idx], flags, pstrChars) : false )
            --idx;

        ++nWord;
    }

    return -1;
}

template< class baseString, typename Tchar >
bool StringT<baseString, Tchar>::isFindwordChar( Tchar ch, DWORD flags, const Tchar *pstrChars )
{
    if( !ch ) return false;

    bool search_word_chars = ( findword_WordChars == (flags & (findword_SeparatorChars | findword_WordChars)) );

    if( _strchr_(pstrChars, ch) ) return search_word_chars;
    if( flags & findword_inclSpaces ) if( _isspace_( ch ) ) return search_word_chars;

    return !search_word_chars;
}

///<summary> Converts single LineFeeds into
///  Carriage-Return+LineFeed sequences, but only if a LF does
///  not belong already to a CR+LF sequence. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::MakeCRLF()
{
    _idx_ i = 0;
    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    while( ptr[i] )
    {
        if( ptr[i] == _r ? ptr[i + 1] == _n : false )
        {
            // nothing to do since CR+LF already present
            i += 2;
        }
        else if( ptr[i] == _r )
        {
            // only a CR -> add a LF

            Insert( i + 1, _n );
            ptr = *this;

            i += 2;
        }
        else if( ptr[i] == _n )
        {
            // a LF found -> insert a CR before it

            Insert( i, _r );
            ptr = *this;

            i += 2;
        }
        else
            ++i;
    }

    return *this;
}

///<summary> Converts Carriage-Return+LineFeed sequences
///  into single LineFeeds.
///  This is used by Unix-like OS. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::MakeLF()
{
    _idx_ i = 0;
    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    while( ptr[i] )
    {
        if( ptr[i] == _r ? ptr[i + 1] == _n : false )
        {
            // CR+LF found -> remove the CR

            Delete( i, 1 );
            ptr = *this;

            ++i;
        }
        else if( ptr[i] == _r )
        {
            // only a CR -> change it to a LF

            Tchar *p = GetBuffer();
            if( p ) { p[i] = _n; ReleaseBuffer(); }
            ptr = *this;

            ++i;
        }
        else
            ++i;
    }

    return *this;
}

///<summary> Converts Carriage-Return+LineFeed sequences
///  into single Carriage-Returns.
///  This is used by RichEdit version 2.0 and higher
///  as well as Mac OS. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::MakeCR()
{
    _idx_ i = 0;
    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    while( ptr[i] )
    {
        if( ptr[i] == _r ? ptr[i + 1] == _n : false )
        {
            // CR+LF found -> remove the LF

            Delete( i + 1, 1 );
            ptr = *this;

            ++i;
        }
        else if( ptr[i] == _n )
        {
            // only a LF -> change it to a CR

            Tchar *p = GetBuffer();
            if( p ) { p[i] = _r; ReleaseBuffer(); }
            ptr = *this;

            ++i;
        }
        else
            ++i;
    }

    return *this;
}

///<remarks>
///  An empty row will also be interpreted as a row.
///  So for a completely empty string, this method will return 1. <br/>
///  Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)).
///</remarks>
template< class baseString, typename Tchar >
_size_t_ StringT<baseString, Tchar>::GetNumRows()
{
    const Tchar *ptr = *this;

    _size_t_ count = 1;
    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    for( ; *ptr ; ++ptr )
    {
        if( *ptr == _r ) {
            if( ptr[1] == _n )
                ++ptr;
            ++count;
        } else if( *ptr == _n ) {
            ++count;
        }
    }

    return count;
}

///<summary> Returns index of the row where character position is in. </summary>
///<returns> -1 if the char index is out of string range. </returns>
///<remkars> Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)). </remarks>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::GetRowindexFromChar( _idx_ idxChar )
{
    if( idxChar <  0 || (_size_t_) idxChar > length() ) return -1;
    if( idxChar == 0 ) return 0;

    const Tchar *ptr = *this;

    _idx_ idxRow = 0;
    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    for( --idxChar ; idxChar >= 0 ; --idxChar )
    {
        if( ptr[idxChar] == _n ) {
            if( idxChar > 0 )
                if( ptr[idxChar - 1] == _r )
                    --idxChar;
            ++idxRow;
        } else if( ptr[idxChar] == _r ) {
            ++idxRow;
        }
    }

    return idxRow;
}

///<summary> Returns starting character index of the specified row. </summary>
///<returns> -1 if the row index is out of string range. </returns>
///<remkars> Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)). </remarks>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::GetCharindexFromRow( _idx_ idxRow )
{
    if( idxRow <  0 ) return -1;
    if( idxRow == 0 ) return  0;

    const Tchar *ptr = *this;

    if( !*ptr )
        return idxRow == 0 ? 0 : -1;

    _idx_ idxChar = 0;
    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    for( ; ptr[idxChar] ; ++idxChar )
    {
        if( ptr[idxChar] == _r ) {
            if( ptr[idxChar + 1] == _n )
                ++idxChar;
            if( --idxRow == 0 )
                return ++idxChar;
        } else if( ptr[idxChar] == _n ) {
            if( --idxRow == 0 )
                return ++idxChar;
        }
    }

    return -1;
}

///<summary> Returns the character index at the start of the
///  row of text where the specified character index is in. </summary>
///<returns> -1 if the specified character index is out of string range. </returns>
///<remkars> Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)). </remarks>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::GetCharindexAtStartofrow( _idx_ idxChar )
{
    if( idxChar <  0 || (_size_t_) idxChar > length() ) return -1;
    if( idxChar == 0 ) return 0;

    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    for( --idxChar ; idxChar >= 0 ; --idxChar )
    {
        if( ptr[idxChar] == _r || ptr[idxChar] == _n )
            return ++idxChar;
    }

    return ++idxChar;
}

///<summary> Returns the character index after the last character of the
///  row of text where the specified character index is in. </summary>
///<returns> -1 if the specified character index is out of string range. </returns>
///<remkars> The returned index will point the the CR or LF character at the end of the row. <br/>
///  Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)). </remarks>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::GetCharindexAfterEndofrow( _idx_ idxChar )
{
    if( idxChar < 0 || (_size_t_) idxChar > length() ) return -1;

    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    for( ; ptr[idxChar] ; ++idxChar )
    {
        if( ptr[idxChar] == _r || ptr[idxChar] == _n )
            return idxChar;
    }

    return idxChar;
}

///<summary> Moves <b>nRowOffset</b> rows up or down beginning at the current row
///  where <b>idxChar</b> character index is in and returns the start of row
///  char index. </summary>
///<returns> -1 if the specified character index is out of string range. </returns>
///<remkars> Handles all 3 kinds of text types (windows (CRLF), linux (LF), mac (CR)). </remarks>
template< class baseString, typename Tchar >
_idx_ StringT<baseString, Tchar>::GetStartcharindexOffsetRow( _idx_ idxChar, int nRowOffset, _size_t_ overrideLength )
{
    if( idxChar < 0 || (_size_t_) idxChar > overrideLength ) return -1;

    const Tchar *ptr = *this;

    Tchar _r = _Tchar('\r');
    Tchar _n = _Tchar('\n');

    if( nRowOffset <= 0 )
    {
        if( idxChar == 0 ) return nRowOffset == 0 ? 0 : -1;

        for( --idxChar ; idxChar >= 0 ; --idxChar )
        {
            if( ptr[idxChar] == _n ) {
                if( nRowOffset++ == 0 )
                    return ++idxChar;
                if( idxChar > 0 )
                    if( ptr[idxChar - 1] == _r )
                        --idxChar;
            } else if( ptr[idxChar] == _r ) {
                if( nRowOffset++ == 0 )
                    return ++idxChar;
            }
        }

        ++idxChar;
    }
    else
    {
        for( ; ptr[idxChar] ; ++idxChar )
        {
            if( ptr[idxChar] == _r ) {
                if( ptr[idxChar + 1] == _n )
                    ++idxChar;
                if( --nRowOffset == 0 )
                    return ++idxChar;
            } else if( ptr[idxChar] == _n ) {
                if( --nRowOffset == 0 )
                    return ++idxChar;
            }
        }
    }

    return nRowOffset == 0 ? idxChar : -1;
}

///<summary>
///  Increases indentation by inserting <b>cColumns</b> number of space or tab
///  characters into each line of text.
///</summary>
///<remarks>
///  If the <b>iEndRow</b> index points to the *first* character of
///  a line of text, then that line will *not* be included in the
///  indentation. If it points to any following character in that
///  line of text, then the line of text *will* be included in the
///  indentation.
///</remarks>
///<param name="cColumns">
///  Number of column chars to decrease the indentation. This is counted in normal
///  single space characters, not in tab characters.
///</param>
///<param name="cTab">
///  Width of a tab character in number of normal single space characters.
///</param>
///<param name="bAllowTabs">
///  If for the indentation it is allowed to use tab characters (besides of space chars)
///  or if only space chars are allowed.
///</param>
///<param name="iStartChar">
///  Index of the first character where to begin indenting. 0 = comletely from beginning.
///</param>
///<param name="iEndChar">
///  Index (exclusive) of the character after the last character
///  that should be involved in the indenting process.
///  -1 = run till last character in the string.
///</param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Indent( _size_t_ cColumns, _size_t_ cTab, bool bAllowTabs, _idx_ iStartChar, _idx_ iEndChar )
{
    if( cColumns < 0 || cTab <= 0 ) return *this;
    if( cColumns == 0 ) return *this;

    _size_t_ len = length();

    if( !CheckIndexRange_Special( iStartChar, iEndChar, len, 0 ) )
        return *this;

    iStartChar = GetCharindexAtStartofrow( iStartChar );
    if( iStartChar < 0 ) return *this; // just for savety. should never occur

    _idx_ iCurr = iStartChar;
    Tchar *ptr = (Tchar*) (const Tchar*) *this;   // be careful in here, when you may write to the buffer

    Tchar _r   = _Tchar('\r');
    Tchar _n   = _Tchar('\n');
    Tchar _t   = _Tchar('\t');
    Tchar _spc = _Tchar(' ');

    bool gotbuffer = false;

    while( ptr[iCurr] )
    {
        // First we check if the current line of text only consists of
        // space typical characters. If so, we do not indent the row.
        bool bOnlySpacesInRow = true;
        _idx_ iEndSpaces;           // character position in the string where the leading spaces sequence ended
        _idx_ nColumnPos = 0;       // and the corresponding column position (with expanded tabs counted)
        for( iEndSpaces = iCurr ; ptr[iEndSpaces] ; ++iEndSpaces )
        {
            if( ptr[iEndSpaces] == _r || ptr[iEndSpaces] == _n )
                break;

            if( !_isspace_( ptr[iEndSpaces] ) )
                { bOnlySpacesInRow = false; break; }

            if( ptr[iEndSpaces] == _t )
                nColumnPos = ((nColumnPos + cTab) / cTab) * cTab;
            else
                ++nColumnPos;
        }

        _idx_ cInsertedChars = 0;    // could be negative if chars have been removed

        if( !bOnlySpacesInRow )
        {
            // Now we indent the current line of text.

            _idx_ nDesiredColumnPos = nColumnPos + cColumns;

            if( bAllowTabs )
            {
                _ssize_t_ cDesiredTabs  = nDesiredColumnPos / cTab;
                _ssize_t_ cDesiredChars = cDesiredTabs + (nDesiredColumnPos - cDesiredTabs * cTab);

                _ssize_t_ cInsert = cDesiredChars - (iEndSpaces - iCurr);
                if( cInsert > 0 )
                {
                    if( !( ptr = GetBuffer( len + cInsert ) ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }
                else if( cInsert < 0 )
                {
                    if( !gotbuffer )
                        if( !( ptr = GetBuffer() ) )
                            { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }

                if( !gotbuffer )
                {
                    if( !( ptr = GetBuffer() ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                }

                _idx_ iEndPos = iCurr + cDesiredTabs;
                _idx_ i = iCurr;
                while( i < iEndPos ) ptr[i++] = _t;
                iEndPos = iCurr + cDesiredChars;
                while( i < iEndPos ) ptr[i++] = _spc;

                cInsertedChars = cInsert;
            }
            else
            {
                _ssize_t_ cInsert = nDesiredColumnPos - (iEndSpaces - iCurr);
                if( cInsert > 0 )
                {
                    if( !( ptr = GetBuffer( len + cInsert ) ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }
                else if( cInsert < 0 )
                {
                    if( !gotbuffer )
                        if( !( ptr = GetBuffer() ) )
                            { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }

                if( !gotbuffer )
                {
                    if( !( ptr = GetBuffer() ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                }

                _idx_ iEndPos = iCurr + nDesiredColumnPos;
                _idx_ i = iCurr;
                while( i < iEndPos ) ptr[i++] = _spc;

                cInsertedChars = cInsert;
            }
        }

        iEndChar += cInsertedChars;

        // go on to the next row
        iCurr = GetStartcharindexOffsetRow( iCurr, 1, len );
        if( iCurr < 0 ) break;

        if( iCurr >= iEndChar ) break;
    }

    if( gotbuffer )
        ReleaseBuffer();

    return *this;
}

///<summary>
///  Decreases indentation by removing <b>cColumns</b> number of space or tab
///  characters into each line of text.
///</summary>
///<remarks>
///  If the <b>iEndRow</b> index points to the *first* character of
///  a line of text, then that line will *not* be included in the
///  indentation. If it points to any following character in that
///  line of text, then the line of text *will* be included in the
///  indentation.
///</remarks>
///<param name="cColumns">
///  Number of column chars to decrease the indentation. This is counted in normal
///  single space characters, not in tab characters.
///</param>
///<param name="cTab">
///  Width of a tab character in number of normal single space characters.
///</param>
///<param name="bAllowTabs">
///  If for the indentation it is allowed to use tab characters (besides of space chars)
///  or if only space chars are allowed.
///</param>
///<param name="iStartChar">
///  Index of the first character where to begin indenting. 0 = comletely from beginning.
///</param>
///<param name="iEndChar">
///  Index (exclusive) of the character after the last character
///  that should be involved in the indenting process.
///  -1 = run till last character in the string.
///</param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Unindent( _size_t_ cColumns, _size_t_ cTab, bool bAllowTabs, _idx_ iStartChar, _idx_ iEndChar )
{
    if( cColumns < 0 || cTab <= 0 ) return *this;
    if( cColumns == 0 ) return *this;

    _size_t_ len = length();

    if( !CheckIndexRange_Special( iStartChar, iEndChar, len, 0 ) )
        return *this;

    iStartChar = GetCharindexAtStartofrow( iStartChar );
    if( iStartChar < 0 )return *this; // just for savety. should never occur

    _idx_ iCurr = iStartChar;
    Tchar *ptr = (Tchar*) (const Tchar*) *this;   // be careful in here, when you may write to the buffer

    Tchar _r   = _Tchar('\r');
    Tchar _n   = _Tchar('\n');
    Tchar _t   = _Tchar('\t');
    Tchar _spc = _Tchar(' ');

    bool gotbuffer = false;

    while( ptr[iCurr] )
    {
        _idx_ iEndSpaces;           // character position in the string where the leading spaces sequence ended
        _idx_ nColumnPos = 0;       // and the corresponding column position (with expanded tabs counted)
        for( iEndSpaces = iCurr ; ptr[iEndSpaces] ; ++iEndSpaces )
        {
            if( ptr[iEndSpaces] == _r || ptr[iEndSpaces] == _n )
                break;

            if( !_isspace_( ptr[iEndSpaces] ) )
                break;

            if( ptr[iEndSpaces] == _t )
                nColumnPos = ((nColumnPos + cTab) / cTab) * cTab;
            else
                ++nColumnPos;
        }

        _idx_ cInsertedChars = 0;    // could be negative if chars have been removed

        // Now we indent the current line of text.

        _idx_ nDesiredColumnPos = nColumnPos - cColumns;
        if( nDesiredColumnPos < 0 ) nDesiredColumnPos = 0;

        if( nDesiredColumnPos != nColumnPos )
        {
            if( bAllowTabs )
            {
                _ssize_t_ cDesiredTabs  = nDesiredColumnPos / cTab;
                _ssize_t_ cDesiredChars = cDesiredTabs + (nDesiredColumnPos - cDesiredTabs * cTab);

                _ssize_t_ cInsert = cDesiredChars - (iEndSpaces - iCurr);
                if( cInsert > 0 )
                {
                    if( !( ptr = GetBuffer( len + cInsert ) ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }
                else if( cInsert < 0 )
                {
                    if( !gotbuffer )
                        if( !( ptr = GetBuffer() ) )
                            { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }

                if( !gotbuffer )
                {
                    if( !( ptr = GetBuffer() ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                }

                _idx_ iEndPos = iCurr + cDesiredTabs;
                _idx_ i = iCurr;
                while( i < iEndPos ) ptr[i++] = _t;
                iEndPos = iCurr + cDesiredChars;
                while( i < iEndPos ) ptr[i++] = _spc;

                cInsertedChars = cInsert;
            }
            else
            {
                _ssize_t_ cInsert = nDesiredColumnPos - (iEndSpaces - iCurr);
                if( cInsert > 0 )
                {
                    if( !( ptr = GetBuffer( len + cInsert ) ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }
                else if( cInsert < 0 )
                {
                    if( !gotbuffer )
                        if( !( ptr = GetBuffer() ) )
                            { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                    memmove( ptr + iEndSpaces + cInsert, ptr + iEndSpaces, (len + 1 - iEndSpaces) * sizeof(Tchar) );
                    len += cInsert;
                }

                if( !gotbuffer )
                {
                    if( !( ptr = GetBuffer() ) )
                        { if( gotbuffer ) ReleaseBuffer(); return *this; }
                    gotbuffer = true;
                }

                _idx_ iEndPos = iCurr + nDesiredColumnPos;
                _idx_ i = iCurr;
                while( i < iEndPos ) ptr[i++] = _spc;

                cInsertedChars = cInsert;
            }
        }

        iEndChar += cInsertedChars;

        // go on to the next row
        iCurr = GetStartcharindexOffsetRow( iCurr, 1, len );
        if( iCurr < 0 ) break;

        if( iCurr >= iEndChar ) break;
    }

    if( gotbuffer )
        ReleaseBuffer();

    return *this;
}

///<summary> The string is expected to be a floating point value, both either
///  with mantisse/exponent syntax or not. </summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::BeautifyFloatValue( DWORD flags )
{
    _idx_ ix = Strpbrk( _Tstr("eE ") );

    if( ix >= 0 ? ( (*this)[ix] == _Tchar('e') || (*this)[ix] == _Tchar('E') ) : false )
    {
        // The float value has the form of "....e..."

        _idx_ i = ix + 1;
        if( (*this)[i] == _Tchar('-') || (*this)[i] == _Tchar('+') )
            ++i;

        while( (*this)[i] == _Tchar('0') )
            Delete( i );

        if( !_isdigit_( (*this)[i] ) )
            Insert( i, _Tchar('0') );
    }

    _size_t_ len = length();
    if( ix < 0 ) ix = len;

    // Now we get the current language's comma value.
    Tchar comma = _Tchar('.');
    if( !(flags & BeautyF_CommaIsDot) )
    {
        Tchar s[10];
        _sprintf_( s, _Tstr("%f"), 1.0 );
        comma = s[1];
        if( _isalnum_(comma) ) comma = _Tchar('.');
    }

    Tchar szcomma[2] = { comma, (Tchar) 0 };

    StringT front = Left( ix );
    _idx_ iComma = front.Strchr( comma );
    if( iComma >= 0 )
    {
        front.TrimRight( _Tstr("0") );

        if( flags & BeautyF_TrimComma )
        {
            if( front.EndsWith(szcomma) )
                front.TrimRight( szcomma );
        }
        else
        {
            if( flags & BeautyF_AutoAddCommaZero )
                if( front.EndsWith(szcomma) ) front += _Tchar('0');
        }

        if( flags & BeautyF_2DigitsBeforeComma ) {
            if     ( iComma == 0 ) front = _Tstr("00") + front;
            else if( iComma == 1 ) front = _Tchar('0') + front;
        }
    }
    else
    {
        if( flags & BeautyF_AutoAddCommaZero ) {
            front += comma;
            front += _Tchar('0');
        }

        if( flags & BeautyF_2DigitsBeforeComma ) {
            if     ( ix == 0 ) front = _Tstr("00") + front;
            else if( ix == 1 ) front = _Tchar('0') + front;
        }
    }

    if( front.IsEmpty() ) front = _Tstr("0");

    if( (_idx_) len > ix )
        *this = front + Right( len - ix );
    else
        *this = front;

    return *this;
}

///<summary> Expects the string to be a version string with several numbers
///  and returns it in a more human readable version. It allows also removing the build number.
///  The numbers must be decimal integer values and can be separated by any
///  non-decimal-digit characters. </summary>
///<param name="nMinDisplayNums"> Minimum number of numbers to display.
///  So if 2, "1, 0, 0, 1" will be converted into "1.0" </param>
///<param name="nMaxDisplayNums"> Maximum number of numbers to display.
///  So if 3, "1, 0, 0, 1" will be converted into "1.0".
///  But if 4, "1, 0, 0, 1" will be converted into "1.0.0.1". </param>
///<param name="nVersionNrNums"> The number of numbers the whole version information
///  consists of, independent from how many finally will be displayed.
///  The samples shown in params <b>nMinDisplayNums</b> and <b>nMinDisplayNums</b>
///  are using a value <b>nVersionNrNums</b> of 4. But you can set it to any value desired. </param>
///<param name="pstrNewSep"> When the beautified version number is generated, this
///  separation string will be used as separator string between 2 consecutive numbers. </param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::BeautifyVersionNr( int nMinDisplayNums, int nMaxDisplayNums, int nVersionNrNums, const Tchar *pstrNewSep )
{
    if( nMaxDisplayNums > nVersionNrNums ) nMaxDisplayNums = nVersionNrNums;
    if( nMinDisplayNums > nMaxDisplayNums ) nMinDisplayNums = nMaxDisplayNums;
    if( nMaxDisplayNums <= 0 || nVersionNrNums <= 0 )
        { Empty(); return *this; }

    int mem_numbers[10];
    bool alloc_mem = nMaxDisplayNums > _countof(mem_numbers);
    int *alloc_numbers = alloc_mem ? (int*) malloc( nMaxDisplayNums * sizeof(int) ) : NULL;

    if( alloc_mem && !alloc_numbers )
        { Empty(); return *this; }

    int *pnumbers = alloc_numbers ? alloc_numbers : mem_numbers;
    // Now #pnumbers# is the array where to write the numbers into.

    const Tchar *ptr = *this;

    _idx_ iparse = 0;
    int i;
    for( i = 0 ; i < nMaxDisplayNums ; ++i )
    {
        pnumbers[i] = 0;

        _idx_ iS, iE;
        iparse = Findnextword( iS, iE, iparse, 1, findword_WordChars, _Tstr("0123456789") );

        if( iparse >= 0 && iS >= 0 )
            pnumbers[i] = _stoi_( ptr + iS );
    }

    int cDisplayNums = nMinDisplayNums;
    for( i = nMaxDisplayNums - 1 ; i >= nMinDisplayNums ; --i )
        if( pnumbers[i] != 0 )
            { cDisplayNums = i + 1; break; }

    StringT s;
    for( i = 0 ; i < cDisplayNums ; ++i )
    {
        if( !s.IsEmpty() ) s += pstrNewSep;
        s += fromInt( pnumbers[i] );
    }

    *this = s;

    if( alloc_numbers ) free( alloc_numbers );
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::ParentPath()
{
    Tchar *pdest = GetBuffer();
    if( !pdest ) return *this;

    _parentpath_( pdest );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::RootPath()
{
    const Tchar *pstr = *this;
    _size_t_ len = length();

    _size_t_ newlen = _rootpathlen_( pstr );
    if( len == newlen ) return *this;

    Tchar *pdest = GetBuffer( newlen );
    if( !pdest ) return *this;

    _rootpath_( pdest, newlen + 1 );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::StripPath()
{
    Tchar *pdest = GetBuffer();
    if( !pdest ) return *this;

    _strippath_( pdest );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::AsPath()
{
    const Tchar *pstr = *this;
    _size_t_ len = length();

    _size_t_ newlen = _aspathlen_( pstr );
    if( len == newlen ) return *this;

    Tchar *pdest = GetBuffer( newlen );
    if( !pdest ) return *this;

    _aspath_( pdest, newlen + 1 );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::OsPath()
{
    Tchar *pdest = GetBuffer();
    if( !pdest ) return *this;

    _ospath_( pdest );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::UnixPath()
{
    Tchar *pdest = GetBuffer();
    if( !pdest ) return *this;

    _unixpath_( pdest );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::WindowsPath()
{
    Tchar *pdest = GetBuffer();
    if( !pdest ) return *this;

    _windowspath_( pdest );

    ReleaseBuffer();
    return *this;
}

// This imagines #this# to contain a path (maybe with a filename as well).
// The #this# path may be absolute or relative.
// If #this# contains a path that is a subdirectory of #pstrRelPath#,
// then the resulting string will be a path that is relative to the
// #pstrRelPath#, which means that it is exactly that subdirectory (w/ or w/o file).
// It is done that way, if #this# path starts exactly with the #pstrRelPath#,
// then it will be made relative.
// If making relative is not possible, then the resulting string will be
// exactly #this# again.
//  > pbSucceeded : will be set to TRUE if it is made relative or FALSE otherwise
//                  but only if this pointer here is not NULL.
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::GetSubpathRelativeTo( const Tchar *pstrRelPath, bool *pbSucceeded )
{
    if( pbSucceeded ) *pbSucceeded = false;
    if( ! pstrRelPath ) return *this;
    if( !*pstrRelPath ) return *this;

    Tchar path[_MAX_PATH_ + 1];
    _aspath_( _strcpy_( path, pstrRelPath, _countof(path) ), _countof(path) );
    if( !_pathishere_( *this, path ) ) return *this;

    if( pbSucceeded ) *pbSucceeded = true;
    return Right( length() - _strlen_(path) );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetSystemerrorMessage( DWORD errorcode, DWORD langid )
{
    Tchar *pMsgBuf = NULL;

    DWORD ret = privateString::char_T<Tchar>::
                __FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                                 FORMAT_MESSAGE_FROM_SYSTEM |
                                 FORMAT_MESSAGE_IGNORE_INSERTS,
                                 NULL,
                                 errorcode,
                                 langid,
                                 (Tchar*) &pMsgBuf,
                                 0,
                                 NULL );

    if( ret > 0 && pMsgBuf ) *this = pMsgBuf;
    else Empty();

    if( pMsgBuf )
        LocalFree( pMsgBuf );

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Loadstring( UINT resid )
{
    __if_exists    ( baseString::LoadString ) { baseString::LoadString( resid ); return *this; }
    __if_not_exists( baseString::LoadString ) { Loadstring( Get_Resource_Instance(), resid ); return *this; }
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Loadstring( HINSTANCE hInstance, UINT resid )
{
    __if_exists( baseString::LoadString )
    {
        baseString::LoadString( hInstance, resid );
        return *this;
    }

    __if_not_exists( baseString::LoadString )
    {
        Empty();

        if( resid == 0 ) return *this;

        _size_t_ step = 256;
        _size_t_ count = step;
        _size_t_ len = 0;

        for( ;; )
        {
            Tchar *ptr = GetBuffer( count - 1 );
            if( !ptr ) return *this;

            len = (_size_t_) privateString::char_T<Tchar>::__LoadString( hInstance, resid, ptr, count );
            if( len < count - 1 ) break;

            count += step;
        }

        ReleaseBuffer();
        return *this;
    }
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Loadstring( HINSTANCE hInstance, UINT resid, LANGID language )
{
    Empty();

    if( resid == 0 ) return *this;

    #ifndef _WIN32_WCE
    HRSRC hRes = ::FindResourceEx( hInstance, RT_STRING, MAKEINTRESOURCE((resid >> 4) + 1), language );
    #else
    HRSRC hRes = ::FindResource( hInstance, RT_STRING, MAKEINTRESOURCE((resid >> 4) + 1) );
    #endif

    #ifndef _WIN32_WCE
    if( !hRes && SUBLANGID(language) == SUBLANG_NEUTRAL && language != LANG_NEUTRAL )
    {
        // If we have no sublanguage specified but we desire a specific language
        // (so we only specified the primary language), we enumerate all sublanguages
        // until we get a resource handle.
        LANGID primlang = PRIMARYLANGID(language);
        WORD maxsublang = SUBLANGID(0xffff);
        for( WORD sublang = 0 ; sublang <= maxsublang && !hRes ; ++sublang )
            hRes = ::FindResourceEx( hInstance, RT_STRING, MAKEINTRESOURCE((resid >> 4) + 1), MAKELANGID(primlang, sublang) );
    }
    #endif

    if( !hRes ) return *this;

    HGLOBAL hGlo = ::LoadResource( hInstance, hRes );
    if( !hGlo ) return *this;

    // A string resource contains a WORD that indicates the length in chars
    // followed by the string in WCHARs (*not* null-terminated).
    PBYTE ptr = (PBYTE) ::LockResource( hGlo );
    if( !ptr ) return *this;

    DWORD siz = ::SizeofResource( hInstance, hRes );
    PBYTE ptrend = ptr + siz;
    UINT idx = resid & 0xf;

    while( idx > 0 && ptr < ptrend )
    {
        ptr += sizeof(WORD) + (*(WORD*) ptr) * sizeof(WCHAR);
        --idx;
    }

    if( ptr >= ptrend ) return *this;

    WORD len = *(WORD*) ptr;
    if( len <= 0 ) return *this;

    PCWSTR pstrRes = (PCWSTR) (ptr + sizeof(WORD));

    Strcat( pstrRes, (_size_t_) len );
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetWindowtext( HWND hwnd )
{
    _size_t_ len = (_size_t_) ::GetWindowTextLength( hwnd );
    if( !hwnd || len <= 0 )
        { Empty(); return *this; }

    Tchar *pdest = GetBuffer( len );
    if( !pdest ) return *this;

    // #::GetWindowText()# will *always* append the null-terminator
    // even if the string buffer is too small.
    // The specified buffer size is counted in characters *including*
    // the null-terminator.
    if( privateString::char_T<Tchar>::__GetWindowText( hwnd, pdest, len + 1 ) == 0 )
        { ReleaseBuffer(); return *this; }

    ReleaseBuffer();
    return *this;
}

///<summary>
///  <b>Wm_Gettext()</b> acts similar like <b>GetWindowtext()</b>, except that
///  it sends the messages WM_GETTEXTLENGTH and WM_GETTEXT directly. Although
///  GetWindowText() sens a WM_GETTEXT as part of its implementation, there
///  is a difference if the window whos text is to be retrieved is in another
///  process. In that case, sending WM_GETTEXT directly will work because
///  Windows will automatically do the necessary address space translation.
///  (To be exact, it is not an address space translation. Windows uses memory
///  mapped files in the background to copy the text contents automatically.)
///</summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::Wm_Gettext( HWND hwnd )
{
    if( !hwnd )
        { Empty(); return *this; }

    _size_t_ len = (_size_t_) privateString::char_T<Tchar>::__SendMessage( hwnd, WM_GETTEXTLENGTH, 0, 0 );
    if( len <= 0 )
        { Empty(); return *this; }

    Tchar *pdest = GetBuffer( len );
    if( !pdest ) return *this;

    // The specified buffer size is counted in characters *including*
    // the null-terminator.
    if( privateString::char_T<Tchar>::__SendMessage( hwnd, WM_GETTEXT, (WPARAM) (len + 1), (LPARAM) pdest ) == 0 )
        { ReleaseBuffer(); return *this; }

    ReleaseBuffer();
    return *this;
}

///<param name="fuFlags"> combination of SMTO_* flags as for <b>::SendMessageTimeout()</b> </param>
///<returns> False on timeout or memory error or message error, otherwise always true. </returns>
template< class baseString, typename Tchar >
bool StringT<baseString, Tchar>::Wm_Gettext_timeout( HWND hwnd, UINT timeout, UINT fuFlags )
{
    if( !hwnd )
        { Empty(); return true; }

    DWORD_PTR result = 0;
    LRESULT lr = privateString::char_T<Tchar>::__SendMessageTimeout( hwnd, WM_GETTEXTLENGTH, 0, 0, fuFlags, timeout, &result );
    if( !lr )
        { Empty(); return false; }
    _size_t_ len = (_size_t_) result;
    if( len <= 0 )
        { Empty(); return true; }

    Tchar *pdest = GetBuffer( len );
    if( !pdest ) return false;

    // The specified buffer size is counted in characters *including*
    // the null-terminator.
    // Note: Because WM_GETTEXT uses a pointer, we do not use SendMessageTimeout.
    //  And it is not necessary because we used it in the call above already.
    //  Let's hope that between these 2 message sends, the destination process
    //  does not get blocked suddenly.
    if( privateString::char_T<Tchar>::__SendMessage( hwnd, WM_GETTEXT, (WPARAM) (len + 1), (LPARAM) pdest ) == 0 )
        { ReleaseBuffer(); return false; }

    ReleaseBuffer();
    return true;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetClassname( HWND hwnd )
{
    if( !hwnd )
        { Empty(); return *this; }

    Tchar s[512];
    // #::GetClassName()# will *always* append the null-terminator
    // even if the string buffer is too small.
    // The specified buffer size is counted in characters *including*
    // the null-terminator.
    if( privateString::char_T<Tchar>::__GetClassName( hwnd, s, _countof(s) ) == 0 )
        return *this;

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetListboxItemtext( HWND hwnd, int iItem )
{
    if( !hwnd )
        { Empty(); return *this; }

    _size_t_ len = (_size_t_) privateString::char_T<Tchar>::__SendMessage( hwnd, LB_GETTEXTLEN, iItem, 0 );
    if( len <= 0 || (int) len == LB_ERR ) return *this;

    Tchar *pdest = GetBuffer( len );
    if( !pdest ) return *this;

    privateString::char_T<Tchar>::__SendMessage( hwnd, LB_GETTEXT, iItem, (LPARAM) pdest );
    pdest[len] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetComboboxLBItemtext( HWND hwnd, int iItem )
{
    if( !hwnd )
        { Empty(); return *this; }

    _size_t_ len = (_size_t_) privateString::char_T<Tchar>::__SendMessage( hwnd, CB_GETLBTEXTLEN, iItem, 0 );
    if( len <= 0 || (int) len == CB_ERR ) return *this;

    Tchar *pdest = GetBuffer( len );
    if( !pdest ) return *this;

    privateString::char_T<Tchar>::__SendMessage( hwnd, CB_GETLBTEXT, iItem, (LPARAM) pdest );
    pdest[len] = (Tchar) 0;

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetTreeviewItemtext( HWND hwndTV, HTREEITEM htiItem )
{
    if( !hwndTV || !htiItem )
        { Empty(); return *this; }

    TCHAR s[512] = _T("");

    TVITEM tvi;
    memset( &tvi, 0, sizeof(tvi) );
    tvi.mask = TVIF_HANDLE | TVIF_TEXT;
    tvi.hItem = htiItem;
    tvi.pszText = s;
    tvi.cchTextMax = _countof(s) - 1;

    if( !TreeView_GetItem( hwndTV, &tvi ) )
        { Empty(); return *this; }

    s[_countof(s) - 1] = (TCHAR) 0;
    *this = s;

    return *this;
}

#ifndef _WIN32_WCE

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetCurrentdirectory()
{
    _size_t_ zlen = (_size_t_) privateString::char_T<Tchar>::__GetCurrentDirectory( 0, (Tchar*) NULL );
    // The returned size is counted in characters and *includes* the
    // null-terminator.
    if( zlen <= 1 )
        { Empty(); return *this; }

    Tchar *pdest = GetBuffer( zlen - 1 );
    if( !pdest ) return *this;

    if( 0 == privateString::char_T<Tchar>::__GetCurrentDirectory( zlen, pdest ) )
        { ReleaseBuffer(); return *this; }

    ReleaseBuffer();
    return *this;
}

#endif

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetModulefilename( HMODULE hModule )
{
    Tchar s[_MAX_PATH_ + 1];
    // #::GetModuleFileName()# will *not* append a null-terminator
    // if the passed string buffer is too short (otherwise it will).
    if( 0 == privateString::char_T<Tchar>::__GetModuleFileName( hModule, s, _countof(s) ) )
        { Empty(); return *this; }
    s[_MAX_PATH_] = (Tchar) 0;  // this *is* necessary
    // As I figured out, the returned path could be not case-correct
    // for case-insensitive file systems on some operating systems.
    // On WindowsXP, it *is* case-correct.

    _ospath_( s );

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetFullpathname( const Tchar *pstrRelFile )
{
    if( _isstrempty_( pstrRelFile ) )
        { Empty(); return *this; }

    Tchar fullfile[_MAX_PATH_ + 1];
    Tchar *filename;

    if( !privateString::char_T<Tchar>::__GetFullPathName( pstrRelFile, _countof(fullfile), fullfile, &filename ) )
        { Empty(); return *this; }

    _ospath_( fullfile );

    *this = fullfile;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetUsername()
{
    Tchar s[UNLEN + 1];
    DWORD zlen = (DWORD) _countof(s);

    if( 0 == privateString::char_T<Tchar>::__GetUserName( s, &zlen ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

#ifndef _WIN32_WCE

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetComputername()
{
    Tchar s[MAX_COMPUTERNAME_LENGTH + 1 + 128]; // +128 for savety portion
    DWORD zlen = (DWORD) _countof(s);

    if( 0 == privateString::char_T<Tchar>::__GetComputerName( s, &zlen ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetWindowsdirectory()
{
    Tchar s[_MAX_PATH_ + 1];

    if( 0 == privateString::char_T<Tchar>::__GetWindowsDirectory( s, _countof(s) ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetSystemdirectory()
{
    Tchar s[_MAX_PATH_ + 1];

    if( 0 == privateString::char_T<Tchar>::__GetSystemDirectory( s, _countof(s) ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetSpecialFolderPath( HWND hwndOwner, int nFolderCSIDL, bool bCreate )
{
    Tchar s[MAX_PATH + 1];

    if( !privateString::char_T<Tchar>::__SHGetSpecialFolderPath( hwndOwner, s, nFolderCSIDL, bCreate ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetProgramFilesFolder()
{
    Tchar s[MAX_PATH + 1];

    if( privateString::char_T<Tchar>::__SHGetSpecialFolderPath( NULL, s, CSIDL_PROGRAM_FILES, FALSE ) )
    {
        *this = s;
        return *this;
    }

    // If retrieving CSIDL_PROGRAM_FILES failed,
    // we try to get it from the registry directly.
    return RegQueryStringVal( HKEY_LOCAL_MACHINE,
                              _Tstr("ProgramFilesDir"),
                              _Tstr("Software\\Microsoft\\Windows\\CurrentVersion") );   // key does not support '/' path delimiters
}

#endif

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetTempDir()
{
    Tchar s[_MAX_PATH_ + 1];

    if( 0 == privateString::char_T<Tchar>::__GetTempPath( _countof(s), s ) )
        { Empty(); return *this; }

    *this = s;

    return *this;
}

///<remarks> If <b>pstrSubkey</b> is NULL, then it is imagined that <b>hkey</b>
///  is already the key where you want to retrieve the string value from. <br/>
///  Don't forget that <b>pstrSubkey</b> does not support '/' chars as path delimiters.
///  Use '\\' chars only. </remarks>
template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::RegQueryStringVal( HKEY hkey, const Tchar *pstrValue, const Tchar *pstrSubkey )
{
    HKEY hk = NULL;
    if( pstrSubkey ) {
        LONG err = privateString::char_T<Tchar>::
                   __RegOpenKeyEx( hkey,
                                   pstrSubkey,   // does not support '/' path delimiters
                                   0,
                                   KEY_QUERY_VALUE,
                                   &hk );
        if( err != ERROR_SUCCESS )
            { Empty(); return *this; }
    } else {
        hk = hkey;
    }

    DWORD dwType = 0;
    DWORD dwSize = 0;

    LONG err = privateString::char_T<Tchar>::
               __RegQueryValueEx( hk,
                                  pstrValue,
                                  NULL,
                                  &dwType,
                                  NULL,
                                  &dwSize );
    if( err != ERROR_SUCCESS )
        { RegCloseKey( hkey ); Empty(); return *this; }

    if( dwType != REG_SZ && dwType != REG_EXPAND_SZ )
        { RegCloseKey( hkey ); Empty(); return *this; }

    if( dwSize <= sizeof(Tchar) )
        { RegCloseKey( hkey ); Empty(); return *this; }

    _size_t_ zlen = (_size_t_) dwSize / sizeof(Tchar);

    Tchar *pdest = GetBuffer( zlen - 1 );
    if( !pdest )
        { RegCloseKey( hkey ); Empty(); return *this; }

    err = privateString::char_T<Tchar>::
          __RegQueryValueEx( hk,
                             pstrValue,
                             NULL,
                             NULL,
                             (PBYTE)pdest,
                             &dwSize );
    if( err != ERROR_SUCCESS )
        { RegCloseKey( hkey ); ReleaseBuffer(); Empty(); return *this; }

    RegCloseKey( hkey );

    ReleaseBuffer();
    return *this;
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::GetLocaleinfo( LCID locale, LCTYPE lctype )
{
    int zlen = privateString::char_T<Tchar>::__GetLocaleInfo( locale, lctype, NULL, 0 );
    if( zlen <= 1 )
        { Empty(); return *this; }

    Tchar *pdest = GetBuffer( zlen - 1 );
    if( !pdest )
        { Empty(); return *this; }

    if( privateString::char_T<Tchar>::__GetLocaleInfo( locale, lctype, pdest, zlen ) == 0 )
        { pdest[0] = (Tchar) 0; ReleaseBuffer(); Empty(); return *this; }

    pdest[zlen - 1] = (Tchar) 0;    // just for savety. is not necessary
    ReleaseBuffer();
    return *this;
}

#ifndef _WIN32_WCE

template< class baseString, typename Tchar >
StringT<baseString, Tchar>& StringT<baseString, Tchar>::StrftimeEx( const Tchar *format, const struct tm *ptm, int nanos, LCID Locale )
{
    if( !format || !ptm )
        { Empty(); return *this; }

    StringT str;

    const Tchar *p0 = format;
    const Tchar *p  = format;
    while( p = _strstr_( p, _Tstr("%(") ) )
    {
        const Tchar *pf = p;
        p += 2;

        bool bracket_format_found = true;
        if( pf > format )
            if( pf[-1] == _Tchar('%') )     // if we found "%%(", then this is *not* our formatting
                bracket_format_found = false;

        const Tchar *pcmd = p;

        p  = _strchr_( p, _Tchar(')') );
        if( !p ) break;
        ++p;

        if( bracket_format_found )
        {
            if( pf > p0 )
                str.Strcat( p0, (_size_t_)(pf - p0) );

            str += GetTimeStringFromSpecialFormat( pcmd, ptm, nanos, Locale ).DuplicateChar( _Tchar('%') );
            // Note: The string returned from #GetTimeStringFromSpecialFormat()#
            //  is already expanded and could contain '%' chars which must not
            //  be treated as command chars for #strftime()# below.
            //  So we duplicate them so that #strftime()# expands each "%%"
            //  sequence back to one single '%'.
        }
        else
            str.Strcat( p0, (_size_t_)(p - p0) );

        p0 = p;
    }

    str += p0;

    _size_t_ zlen = str.length() + 256 + 1;

    Tchar *pdest = GetBuffer( zlen - 1 );
    if( !pdest )
        { Empty(); return *this; }

    if( !_strftime_( pdest, zlen, str, ptm ) )
        pdest[0] = 0;

    pdest[zlen - 1] = (Tchar) 0;    // just for savety. is not necessary
    ReleaseBuffer();
    return *this;
}

// Forms a time string in local time.
// For format of #format# string, see C-Runtime documentation
// on descriptions to #strftime()#.
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::timeFmt( time_t t, const Tchar *format )
{
    Tchar s[256];
    struct tm _tm;
    struct tm *ptmtmp = _localtime_( &_tm, &t );

    if( !ptmtmp ) s[0] = (Tchar) 0;
    else if( !_strftime_( s, _countof(s), format, ptmtmp ) ) s[0] = (Tchar) 0;

    return s;
}

// Forms a time string in UTC time.
// For format of #format# string, see C-Runtime documentation
// on descriptions to #strftime()#.
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::timeFmtGmt( time_t t, const Tchar *format )
{
    Tchar s[256];
    struct tm _tm;
    struct tm *ptmtmp = _gmtime_( &_tm, &t );

    if( !ptmtmp ) s[0] = (Tchar) 0;
    else if( !_strftime_( s, _countof(s), format, ptmtmp ) ) s[0] = (Tchar) 0;

    return s;
}

#endif

template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::getWeekdayFullName( int ix )
{
    int year = 2001, month = 1, day = (ix + 6) % 7 + 1, hour = 12, min = 0, sec = 0, dst = -1;

    struct tm atm;
    atm.tm_sec = sec;
    atm.tm_min = min;
    atm.tm_hour = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon = month - 1;
    // (year >= 1900) must be
    atm.tm_year = year - 1900;
    atm.tm_isdst = dst;

    time_t t = mktime( &atm );

    return timeFmt( t, _Tstr("%A") );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::getWeekdayAbbrevName( int ix )
{
    int year = 2001, month = 1, day = (ix + 6) % 7 + 1, hour = 12, min = 0, sec = 0, dst = -1;

    struct tm atm;
    atm.tm_sec = sec;
    atm.tm_min = min;
    atm.tm_hour = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon = month - 1;
    // (year >= 1900) must be
    atm.tm_year = year - 1900;
    atm.tm_isdst = dst;

    time_t t = mktime( &atm );

    return timeFmt( t, _Tstr("%a") );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::getMonthFullName( int ix )
{
    int year = 2001, month = (ix + 155) % 12 + 1, day = 1, hour = 12, min = 0, sec = 0, dst = -1;

    struct tm atm;
    atm.tm_sec = sec;
    atm.tm_min = min;
    atm.tm_hour = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon = month - 1;
    // (year >= 1900) must be
    atm.tm_year = year - 1900;
    atm.tm_isdst = dst;

    time_t t = mktime( &atm );

    return timeFmt( t, _Tstr("%B") );
}

template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::getMonthAbbrevName( int ix )
{
    int year = 2001, month = (ix + 155) % 12 + 1, day = 1, hour = 12, min = 0, sec = 0, dst = -1;

    struct tm atm;
    atm.tm_sec = sec;
    atm.tm_min = min;
    atm.tm_hour = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon = month - 1;
    // (year >= 1900) must be
    atm.tm_year = year - 1900;
    atm.tm_isdst = dst;

    time_t t = mktime( &atm );

    return timeFmt( t, _Tstr("%b") );
}

///<param name="flags"> a combination of values from <b>strplaceholder_t</b>. Only
///  StrNumT_* and StrDigNumF_* values are supported. </param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromInt( int val, strplaceholder_v flags )
{
    static const Tchar *apstrFormats1[] = { _Tstr("%d"), _Tstr("0x%X"), _Tstr("0%o") };
    static const Tchar *apstrFormats2[] = { _Tstr("%d"), _Tstr("0x%08X"), _Tstr("0%011o") };

    int numt = flags & StrNumT__Mask;
    if( numt > StrNumT_Octal ) return StringT();

    StringT s;
    s.Sprintf( flags & StrDigNumF_PadWithZeros ?
        apstrFormats2[numt] : apstrFormats1[numt], val );
    return s;
}

///<param name="flags"> a combination of values from <b>strplaceholder_t</b>. Only
///  StrNumT_* and StrDigNumF_* values are supported. </param>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromUint( UINT val, strplaceholder_v flags )
{
    static const Tchar *apstrFormats1[] = { _Tstr("%u"), _Tstr("0x%X"), _Tstr("0%o") };
    static const Tchar *apstrFormats2[] = { _Tstr("%u"), _Tstr("0x%08X"), _Tstr("0%011o") };

    int numt = flags & StrNumT__Mask;
    if( numt > StrNumT_Octal ) return StringT();

    StringT s;
    s.Sprintf( flags & StrDigNumF_PadWithZeros ?
        apstrFormats2[numt] : apstrFormats1[numt], val );
    return s;
}

///<summary>
///  <b>fromGUID()</b> converts a GUID into a string.
///  The implementation is an own implementation and returns
///  uppercase letters.
///</summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromGUID( const GUID& guid )
{
    StringT s;
    s.Sprintf( _Tstr("{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}"),
               guid.Data1, guid.Data2, guid.Data3,
               guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3],
               guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7] );
    return s;
}

///<summary>
///  <b>fromGUID2()</b> converts a GUID into a string using the
///  system's <b>StringFromGUID2()</b> function.
///</summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromGUID2( const GUID& guid )
{
    OLECHAR s[128];
    s[0] = (OLECHAR) 0;
    StringFromGUID2( guid, s, _countof(s) );
    return s;
}

///<summary>
///  <b>fromCLSID()</b> converts a CLSID into a string using the
///  system's <b>StringFromCLSID()</b> function.
///</summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromCLSID( const CLSID& clsid )
{
    StringT s;
    LPOLESTR *ps = NULL;

    if( StringFromCLSID( clsid, &ps ) == S_OK )
    {
        if( ps ) s = *ps;
        CoTaskMemFree( ps );
    }

    return s;
}

///<summary>
///  <b>fromIID()</b> converts an IID into a string using the
///  system's <b>StringFromIID()</b> function.
///</summary>
template< class baseString, typename Tchar >
StringT<baseString, Tchar> StringT<baseString, Tchar>::fromIID( const IID& iid )
{
    StringT s;
    LPOLESTR *ps = NULL;

    if( StringFromIID( iid, &ps ) == S_OK )
    {
        if( ps ) s = *ps;
        CoTaskMemFree( ps );
    }

    return s;
}

///<summary> Hashkey value calculation for <b>HashTable</b> template. </summary>
template< class baseString, typename Tchar >
_hashval_t_ StringT<baseString, Tchar>::EvalHashKey() const
{
    // We take up to the first 64 characters of the string
    // and form a simple hash key value from it.
    DWORD hval = 0;
    _size_t_ maxlen = 64;

    const Tchar *ptr = *this;

    while( *ptr && --maxlen >= 0 )
    {
        hval  = (((DWORD) LOWORD(hval)) << 16) | (DWORD) HIWORD(hval);
        hval += (DWORD) *(ptr++);
    }

    return (_hashval_t_) hval;
}



#ifdef __argonMemory_h
typedef StringT<privateString::BaseStringShared<TCHAR>,   TCHAR>    _string_;
typedef StringT<privateString::BaseStringShared<char>,    char>     _stringA_;
typedef StringT<privateString::BaseStringShared<wchar_t>, wchar_t>  _stringW_;
#endif

#ifndef _argon_no_CString
typedef StringT<CString,  TCHAR>    String;
typedef StringT<CStringA, char>     StringA;
typedef StringT<CStringW, wchar_t>  StringW;
#else
#ifdef __argonMemory_h
typedef _string_    String;
typedef _stringA_   StringA;
typedef _stringW_   StringW;
#else
#error If you don't want to use CString as base class for StringT<>, you need to include "argonMemory.h" *before* "argonString.h" and "argonErrors.h" (and before other include files that include one of these).
#endif
#endif

// Note: Above, the types String, StringA and StringW have been defined.
// These are the main types to be used for strings.
//
// The types _string_, _stringA_ and _stringW_ are defined only if argonMemory.h
// has been included before argonString.h and argonErrors.h (and any other header file
// that includes one of these).
// These types are always based only on the memory handling of argonMemory.

// If you want to use the following string types, you will need to include
// "argonMinistring.h" in your project. Since the following types are only
// defines, it is not necessary to include "argonMinistring.h" before "argonString.h".
#define _ministring_    StringT<miniString<TCHAR>,   TCHAR>
#define _ministringA_   StringT<miniString<char>,    char>
#define _ministringW_   StringT<miniString<wchar_t>, wchar_t>




}

#endif
